/*
 *  Copyright (C) 2007  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <libintl.h>
#include <math.h>
#include <gtk/gtk.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-object.h>
#include <kiba-fixed-object.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>

#define _(string)   gettext (string)
#define PLUGIN_NAME "bounce"

#define OBJECT_SIZE(obj) MAX(GTK_WIDGET(obj)->allocation.width, GTK_WIDGET(obj)->allocation.height)

typedef struct _KibaBounce       KibaBounce;
typedef struct _KibaBounceObject KibaBounceObject;

enum {
	KIBA_BOUNCE_UP = 1,
	KIBA_BOUNCE_DOWN = -1
};

struct _KibaBounceObject {
	KibaBounce  *bounce;
	GtkWidget  *widget;
	float      offset;
	gint        max_height;
	gint        bounce_id;
	gint        state;
	
	gboolean    entered;
	gboolean    urgent;
	
	gboolean    scale_up;
	float      scaleX;
	float      scaleY;
	float      scale_offset;
	gint        scale_id;
};

typedef enum {
	KIBA_BOUNCE_URGENT = 0,
	KIBA_BOUNCE_SCALE,
	KIBA_BOUNCE_HOVERING,
	KIBA_BOUNCE_TRIGGERING
}KibaBounceFlags;

struct _KibaBounce {
	GList *objects;
	KibaBounceFlags flags;
	
	gint height;
	gint exec_height;
	gint step;
};

static KibaBounceObject* kiba_bounce_get_object_for_widget (KibaBounce  *bounce,
							GtkWidget *widget) {
	GList *l;
	
	for (l = bounce->objects; l != NULL; l = l->next) {
		KibaBounceObject *object = l->data;
		
		if (object->widget == widget)
			return object;
	}
	
	return NULL;
}

static gint scale_object_timeout (gpointer data) {
	KibaBounceObject *object = data;
	GtkWidget *dock;
	float factor_step;
	const KibaLayoutData *d;
	
	dock = object->widget->parent;
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	factor_step = (0.04 + (0.04*(((float) object->bounce->step)/50.0)))*(object->scaleX/1.25);
	
	kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(object->widget));
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		if (object->scale_up) {
			object->scaleX -= factor_step;
			object->scaleY += factor_step;
			
			if (object->scaleY >= 1.5) {
				object->scaleY = 1.5;
				object->scaleX = 0.5;
				object->scale_up = FALSE;
				object->state = KIBA_BOUNCE_UP;
			}
		}
		else {
			object->scaleX += factor_step;
			object->scaleY -= factor_step;
			
			if (object->scaleY <= 1.0) {
				object->scaleY = 1.0;
				object->scaleX = 1.0;
				object->scale_up = TRUE;
			}
		}
		
		object->scale_offset = (object->widget->allocation.width*object->scaleX)*(object->scaleX-1.0f)/2.0f;
	}
	else {
		if (object->scale_up) {
			object->scaleY -= factor_step;
			object->scaleX += factor_step;
			
			if (object->scaleX >= 1.5) {
				object->scaleX = 1.5;
				object->scaleY = 0.5;
				object->scale_up = FALSE;
				object->state = KIBA_BOUNCE_UP;
			}
		}
		else {
			object->scaleY += factor_step;
			object->scaleX -= factor_step;
			
			if (object->scaleX <= 1.0) {
				object->scaleX = 1.0;
				object->scaleY = 1.0;
				object->scale_up = TRUE;
			}
		}
		
		object->scale_offset = (object->widget->allocation.height*object->scaleY)*(object->scaleY-1.0f)/2.0f;
	}
	
	kiba_object_update_scale_factor (KIBA_OBJECT(object->widget));
	
	if (object->scaleX == 1.0) {
		object->scale_id = 0;
		return FALSE;
	}
	
	return TRUE;
}

static gint bounce_object_timeout (gpointer data) {
	KibaBounceObject *object = data;
	
	if (object->scale_id)
		return TRUE;
	
	kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(object->widget));
	
	float slow_down_factor = ((float) object->bounce->step)/10.0;
	object->offset +=  object->state *
		(slow_down_factor *(1.5 - object->offset/object->max_height));
	object->offset = MIN (object->offset, (float) object->max_height);
	
	kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(object->widget));
		
	if (object->offset == (float) object->max_height)
		object->state = KIBA_BOUNCE_DOWN;
	
	if (object->state == KIBA_BOUNCE_DOWN && object->offset <= 0.0) {
		gboolean stop_bounce = FALSE;
		gboolean urgent;
		gboolean hovering;
		
		hovering = (GET_FLAG(object->bounce->flags, KIBA_BOUNCE_HOVERING) && object->entered);
		if (!hovering) {
			urgent = (GET_FLAG(object->bounce->flags, KIBA_BOUNCE_URGENT) && object->urgent);
			if (!urgent) {
				stop_bounce = TRUE;
			}
		}
		
		if (object->max_height == object->bounce->exec_height)
			object->max_height = object->bounce->height;
		
		if (GET_FLAG(object->bounce->flags, KIBA_BOUNCE_SCALE)) {
			object->scale_up = TRUE;
			object->scale_id = g_timeout_add (20, scale_object_timeout, object);
		}
		else
			object->state = KIBA_BOUNCE_UP;
		
		if (stop_bounce || !object->max_height ||
		    GET_FLAG(KIBA_OBJECT(object->widget)->state, KIBA_OBJECT_DRAGGING)) {
			object->bounce_id = 0;
			object->offset = 0.0;
			return FALSE;
		}
	}
	
	return TRUE;
}

static void kiba_object_start_bouncing (KibaBounceObject *object,
				 gint              bounce_height) {
	if (bounce_height <= 0.0)
		return;
	
	if (GET_FLAG(KIBA_OBJECT(object->widget)->state, KIBA_OBJECT_DRAGGING))
		return;
	
	object->max_height = bounce_height;
	
	if (!object->bounce_id) {
		object->state = KIBA_BOUNCE_UP;
		object->bounce_id = g_timeout_add (20, bounce_object_timeout, object);
	}
}

static void kiba_object_stop_bouncing (KibaBounceObject *object) {
	object->entered = FALSE;
}

static void object_urgent_event_cb (KibaObject *object,
				    gboolean    urgent,
				    gpointer    data) {
	KibaBounceObject *bounce = data;
	
	if(urgent) {
		bounce->urgent = TRUE;
		if (GET_FLAG(bounce->bounce->flags, KIBA_BOUNCE_URGENT))
			kiba_object_start_bouncing(bounce, bounce->bounce->exec_height);
	} else {
		bounce->urgent = FALSE;
	}
}

static void kiba_prefs_bounce_height_cb (const char *option_id,
				         gpointer    value,
				         gpointer    data) {
	KibaBounce *bounce = data;
	
	bounce->height = GPOINTER_TO_INT(value);
}

static void kiba_prefs_bounce_exec_height_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaBounce *bounce = data;
	
	bounce->exec_height = GPOINTER_TO_INT(value);
}

static void kiba_prefs_bounce_hovering_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaBounce *bounce = data;
	gboolean hovering = GPOINTER_TO_INT(value);
	
	if (hovering) {
		if (!GET_FLAG(bounce->flags, KIBA_BOUNCE_HOVERING))
			SET_FLAG(bounce->flags, KIBA_BOUNCE_HOVERING);
	}
	else if (GET_FLAG(bounce->flags, KIBA_BOUNCE_HOVERING))
		DEL_FLAG(bounce->flags, KIBA_BOUNCE_HOVERING);
}

static void kiba_prefs_bounce_scale_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaBounce *bounce = data;
	gboolean scale = GPOINTER_TO_INT(value);
	
	if (scale) {
		if (!GET_FLAG(bounce->flags, KIBA_BOUNCE_SCALE))
			SET_FLAG(bounce->flags, KIBA_BOUNCE_SCALE);
	}
	else if (GET_FLAG(bounce->flags, KIBA_BOUNCE_SCALE))
		DEL_FLAG(bounce->flags, KIBA_BOUNCE_SCALE);
}

static void kiba_prefs_bounce_urgent_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaBounce *bounce = data;
	gboolean urgent = GPOINTER_TO_INT(value);
	
	if (urgent) {
		if (!GET_FLAG(bounce->flags, KIBA_BOUNCE_URGENT))
			SET_FLAG(bounce->flags, KIBA_BOUNCE_URGENT);
	}
	else if (GET_FLAG(bounce->flags, KIBA_BOUNCE_URGENT))
		DEL_FLAG(bounce->flags, KIBA_BOUNCE_URGENT);
}

static void kiba_prefs_bounce_triggering_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaBounce *bounce = data;
	gboolean triggering = GPOINTER_TO_INT(value);
	
	if (triggering) {
		if (!GET_FLAG(bounce->flags, KIBA_BOUNCE_TRIGGERING))
			SET_FLAG(bounce->flags, KIBA_BOUNCE_TRIGGERING);
	}
	else if (GET_FLAG(bounce->flags, KIBA_BOUNCE_TRIGGERING))
		DEL_FLAG(bounce->flags, KIBA_BOUNCE_TRIGGERING);
}

static void kiba_prefs_bounce_step_cb (const char *option_id,
				     gpointer    value,
				     gpointer    data) {
	KibaBounce *bounce = data;
	
	bounce->step = GPOINTER_TO_INT(value);
}

static gboolean object_release_event_cb (GtkWidget      *widget,
					 GdkEventButton *event,
					 gpointer        data) {
	KibaBounceObject *object = data;
	
	if (GET_FLAG(object->bounce->flags, KIBA_BOUNCE_TRIGGERING))
		kiba_object_start_bouncing(object, object->bounce->exec_height);
	
	return FALSE;
}

static gboolean object_enter_event_cb  (GtkWidget        *widget,
					GdkEventCrossing *event,
					gpointer          data) {
	KibaBounceObject *object = data;
	
	object->entered = TRUE;
	if (GET_FLAG(object->bounce->flags, KIBA_BOUNCE_HOVERING))
		kiba_object_start_bouncing(object, object->bounce->height);
	
	return FALSE;
}

static gboolean object_leave_event_cb  (GtkWidget        *widget,
					GdkEventCrossing *event,
					gpointer          data) {
	KibaBounceObject *object = data;
	
	object->entered = FALSE;
	
	return FALSE;
}

static void kiba_bounce_place_surface (KibaObject *object,
				       float     *x,
				       float     *y,
				       gpointer    data) {
	KibaBounceObject *bounce_object = data;
	GtkWidget *dock;
	const KibaLayoutData *d;
	
	dock = GTK_WIDGET(object)->parent;
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		*x = bounce_object->offset + bounce_object->scale_offset;
		*y = 0.0;
	}
	else {
		*x = 0.0;
		*y = bounce_object->offset + bounce_object->scale_offset;
	}
}

static void kiba_bounce_max_offset_func (KibaObject *object,
				         float     *x,
				         float     *y,
				         gpointer    data) {
	KibaBounceObject *bounce_object = data;
	GtkWidget *dock;
	const KibaLayoutData *d;
	
	dock = GTK_WIDGET(object)->parent;
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		*x = (float) MAX(bounce_object->bounce->height, bounce_object->bounce->exec_height);
		*y = 0.0;
	}
	else {
		*x = 0.0;
		*y = (float) MAX(bounce_object->bounce->height, bounce_object->bounce->exec_height);
	}
}

static void kiba_object_scale_func (KibaObject *object,
				    float     *scale_x,
				    float     *scale_y,
				    gpointer    data) {
	KibaBounceObject *bounce_object = data;
	
	*scale_x = bounce_object->scaleX;
	*scale_y = bounce_object->scaleY;
}

static void kiba_object_max_scale_func (KibaObject *object,
				        float     *scale_x,
				        float     *scale_y,
				        gpointer    data) {
	KibaBounceObject *bounce_object = data;
	GtkWidget *dock;
	const KibaLayoutData *d;
	
	dock = GTK_WIDGET(object)->parent;
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		*scale_x = 1.0;
		*scale_y = 1.25;
	}
	else {
		*scale_x = 1.25;
		*scale_y = 1.0;
	}
}

static void dock_object_add_cb (GtkContainer *container,
				GtkWidget    *widget,
				gpointer      data) {
	KibaBounce *bounce = data;
	KibaBounceObject *object;
	
	if (!KIBA_IS_DRAWABLE_OBJECT(widget))
		return;
	if (kiba_bounce_get_object_for_widget (bounce, widget) != NULL)
		return;
	
	object = g_new0 (KibaBounceObject, 1);
	object->widget = widget;
	object->offset = 0.0;
	object->max_height = 0;
	object->entered = FALSE;
	object->bounce = bounce;
	object->state = 0;
	object->bounce_id = 0;
	object->scaleX = 1.0;
	object->scaleY = 1.0;
	object->scale_offset = 0.0;
	object->urgent = FALSE;
	
	bounce->objects = g_list_append (bounce->objects, object);
	
	kiba_object_add_custom_offset_function (KIBA_OBJECT(widget),
		kiba_bounce_place_surface,
		kiba_bounce_max_offset_func,
		object);
	
	kiba_object_add_custom_scale_function (KIBA_OBJECT(widget),
		kiba_object_scale_func,
		kiba_object_max_scale_func,
		FALSE,
		object);
	
	g_signal_connect (G_OBJECT(widget), "enter-notify-event",
		G_CALLBACK(object_enter_event_cb), object);
	g_signal_connect (G_OBJECT(widget), "leave-notify-event",
		G_CALLBACK(object_leave_event_cb), object);
	g_signal_connect (G_OBJECT(widget), "button-release-event",
		G_CALLBACK(object_release_event_cb), object);
	g_signal_connect (G_OBJECT(widget), "urgent-toggled",
		G_CALLBACK(object_urgent_event_cb), object);
}

static void dock_object_remove_cb (GtkContainer *container,
				   GtkWidget    *widget,
				   gpointer      data) {
	KibaBounce *bounce = data;
	KibaBounceObject *object;
	
	object = kiba_bounce_get_object_for_widget (bounce, widget);
	
	if (!object)
		return;
	
	if (object->bounce_id)
		g_source_remove (object->bounce_id);
	if (object->scale_id)
		g_source_remove (object->scale_id);
	
	kiba_object_remove_custom_offset_function_by_data (KIBA_OBJECT(widget), object);
	kiba_object_remove_custom_scale_function_by_data (KIBA_OBJECT(object->widget), object);
	
	bounce->objects = g_list_remove (bounce->objects, object);
	
	g_free (object);
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaBounce *bounce;
	KibaWin *win;
	KibaDock *dock;
	GList *l;
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	bounce = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	for (l = bounce->objects; l != NULL; l = bounce->objects) {
		KibaBounceObject *object = l->data;
		
		if (object->bounce_id)
			g_source_remove (object->bounce_id);
		if (object->scale_id)
			g_source_remove (object->scale_id);
		
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_enter_event_cb), object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_leave_event_cb), object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_release_event_cb), object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_urgent_event_cb), object);
		
		kiba_object_remove_custom_offset_function_by_data (KIBA_OBJECT(object->widget), object);
		kiba_object_remove_custom_scale_function_by_data (KIBA_OBJECT(object->widget), object);
		
		g_free (object);
		bounce->objects = g_list_remove (bounce->objects, l->data);
	}
	g_list_free (bounce->objects);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_add_cb), bounce);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_remove_cb), bounce);
	
	g_free (bounce);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaBounce *bounce;
	KibaWin *win;
	KibaDock *dock;
	GtkContainer *container;
	GList *l;
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	container = GTK_CONTAINER(dock);
	
	bounce = g_new0(KibaBounce, 1);
	bounce->height = 20;
	bounce->exec_height = 30;
	bounce->step = 5;
	bounce->objects = NULL;
	
	kiba_prefs_add_callback (file, "bounce_height",
		kiba_prefs_bounce_height_cb, bounce, TRUE);
	kiba_prefs_add_callback (file, "bounce_exec_height",
		kiba_prefs_bounce_exec_height_cb, bounce, TRUE);
	kiba_prefs_add_callback (file, "bounce_speed",
		kiba_prefs_bounce_step_cb, bounce, TRUE);
	kiba_prefs_add_callback (file, "bounce_hovering",
		kiba_prefs_bounce_hovering_cb, bounce, TRUE);
	kiba_prefs_add_callback (file, "bounce_triggering",
		kiba_prefs_bounce_triggering_cb, bounce, TRUE);
	kiba_prefs_add_callback (file, "bounce_scale",
		kiba_prefs_bounce_scale_cb, bounce, TRUE);
	kiba_prefs_add_callback (file, "bounce_urgent",
		kiba_prefs_bounce_urgent_cb, bounce, TRUE);
	
	g_signal_connect (G_OBJECT(dock), "remove",
		G_CALLBACK(dock_object_remove_cb), bounce);
	g_signal_connect_after (G_OBJECT(dock), "add",
		G_CALLBACK(dock_object_add_cb), bounce);
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		
		dock_object_add_cb (container, child, bounce);
	}
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) bounce);
	
	return TRUE;
}

static PrefMember bounce_options[] = {
    { "bounce_height", "0", KIBA_INT },
    { "bounce_exec_height", "20", KIBA_INT },
    { "bounce_triggering", "TRUE", KIBA_BOOLEAN },
    { "bounce_hovering", "TRUE", KIBA_BOOLEAN },
    { "bounce_scale", "TRUE", KIBA_BOOLEAN },
    { "bounce_urgent", "TRUE", KIBA_BOOLEAN },
    { "bounce_speed", "5", KIBA_INT }
};

static KibaPluginVtable vtable = {
	"Bounce Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
};

KIBA_INIT_PLUGIN (vtable, bounce_options, NULL, NULL, NULL);
