/*
 *  Copyright (C) 2007  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
 *
 *  Note : based on xfce4-panel clipman plugin from Nick Schermer <nick@xfce.org>
*/

#include "config.h"

#include <string.h>
#include <stdlib.h>
#include <libintl.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>

#include <kiba.h>
#include <kiba-dock.h>
#include <kiba-icon.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>
#include <kiba-dialog.h>
#include <kiba-menu-items.h>
#include <kiba-macros.h>
#include <kiba-types.h>

#define _(String) dgettext (GETTEXT_PACKAGE, String)
#define PLUGIN_NAME "clipman"
/* Milisecond to check the clipboards(s) */
#define TIMER_INTERVAL  500

typedef struct _KibaClipman KibaClipman;

typedef enum {
	PRIMARY = 0,
	DEFAULT
} ClipboardType;

static const char* clipman_behaviour_data[] = {
	"normal",
	"strictly"
};

typedef enum {
	NORMAL = 0,
	STRICTLY
} ClipboardBehaviour;

typedef struct {
	GtkWidget    *menu_item;
	gchar        *text;
	gchar        *title;        /* I've added the title to save
					* some time when opening the menu */
	ClipboardType fromtype;
} ClipmanClip;

typedef struct {
	KibaClipman    *clipman;
	ClipmanClip    *clip;
} ClipmanAction;

static const char* clipman_flag_data[] = {
	"clipman_ignore_select",
	"clipman_prevent_empty",
	"clipman_separate_menu",
	"clipman_item_numbers"
};

typedef enum {
	CLIPMAN_IGNORE_SELECT = 0,
	CLIPMAN_PREVENT_EMPTY,
	CLIPMAN_SEPARATE_MENU,
	CLIPMAN_ITEM_NUMBERS
}KibaClipmanFlags;

struct _KibaClipman {
	GtkWidget           *icon;
	
	GPtrArray           *clips;
	ClipboardBehaviour   behaviour;
	KibaClipmanFlags     flags;
	
	int                  timeout_id;
	
	/* settings */
	gint                 menu_characters;
	gint                 history_items;
};

/* The clipboards */
static GtkClipboard *primaryClip;
static GtkClipboard *defaultClip;

/**
Helper Functions
**/

static void clipman_position_menu (GtkMenu     *menu,
				   int         *x,
				   int         *y,
				   gboolean    *push_in,
				   KibaClipman *clipman) {
	GtkWidget *dock;
	int width, height;
	GtkRequisition  req;
	const KibaLayoutData *d;
	
	dock = GTK_WIDGET(clipman->icon)->parent;
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	gtk_widget_size_request (GTK_WIDGET (menu), &req);
	
	*x = clipman->icon->allocation.x + dock->parent->allocation.x;
	*y = clipman->icon->allocation.y + dock->parent->allocation.y;
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		*x -= d->halignment == 0 ?
			-clipman->icon->allocation.width : req.width;
		*y -= (req.height - clipman->icon->allocation.width) / 2.0;
	}
	else {
		*x -= (req.width - clipman->icon->allocation.width) / 2.0;
		*y -= d->valignment == 0 ?
			-clipman->icon->allocation.width : req.height;
	}
}

static gchar* clipman_create_title (gchar *txt,
				    gint   length) {
	gchar *s, *t, *u, *buf;
	gint   i;
	
	/* rough string copy */
	s = g_strndup (txt, length*8 + 8);
	
	if (G_UNLIKELY (!s))
		return NULL;
	
	if (!g_utf8_validate (s, -1, NULL)) {
		DEBUG_INFO ("Title is not utf8 complaint, we're going to convert it\n");
	
		u = g_locale_to_utf8 (s, -1, NULL, NULL, NULL);
		g_free (s);
		s = u;
	
		/* Check the title again */
		if (!g_utf8_validate (s, -1, NULL)) {
			DEBUG_ERROR ("Title is still not utf8 complaint, we going to drop this clip\n");
			g_free (s);
			return NULL;
		}
	}
	
	/* create string length */
	buf = g_malloc(length*6); /* max length of utf8 char is 6 */
	g_utf8_strncpy (buf, s, length);
	g_free (s);
	s = buf;
	
	/* remove tabs and newlines */
	/* this works when the string is utf8-valid */
	i = 0;
	while (s[i] != '\0') {
		if (s[i] == '\n' || s[i] == '\r' || s[i] == '\t')
		s[i] = ' ';
		i++;
	}
	
	/* remove html characters */
	t = g_markup_escape_text (s, -1);
	g_free (s);
	
	return t;
}

static void clipman_free_clip (ClipmanClip *clip) {
	g_free (clip->text);
	g_free (clip->title);
	
	g_free (clip);
	
	DEBUG_INFO ("Clip successfully freed\n");
}

void clipman_check_array_len (KibaClipman *clipman) {
	ClipmanClip *clip;
	
	while (clipman->clips->len > clipman->history_items) {
		clip = g_ptr_array_index (clipman->clips, 0);
		g_ptr_array_remove (clipman->clips, clip);
		clipman_free_clip (clip);
		
		DEBUG_INFO ("A clip have been removed\n");
	}
}

static void clipman_add_clip (KibaClipman   *clipman,
			      gchar         *txt,
			      ClipboardType  type) {
	ClipmanClip *new_clip;
	
	if (G_LIKELY (txt != NULL) && G_LIKELY (strcmp (txt, ""))) {
		new_clip = g_new0 (ClipmanClip, 1);
	
		new_clip->title = clipman_create_title (txt, clipman->menu_characters);

		/* No valid title could be created, drop it... */
		if (new_clip->title == NULL) {
			g_free (new_clip);
			return;
		}
	
		new_clip->text     = g_strdup (txt);
		new_clip->fromtype = type;
	
		g_ptr_array_add (clipman->clips, new_clip);
	
		DEBUG_INFO ("Added clip %s\n", new_clip->title);
	}
}

static gboolean clipman_exists (KibaClipman   *clipman,
				gchar         *txt,
				ClipboardType  type) {
	guint        i;
	ClipmanClip *clip;
	
	/* Walk through the array backwards, because
	* if the text exists, this will probably be the newest */
	
	for (i = clipman->clips->len; i--; ) {
		clip = g_ptr_array_index (clipman->clips, i);
	
		if (G_LIKELY ((clip->text != NULL) && (strcmp(clip->text, txt) == 0))) {
			switch (clipman->behaviour) {
				case NORMAL:
					if (type == DEFAULT && clip->fromtype == PRIMARY)
						clip->fromtype = DEFAULT;
					return TRUE;
				
				case STRICTLY:
					if (type == clip->fromtype)
						return TRUE;
					return FALSE;
			}
		}
	}
	
	return FALSE;
}

static void clipman_restore_empty (KibaClipman   *clipman,
				   ClipboardType  type) {
	guint        i;
	ClipmanClip *clip;
	
	if (clipman->behaviour == STRICTLY && G_LIKELY (clipman->clips->len > 0)) {
		/* Walk through the array till a clip of it's own
		* type is found, then past it in the clipboard */
		for (i = clipman->clips->len; i--;) {
			clip = g_ptr_array_index (clipman->clips, i);
	
			if (clip->fromtype == type) {
				switch (type) {
					case PRIMARY:
						gtk_clipboard_set_text(primaryClip, clip->text, -1);
						break;
					case DEFAULT:
						gtk_clipboard_set_text(defaultClip, clip->text, -1);
						break;
				}
				
				DEBUG_INFO ("Clipboard restored with a clip from same type\n");
				
				break;
			}
		}
	}
	else if (clipman->clips->len > 0) {
		/* Grap the latest clip and paste it in the clipboard */
		clip = g_ptr_array_index (clipman->clips, (clipman->clips->len-1));
	
		switch (type) {
			case PRIMARY:
				gtk_clipboard_set_text(primaryClip, clip->text, -1);
				break;
			case DEFAULT:
				gtk_clipboard_set_text(defaultClip, clip->text, -1);
				break;
		}
	
		DEBUG_INFO ("Last clip added\n");
	}
}

static gboolean kiba_clipman_time_handler (KibaClipman *clipman) {
	gchar *ptext = NULL, *dtext;
	GdkPixbuf *ppixbuf, *dpixbuf;
	GdkModifierType  state;
	
	if (!GET_FLAG(clipman->flags, CLIPMAN_IGNORE_SELECT)) {
		/* Get mouse button information */
		gdk_window_get_pointer(NULL, NULL, NULL, &state);
	
		ptext = gtk_clipboard_wait_for_text (primaryClip);
	
		if (GET_FLAG(clipman->flags, CLIPMAN_PREVENT_EMPTY) && ptext == NULL) {
			clipman_restore_empty (clipman, PRIMARY);
		}
		else if (ptext != NULL && !(state & GDK_BUTTON1_MASK) &&
		        !clipman_exists (clipman, ptext, PRIMARY)) {
			DEBUG_INFO ("Item added from primary clipboard\n");
			clipman_add_clip (clipman, ptext, PRIMARY);
			clipman_check_array_len (clipman);
		}
	
		g_free (ptext);
	}
	
	dtext = gtk_clipboard_wait_for_text (defaultClip);
	
	/* Check default clipboard */
	if (GET_FLAG(clipman->flags, CLIPMAN_PREVENT_EMPTY) && dtext == NULL) {
		clipman_restore_empty (clipman, DEFAULT);
	}
	else if (G_LIKELY (dtext != NULL) && !clipman_exists (clipman, dtext, DEFAULT)) {
		DEBUG_INFO ("Item added from default clipboard\n");
		clipman_add_clip (clipman, dtext, DEFAULT);
		clipman_check_array_len (clipman);
	}
	
	g_free (dtext);
	
	return TRUE;
}

/**
Signal Handler
**/

static void clipman_destroy_menu (GtkWidget   *menu,
				  gpointer     data) {
	gtk_widget_destroy (menu);
}

static void clear_dialog_response (GtkWidget    *widget,
				   int           response,
				   KibaClipman  *clipman) {
	if (response == GTK_RESPONSE_YES) {
		ClipmanClip *clip;
		
		gtk_clipboard_clear (primaryClip);
		gtk_clipboard_clear (defaultClip);
		
		while (clipman->clips->len > 0) {
			clip = g_ptr_array_index (clipman->clips, 0);
			g_ptr_array_remove (clipman->clips, clip);
			clipman_free_clip (clip);
		}
	
		/* 'Save' the empty clipboard */
// 		clipman_save (clipman->plugin, clipman);
	}
	
	gtk_widget_destroy (widget);
}

static gboolean clipman_clear (GtkWidget      *mi,
			       GdkEventButton *ev,
			       KibaClipman    *clipman) {
	
	GtkWidget   *dialog;
	
	dialog = kiba_dialog_confirm (_("Clear the Clipboard?"),
		_("If you choose to clear the Clipboard, all items in it will be permanently lost. Please note that you can also delete them individually by using Button 3 in the Context Menu."));
	
	g_signal_connect (dialog, "response",
			  G_CALLBACK (clear_dialog_response), clipman);
	
	return FALSE;
}

/**
Clipman Menu
**/

static gboolean clipman_item_clicked (GtkWidget      *mi,
				      GdkEventButton *ev,
				      ClipmanAction  *action) {
	gchar *dtext, *ptext;
	
	if (ev->button == 1 && action->clipman->behaviour == STRICTLY) {
		DEBUG_INFO ("Clip copied to his own clipboard (STRICTLY)");
	
		if (action->clip->fromtype == DEFAULT) {
			gtk_clipboard_clear (defaultClip);
			gtk_clipboard_set_text (defaultClip, action->clip->text, -1);
		}
	
		if (action->clip->fromtype == PRIMARY) {
			gtk_clipboard_clear (primaryClip);
			gtk_clipboard_set_text (primaryClip, action->clip->text, -1);
		}
	}
	else if (ev->button == 1) {
		gtk_clipboard_clear (defaultClip);
		gtk_clipboard_set_text (defaultClip, action->clip->text, -1);
		DEBUG_INFO ("Clip copied to default clipboard");
	
		if (!GET_FLAG(action->clipman->flags, CLIPMAN_IGNORE_SELECT)) {
			gtk_clipboard_clear (primaryClip);
			gtk_clipboard_set_text (primaryClip, action->clip->text, -1);
			DEBUG_INFO ("Clip copied to primary clipboard");
		}
	}
	else if (ev->button == 3) {
		DEBUG_INFO ("Removed the selected clip from the History");
		
		dtext = gtk_clipboard_wait_for_text (defaultClip);
		if (dtext && !strcmp(dtext, action->clip->text)) {
			gtk_clipboard_clear (defaultClip);
			gtk_clipboard_set_text (defaultClip, "", -1);
		}
		g_free (dtext);
	
		ptext = gtk_clipboard_wait_for_text (primaryClip);
		if (ptext && !strcmp(ptext, action->clip->text)) {
			gtk_clipboard_clear (primaryClip);
			gtk_clipboard_set_text (primaryClip, "", -1);
		}
		g_free (ptext);
	
		g_ptr_array_remove (action->clipman->clips, action->clip);
		clipman_free_clip (action->clip);
	}
	
	g_free (action);
	
	return FALSE;
}

static gboolean key_press_handler (GtkWidget    *menu,
				   GdkEventKey  *event,
				   gpointer      data) {
	KibaClipman *clipman = data;
	ClipmanClip *clip;
	gchar       *letter;
	gint         i;
	
	/* we just wanna act on press events */
	if (event->type != GDK_KEY_PRESS)
		return FALSE;
	
	letter = gdk_keyval_name (event->keyval);
	
	/* activate selected menuitem */
	if (event->keyval == GDK_Return) {
		GtkWidget *selected;
		
		//FIXME: get the selected menu item, gtk really should offer a easy function for that
		selected = gtk_menu_get_active (GTK_MENU (menu));
		
		/* walk through the clips */
		for (i = clipman->clips->len; i--; ) {
			clip = g_ptr_array_index (clipman->clips, i);
		
			if (selected == clip->menu_item) {
				ClipmanAction  *action;
				GdkEvent       *event;
				
				event = gdk_event_new (GDK_BUTTON_RELEASE);
				((GdkEventButton*)event)->button = 1;
				
				action = g_new0 (ClipmanAction, 1);
				action->clipman = clipman;
				action->clip = clip;
				
				clipman_item_clicked (selected, (GdkEventButton*)event, action);
				
				return TRUE;
			}
		}
	}
	/* we just wanna check for letters */
	else if (strlen (letter) > 1)
		return FALSE;
	
	/* walk through the clips */
	for (i = clipman->clips->len; i--; ) {
		clip = g_ptr_array_index (clipman->clips, i);
		
		/* we found a clip */
		if (g_str_has_prefix (clip->text, letter)) {
			gtk_menu_shell_select_item (GTK_MENU_SHELL (menu), clip->menu_item);
			return TRUE;
		}
	}
	
	return FALSE;
}

static GtkWidget* clipman_create_menuitem (ClipmanAction *action,
					   guint          width,
					   guint          number,
					   gboolean       bold) {
	GtkWidget *item, *image;
	GtkLabel  *label;
	GdkPixbuf *icon;
	gchar     *title;
	
	item = gtk_image_menu_item_new_with_label  ("");
	
	if (GET_FLAG(action->clipman->flags, CLIPMAN_ITEM_NUMBERS)) {
		if (number < 10)
		title = g_strdup_printf ("<tt><span size=\"smaller\">%d. </span></tt> %s",
				number, action->clip->title);
		else
		title = g_strdup_printf ("<tt><span size=\"smaller\">%d.</span></tt> %s",
				number, action->clip->title);
	}
	else if (bold)
		title = g_strdup_printf("<b>%s</b>", action->clip->title);
	else
		title = g_strdup_printf("%s", action->clip->title);
	
	if (g_str_has_suffix(action->clip->text,".jpg") ||
	    g_str_has_suffix(action->clip->text,".png") ||
	    g_str_has_suffix(action->clip->text,".svg") ||
	    g_str_has_suffix(action->clip->text,".gif")) {
		icon = NULL;
		if (kiba_load_icon (action->clip->text, 24, &icon)) {
			image = gtk_image_new_from_pixbuf (icon);
			gdk_pixbuf_unref (icon);
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM(item), image);
		}
	}
	
	label = GTK_LABEL(GTK_BIN(item)->child);
	gtk_label_set_markup (label, title);
	gtk_label_set_single_line_mode (label, TRUE);
	gtk_label_set_ellipsize (label, PANGO_ELLIPSIZE_END);
	gtk_label_set_max_width_chars (label, width);
	
	return item;
}

static void clipman_clicked_separated (GtkWidget   *menu,
				       KibaClipman *clipman) {
	gchar *ptext, *dtext;
	guint i, j;
	ClipmanAction *action = NULL;
	ClipmanClip *clip;
	GtkWidget *mi;
	
	/* Default Clips */
	dtext = gtk_clipboard_wait_for_text (defaultClip);
	j = 0;
	
	for (i = clipman->clips->len; i--; ) {
		clip = g_ptr_array_index (clipman->clips, i);
	
		if (clip->fromtype == DEFAULT) {
			j++;
			
			action = g_new0 (ClipmanAction, 1);
			action->clipman = clipman;
			action->clip = clip;
		
			if (dtext != NULL && G_LIKELY (clip->text != NULL) &&
			    strcmp(clip->text, dtext) == 0) {
				mi = clipman_create_menuitem (action, clipman->menu_characters, j, TRUE);
				g_free (dtext);
				dtext = NULL;
			}
			else {
				mi = clipman_create_menuitem (action, clipman->menu_characters, j, FALSE);
			}
			clip->menu_item = mi;
			
			g_signal_connect (G_OBJECT(mi), "button_release_event",
				G_CALLBACK(clipman_item_clicked), action);
	
			gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
		}
	}
	
	g_free (dtext);
	
	if (j == 0) {
		mi = gtk_menu_item_new_with_label (_("< Default History Empty >"));
		gtk_widget_set_sensitive (mi, FALSE);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
	}
	
	/* Primary Clips */
	kiba_menu_add_separator (menu);
	
	ptext = gtk_clipboard_wait_for_text (primaryClip);
	j = 0;
	
	for (i = clipman->clips->len; i--; ) {
		clip = g_ptr_array_index (clipman->clips, i);
	
		if (clip->fromtype == PRIMARY) {
			j++;
	
			action = g_new0 (ClipmanAction, 1);
			action->clipman = clipman;
			action->clip = clip;
	
			if (ptext != NULL && G_LIKELY (clip->text != NULL) && strcmp(clip->text, ptext) == 0) {
				mi = clipman_create_menuitem (action, clipman->menu_characters, j, TRUE);
				g_free (ptext);
				ptext = NULL;
			}
			else {
				mi = clipman_create_menuitem (action, clipman->menu_characters, j, FALSE);
			}
			clip->menu_item = mi;
	
			g_signal_connect (G_OBJECT(mi), "button_release_event",
				G_CALLBACK(clipman_item_clicked), action);
	
			gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
		}
	}
	
	g_free (ptext);
	
	if (j == 0) {
		mi = gtk_menu_item_new_with_label (_("< Selection History Empty >"));
		gtk_widget_set_sensitive (mi, FALSE);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
	}
	
	kiba_menu_add_separator (menu);
	
	mi = gtk_menu_item_new_with_label (_("Clear History"));
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
	
	g_signal_connect (G_OBJECT (mi), "button_release_event",
		G_CALLBACK (clipman_clear), clipman);
}

static void clipman_clicked_not_separated (GtkWidget   *menu,
					   KibaClipman *clipman) {
	gchar         *ptext, *dtext;
	guint          i;
	ClipmanAction *action = NULL;
	ClipmanClip   *clip;
	GtkWidget     *mi;
	
	ptext = gtk_clipboard_wait_for_text (primaryClip);
	dtext = gtk_clipboard_wait_for_text (defaultClip);
	
	for (i = clipman->clips->len; i--;) {
		clip = g_ptr_array_index (clipman->clips, i);
	
		action = g_new0 (ClipmanAction, 1);
		action->clipman = clipman;
		action->clip = clip;
	
		if (dtext != NULL && G_LIKELY (clip->text != NULL) && strcmp(clip->text, dtext) == 0) {
			mi = clipman_create_menuitem (action, clipman->menu_characters, clipman->clips->len-i, TRUE);
			g_free (dtext);
			dtext = NULL;
		}
		else if (ptext != NULL && G_LIKELY (clip->text != NULL) && strcmp(clip->text, ptext) == 0) {
			mi = clipman_create_menuitem (action, clipman->menu_characters, clipman->clips->len-i, TRUE);
			g_free (ptext);
			ptext = NULL;
		}
		else {
			mi = clipman_create_menuitem (action, clipman->menu_characters, clipman->clips->len-i, FALSE);
		}
		clip->menu_item = mi;
		
		g_signal_connect (G_OBJECT(mi), "button_release_event",
			G_CALLBACK(clipman_item_clicked), action);
	
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
	}
	
	g_free (ptext);
	g_free (dtext);
	
	kiba_menu_add_separator (menu);
	
	mi = gtk_menu_item_new_with_label (_("Clear History"));
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
	
	g_signal_connect (G_OBJECT (mi), "button_release_event",
		G_CALLBACK (clipman_clear), clipman);
}

/**
KibaObject Callbacks
**/

static gboolean kiba_object_button_release_cb (GtkWidget      *widget,
					       GdkEventButton *event,
					       gpointer        data) {
	GtkWidget *mi;
	GtkWidget *menu;
	KibaClipman *clipman = data;
	gchar     *title;
	
	menu = gtk_menu_new ();
	
	title = g_strdup_printf("<span weight=\"bold\">%s</span>", _("Clipman History"));
	mi = gtk_menu_item_new_with_label  ("");
	gtk_label_set_markup(GTK_LABEL(GTK_BIN(mi)->child), title);
	g_free (title);
	gtk_widget_set_sensitive (mi, FALSE);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
	
	kiba_menu_add_separator (menu);
	
	if (GET_FLAG(clipman->flags, CLIPMAN_SEPARATE_MENU) &&
	   !GET_FLAG(clipman->flags, CLIPMAN_IGNORE_SELECT) &&
	    G_LIKELY (clipman->clips->len > 0)) {
		clipman_clicked_separated (menu, clipman);
	}
	else if (G_LIKELY (clipman->clips->len > 0)) {
		clipman_clicked_not_separated (menu, clipman);
	}
	else {
		mi = gtk_menu_item_new_with_label (_("< Clipboard Empty >"));
		gtk_widget_set_sensitive (mi, FALSE);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
	}
	
	gtk_widget_show_all (menu);
	
	/* Also destroy the menu items when nothing is clicked */
	g_signal_connect (G_OBJECT(menu), "deactivate",
		G_CALLBACK(clipman_destroy_menu), clipman);
	
	/* select item when it starts with the letter of the pressed key,
	   and activate previously selected items on Return keypress */
	g_signal_connect (G_OBJECT (menu),
			  "key-press-event",
			  G_CALLBACK (key_press_handler),
			  clipman);
	
	gtk_menu_popup (GTK_MENU (menu), NULL, NULL,
				(GtkMenuPositionFunc) clipman_position_menu,
				clipman, 0,
				event->time);
	
	return FALSE;
}

/**
KibaPrefs Callbacks
**/

static void object_size_factor_cb (const char *option_id,
			           gpointer    value,
			           gpointer    data) {
	KibaClipman  *clipman = data;
	float factor;
	
	factor = g_ascii_strtod((char*) value, NULL);
	
	kiba_object_set_requisition_factor (KIBA_OBJECT(clipman->icon), factor, factor);
	KIBA_DRAWABLE_OBJECT(clipman->icon)->surface_data->width  = clipman->icon->allocation.width;
	KIBA_DRAWABLE_OBJECT(clipman->icon)->surface_data->height = clipman->icon->allocation.height;
	kiba_icon_set_file (KIBA_ICON(clipman->icon), "gtk-paste");
	gtk_widget_queue_resize (clipman->icon->parent);
}

static void clipman_flags_cb (const char *option_id,
			      gpointer    value,
			      gpointer    data) {
	KibaClipman *clipman = data;
	KibaClipmanFlags modified_flag;
	gint i;
	
	modified_flag = -1;
	for (i = 0; i < G_N_ELEMENTS(clipman_flag_data); i++) {
		if (strcmp (option_id, clipman_flag_data[i]) == 0)
			modified_flag = i;
	}
	if (modified_flag == -1)
		return;
	
	if ((gboolean) GPOINTER_TO_INT(value)) {
		if (!GET_FLAG (clipman->flags, modified_flag))
			SET_FLAG  (clipman->flags, modified_flag);
	}
	else if (GET_FLAG (clipman->flags, modified_flag))
		DEL_FLAG  (clipman->flags, modified_flag);
}

static void object_alpha_cb (const char *option_id,
			     gpointer    value,
			     gpointer    data) {
	KibaClipman *clipman = data;
	float alpha;
	
	alpha = g_ascii_strtod((char*) value, NULL);
	
	kiba_drawable_object_set_alpha (KIBA_DRAWABLE_OBJECT(clipman->icon), alpha);
}

static void clipman_behaviour_cb (const char *option_id,
			          gpointer    value,
			          gpointer    data) {
	KibaClipman *clipman = data;
	ClipboardBehaviour new_behaviour;
	gint i;
	const char *behaviour_string = (char*) value;
	
	new_behaviour = -1;
	for (i = 0; i < G_N_ELEMENTS(clipman_behaviour_data); i++) {
		if (strcmp (behaviour_string, clipman_behaviour_data[i]) == 0)
			new_behaviour = i;
	}
	if (new_behaviour == -1)
		return;
	
	clipman->behaviour = new_behaviour;
}

static void clipman_character_cb (const char *option_id,
				  gpointer    value,
				  gpointer    data) {
	KibaClipman *clipman = data;
	
	clipman->menu_characters = GPOINTER_TO_INT(value);
}

static void clipman_history_cb (const char *option_id,
				gpointer    value,
				gpointer    data) {
	KibaClipman *clipman = data;
	
	clipman->history_items = GPOINTER_TO_INT(value);
}

static void icon_theme_changed_cb (GtkIconTheme *icon_theme,
				   gpointer      user_data) {
	KibaClipman *clipman = user_data;
	
	kiba_icon_set_file (KIBA_ICON(clipman->icon), "gtk-paste");
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaClipman *clipman;
	ClipmanClip *clip;
	gint i;
	
	clipman = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_remove_callback_by_data (file, clipman);
	kiba_object_destroy (KIBA_OBJECT(clipman->icon));
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(gtk_icon_theme_get_default()),
		G_CALLBACK(icon_theme_changed_cb), clipman);
	
	/* Free the clipboards */
	gtk_clipboard_clear (primaryClip);
	gtk_clipboard_clear (defaultClip);
	
	g_source_remove (clipman->timeout_id);
	
	/* Remove clipboard items */
	for (i = clipman->clips->len; i--;) {
		clip = g_ptr_array_index (clipman->clips, i);
		g_ptr_array_remove_fast (clipman->clips, clip);
		clipman_free_clip (clip);
	}
	g_ptr_array_free (clipman->clips, TRUE);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaClipman *clipman;
	GtkObject *adj;
	gint right_clipman = 0, left_clipman = 0;
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");	
#endif
	
	clipman = g_new0 (KibaClipman, 1);
	
	/* add icon to the dock */
	clipman->icon = kiba_icon_new_from_file (PLUGIN_NAME, "gtk-paste");
	kiba_add_plugin_object(kiba, plugin, KIBA_OBJECT(clipman->icon));
	gtk_widget_show (clipman->icon);
	
	kiba_object_set_title (KIBA_OBJECT(clipman->icon), _("Clipboard Manager"));
	
	clipman->clips = g_ptr_array_new ();
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_add_callback (file, "clipman_ignore_select",
		clipman_flags_cb, clipman, TRUE);
	kiba_prefs_add_callback (file, "clipman_prevent_empty",
		clipman_flags_cb, clipman, TRUE);
	kiba_prefs_add_callback (file, "clipman_separate_menu",
		clipman_flags_cb, clipman, TRUE);
	kiba_prefs_add_callback (file, "clipman_item_numbers",
		clipman_flags_cb, clipman, TRUE);
	kiba_prefs_add_callback (file, "clipman_behaviour",
		clipman_behaviour_cb, clipman, TRUE);
	kiba_prefs_add_callback (file, "clipman_menu_characters",
		clipman_character_cb, clipman, TRUE);
	kiba_prefs_add_callback (file, "clipman_behaviour",
		clipman_history_cb, clipman, TRUE);
	
	clipman->timeout_id = g_timeout_add (TIMER_INTERVAL,
		(GSourceFunc) kiba_clipman_time_handler, (gpointer) clipman);
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) clipman);
	
	/* add object signal handler */
	g_signal_connect (G_OBJECT (clipman->icon), "button-release-event",
			G_CALLBACK (kiba_object_button_release_cb), clipman);
	
	g_signal_connect (G_OBJECT (gtk_icon_theme_get_default()), "changed",
			G_CALLBACK (icon_theme_changed_cb), clipman);
	
	kiba_prefs_add_callback (file, "clipman_size_factor",
		object_size_factor_cb, clipman, TRUE);
	kiba_prefs_add_callback (file, "clipman_alpha",
		object_alpha_cb, clipman, TRUE);
	
	/* Connect to the clipboards */
	defaultClip = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
	primaryClip = gtk_clipboard_get (GDK_SELECTION_PRIMARY);
	
	return TRUE;
}

static PrefMember clipman_options[] = {
	{ "clipman_ignore_select", "TRUE", KIBA_BOOLEAN },
	{ "clipman_prevent_empty", "TRUE", KIBA_BOOLEAN },
	{ "clipman_separate_menu", "FALSE", KIBA_BOOLEAN },
	{ "clipman_item_numbers", "TRUE", KIBA_BOOLEAN },
	{ "clipman_behaviour", "normal", KIBA_STRING },
	{ "clipman_menu_characters", "30", KIBA_INT },
	{ "clipman_history_items", "20", KIBA_INT },
	{ "clipman_alpha", "1.0", KIBA_DOUBLE },
	{ "clipman_size_factor", "1.0", KIBA_DOUBLE }
};

static KibaPluginVtable vtable = {
	"Clipman Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
};

KIBA_INIT_PLUGIN (vtable, clipman_options, NULL, NULL, NULL);
