/*
 *  Copyright (C) 2006-2008 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Authors : Daniel Bessler <danielb@kiba-dock.org>
 *            Sandipan <sandi.newton@gmail.com>
*/

#include "config.h"

#include <time.h>
#include <string.h>
#include <libintl.h>
#include <gtk/gtk.h>

#include <kiba-prefs.h>
#include <kiba-screen.h>
#include <kiba-debug.h>
#include <kiba-dialog.h>
#include <kiba-macros.h>
#include <kiba-info-win.h>
#include <kiba-icon-view-win.h>

#include "calendar.h"

#define _(String) dgettext (GETTEXT_PACKAGE, String)

#define REMINDER_DIR "/.kiba-dock/reminder"

#define KIBA_CALENDAR_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_CALENDAR, KibaCalendarPrivate))

G_DEFINE_TYPE (KibaCalendar, kiba_calendar, KIBA_TYPE_INFO_WIN)

typedef struct {
	GtkWidget *calendar;
	GtkWidget *reminderLabel;
	KibaIconViewWin *reminderPopup;
	GtkWidget *view;
	GtkListStore *store;
	
	GList *expired_reminders;
	gint reminder_id;
}KibaCalendarPrivate;

enum {
	COL_CHECKBOX = 0,
	COL_PERCENT,
	COL_DESC,
	NUM_COLS
};

static void __calendar_day_click_cb (GtkCalendar *realCalendar,
				   gpointer     user_data);

static void __destroy_widget_cb (GtkWidget *widget,
			       gpointer   data) {
	gtk_widget_destroy (GTK_WIDGET(data));
}

/*** Reminder ***/

static gint kiba_reminder_sort_func (gconstpointer a,
				     gconstpointer b) {
	const KibaCalendarReminder *areminder = a;
	const KibaCalendarReminder *breminder = b;
	
	if (areminder->time.year != breminder->time.year)
		return areminder->time.year - breminder->time.year;
	
	if (areminder->time.month != breminder->time.month)
		return areminder->time.month - breminder->time.month;
	
	if (areminder->time.hour != breminder->time.hour)
		return areminder->time.hour - breminder->time.hour;
	
	return areminder->time.minute - breminder->time.minute;
}

static gboolean _write_reminder_file (KibaCalendarReminder *reminder) {
	FILE *new_file;
	
	if (g_file_test(reminder->desktop_uri, G_FILE_TEST_EXISTS) &&
	   !g_file_test(reminder->desktop_uri, G_FILE_TEST_IS_REGULAR))
		return FALSE;
	
	new_file = fopen (reminder->desktop_uri, "w");
	if (!new_file)
		return FALSE;
	
	fprintf (new_file,
		"[Desktop Entry]\n"
		"Type=Reminder\n"
		"Version=1.0\n"
		"Name=%s\n"
		"Description=%s\n"
		"Year=%i\n"
		"Month=%i\n"
		"Day=%i\n"
		"Hour=%i\n"
		"Minute=%i\n"
		"Expire=%i\n"
		"Exec=evolution calendar:///?startdate=%i.%i.%i\n"
		"Icon=evolution\n"
		"LauncherGroup=Reminder\n",
		reminder->summary != NULL ? reminder->summary : _("No Name"),
		reminder->description != NULL ? reminder->description : _("No Description"),
		reminder->time.year, reminder->time.month+1,
		reminder->time.day, reminder->time.hour, reminder->time.minute, reminder->expire_time,
		reminder->time.year, reminder->time.month+1, reminder->time.day);
		
	fclose (new_file);
	
	return TRUE;
}

static gboolean _load_reminder_file (KibaCalendarReminder *reminder,
					       const char        *uri,
					       GError          **error) {
	static const char file_uri_prefix[] = "file://";
	static const char desktop_entry_group[] = "Desktop Entry";
	static const char desktop_file_suffix[] = ".desktop";
	static const struct { const char *string; KibaType type; size_t offset; } desktop_data[] = {
		{ "Name",        KIBA_STRING, G_STRUCT_OFFSET (KibaCalendarReminder, summary) },
		{ "Description", KIBA_STRING, G_STRUCT_OFFSET (KibaCalendarReminder, description) },
		{ "Expire",      KIBA_INT,    G_STRUCT_OFFSET (KibaCalendarReminder, expire_time) },
		{ "Year",        KIBA_INT,    G_STRUCT_OFFSET (KibaCalendarReminder, time.year) },
		{ "Month",       KIBA_INT,    G_STRUCT_OFFSET (KibaCalendarReminder, time.month) },
		{ "Day",         KIBA_INT,    G_STRUCT_OFFSET (KibaCalendarReminder, time.day) },
		{ "Hour",        KIBA_INT,    G_STRUCT_OFFSET (KibaCalendarReminder, time.hour) },
		{ "Minute",      KIBA_INT,    G_STRUCT_OFFSET (KibaCalendarReminder, time.minute) },
	};
	static const char *desktop_keys[G_N_ELEMENTS (desktop_data)];
	
	GKeyFile *key_file;
	const char *filename;
	gint i;
	
	for (i = 0; i < G_N_ELEMENTS (desktop_keys); i++) {
		desktop_keys[i] = desktop_data[i].string;
	}
	
	if (!g_file_test(uri, G_FILE_TEST_IS_REGULAR | G_FILE_TEST_IS_SYMLINK)) {
		DEBUG_ERROR ("'%s' is not a valid File\n", uri);
		return FALSE;
	}
	if (!g_str_has_suffix(uri, desktop_file_suffix)) {
		DEBUG_ERROR ("'%s' is not a .desktop File\n", uri);
		return FALSE;
	}
	if (g_str_has_prefix (uri, file_uri_prefix))
		filename = uri + strlen (file_uri_prefix);
	else
		filename = uri;
	
	key_file = g_key_file_new ();
	if (!g_key_file_load_from_file (key_file, filename, 0, error))
		return FALSE;
	
	for (i = 0; i < G_N_ELEMENTS (desktop_data); i++) {
		switch (desktop_data[i].type) {
			case KIBA_INT: {
				gint value;
				
				value = g_key_file_get_integer (key_file,
						desktop_entry_group,
						desktop_data[i].string,
						error);
				g_clear_error (error);
				*error = NULL;
				
				G_STRUCT_MEMBER (int , reminder, desktop_data[i].offset) = value;
				break;
			}
			case KIBA_STRING: {
				char *value;
				
				value = g_key_file_get_string (key_file,
						desktop_entry_group,
						desktop_data[i].string,
						error);
				g_clear_error (error);
				*error = NULL;
				
				G_STRUCT_MEMBER (char* , reminder, desktop_data[i].offset) = value;
				break;
			}
			default:
				break;
		}
	}
	
	reminder->time.month--;
	reminder->desktop_uri = g_strdup (filename);
	
	return TRUE;
}

static void __delete_reminder_cb (GtkButton         *button,
			        KibaCalendarReminder *reminder) {
	KibaCalendar *calendar = reminder->calendar;
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GList *l;
	guint year, month, day;
	
	if (!reminder->desktop_uri)
		return;
	
	if (reminder->dialog.win) {
		gtk_widget_destroy (reminder->dialog.win);
	}
	
	calendar->reminders = g_list_remove (calendar->reminders, reminder);
	if (g_list_find(priv->expired_reminders, reminder))
		priv->expired_reminders = g_list_remove (priv->expired_reminders, reminder);
	g_remove (reminder->desktop_uri);
	g_free (reminder->description);
	g_free (reminder->summary);
	g_free (reminder);
	
	if (calendar) {
		gtk_calendar_get_date (GTK_CALENDAR(priv->calendar), &year, &month, &day);
		if (reminder->time.month == month && reminder->time.year == year) {
			gboolean unmark;
			
			unmark = TRUE;
			for (l = calendar->reminders; l != NULL; l = l->next) {
				KibaCalendarReminder *check_reminder = l->data;
				if (check_reminder->time.month == month &&
				    check_reminder->time.year == year) {
					unmark = FALSE;
					break;
				}
			}
			if (unmark);
				gtk_calendar_unmark_day (GTK_CALENDAR(priv->calendar),
					reminder->time.day);
			
			if (reminder->time.day == day) {
				__calendar_day_click_cb (GTK_CALENDAR(priv->calendar), calendar);
			}
		}
	}
}

static void __save_reminder_cb (GtkButton *button,
			      gpointer   data) {
	KibaCalendarReminder *reminder = data;
	KibaCalendar *calendar = reminder->calendar;
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GtkTextBuffer *buffer;
	GtkTextIter start, end;
	guint hour, minute, expire_time;
	const char *homedir;
	char *uri;
	char *description = NULL;
	
	homedir = g_get_home_dir ();
	
	char reminder_dir[strlen(homedir)+strlen(REMINDER_DIR)];
	sprintf (reminder_dir, "%s" REMINDER_DIR, homedir);
	
	const char *summary = gtk_entry_get_text(GTK_ENTRY(reminder->dialog.summary));
	
	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW(reminder->dialog.textview));
	gtk_text_buffer_get_start_iter (buffer, &start);
	gtk_text_buffer_get_end_iter (buffer, &end);
	description = gtk_text_buffer_get_text (buffer, &start, &end, FALSE);
	
	hour = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(reminder->dialog.hour_spin));
	minute = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(reminder->dialog.minute_spin));
	expire_time = 60*gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(reminder->dialog.expire_hour_spin)) +
		gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(reminder->dialog.expire_minute_spin));
	
	if (!reminder->desktop_uri) {
		char filename[28];
		sprintf (filename, "reminder_%i_%i_%i",
				reminder->time.day,
				reminder->time.month,
				reminder->time.year);
		
		uri = g_strdup_printf ("%s/%s-0.desktop", reminder_dir, filename);
		gint i = 1;
		while (g_file_test(uri, G_FILE_TEST_EXISTS)) {
			g_free (uri);
			uri = g_strdup_printf ("%s/%s-%i.desktop", reminder_dir, filename, i);
			i++;
		}
		reminder->desktop_uri = uri;
	}
	
	if (!g_list_find(calendar->reminders, reminder)) {
		calendar->reminders = g_list_append (calendar->reminders, reminder);
	}
	if (reminder->summary)
		g_free (reminder->summary);
	if (reminder->description)
		g_free (reminder->description);
	reminder->summary     = g_strdup(summary);
	reminder->description = description;
	reminder->time.hour   = hour;
	reminder->time.minute = minute;
	reminder->expire_time = expire_time;
	
	calendar->reminders = g_list_sort (calendar->reminders, kiba_reminder_sort_func);
	_write_reminder_file (reminder);
	
	if (GTK_IS_CALENDAR(priv->calendar)) {
		guint year, month, day;
		
		gtk_calendar_get_date (GTK_CALENDAR(priv->calendar),
			&year, &month, &day);
		if (reminder->time.month == month && reminder->time.year == year) {
			gtk_calendar_mark_day (GTK_CALENDAR(priv->calendar), reminder->time.day);
			
			if (reminder->time.day == day) {
				__calendar_day_click_cb (GTK_CALENDAR(priv->calendar), calendar);
			}
		}
	}
}

static void __reminder_dialog_destroy_cb (GtkWidget *widget,
					gpointer   data) {
	KibaCalendarReminder *reminder = data;
	
	reminder->dialog.win = NULL;
	reminder->dialog.expire_hour_spin = NULL;
	reminder->dialog.expire_minute_spin = NULL;
	reminder->dialog.minute_spin = NULL;
	reminder->dialog.hour_spin = NULL;
	reminder->dialog.textview = NULL;
	reminder->dialog.summary = NULL;
}

static void __create_reminder_window (KibaCalendar         *calendar,
				    KibaCalendarReminder *reminder,
				    unsigned int       year,
				    unsigned int       month,
				    unsigned int       day) {
	GtkWidget *table;
	GtkWidget *label;
	GtkObject *adjustment;
	GtkWidget *hbox;
	GtkWidget *button;
	GtkWidget *scrolled_window;
	GdkPixbuf *pixbuf = NULL;
	gchar *title;
	gchar *month_string;
	gchar *summary = NULL,*description = NULL;
	gint hour, minute, expire_minutes, expire_hours;
	gint i = 0;
	
	if(reminder) {
		if (GTK_IS_WINDOW(reminder->dialog.win)) {
			gtk_window_present (GTK_WINDOW(reminder->dialog.win));
			return;
		}
		
		year = reminder->time.year;
		month = reminder->time.month;
		day = reminder->time.day;
		minute = reminder->time.minute;
		hour = reminder->time.hour;
		summary = reminder->summary;
		description = reminder->description;
		expire_hours = reminder->expire_time/60;
		expire_minutes = reminder->expire_time%60;
	}
	else {
		reminder = g_new0 (KibaCalendarReminder, 1);
		reminder->calendar = calendar;
		reminder->desktop_uri = NULL;
		reminder->description = NULL;
		reminder->summary = NULL;
		reminder->time.year  = year;
		reminder->time.month = month;
		reminder->time.day   = day;
		hour = 12;
		minute = 0;
		expire_hours = 0;
		expire_minutes = 30;
	}
	
	switch (month) {
		case 0:  month_string = _("January"); break;
		case 1:  month_string = _("February"); break;
		case 2:  month_string = _("March"); break;
		case 3:  month_string = _("April"); break;
		case 4:  month_string = _("May"); break;
		case 5:  month_string = _("June"); break;
		case 6:  month_string = _("July"); break;
		case 7:  month_string = _("August"); break;
		case 8:  month_string = _("September"); break;
		case 9:  month_string = _("October"); break;
		case 10: month_string = _("November"); break;
		default: month_string = _("December"); break;
	}
	month++;
	
	{
		reminder->dialog.win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
		gtk_window_set_role (GTK_WINDOW(reminder->dialog.win), "kiba-reminder");
	}
	/* { FIXME: no input focus :/
		reminder_window = kiba_info_win_new();
		gtk_window_set_type_hint (GTK_WINDOW(reminder_window), GDK_WINDOW_TYPE_HINT_DOCK);
	} */
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW(reminder->dialog.win), TRUE);
	gtk_window_set_position (GTK_WINDOW(reminder->dialog.win), GTK_WIN_POS_CENTER);
	gtk_widget_set_size_request (reminder->dialog.win, 450, 240);
	/* Translators: %s, %i %i is month, day year */
	title = g_strdup_printf(_("Reminder for %s, %i %i"), month_string, day, year);
	gtk_window_set_title (GTK_WINDOW(reminder->dialog.win), title);
	g_free(title);
	
	if (kiba_load_icon ("kiba-dock", 16, &pixbuf)) {
		gtk_window_set_icon (GTK_WINDOW(reminder->dialog.win), pixbuf);
		g_object_unref (G_OBJECT(pixbuf));
	}
	
	table = gtk_table_new (4, 2, FALSE);
	gtk_container_add (GTK_CONTAINER(reminder->dialog.win), table);
	
	{
		label = gtk_label_new (_("Summary"));
		gtk_misc_set_alignment (GTK_MISC (label), 0, 0);
		gtk_table_attach (GTK_TABLE (table), label,
				0, 1, i, i+1,
				GTK_FILL,
				GTK_FILL,
				5, 5);
		gtk_widget_show (label);
			
		reminder->dialog.summary = gtk_entry_new ();
		gtk_widget_show (reminder->dialog.summary);
		gtk_table_attach (GTK_TABLE (table), reminder->dialog.summary,
				1, 2, i, i+1,
				GTK_EXPAND | GTK_FILL,
				0,
				0, 0);
		i++;
		if(summary)
			gtk_entry_set_text(GTK_ENTRY(reminder->dialog.summary), summary);
	}
	
	label = gtk_label_new (_("Time"));
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0);
	gtk_table_attach (GTK_TABLE (table), label,
			0, 1, i, i+1,
			GTK_FILL,
			GTK_FILL,
			5, 5);
	gtk_widget_show (label);
	
	hbox = gtk_hbox_new (FALSE, 0);
	gtk_table_attach (GTK_TABLE (table), hbox,
			1, 2, i, i+1,
			GTK_FILL,
			GTK_FILL,
			0, 0);
	i++;
		
		adjustment = gtk_adjustment_new (0, 0, 23, 1, 10, 10);
		reminder->dialog.hour_spin = gtk_spin_button_new (GTK_ADJUSTMENT (adjustment), 1, 0);
		gtk_spin_button_set_value (GTK_SPIN_BUTTON(reminder->dialog.hour_spin), hour);
		gtk_box_pack_start (GTK_BOX (hbox), reminder->dialog.hour_spin, FALSE, TRUE, 0);
		gtk_widget_show (reminder->dialog.hour_spin);
	
		label = gtk_label_new (" : ");
		gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
		gtk_widget_show (label);
	
		adjustment = gtk_adjustment_new (0, 0, 59, 1, 10, 10);
		reminder->dialog.minute_spin = gtk_spin_button_new (GTK_ADJUSTMENT (adjustment), 1, 0);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(reminder->dialog.minute_spin), minute);
		gtk_box_pack_start (GTK_BOX (hbox), reminder->dialog.minute_spin, FALSE, TRUE, 0);
		gtk_widget_show (reminder->dialog.minute_spin);
	gtk_widget_show (hbox);

	label = gtk_label_new (_("Expire Time"));
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0);
	gtk_table_attach (GTK_TABLE (table), label,
			0, 1, i, i+1,
			GTK_FILL,
			GTK_FILL,
			5, 5);
	gtk_widget_show (label);
	hbox = gtk_hbox_new (FALSE, 0);
	gtk_table_attach (GTK_TABLE (table), hbox,
			1, 2, i, i+1,
			GTK_FILL,
			GTK_FILL,
			0, 0);
	i++;
		adjustment = gtk_adjustment_new (expire_hours, 0, 72, 1, 10, 10);
		reminder->dialog.expire_hour_spin =
			gtk_spin_button_new (GTK_ADJUSTMENT (adjustment), 1, 0);
		gtk_box_pack_start (GTK_BOX (hbox), reminder->dialog.expire_hour_spin, FALSE, TRUE, 0);
		gtk_widget_show (reminder->dialog.expire_hour_spin);
		
		label = gtk_label_new (_("Hours"));
		gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, TRUE, 0);
		gtk_widget_show (label);
		
		adjustment = gtk_adjustment_new (expire_minutes, 0, 60, 1, 10, 10);
		reminder->dialog.expire_minute_spin =
			gtk_spin_button_new (GTK_ADJUSTMENT (adjustment), 1, 0);
		gtk_box_pack_start (GTK_BOX (hbox), reminder->dialog.expire_minute_spin, FALSE, TRUE, 0);
		gtk_widget_show (reminder->dialog.expire_minute_spin);
		
		label = gtk_label_new (_("Minutes"));
		gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, TRUE, 0);
		gtk_widget_show (label);
	gtk_widget_show (hbox);
	
	label = gtk_label_new (_("Description"));
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0);
	gtk_table_attach (GTK_TABLE (table),
			label,
			0, 1, i, i+1,
			GTK_FILL,
			GTK_FILL,
			5, 5);
	gtk_widget_show (label);
	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
	gtk_table_attach (GTK_TABLE (table), scrolled_window,
			1, 2, i, i+1,
			GTK_FILL,
			GTK_EXPAND | GTK_FILL,
			0, 0);
	i++;
		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window), 
						GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
		gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolled_window), GTK_SHADOW_OUT);
	
		reminder->dialog.textview = gtk_text_view_new ();
		gtk_text_view_set_right_margin (GTK_TEXT_VIEW(reminder->dialog.textview), 5);
		gtk_text_view_set_left_margin (GTK_TEXT_VIEW(reminder->dialog.textview), 5);
		gtk_text_view_set_pixels_below_lines (GTK_TEXT_VIEW(reminder->dialog.textview), 5);
		gtk_text_view_set_pixels_above_lines (GTK_TEXT_VIEW(reminder->dialog.textview), 5);
		if(description) {
			GtkTextBuffer *tb = gtk_text_view_get_buffer (GTK_TEXT_VIEW(reminder->dialog.textview));
			gtk_text_buffer_set_text (tb,description, g_utf8_strlen(description, -1));
		}
		gtk_widget_show (reminder->dialog.textview);
		gtk_container_add (GTK_CONTAINER (scrolled_window), reminder->dialog.textview);
		gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (reminder->dialog.textview), GTK_WRAP_WORD);
	gtk_widget_show (scrolled_window);

	hbox = gtk_hbox_new (TRUE, 0);
	gtk_table_attach (GTK_TABLE (table), hbox,
			1, 2, i, i+1,
			GTK_FILL,
			GTK_FILL,
			0, 0);
	i++;
	
		button = gtk_button_new_from_stock ("gtk-save");
		g_signal_connect(G_OBJECT(button),"clicked",
			G_CALLBACK(__save_reminder_cb), reminder);
		gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
		gtk_widget_show (button);
		
		button = gtk_button_new_from_stock ("gtk-delete");
		g_signal_connect (G_OBJECT(button), "clicked",
			G_CALLBACK(__delete_reminder_cb), reminder);
		g_signal_connect_after (G_OBJECT(button), "clicked",
			G_CALLBACK(__destroy_widget_cb), reminder->dialog.win);
		gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
		gtk_widget_show (button);
		
		button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
		g_signal_connect (G_OBJECT(button), "clicked",
			G_CALLBACK(__destroy_widget_cb), reminder->dialog.win);
		gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
		gtk_widget_show (button);
	gtk_widget_show (hbox);
	
	g_signal_connect (G_OBJECT(reminder->dialog.win), "destroy",
		G_CALLBACK(__reminder_dialog_destroy_cb), reminder);
	
	gtk_widget_show (table);
	gtk_widget_show (reminder->dialog.win);
}

static KibaCalendarReminder* __get_selected_reminder (KibaCalendar    *calendar,
						 GtkTreeModel *model,
						 GtkTreeIter  *iter) {
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	KibaCalendarReminder *reminder = NULL;
	GList *l;
	gint count = 0;
	
	gtk_tree_model_get (model,
		iter,
		COL_CHECKBOX,
		&count, -1);
	
	for (l = calendar->reminders; l != NULL; l = l->next) {
		reminder = l->data;
		
		if (!reminder->active)
			continue;
		
		count--;
		if (count > 0)
			continue;
		
		return reminder;
	}
	
	return reminder;
}

static void __reminder_drag_data_get_cb (GtkWidget        *widget,
					    GdkDragContext   *drag_context,
					    GtkSelectionData *data,
					    guint             info,
					    guint             time,
					    gpointer          user_data) {
	KibaCalendar *calendar = user_data;
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	KibaCalendarReminder *selected;
	GtkTreeSelection *selection;
	GtkTreeIter iter;
	GtkTreeModel *model;
	
	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(priv->view));
	gtk_tree_selection_get_selected (selection, &model, &iter);
	
	selected = __get_selected_reminder (calendar,
			model,
			&iter);
	
	gtk_selection_data_set(data,
		GDK_SELECTION_TYPE_STRING,
		8,	/* 8 bits per character. */
		(guchar *)selected->desktop_uri,
		strlen(selected->desktop_uri));
}

static void __reminder_activate (KibaCalendar         *calendar,
					  KibaCalendarReminder *reminder,
					  gint               count) {
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GtkTreeIter iter;
	char *text;
	char summary[strlen(reminder->summary)];
	
	text = g_strdup_printf ("%d:%02d %s",
		reminder->time.hour > 12 ? (reminder->time.hour - 12) : reminder->time.hour,
		reminder->time.minute,
		reminder->time.hour >= 12 ? "PM" : "AM");
	
	sprintf (summary, "%s", reminder->summary);
	if (strlen (summary) > 16) {
		summary[13] = '.';
		summary[14] = '.';
		summary[15] = '.';
		summary[16] = '\0';
	}
	gtk_list_store_append (priv->store, &iter);
	gtk_list_store_set (priv->store, &iter,
		COL_CHECKBOX, count,
		COL_PERCENT, text,
		COL_DESC, summary, -1);
	
	reminder->active = TRUE;
}

static gint __activate_reminders (KibaCalendar *calendar,
					   unsigned int year,
					   unsigned int month,
					   unsigned int day) {
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GList *l;
	gint count;
	
	count = 1;
	for (l = calendar->reminders; l != NULL; l = l->next) {
		KibaCalendarReminder *reminder = l->data;
		
		if (year != reminder->time.year || month != reminder->time.month || day != reminder->time.day)
			continue;
		
		__reminder_activate (calendar, reminder, count);
		
		count++;
	}
	
	return count;
}

static void __deactivate_reminders (KibaCalendar *calendar) {
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GList *l;
	
	for (l = calendar->reminders; l != NULL; l = l->next) {
		KibaCalendarReminder *reminder = l->data;
		
		reminder->active = FALSE;
	}
	gtk_list_store_clear (priv->store);
}

static void __load_reminders (KibaCalendar *calendar) {
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GError *error = NULL;
	const char *homedir;
	char *file_name;
	GDir *dir;
	
	homedir = g_get_home_dir ();
	
	char reminder_dir[strlen(homedir)+strlen(REMINDER_DIR)];
	sprintf (reminder_dir, "%s" REMINDER_DIR, homedir);
	if (!g_file_test (reminder_dir, G_FILE_TEST_IS_DIR)) {
		if (g_mkdir (reminder_dir, 7664) != 0) {
			DEBUG_ERROR ("Failed to Create %s\n", reminder_dir);
			return;
		}
	}
	
	dir = g_dir_open (reminder_dir, 0, &error);
	if (error != NULL) {
		DEBUG_ERROR ("Failed to open '%s': %s\n",
			reminder_dir, error->message);
		g_error_free (error);
		return;
	}
	
	while ((file_name = (char*) g_dir_read_name (dir)) != NULL) {
		KibaCalendarReminder *reminder;
		GError *error = NULL;
		
		if (!g_str_has_suffix (file_name, ".desktop"))
			continue;
		
		char uri[strlen(reminder_dir) + strlen(file_name) + 1];
		sprintf (uri, "%s/%s", reminder_dir, file_name);
		
		reminder = g_new0 (KibaCalendarReminder, 1);
		reminder->calendar = calendar;
		if (!_load_reminder_file (reminder, uri, &error)) {
			g_free (reminder);
			continue;
		}
		
		calendar->reminders = g_list_append (calendar->reminders, reminder);
		calendar->reminders = g_list_sort (calendar->reminders,
					kiba_reminder_sort_func);
	}
	g_dir_close (dir);
}

static void __reminder_popup_destroy_cb (GtkWidget *widget,
				       KibaCalendar *calendar) {
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	priv->reminderPopup = NULL;
}

static void __remove_reminder_popup_entry_cb (GtkButton *button,
					    KibaCalendar *calendar) {
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GtkWidget *widget;
	
	widget = g_object_get_data (G_OBJECT(button), "icon-view-widget");
	
	kiba_icon_view_win_remove (priv->reminderPopup, widget);
	
	if (gtk_container_get_children(GTK_CONTAINER(priv->reminderPopup->icon_view)))
		gtk_widget_destroy (priv->reminderPopup->win);
}

static void __defer_reminder_popup_entry_cb (GtkButton         *button,
					   KibaCalendarReminder *reminder) {
	__create_reminder_window (reminder->calendar, reminder, 0, 0, 0);
}

static GtkWidget* __create_reminder_popup_entry (KibaCalendarReminder *reminder) {
	GtkWidget *ret;
	GtkWidget *label;
	GtkWidget *button;
	char *markup;
	char *date;
	
	ret = gtk_vbox_new (FALSE, 5);
	
	label = gtk_label_new(reminder->summary);
	gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (ret), label, FALSE, FALSE, 0);
	
	date = g_strdup_printf ("%s: %i.%i.%i %s%i:%s%i",
		_("Date"),
		reminder->time.day,
		reminder->time.month+1,
		reminder->time.year,
		reminder->time.hour   < 10 ? "0" : "", reminder->time.hour,
		reminder->time.minute < 10 ? "0" : "", reminder->time.minute);
	label = gtk_label_new(date);
	g_free (date);
	gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (ret), label, FALSE, FALSE, 0);
	
	button = gtk_button_new_from_stock (GTK_STOCK_DELETE);
	gtk_widget_show (button);
	g_object_set_data (G_OBJECT(button), "icon-view-widget", ret);
	g_signal_connect (G_OBJECT(button), "clicked",
		G_CALLBACK(__delete_reminder_cb), reminder);
	g_signal_connect (G_OBJECT(button), "clicked",
		G_CALLBACK(__remove_reminder_popup_entry_cb), reminder->calendar);
	gtk_box_pack_start (GTK_BOX (ret), button, FALSE, FALSE, 0);
	
	button = gtk_button_new_with_label (_("Defer"));
	gtk_widget_show (button);
	g_object_set_data (G_OBJECT(button), "icon-view-widget", ret);
	g_signal_connect (G_OBJECT(button), "clicked",
		G_CALLBACK(__defer_reminder_popup_entry_cb), reminder);
	g_signal_connect (G_OBJECT(button), "clicked",
		G_CALLBACK(__remove_reminder_popup_entry_cb), reminder->calendar);
	gtk_box_pack_start (GTK_BOX (ret), button, FALSE, FALSE, 0);
	
	return ret;
}

static KibaIconViewWin* __create_reminder_popup (KibaCalendar *calendar) {
	KibaIconViewWin *view_win;
	GtkWidget *label;
	GList *l;
	GList *children;
	char *markup;
	
	view_win = kiba_icon_view_win_new(KIBA_ICON_VIEW_BOX_POSITION_RIGHT, 1, 3, TRUE);
	gtk_window_set_type_hint (GTK_WINDOW(view_win->win), GDK_WINDOW_TYPE_HINT_DOCK);
	gtk_window_set_role (GTK_WINDOW(view_win->win), "kiba-calendar");
	kiba_icon_view_set_prelightable (KIBA_ICON_VIEW(view_win->icon_view), FALSE);
	
	label = gtk_label_new(NULL);
	markup = g_markup_printf_escaped (
			"<span weight=\"bold\">%s</span>",
			_("Expired Reminders"));
	gtk_label_set_markup (GTK_LABEL (label), markup);
	g_free (markup);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (view_win->top_box), label, FALSE, FALSE, 5);
	
	g_signal_connect (G_OBJECT(view_win->win), "destroy",
		G_CALLBACK(__reminder_popup_destroy_cb), calendar);
	
	return view_win;
}

static gint __reminder_timeout_cb (gpointer data) {
	KibaCalendar *calendar = data;
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GList *new_expired = NULL;
	GList *children;
	GList *l;
	guint year, month, day, hours, minutes;
	static time_t systime;
	struct tm* g_pTime;
	
	time (&systime);
	g_pTime = localtime (&systime);
	minutes = g_pTime->tm_min;
	hours   = g_pTime->tm_hour;
	day     = g_pTime->tm_mday;
	month   = g_pTime->tm_mon + 1;
	year    = g_pTime->tm_mon + 1999;
	
	if (calendar == NULL)
		return FALSE;
	
	for (l = calendar->reminders; l != NULL; l = l->next) {
		KibaCalendarReminder *reminder = l->data;
		gboolean expired = FALSE;
		gint expire_hours;
		gint expire_day;
		gint expire_minutes;
		
		if (g_list_find(priv->expired_reminders, reminder))
			continue;
		
		expire_hours   = reminder->expire_time/60;
		expire_minutes = reminder->expire_time%60;
		
		if (reminder->time.minute < expire_minutes) {
			expire_minutes -= 60;
			expire_hours++;
		}
		
		if (reminder->time.hour < expire_hours) {
			expire_hours -= 24;
			expire_day = 1;
		}
		else
			expire_day = 0;
		
		if (reminder->time.year+1 < year)
			expired = TRUE;
		else if (reminder->time.year+1 > year)
			continue;
		else if (reminder->time.month+1 < month)
			expired = TRUE;
		else if (reminder->time.month+1 > month)
			continue;
		else if (reminder->time.day - expire_day < day)
			expired = TRUE;
		else if (reminder->time.day - expire_day > day)
			continue;
		else if (reminder->time.hour - expire_hours < hours)
			expired = TRUE;
		else if (reminder->time.hour - expire_hours > hours)
			continue;
		else if (reminder->time.minute - expire_minutes < minutes)
			expired = TRUE;
		else if (reminder->time.minute - expire_minutes > minutes)
			continue;
		if (expired) {
			priv->expired_reminders = g_list_append (priv->expired_reminders, reminder);
			new_expired = g_list_append (new_expired, reminder);
		}
	}
	
	if (!new_expired)
		return TRUE;
	
	/* create/present the info win */
	if (!priv->reminderPopup) {
		priv->reminderPopup = __create_reminder_popup (calendar);
		gtk_widget_show (priv->reminderPopup->win);
		gdk_window_set_keep_above (priv->reminderPopup->win->window, TRUE);
	}
	else {
		gtk_window_present (GTK_WINDOW(priv->reminderPopup->win));
	}
	
	/* adding newly expired reminders to the info win */
	for (l = new_expired; l != NULL; l = new_expired) {
		KibaCalendarReminder *reminder = l->data;
		GtkWidget *box;
		
		box = __create_reminder_popup_entry (reminder);
		kiba_icon_view_win_add (priv->reminderPopup, box);
		
		new_expired = g_list_remove (new_expired, l->data);
	}
	g_list_free (new_expired);
	
	children = gtk_container_get_children(GTK_CONTAINER(priv->reminderPopup->icon_view));
	for (l = children; l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		gtk_widget_set_usize (child, priv->reminderPopup->icon_view->allocation.width, -1);
	}
	
	return TRUE;
}

static void __calendar_day_click_cb (GtkCalendar *realCalendar,
				   gpointer     user_data) {
	KibaCalendar *calendar = user_data;
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	guint day, month, year;
	
	gtk_calendar_get_date (GTK_CALENDAR(priv->calendar), &year, &month, &day);
	__deactivate_reminders (calendar);
	if (__activate_reminders (calendar, year, month, day) == 1) {
		gtk_widget_hide (priv->view);
		gtk_widget_hide (priv->reminderLabel);
	}
	else {
		gtk_widget_show (priv->view);
		gtk_widget_show (priv->reminderLabel);
	}
}

static void __reminder_row_activated_real (GtkTreeModel  *model,
					    GtkTreePath   *path,
					    GtkTreeIter   *iter,
					    gpointer       data) {
	KibaCalendar *calendar = data;
	KibaCalendarReminder *reminder;
	
	reminder = __get_selected_reminder(calendar, model, iter);
	if (!reminder)
		return;
	
	__create_reminder_window (calendar, reminder, 0, 0, 0);
}

static void __reminder_row_activated (GtkTreeView       *treeview,
				    GtkTreePath       *path,
				    GtkTreeViewColumn *col,
				    gpointer           user_data) {
	GtkTreeSelection  *selection;
	
	selection = gtk_tree_view_get_selection (treeview);
	gtk_tree_selection_selected_foreach (selection,
		__reminder_row_activated_real, user_data);
}

static void __calendar_day_activate_cb (gpointer user_data,
				      GtkCalendar *realCalendar) {
	KibaCalendar *calendar = user_data;
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	unsigned int month, year, day;
	
	gtk_calendar_get_date (GTK_CALENDAR(priv->calendar),
		&year, &month, &day);
	__create_reminder_window (calendar, NULL, year, month, day);
}

static void __calendar_month_changed_cb (GtkCalendar *realCalendar,
				       gpointer     user_data) {
	KibaCalendar *calendar = user_data;
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GList *l;
	unsigned int month;
	unsigned int year;
	
	gtk_calendar_get_date (GTK_CALENDAR(priv->calendar), &year, &month, NULL);
	
	gtk_calendar_clear_marks (GTK_CALENDAR(priv->calendar));
	for (l = calendar->reminders; l != NULL; l = l->next) {
		KibaCalendarReminder *reminder = l->data;
		
		if (reminder->time.month == month && reminder->time.year == year) {
			gtk_calendar_mark_day (GTK_CALENDAR(priv->calendar), reminder->time.day);
		}
	}
	
	__calendar_day_click_cb (GTK_CALENDAR(priv->calendar), user_data);
}

static GtkWidget* __create_view_and_model (KibaCalendar *calendar) {
	static const char *note_icons[] = {
		"notes", "stock_note"
	};
	
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GtkCellRenderer *renderer;
	GtkTreeModel *model;
	GtkListStore *store;
	GtkWidget *view;
	GdkPixbuf *pixbuf = NULL;
	gint i;
	
	view = gtk_tree_view_new ();
	
	renderer = gtk_cell_renderer_pixbuf_new();
	
	gtk_tree_selection_set_mode (
		gtk_tree_view_get_selection (GTK_TREE_VIEW(view)),
		GTK_SELECTION_SINGLE);
	
	for (i = 0; i < G_N_ELEMENTS(note_icons); i++) {
		if (kiba_load_icon (note_icons[i], 16, &pixbuf))
			break;
	}
	
	if (!pixbuf) {
		pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 16, 16);
		gdk_pixbuf_fill (pixbuf, 0x00123008 << 8);
	}
	g_object_set (renderer, "visible", pixbuf != NULL, "pixbuf", pixbuf, NULL);
	
	gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
		-1, NULL, renderer, NULL);
	
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (view), FALSE);
	
	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
		-1, NULL, renderer, "text", COL_PERCENT, NULL);
	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
		-1, NULL, renderer, "text", COL_DESC, NULL);
	g_object_set (renderer, "weight", PANGO_WEIGHT_BOLD,
				"weight-set", TRUE, NULL);
	
	priv->store = gtk_list_store_new (NUM_COLS,
			G_TYPE_INT,
			G_TYPE_STRING,
			G_TYPE_STRING);
	gtk_tree_view_set_model (GTK_TREE_VIEW (view),
			GTK_TREE_MODEL(priv->store));
	
	return view;
}

static void __calendar_destroy_cb (GtkWidget *widget, gpointer data) {
	KibaCalendar *calendar = KIBA_CALENDAR(widget);
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GList *l;
	
	g_source_remove(priv->reminder_id);
	
	for (l = calendar->reminders; l != NULL; l = calendar->reminders) {
		KibaCalendarReminder *reminder = l->data;
		__delete_reminder_cb(NULL, reminder);
	}
	g_list_free(calendar->reminders);
}

/** Public methods **/

void kiba_calendar_show_reminder_for_date (KibaCalendar *calendar,
                                           KibaCalendarReminder *reminder,
                                           unsigned int year,
                                           unsigned int month,
                                           unsigned int day) {
	__create_reminder_window (calendar, reminder, year, month, day);
}

/** Class Init/Fini **/

static void kiba_calendar_class_init(KibaCalendarClass *class) {
	g_type_class_add_private (class, sizeof (KibaCalendarPrivate));
}

static void kiba_calendar_init(KibaCalendar *calendar) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	KibaCalendarPrivate *priv = KIBA_CALENDAR_GET_PRIVATE(calendar);
	GtkWidget *vbox;
	GtkWidget *label;
	GtkCalendarDisplayOptions options;
	time_t sys_time;
	struct tm *tm;
	char *markup;
	
	gtk_window_set_role (GTK_WINDOW(calendar), "kiba-calendar");
	gtk_window_set_screen (GTK_WINDOW(calendar), KIBA_GDK_SCREEN);
// 	gtk_window_set_type_hint (GTK_WINDOW(calendar), GDK_WINDOW_TYPE_HINT_DOCK);
	gtk_window_set_type_hint (GTK_WINDOW(calendar), GDK_WINDOW_TYPE_HINT_DIALOG);
	gtk_window_set_decorated (GTK_WINDOW(calendar), FALSE);
	gtk_window_set_resizable (GTK_WINDOW(calendar), FALSE);
	gtk_window_stick (GTK_WINDOW(calendar));
	gtk_window_set_title (GTK_WINDOW(calendar), _("Calendar"));
	
	vbox = gtk_vbox_new (FALSE, 6);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 6);
	gtk_container_add (GTK_CONTAINER (calendar), vbox);
	gtk_widget_show (vbox);
	
	priv->reminderLabel = gtk_label_new(_("Reminders"));
	gtk_misc_set_alignment (GTK_MISC(priv->reminderLabel), 0, 1);
	priv->view = __create_view_and_model(calendar);
	markup = g_markup_printf_escaped ("<span weight=\"bold\">%s</span>", _("Reminders"));
	gtk_label_set_markup (GTK_LABEL(priv->reminderLabel), markup);
	g_free(markup);
	
	gtk_box_pack_start (GTK_BOX (vbox), priv->reminderLabel, TRUE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), priv->view, TRUE, FALSE, 0);
	g_signal_connect_after (G_OBJECT(priv->view), "row-activated",
		G_CALLBACK(__reminder_row_activated), calendar);
	
	gtk_drag_source_set (priv->view,
		GDK_BUTTON1_MASK | GDK_BUTTON2_MASK,
		targets, G_N_ELEMENTS (targets),
		GDK_ACTION_LINK);
	g_signal_connect (GTK_OBJECT(priv->view), "drag-data-get",
		G_CALLBACK(__reminder_drag_data_get_cb), calendar);
	
	priv->calendar = gtk_calendar_new ();
	g_object_add_weak_pointer (G_OBJECT (priv->calendar),
				(gpointer *) &priv->calendar);
	
	options = gtk_calendar_get_display_options (GTK_CALENDAR (priv->calendar));
	options |= GTK_CALENDAR_SHOW_WEEK_NUMBERS;
	gtk_calendar_set_display_options (GTK_CALENDAR (priv->calendar), options);
	
	time (&sys_time);
	tm = localtime (&sys_time);
	gtk_calendar_select_month (GTK_CALENDAR (priv->calendar), tm->tm_mon, tm->tm_year + 1900);
	gtk_calendar_select_day (GTK_CALENDAR (priv->calendar), tm->tm_mday);
	gtk_box_pack_start (GTK_BOX (vbox), priv->calendar, TRUE, FALSE, 0);
	gtk_widget_show (priv->calendar);
	
	__load_reminders(calendar);
	
	/* check for expired reminders */
	priv->reminder_id = g_timeout_add (MINUTE_INTERVAL, (GSourceFunc)(__reminder_timeout_cb), calendar);
	__reminder_timeout_cb (calendar);

	g_signal_connect_swapped (G_OBJECT(priv->calendar), "day-selected-double-click",
		G_CALLBACK (__calendar_day_activate_cb), calendar);
	g_signal_connect_after (G_OBJECT(priv->calendar), "month-changed",
		G_CALLBACK (__calendar_month_changed_cb), calendar);
	g_signal_connect_after (G_OBJECT(priv->calendar), "day-selected",
		G_CALLBACK (__calendar_day_click_cb), calendar);
	__calendar_month_changed_cb (GTK_CALENDAR(priv->calendar), calendar);
	
	g_signal_connect_after (G_OBJECT(calendar), "destroy",
		G_CALLBACK (__calendar_destroy_cb), NULL);
}

KibaCalendar* kiba_calendar_new () {
	return g_object_new (KIBA_TYPE_CALENDAR, NULL);
}

