/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Authors : Daniel Bessler <danielb@kiba-dock.org>
 *            Sandipan <sandi.newton@gmail.com>
*/

#include "config.h"

#include <math.h>
#include <time.h>
#include <locale.h>
#include <libintl.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-icon.h>
#include <kiba-plugin.h>
#include <kiba-plugin-loader.h>
#include <kiba-prefs.h>
#include <kiba-screen.h>
#include <kiba-debug.h>
#include <kiba-dialog.h>
#include <kiba-macros.h>
#include <kiba-info-win.h>
#include <kiba-svg-macro.h>
#ifdef USE_SVG
  #include <librsvg/rsvg.h>
#endif

#include "calendar.h"

#define _(String)             dgettext (GETTEXT_PACKAGE, String)
#define PLUGIN_NAME           "clock"
#define CLOCK_DIR             "/.kiba-dock/"PLUGIN_NAME
#define CAIRO_CLOCK_THEME_DIR "/usr/share/cairo-clock/themes"

typedef struct _KibaClock KibaClock;

static const char* kiba_clock_type_ext[] = {
	".png",
	".svg"
};

typedef enum {
	KIBA_CLOCK_PIXBUF_THEME = 0,
	KIBA_CLOCK_SVG_THEME
}KibaClockThemeType;

struct _KibaClock {
	GtkWidget *object;
	
	KibaClockThemeType theme_type;
	CairoSurfaceData   foreground_surface;
	CairoSurfaceData   hour_surface;
	CairoSurfaceData   minute_surface;
	CairoSurfaceData   second_surface;
	PangoLayout       *pango_layout; /*used for digital clock */
	
	gint               timeout_id;
	
	char               *theme;
	char               *font;
	KibaRGBA            text_rgba;
	KibaRGBA            text_shadow_rgba;
	float              size_factor;
	gboolean            show_seconds;
	gboolean            digital;
	gboolean            twentyfour_mode;
	
	KibaPrefsFile *file;
	KibaCalendar *calendar;
};

static const char *clock_config_tools [] = {
	"system-config-date",
	"redhat-config-date",
        "/sbin/yast2 timezone",
	"time-admin",
};

typedef enum {
	KIBA_CLOCK_BACKGROUND = 0,
	KIBA_CLOCK_HOUR_SURFACE,
	KIBA_CLOCK_MINUTE_SURFACE,
	KIBA_CLOCK_SECOND_SURFACE,
	KIBA_CLOCK_FOREGROUND
}KibaClockIconLevel;

static const struct { const char *name; KibaClockIconLevel level; } clock_theme_data[] = {
	{ "clock-hour-hand", KIBA_CLOCK_HOUR_SURFACE },
	{ "clock-minute-hand", KIBA_CLOCK_MINUTE_SURFACE },
	{ "clock-second-hand", KIBA_CLOCK_SECOND_SURFACE },
	{ "clock-drop-shadow", KIBA_CLOCK_BACKGROUND },
	{ "clock-face-shadow", KIBA_CLOCK_BACKGROUND },
	{ "clock-face", KIBA_CLOCK_BACKGROUND },
	{ "clock-marks", KIBA_CLOCK_BACKGROUND },
	{ "clock-frame", KIBA_CLOCK_BACKGROUND },
	{ "clock-glass", KIBA_CLOCK_FOREGROUND }
};

enum {
	COL_CHECKBOX = 0,
	COL_PERCENT,
	COL_DESC,
	NUM_COLS
};

static void kiba_clock_set_surface_size (KibaClock *clock);
static void kiba_clock_update_pango_layout (KibaClock *clock,
					    gint       hours,
					    gint       minutes,
					    gint       seconds,
					    const char *pm);
static void kiba_clock_update_surface (KibaClock *clock);

/**
Helper Functions
**/

static char* kiba_clock_get_theme_path (KibaClock *clock) {
	char *dir;
	
	if (!clock->theme || strcmp (clock->theme, "none") == 0)
		clock->theme = "default";
	
	if (g_strrstr (clock->theme, "/") && g_file_test (clock->theme, G_FILE_TEST_IS_DIR))
		return g_strdup (clock->theme);
	
	dir = g_strdup_printf ("%s" CLOCK_DIR "/%s",
			g_get_home_dir (), clock->theme);
	
	if (g_file_test (dir, G_FILE_TEST_IS_DIR))
		return dir;
	g_free (dir);
	dir = NULL;
	
	dir = g_strdup_printf (CAIRO_CLOCK_THEME_DIR "/%s", clock->theme);
	
	if (g_file_test (dir, G_FILE_TEST_IS_DIR))
		return dir;
	g_free (dir);
	
	if (g_file_test (KIBA_CLOCK_THEME_DIR, G_FILE_TEST_IS_DIR))
		return g_strdup(KIBA_CLOCK_THEME_DIR);
	
	if (g_file_test (KIBA_PLUGINS_TOP_SRCDIR "/files/clock_theme", G_FILE_TEST_IS_DIR))
		return g_strdup(KIBA_PLUGINS_TOP_SRCDIR "/files/clock_theme");
	
	return NULL;
}

static KibaClockThemeType kiba_clock_get_theme_type (KibaClock *clock) {
	GDir *dir;
	GError *error = NULL;
	gchar *path;
	KibaClockThemeType type;
	
	path = kiba_clock_get_theme_path (clock);
	dir = g_dir_open (path, 0, &error);
	if (error != NULL) {
		g_free (path);
		g_error_free (error);
		return KIBA_CLOCK_PIXBUF_THEME;
	}
	
	const char *file = g_dir_read_name (dir);
	
	if (g_str_has_suffix (file, kiba_clock_type_ext[KIBA_CLOCK_SVG_THEME]))
		return KIBA_CLOCK_SVG_THEME;
	
	g_dir_close (dir);
	g_free (path);
	
	return KIBA_CLOCK_PIXBUF_THEME;
}

static void kiba_clock_place_window (KibaClock *clock,
				     GtkWidget *widget) {
	GtkWidget *dock, *win;
	gint w, h, x, y;
	const KibaLayoutData *d;
	
	dock = GTK_WIDGET(clock->object)->parent;
	win = dock->parent;
	
	if (!KIBA_IS_WIN(win))
		return;
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	struct {
		gint  *x, *y;
		gint  *w, *h;
		gint  alignment;
	} magic[2] = {
		{ &x, &y , &w, &h, d->valignment},
		{ &y, &x , &h, &w, d->halignment}
	};
	w = widget->allocation.width;
	h = widget->allocation.height;
	x = clock->object->allocation.x + GTK_WIDGET(win)->allocation.x;
	y = clock->object->allocation.y + GTK_WIDGET(win)->allocation.y;
	
	*magic[d->orientation].x -= (*magic[d->orientation].w - clock->object->allocation.width)/2;
	*magic[d->orientation].y -= (magic[d->orientation].alignment == 0 ?
			- clock->object->allocation.height : *magic[d->orientation].h);
	
	if (x + w > KIBA_SCREEN_DEFAULT->geometry.x + KIBA_SCREEN_DEFAULT->geometry.width)
		x = KIBA_SCREEN_DEFAULT->geometry.x + KIBA_SCREEN_DEFAULT->geometry.width - w;
	else if (x < KIBA_SCREEN_DEFAULT->geometry.x)
		x = KIBA_SCREEN_DEFAULT->geometry.x;
	if (y + h > KIBA_SCREEN_DEFAULT->geometry.y + KIBA_SCREEN_DEFAULT->geometry.height)
		y = KIBA_SCREEN_DEFAULT->geometry.y + KIBA_SCREEN_DEFAULT->geometry.height - h;
	else if (y < KIBA_SCREEN_DEFAULT->geometry.y)
		y = KIBA_SCREEN_DEFAULT->geometry.y;
	
	gtk_window_move (GTK_WINDOW (widget), x, y);
}

static gboolean kiba_clock_time_handler (KibaClock *clock) {
	char *new_string;
	int seconds, minutes, hours, day, month;
	char *month_string, *locale;
	static time_t systime;
	struct tm* g_pTime;
	const char *pm;
	
	if (!clock->object)
		return FALSE;
	
	time (&systime);
	g_pTime = localtime (&systime);
	minutes = g_pTime->tm_min;
	hours   = g_pTime->tm_hour;
	seconds = g_pTime->tm_sec;
	day     = g_pTime->tm_mday;
	month   = g_pTime->tm_mon + 1;
	
	if (clock->twentyfour_mode)
		pm = "";
	else {
		if (hours >= 12) {
			pm = " PM";
			if (hours > 12) {
				hours = hours - 12;
			}
		}
		else {
			pm = " AM";
		}
	}
	
	switch (month) {
		case 1:  month_string = _("Jan"); break;
		case 2:  month_string = _("Feb"); break;
		case 3:  month_string = _("Mar"); break;
		case 4:  month_string = _("Apr"); break;
		case 5:  month_string = _("May"); break;
		case 6:  month_string = _("Jun"); break;
		case 7:  month_string = _("Jul"); break;
		case 8:  month_string = _("Aug"); break;
		case 9:  month_string = _("Sep"); break;
		case 10: month_string = _("Oct"); break;
		case 11: month_string = _("Nov"); break;
		default: month_string = _("Dec"); break;
	}
	
	if (clock->show_seconds)
		new_string = g_strdup_printf ("%i. %s %s%i:%s%i:%s%i%s", day, month_string,
			(hours   < 10) ? "0" : "", hours,
			(minutes < 10) ? "0" : "", minutes,
			(seconds   < 10) ? "0" : "", seconds, pm);
	else
		new_string = g_strdup_printf ("%i. %s %s%i:%s%i%s", day, month_string,
			(hours   < 10) ? "0" : "", hours,
			(minutes < 10) ? "0" : "", minutes, pm);
	
	kiba_object_set_title (KIBA_OBJECT(clock->object), new_string);
	g_free (new_string);
	
	if (clock->digital) {
		kiba_clock_update_pango_layout (clock, hours, minutes, seconds, pm);
		kiba_clock_update_surface (clock);
	}
	
	kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(clock->object));
	
	return TRUE;
}

/**
Time configuration
**/

static gboolean try_config_tool (GdkScreen  *screen,
				 const char *tool) {
	GError    *err;
	char     **argv;
	char      *path;
	
	if (!tool || tool[0] == '\0')
		return FALSE;
	
	if (!g_shell_parse_argv (tool, NULL, &argv, NULL))
		return FALSE;
	
	if (!(path = g_find_program_in_path (argv [0]))) {
		g_strfreev (argv);
		return FALSE;
	}
	
	g_free (path);
	
	err = NULL;
	if (gdk_spawn_on_screen (screen,
				NULL,
				argv,
				NULL,
				G_SPAWN_SEARCH_PATH,
				NULL,
				NULL,
				NULL,
				&err)) {
		g_strfreev (argv);
		return TRUE;
	}
	
	g_strfreev (argv);
	
	kiba_dialog_error (_("Failed to launch Programm"),
			_("There was an error executing '%s': %s"),
			tool, err->message);
	g_error_free (err);	
		
	return TRUE;
}

static void config_date (GtkWidget *button,
			 gpointer   data) {
	KibaClock *clock = data;
	int        i;
	
	for (i = 0; i < G_N_ELEMENTS (clock_config_tools); i++) {
		if (try_config_tool (KIBA_GDK_SCREEN, clock_config_tools [i]))
			return;
	}
		
	kiba_dialog_error (_("Failed to launch Programm"),
		"%s. %s", _("Failed to locate a program for configuring the date and time"),
		_("Perhaps none is installed?"));
}

/**
Rendering
**/

static gint current_level = 0;

static void kiba_clock_update_pango_layout (KibaClock *clock,
					    gint       hours,
					    gint       minutes,
					    gint       seconds,
					    const char *pm) {
	GtkWidget *dock, *win;
	cairo_t *cr;
	char *surface_text;
	
	if (clock->show_seconds)
		surface_text = g_strdup_printf ("%s%i:%s%i:%s%i%s",
			hours   < 10 ? "0" : "", hours,
			minutes < 10 ? "0" : "", minutes,
			seconds < 10 ? "0" : "", seconds, pm);
	else
		surface_text = g_strdup_printf ("%s%i:%s%i%s",
			hours   < 10 ? "0" : "", hours,
			minutes < 10 ? "0" : "", minutes, pm);
	
	dock = GTK_WIDGET(clock->object)->parent;
	win = dock->parent;
	cr = kiba_win_cairo_create (KIBA_WIN(win));
	cairo_update_pango_layout (surface_text, clock->font,
		cr, &clock->pango_layout);
	cairo_destroy (cr);
	g_free (surface_text);
}

static void kiba_clock_draw_digital_surface (cairo_t          *cr,
					     cairo_surface_t  *surface,
					     gint              width,
					     gint              height,
					     void             *data) {
	KibaClock *clock = data;
	gint i;
	
	g_return_if_fail (PANGO_IS_LAYOUT(clock->pango_layout));
	
	cairo_translate (cr, 2, 2);
	
	cairo_set_source_rgba (cr,
		clock->text_shadow_rgba.r,
		clock->text_shadow_rgba.g,
		clock->text_shadow_rgba.b,
		clock->text_shadow_rgba.a);
	for (i = 0; i < 4; i++) {
		cairo_save (cr);
			cairo_translate (cr, (1 - (i&2)), (1 - 2*(i&1)));
			pango_cairo_show_layout (cr, clock->pango_layout);
		cairo_restore (cr);
	}
	
	cairo_set_source_rgba (cr,
		clock->text_rgba.r,
		clock->text_rgba.g,
		clock->text_rgba.b,
		clock->text_rgba.a);
	pango_cairo_show_layout (cr, clock->pango_layout);
}

static void kiba_clock_draw_surface (cairo_t          *cr,
				     cairo_surface_t  *surface,
				     gint              width,
				     gint              height,
				     void             *data) {
	KibaClock *clock = data;
	char *theme_path;
	gint i;
	
	theme_path = kiba_clock_get_theme_path(clock);
	if (!theme_path) {
		DEBUG_ERROR ("KibaClock: failed to load theme\n");
		return;
	}
	
	for (i = 0; i < G_N_ELEMENTS(clock_theme_data); i++) {
		GdkPixbuf  *pixbuf = NULL;
#ifdef USE_SVG
		RsvgHandle *handle = NULL;
#endif
		char *icon_file;
		
		if (clock_theme_data[i].level != current_level)
			continue;
		
		icon_file = g_strdup_printf ("%s/%s%s",
				theme_path,
				clock_theme_data[i].name,
				kiba_clock_type_ext[clock->theme_type]);
		
		if (!g_file_test(icon_file, G_FILE_TEST_EXISTS)) {
			g_free (icon_file);
			continue;
		}
		
		cairo_save (cr);
		
		if (clock_theme_data[i].level != KIBA_CLOCK_FOREGROUND &&
		    clock_theme_data[i].level != KIBA_CLOCK_BACKGROUND) {
			cairo_translate (cr,
				clock->object->allocation.width/2,
				clock->object->allocation.height/2);
			cairo_rotate (cr, -M_PI/2.0f);
		}
		
#ifdef USE_SVG
		if (clock->theme_type == KIBA_CLOCK_SVG_THEME &&
		    kiba_load_svg (icon_file, &clock->object->allocation.width, &handle)) {
			rsvg_handle_render_cairo (handle, cr);
			rsvg_handle_free (handle);
		}
#endif
		if (clock->theme_type == KIBA_CLOCK_PIXBUF_THEME &&
		    kiba_load_icon (icon_file, clock->object->allocation.width, &pixbuf)) {
			gdk_cairo_set_source_pixbuf (cr, pixbuf, 0.0, 0.0);
			cairo_paint (cr);
			g_object_unref (pixbuf);
		}
		
		cairo_restore (cr);
		
		g_free (icon_file);
	}
	
	g_free (theme_path);
}

static void kiba_clock_update_surface (KibaClock *clock) {
	KibaObject *object = KIBA_OBJECT(clock->object);
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(clock->object);
	GtkWidget *dock, *win;
	cairo_t *cr;
	gint i;
	
	CairoSurfaceData *surface_data[KIBA_CLOCK_FOREGROUND] = {
		&clock->hour_surface,
		&clock->minute_surface,
		&clock->second_surface,
		&clock->foreground_surface
	};
	
	dock = GTK_WIDGET(clock->object)->parent;
	win = dock->parent;
	cr = kiba_win_cairo_create (KIBA_WIN(win));
	
	if (clock->digital) {
		cairo_save (cr);
			current_level = KIBA_CLOCK_BACKGROUND;
			cairo_update_surface (cr, drawable_object->surface_data,
					kiba_clock_draw_digital_surface, clock);
		cairo_restore (cr);
	}
	else {
		cairo_save (cr);
			current_level = KIBA_CLOCK_BACKGROUND;
			cairo_update_surface (cr, drawable_object->surface_data,
					kiba_clock_draw_surface, clock);
		cairo_restore (cr);
		for (i = 0; i < KIBA_CLOCK_FOREGROUND; i++) {
			current_level = i+1;
			cairo_save (cr);
				cairo_update_surface (cr, surface_data[i],
						kiba_clock_draw_surface, clock);
			cairo_restore (cr);
		}
	}
	
	cairo_destroy (cr);
	
	kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(clock->object));
}

static void kiba_clock_set_surface_size (KibaClock *clock) {
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(clock->object);
	gint i;
	
	drawable_object->surface_data->width  = clock->object->allocation.width;
	drawable_object->surface_data->height = clock->object->allocation.width;
	
	CairoSurfaceData *surface_data[KIBA_CLOCK_FOREGROUND] = {
		&clock->hour_surface,
		&clock->minute_surface,
		&clock->second_surface,
		&clock->foreground_surface
	};
	
	for (i = 0; i < KIBA_CLOCK_FOREGROUND; i++) {
		surface_data[i]->width  = clock->object->allocation.width;
		surface_data[i]->height = clock->object->allocation.width;
	}
}

static void kiba_clock_paint (KibaDrawableObject *drawable_object,
			      cairo_t    *cr,
			      gpointer    data) {
	KibaObject *object = KIBA_OBJECT(drawable_object);
	KibaClock *clock = data;
	time_t g_timeOfDay;
	struct tm*  g_pTime;
	gint seconds, minutes, hours;
	gint unscaled_width;
	gint unscaled_height;
	
	g_return_if_fail (KIBA_IS_OBJECT (object));
	g_return_if_fail (drawable_object->surface_data->surface != NULL);
	
	if (!drawable_object->scroll_alpha)
		return;
	
	unscaled_width  = (int) (object->size.width * object->scale_x);
	unscaled_height = (int) (object->size.height * object->scale_y);
	
	cairo_set_source_surface (cr, drawable_object->surface_data->surface, 0, 0);
	kiba_drawable_object_paint_with_prelight (drawable_object, drawable_object->surface_data->surface, cr);
	
	if (!clock->digital) {
		time (&g_timeOfDay);
		g_pTime = localtime (&g_timeOfDay);
		seconds = g_pTime->tm_sec;
		minutes = g_pTime->tm_min;
		hours   = g_pTime->tm_hour;
		
		if (!clock->twentyfour_mode)
			hours = hours >= 12 ? hours - 12 : hours;
		
		if (clock->show_seconds) {
			cairo_save (cr);
				cairo_rotate_centered (cr, seconds/30.0f,
					unscaled_width/2,
					unscaled_height/2);
				cairo_set_source_surface (cr,
					clock->second_surface.surface, 0.0, 0.0);
				cairo_paint_with_alpha (cr, drawable_object->scroll_alpha);
			cairo_restore (cr);
		}
		
		cairo_save (cr);
			cairo_rotate_centered (cr, minutes/30.0f,
				unscaled_width/2,
				unscaled_height/2);
			cairo_set_source_surface (cr,
				clock->minute_surface.surface, 0.0, 0.0);
			cairo_paint_with_alpha (cr, drawable_object->scroll_alpha);
		cairo_restore (cr);
		
		cairo_save (cr);
			cairo_rotate_centered (cr,
				hours/(clock->twentyfour_mode ? 12.0f : 6.0f) +
				minutes/(clock->twentyfour_mode ? 720.0f : 360.0f),
				unscaled_width/2,
				unscaled_height/2);
			cairo_set_source_surface (cr,
				clock->hour_surface.surface, 0.0, 0.0);
			cairo_paint_with_alpha (cr, drawable_object->scroll_alpha);
		cairo_restore (cr);
		
		cairo_set_source_surface (cr, clock->foreground_surface.surface, 0, 0);
		cairo_paint_with_alpha (cr, drawable_object->scroll_alpha);
	}
}

/**
KibaPrefs Callbacks
**/

static void kiba_clock_update_size_from_layout (KibaClock *clock) {
	GtkWidget *dock, *win;
	PangoLayout *layout = NULL;
	PangoRectangle log;
	const char *text;
	cairo_t *cr;
	
	if (!clock->object)
		return;
	
	dock = GTK_WIDGET(clock->object)->parent;
	win = dock->parent;
	
	if (clock->twentyfour_mode) {
		if (clock->show_seconds)
			text = "00:00:00";
		else
			text = "00:00";
	}
	else {
		if (clock->show_seconds)
			text = "00:00:00 AM";
		else
			text = "00:00 AM";
	}
	
	cr = kiba_win_cairo_create (KIBA_WIN(win));
	cairo_update_pango_layout (text, clock->font,
		cr, &layout);
	cairo_destroy (cr);
	pango_layout_get_pixel_extents (layout, NULL, &log);
	g_object_unref (layout);
	
	kiba_object_set_requisition_factor (KIBA_OBJECT(clock->object),
		(float) (log.width+4)/KIBA_OBJECT(clock->object)->size.width,
		(float) (log.height+4)/KIBA_OBJECT(clock->object)->size.height);
	gtk_widget_queue_resize (clock->object->parent);
	
	kiba_clock_set_surface_size (clock);
}

static void dock_object_size_cb (const char *option_id,
			         gpointer    value,
			         gpointer    data) {
	KibaClock *clock = data;
	
	if (clock->digital) {
		kiba_clock_update_size_from_layout (clock);
	}
	else {
		kiba_clock_set_surface_size (clock);
		kiba_clock_update_surface (clock);
	}
}

static void object_size_factor_cb (const char *option_id,
			           gpointer    value,
			           gpointer    data) {
	KibaClock *clock = data;
	
	clock->size_factor = g_ascii_strtod((char*) value, NULL);
	
	if (clock->digital)
		return;
	
	kiba_object_set_requisition_factor (KIBA_OBJECT(clock->object),
		clock->size_factor, clock->size_factor);
	gtk_widget_queue_resize (clock->object);
	kiba_clock_set_surface_size (clock);
	kiba_clock_update_surface (clock);
	gtk_widget_queue_resize (clock->object->parent);
}

static void metric_mode_cb (const char *option_id,
			    gpointer    value,
			    gpointer    data) {
	KibaClock *clock = data;
	
	clock->twentyfour_mode = (gboolean) GPOINTER_TO_INT(value);
	
	if (!KIBA_IS_OBJECT(clock->object))
		return;
	
	if (clock->digital) {
		kiba_clock_update_size_from_layout (clock);
		kiba_clock_update_surface (clock);
	}
}

static void show_seconds_cb (const char *option_id,
			     gpointer    value,
			     gpointer    data) {
	KibaClock *clock = data;
	
	if (clock->timeout_id)
		g_source_remove (clock->timeout_id);
	
	clock->show_seconds = (gboolean) GPOINTER_TO_INT(value);
	if (clock->digital) {
		kiba_clock_update_size_from_layout (clock);
	}
	if (clock->show_seconds)
		clock->timeout_id = g_timeout_add (SECOND_INTERVAL,
			(GSourceFunc) kiba_clock_time_handler, (gpointer) clock);
	else {
		clock->timeout_id = g_timeout_add (MINUTE_INTERVAL,
			(GSourceFunc) kiba_clock_time_handler, (gpointer) clock);
		kiba_clock_time_handler (clock);
	}
}

static void clock_digital_cb (const char *option_id,
			      gpointer    value,
			      gpointer    data) {
	KibaClock *clock = data;
	
	clock->digital = (gboolean) GPOINTER_TO_INT(value);
	if (clock->object) {
		if (!clock->digital) {
			kiba_object_set_requisition_factor (KIBA_OBJECT(clock->object),
				clock->size_factor, clock->size_factor);
			gtk_widget_queue_resize (clock->object->parent);
			kiba_clock_set_surface_size (clock);
		}
		else {
			kiba_clock_update_size_from_layout (clock);
		}
		kiba_clock_update_surface (clock);
	}
}

static void clock_text_font_cb (const char *option_id,
			        gpointer    value,
			        gpointer    data) {
	KibaClock *clock = data;
	
	clock->font = (char*) value;
	
	kiba_clock_update_size_from_layout (clock);
	if (PANGO_IS_LAYOUT(clock->pango_layout)) {
		kiba_clock_update_surface (clock);
		kiba_clock_time_handler (clock);
	}
}

static void clock_text_rgba_cb (const char *option_id,
			        gpointer    value,
			        gpointer    data) {
	KibaClock *clock = data;
	
	clock->text_rgba = kiba_prefs_rgba_from_string ((char*) value);
	
	if (PANGO_IS_LAYOUT(clock->pango_layout)) {
		kiba_clock_update_surface (clock);
		kiba_clock_time_handler (clock);
	}
}

static void clock_text_shadow_rgba_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaClock *clock = data;
	
	clock->text_shadow_rgba = kiba_prefs_rgba_from_string ((char*) value);
	
	if (PANGO_IS_LAYOUT(clock->pango_layout)) {
		kiba_clock_update_surface (clock);
		kiba_clock_time_handler (clock);
	}
}

static void clock_theme_cb (const char *option_id,
			    gpointer    value,
			    gpointer    data) {
	KibaClock *clock = data;
	
	clock->theme = (char*) value;
	
	clock->theme_type = kiba_clock_get_theme_type (clock);
	if (clock->object && !clock->digital)
		kiba_clock_update_surface (clock);
}

/**
KibaWin DND
**/

static void kiba_clock_drag_dest_func (GdkDragContext *drag_context,
				       const char     *drag_data,
				       gpointer        user_data,
				       char          **action_ret,
				       char          **icon_ret) {
	KibaClock *clock = user_data;
	unsigned int year, month, day;
	
	year = month = day = 0;
	
	char data[strlen(drag_data)];
	sprintf(data, "%s", drag_data);
	g_strstrip (data);
	
	if (sscanf (data, "%02u.%02u.%04u", &day, &month, &year) < 3)
		return;
	
	*action_ret = g_strdup(_("Create Clock Reminder"));
	*icon_ret = g_strdup("clock");
}

static gboolean kiba_clock_drag_receive_func (GdkDragContext *drag_context,
					      const char     *drag_data,
					      gpointer        user_data) {
	KibaClock *clock = user_data;
	unsigned int year, month, day;
	
	year = month = day = 0;
	
	char data[strlen(drag_data)];
	sprintf(data, "%s", drag_data);
	g_strstrip (data);
	
	if (sscanf (data, "%02u.%02u.%04u", &day, &month, &year) < 3)
		return FALSE;
	
	if (year == 0 || month == 0 || day == 0)
		return FALSE;
	
	kiba_calendar_show_reminder_for_date (clock->calendar, NULL, year, month, day);
	
	return FALSE;
}

/**
KibaObject Callbacks
**/

static void kiba_object_drag_received_cb (GtkWidget        *widget,
					  GdkDragContext   *drag_context,
					  gint              x,
					  gint              y,
					  GtkSelectionData *selection_data,
					  guint             info,
					  guint             time,
					  gpointer          data) {
	kiba_clock_drag_receive_func (drag_context,
			(char *) selection_data->data, data);
}

static gboolean kiba_object_button_press_cb (GtkWidget      *widget,
					     GdkEventButton *event,
					     gpointer        data) {
	KibaPlugin *plugin = data;
	KibaClock *clock;
	GtkWidget *menu;
	GtkWidget *item;
	
	clock = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	menu = gtk_menu_new ();
	
	item = gtk_image_menu_item_new ();
	kiba_menu_item_setup_with_icon (item, GTK_ICON_SIZE_MENU,
				   GTK_STOCK_PREFERENCES, NULL, _("Modify Date and Time"), TRUE);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
	gtk_widget_show (item);
	g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (config_date), clock);
	
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
		event->button, event->time);
	
	return TRUE;
}

static void calendar_window_destroy_cb (GtkWidget *widget, gpointer data) {
	KibaClock *clock = data;
	clock->calendar = NULL;
}

static void calendar_window_size_allocate_cb (GtkWidget *widget, GtkAllocation *allocation, gpointer data) {
	KibaClock *clock = data;
	kiba_clock_place_window (clock, GTK_WIDGET(clock->calendar));
}

static gboolean kiba_object_button_release_cb (GtkWidget      *widget,
					       GdkEventButton *event,
					       gpointer        data) {
	KibaClock *clock = data;
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	
	if (event->button == 3)
		return FALSE;
	
	if (!GTK_IS_WIDGET(clock->calendar)) {
		clock->calendar = kiba_calendar_new ();
		clock->calendar->file = clock->file;
		g_object_add_weak_pointer (G_OBJECT (clock->calendar), (gpointer *) &clock->calendar);
	}
	
	if (!GTK_WIDGET_MAPPED(clock->calendar)) {
		if (!GTK_WIDGET_REALIZED(clock->calendar))
			gtk_widget_realize(GTK_WIDGET(clock->calendar));
		kiba_clock_place_window (clock, GTK_WIDGET(clock->calendar));
		gtk_window_present(GTK_WINDOW(clock->calendar));
		gtk_widget_map(GTK_WIDGET(clock->calendar));
		gdk_window_set_keep_above (GTK_WIDGET (clock->calendar)->window, TRUE);
		
		g_signal_connect_after (G_OBJECT(clock->calendar), "size-allocate",
			G_CALLBACK(calendar_window_size_allocate_cb), clock);
		g_signal_connect_after (G_OBJECT(clock->object), "size-allocate",
			G_CALLBACK(calendar_window_size_allocate_cb), clock);
		g_signal_connect_after (G_OBJECT(GTK_WIDGET(clock->object)->parent), "size-allocate",
			G_CALLBACK(calendar_window_size_allocate_cb), clock);
		g_signal_connect_after (G_OBJECT(GTK_WIDGET(clock->object)->parent->parent), "size-allocate",
			G_CALLBACK(calendar_window_size_allocate_cb), clock);
	}
	else {
		gtk_widget_unmap(GTK_WIDGET(clock->calendar));
		g_signal_handlers_disconnect_by_func (G_OBJECT(clock->calendar),
			G_CALLBACK(calendar_window_size_allocate_cb), clock);
		g_signal_handlers_disconnect_by_func (G_OBJECT(clock->object),
			G_CALLBACK(calendar_window_size_allocate_cb), clock);
		g_signal_handlers_disconnect_by_func (G_OBJECT(GTK_WIDGET(clock->object)->parent->parent),
			G_CALLBACK(calendar_window_size_allocate_cb), clock);
		g_signal_handlers_disconnect_by_func (G_OBJECT(GTK_WIDGET(clock->object)->parent),
			G_CALLBACK(calendar_window_size_allocate_cb), clock);
	}
	
	return FALSE;
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaClock  *clock;
	GList *l;
	
	clock = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	if (clock->timeout_id)
		g_source_remove (clock->timeout_id);
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_remove_callback_by_data (file, clock);
	
	kiba_win_remove_drag_func (kiba_get_plugin_win(kiba, plugin),
			     kiba_clock_drag_receive_func);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(clock->object),
		  G_CALLBACK(kiba_object_button_press_cb), plugin);
	g_signal_handlers_disconnect_by_func (G_OBJECT(clock->object),
		  G_CALLBACK(kiba_object_button_release_cb), clock);
	g_signal_handlers_disconnect_by_func (G_OBJECT(clock->object),
		  G_CALLBACK(kiba_object_drag_received_cb), clock);
	
	kiba_object_destroy (KIBA_OBJECT(clock->object));
	
	if (clock->foreground_surface.surface)
		cairo_surface_destroy (clock->foreground_surface.surface);
	if (clock->hour_surface.surface)
		cairo_surface_destroy (clock->hour_surface.surface);
	if (clock->minute_surface.surface)
		cairo_surface_destroy (clock->minute_surface.surface);
	if (clock->second_surface.surface)
		cairo_surface_destroy (clock->second_surface.surface);
	
	if (clock->calendar) {
		gtk_widget_destroy (GTK_WIDGET(clock->calendar));
	}
	
	if (PANGO_IS_LAYOUT(clock->pango_layout)) {
		g_object_unref (G_OBJECT(clock->pango_layout));
	}
	
	g_free (clock);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaClock *clock;
	KibaObject *object;
	GList *l;
	gint i;
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");	
#endif
	
	clock = g_new0 (KibaClock, 1);
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_add_callback (file, "clock_24mode",
		metric_mode_cb, clock, TRUE);
	kiba_prefs_add_callback (file, "clock_digital",
		clock_digital_cb, clock, TRUE);
	kiba_prefs_add_callback (file, "clock_text_font",
		clock_text_font_cb, clock, TRUE);
	kiba_prefs_add_callback (file, "clock_text_rgba",
		clock_text_rgba_cb, clock, TRUE);
	kiba_prefs_add_callback (file, "clock_text_shadow_rgba",
		clock_text_shadow_rgba_cb, clock, TRUE);
	kiba_prefs_add_callback (file, "clock_theme",
		clock_theme_cb, clock, TRUE);
	
	clock->object = kiba_drawable_object_new (PLUGIN_NAME);
	kiba_add_plugin_object(kiba, plugin, KIBA_OBJECT(clock->object));
	object = KIBA_OBJECT(clock->object);
	
	kiba_drawable_object_set_paint_function (KIBA_DRAWABLE_OBJECT(object),
		kiba_clock_paint, clock);
	gtk_widget_show (clock->object);
	
	CairoSurfaceData *surface_data[KIBA_CLOCK_FOREGROUND] = {
		&clock->hour_surface,
		&clock->minute_surface,
		&clock->second_surface,
		&clock->foreground_surface
	};
	
	for (i = 0; i < KIBA_CLOCK_FOREGROUND; i++)
		surface_data[i]->surface = NULL;
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) clock);
	
	clock->calendar = kiba_calendar_new ();
	clock->calendar->file = file;
	clock->file = file;
	g_object_add_weak_pointer (G_OBJECT (clock->calendar), (gpointer *) &clock->calendar);
	
	g_signal_connect (G_OBJECT (clock->object), "button-press-event",
			G_CALLBACK (kiba_object_button_press_cb), plugin);
	g_signal_connect (G_OBJECT (clock->object), "button-release-event",
			G_CALLBACK (kiba_object_button_release_cb), clock);
	g_signal_connect (G_OBJECT (clock->object), "drag-data-received",
			G_CALLBACK (kiba_object_drag_received_cb), clock);
	g_signal_connect (G_OBJECT(clock->calendar), "destroy",
		G_CALLBACK(calendar_window_destroy_cb), clock);
	
	kiba_prefs_add_callback (file, "dock_object_size",
		dock_object_size_cb, clock, FALSE);
	kiba_prefs_add_callback (file, "clock_size_factor",
		object_size_factor_cb, clock, TRUE);
	kiba_prefs_add_callback (file, "clock_show_seconds",
		show_seconds_cb, clock, TRUE);
	
	kiba_win_add_drag_func (kiba_get_plugin_win(kiba, plugin),
			kiba_clock_drag_receive_func,
			kiba_clock_drag_dest_func,
			clock);
	
	return TRUE;
}

static PrefMember clock_options[] =  {
	{ "clock_size_factor", "1.0", KIBA_DOUBLE },
	{ "clock_theme", "none", KIBA_STRING },
	{ "clock_show_seconds", "TRUE", KIBA_BOOLEAN },
	{ "clock_digital", "TRUE", KIBA_BOOLEAN },
	{ "clock_text_font", "Sans 12", KIBA_STRING },
	{ "clock_text_rgba", "#000000AA", KIBA_STRING },
	{ "clock_text_shadow_rgba", "#FFFFFFAA", KIBA_STRING },
	{ "clock_24mode", "TRUE", KIBA_BOOLEAN }
};

static KibaPluginVtable vtable = {
	"Clock Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
};

KIBA_INIT_PLUGIN (vtable, clock_options, NULL, NULL, NULL);
