/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <string.h>
#include <locale.h>
#include <stdlib.h>
#include <signal.h>
#include <libintl.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <glib.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>

#include <kiba.h>
#include <kiba-dock.h>
#include <kiba-icon.h>
#include <kiba-plugin.h>
#include <kiba-plugin-loader.h>
#include <kiba-prefs.h>
#include <kiba-screen.h>
#include <kiba-desktop-icon.h>
#include <kiba-x11.h>
#include <kiba-debug.h>
#include <kiba-macros.h>
#include <kiba-inline-pixbufs.h>
#include <kiba-svg-macro.h>

#include "marshallers.h"

#define KIBA_TYPE_DBUS            (kiba_dbus_get_type ())
#define KIBA_DBUS(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), KIBA_TYPE_DBUS, KibaDbus))
#define KIBA_DBUS_CLASS(obj)      (G_TYPE_CHECK_CLASS_CAST ((obj), KIBA_DBUS,  KibaDbusClass))
#define KIBA_IS_DBUS(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), KIBA_TYPE_DBUS))
#define KIBA_IS_DBUS_CLASS(obj)   (G_TYPE_CHECK_CLASS_TYPE ((obj), KIBA_TYPE_DBUS))
#define KIBA_DBUS_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), KIBA_TYPE_DBUS, KibaDbusClass))

#define PLUGIN_NAME             "dbus"
#define _(String)               dgettext (GETTEXT_PACKAGE, String)
#define KIBA_DBUS_PATH          "/org/kiba/dock/Kiba"
#define KIBA_NAMESPACE          "org.kiba.dock.Kiba"
#define BUBBLE_HIDE_INTERVAL    SECOND_INTERVAL*30

typedef struct _KibaDbus         KibaDbus;
typedef struct _KibaDbusClass    KibaDbusClass;

typedef enum {
	KIBA_DBUS_ICON_TYPE_DESKTOP = 0,
	KIBA_DBUS_ICON_TYPE_WIDGET
}KibaDbusIconType;

typedef struct {
	KibaDbus                *dbus;
	
	GQuark                   object_id; /* integer identifier for object */
	KibaDbusIconType         type;
	
	union {
		GtkWidget       *widget;
		KibaDesktopIcon *icon;
	}object;
	
	char                    *wmclass;
}KibaDbusIcon;

struct _KibaDbus {
	GObject parent_instance;
	
	KibaPlugin *plugin;
	
	GList *dbus_icons;   /* list with all DBus KibaDesktopIcon's */
	
	float scale;
};

struct _KibaDbusClass {
	GObjectClass parent_class;
	
	void (*initialled)    (KibaDbus *dbus);
	void (*icon_menu)     (KibaDbus    *dbus,
			       const gchar *id);
	void (*icon_released) (KibaDbus    *dbus,
			       const gchar *id);
	void (*icon_scrolled) (KibaDbus *dbus,
			       const gchar *id,
			       gint direction);
	void (*closed)        (KibaDbus *dbus);
};

enum {
	SIG_INIT = 0,
	SIG_FINI,
	SIG_MENU_REQUEST,
	SIG_ICON_RELEASED,
	SIG_ICON_SCROLLED,
	SIG_LAST_SIGNAL
};

static guint kiba_dbus_signals[SIG_LAST_SIGNAL] = { 0 };

static GList *script_pids;

static void     kiba_dbus_init              (KibaDbus      *self);
static void     kiba_dbus_class_init        (KibaDbusClass *klass);

static gpointer kiba_dbus_parent_class = NULL;
static GType    kiba_dbus_type_id = 0;

/**
Helper Functions
**/

static KibaDbusIcon* kiba_dbus_get_icon_for_object_id (KibaDbus   *dbus,
						    const char    *object_id,
						    GError       **error) {
	GList   *l;
	GQuark  error_quark;
	
	for (l = dbus->dbus_icons; l != NULL; l = l->next) {
		KibaDbusIcon *icon = l->data;
	
		if (icon->object_id == g_quark_from_string(object_id))
			return icon;
	}
	
	error_quark = g_quark_from_string ("kiba-dock");
	g_set_error (error, error_quark, 1, "No Previously added Object Found");
	
	return NULL;
}

static void kiba_dbus_scripts_launch (void) {
	GDir    *dir;
	GError  *error = NULL;
	gchar   *file_name;
	
	const char* scriptdir = DATADIR "/kiba-dock/dbus_scripts";
	if (!g_file_test (scriptdir, G_FILE_TEST_IS_DIR))
		return;
	
	dir = g_dir_open (scriptdir, 0, &error);
	if (error != NULL) {
		g_error_free (error);
		return;
	}
	
	DEBUG_MISC("Try to launch Python Scripts\n");
	
	while (file_name = (char *) g_dir_read_name (dir)) {
		gchar *script_path;
		gint   pid;
		
		if (!g_str_has_suffix (file_name, ".py") &&
		    !g_str_has_suffix (file_name, ".sh"))
			continue;
		
		script_path = g_strdup_printf ("%s/%s", scriptdir, file_name);
		DEBUG_MISC("Launching %s\n", script_path);
		pid = kiba_launch_programm (script_path);
		script_pids = g_list_append (script_pids, GINT_TO_POINTER (pid));
		g_free (script_path);
	}
	g_dir_close (dir);
}

static void kiba_dbus_scripts_kill (void) {
	GList *l;
	
	for (l = script_pids; l != NULL; l = l->next) {
		gint   pid = GPOINTER_TO_INT (l->data);
		
		kill (pid, SIGTERM);
	}
	
	for (l = script_pids; l != NULL; l = script_pids)
		script_pids = g_list_remove (script_pids, l->data);
}

/**
KibaObject callbacks
**/

static gboolean kiba_object_button_press_cb (GtkWidget      *widget,
					     GdkEventButton *event,
					     gpointer        data) {
	KibaDbusIcon *icon = data;
	
	if (event->button != 3)
		return FALSE;
	
	if (icon->type == KIBA_DBUS_ICON_TYPE_DESKTOP &&
	    icon->object.icon->state != KIBA_DESKTOP_STATE_ACTIVE)
		return FALSE;
	
	g_signal_emit (G_OBJECT (icon->dbus),
		kiba_dbus_signals[SIG_MENU_REQUEST], 0,
		g_quark_to_string(icon->object_id));
	
	return TRUE;
}

static gboolean kiba_object_button_release_cb (GtkWidget      *widget,
					       GdkEventButton *event,
					       gpointer        data) {
	KibaDbusIcon *icon = data;
	
	if (event->button != 1)
		return FALSE;
	
	if (icon->type == KIBA_DBUS_ICON_TYPE_DESKTOP &&
	    icon->object.icon->state != KIBA_DESKTOP_STATE_ACTIVE)
		return FALSE;
	
	g_signal_emit (G_OBJECT (icon->dbus),
		kiba_dbus_signals[SIG_ICON_RELEASED], 0,
		g_quark_to_string(icon->object_id));
	
	return TRUE;
}

static void kiba_object_scroll_event (GtkWidget      *widget,
				      GdkEventScroll *event,
				      gpointer        data) {
	KibaDbusIcon *icon = data;
	
	if (icon->type == KIBA_DBUS_ICON_TYPE_DESKTOP)
		return;
	
	g_signal_emit (G_OBJECT (icon->dbus),
		kiba_dbus_signals[SIG_ICON_SCROLLED], 0,
		g_quark_to_string(icon->object_id));
}

/**
DBus Methods
**/

static gboolean kiba_dbus_change_win_state (KibaDbus  *dbus,
					    gint       XID,
					    GError   **error) {
	GdkWindow *window;
	Window focus_window;
	gint state;
	GQuark  error_quark;
	
	error_quark = g_quark_from_string ("kiba-dock");
	
	window = GDK_WINDOW(gdk_xid_table_lookup (XID));
	if (!window) {
		g_set_error (error, error_quark, 1, "No GdkWindow found for given XID");
		return TRUE;
	}
	
	state = kiba_get_winstate (gdk_display, window);
	focus_window = kiba_get_active_window (gdk_display);
	
	/* window is iconified */
	if (state == IconicState) {
		kiba_deiconify_window (NULL, window);
		kiba_focus_window (NULL, window);
	}
	/* window is focussed */
	else if (state == NormalState && (Window) XID == focus_window) {
		kiba_iconify_window (NULL, window);
	}
	/* window is below others */
	else if (state == NormalState)
		kiba_focus_window (NULL, window);
	
	return TRUE;
}

static gboolean kiba_dbus_launch_command (KibaDbus  *dbus,
					  gchar     *command,
					  GError   **error) {
	kiba_launch_programm (command);
	
	return TRUE;
}

static gboolean kiba_dbus_set_icon (KibaDbus *dbus,
				    gchar    *object_id,
				    gchar    *icon,
				    GError  **error) {
	KibaDbusIcon *dbus_obj;
	KibaObject *object;
	
	dbus_obj = kiba_dbus_get_icon_for_object_id (dbus, object_id, error);
	if (!dbus_obj)
		return FALSE;
	
	if (dbus_obj->type == KIBA_DBUS_ICON_TYPE_WIDGET) {
		object = KIBA_OBJECT(dbus_obj->object.widget);
		
		kiba_icon_set_file (KIBA_ICON(object), icon);
	}
	else {
		kiba_desktop_icon_set_icon (dbus_obj->object.icon, icon);
	}
	
	return TRUE;
}

static gboolean kiba_dbus_set_titel (KibaDbus *dbus,
				     gchar    *object_id,
				     gchar    *titel,
				     GError  **error) {
	KibaDbusIcon *icon;
	KibaObject *object;
	
	icon = kiba_dbus_get_icon_for_object_id (dbus, object_id, error);
	if (!icon)
		return FALSE;
	
	if (icon->type == KIBA_DBUS_ICON_TYPE_WIDGET)
		object = KIBA_OBJECT(icon->object.widget);
	else {
		if (icon->object.icon->entry.name)
			g_free (icon->object.icon->entry.name);
		icon->object.icon->entry.name = g_strdup (titel);
		
		if (icon->object.icon->state == KIBA_DESKTOP_STATE_ACTIVE)
			object = KIBA_OBJECT(icon->object.icon->object);
		else
			return TRUE;
	}
	
	kiba_object_set_title (object, titel);
	
	return TRUE;
}

static gboolean kiba_dbus_toggle_active (KibaDbus *dbus,
					 gchar    *object_id,
					 GError  **error) {
	KibaDbusIcon *icon;
	KibaObject *object;
	
	icon = kiba_dbus_get_icon_for_object_id (dbus, object_id, error);
	if (!object)
		return FALSE;
	
	if (icon->type == KIBA_DBUS_ICON_TYPE_WIDGET)
		object = KIBA_OBJECT(icon->object.widget);
	else
		return TRUE;
	
	if (kiba_object_get_state (object, KIBA_OBJECT_ACTIVE))
		kiba_object_unset_state (object, KIBA_OBJECT_ACTIVE);
	else
		kiba_object_set_state (object, KIBA_OBJECT_ACTIVE);
	
	return TRUE;
}

static gboolean kiba_dbus_progress_set_value (KibaDbus *dbus,
					      gchar    *object_id,
					      gint      value,
					      GError  **error) {
	KibaDbusIcon *icon;
	KibaObject *object;
	
	icon = kiba_dbus_get_icon_for_object_id (dbus, object_id, error);
	if (!icon)
		return FALSE;
	
	if (icon->type == KIBA_DBUS_ICON_TYPE_WIDGET)
		object = KIBA_OBJECT(icon->object.widget);
	else
		object = KIBA_OBJECT(icon->object.icon->object);
	
	GtkWidget *widget = GTK_WIDGET(object);
	GtkWidget *win = widget->parent->parent;
	KibaProgress *progress = KIBA_WIN(win)->progress;
	
	if (icon->object.icon->state != KIBA_DESKTOP_STATE_ACTIVE &&
	    !kiba_progress_object_is_running(progress, object))
		return TRUE;
	
	kiba_progress_object_set_value (progress, object, value);
	
	if (value >= 100)
		kiba_progress_object_set_value (progress, object, 100);
	
	return TRUE;
}

static gboolean kiba_dbus_bubble_set_message (KibaDbus *dbus,
					      gchar    *object_id,
					      gchar    *message,
					      GError  **error) {
	KibaDbusIcon *icon;
	KibaObject *object;
	
	icon = kiba_dbus_get_icon_for_object_id (dbus, object_id, error);
	if (!icon)
		return FALSE;
	
	if (icon->type == KIBA_DBUS_ICON_TYPE_WIDGET)
		object = KIBA_OBJECT(icon->object.widget);
	else
		object = KIBA_OBJECT(icon->object.icon->object);
	
	kiba_object_set_tooltip_text(KIBA_OBJECT(object), message);
	
	return TRUE;
}

static gboolean kiba_dbus_bubble_unset_message (KibaDbus *dbus,
						gchar    *object_id,
						GError  **error) {
	KibaDbusIcon *icon;
	KibaObject *object;
	
	icon = kiba_dbus_get_icon_for_object_id (dbus, object_id, error);
	if (!icon)
		return FALSE;
	
	if (icon->type == KIBA_DBUS_ICON_TYPE_WIDGET)
		object = KIBA_OBJECT(icon->object.widget);
	else
		object = KIBA_OBJECT(icon->object.icon->object);
	
	kiba_object_set_tooltip_text(KIBA_OBJECT(object), NULL);
	
	return TRUE;
}

static gboolean kiba_dbus_pulse (KibaDbus *dbus,
				 gchar    *object_id,
				 GError  **error) {
	KibaDbusIcon *icon;
	KibaObject *object;
	
	icon = kiba_dbus_get_icon_for_object_id (dbus, object_id, error);
	if (!icon)
		return FALSE;
	
	if (icon->type == KIBA_DBUS_ICON_TYPE_WIDGET)
		object = KIBA_OBJECT(icon->object.widget);
	else if (icon->object.icon->state == KIBA_DESKTOP_STATE_ACTIVE)
		object = KIBA_OBJECT(icon->object.icon->object);
	else
		return TRUE;
	
	kiba_object_start_pulsing (GTK_WIDGET(object), 2.0);
	kiba_object_stop_pulsing (GTK_WIDGET(object));
	
	return TRUE;
}

static gboolean kiba_dbus_rotate (KibaDbus *dbus,
				  gchar    *object_id,
				  GError  **error) {
	KibaDbusIcon *icon;
	KibaObject *object;
	
	icon = kiba_dbus_get_icon_for_object_id (dbus, object_id, error);
	if (!icon)
		return FALSE;
	
	if (icon->type == KIBA_DBUS_ICON_TYPE_WIDGET)
		object = KIBA_OBJECT(icon->object.widget);
	else if (icon->object.icon->state == KIBA_DESKTOP_STATE_ACTIVE)
		object = KIBA_OBJECT(icon->object.icon->object);
	else
		return TRUE;
	
	kiba_object_start_rotation (GTK_WIDGET(object), 5.0);
	
	return TRUE;
}

static gboolean kiba_dbus_remove_object (KibaDbus *dbus,
					 gchar    *object_id,
					 GError  **error) {
	KibaDbusIcon *icon;
	KibaObject *object;
	
	icon = kiba_dbus_get_icon_for_object_id (dbus, object_id, error);
	if (!icon)
		return FALSE;
	
	dbus->dbus_icons = g_list_remove (dbus->dbus_icons, icon);
	
	if (icon->type == KIBA_DBUS_ICON_TYPE_WIDGET) {
		kiba_object_destroy (KIBA_OBJECT(icon->object.widget));
	}
	else {
		g_signal_handlers_disconnect_by_func (G_OBJECT(icon->object.icon->object),
			G_CALLBACK(kiba_object_button_press_cb), icon);
		g_signal_handlers_disconnect_by_func (G_OBJECT(icon->object.icon->object),
			G_CALLBACK(kiba_object_button_release_cb), icon);
		g_signal_handlers_disconnect_by_func (G_OBJECT(icon->object.icon->object),
			G_CALLBACK(kiba_object_scroll_event), icon);
		
		kiba_desktop_icon_set_file (icon->object.icon, icon->object.icon->entry.uri);
	}
	
	g_free (icon);
	icon = NULL;
	
	return TRUE;
}

static gboolean kiba_dbus_add_object (KibaDbus *dbus,
				      gchar    *object_id,
				      gchar    *wmclass,
				      gchar    *icon,
				      GError  **error) {
	Kiba *kiba = kiba_get();
	GList *l;
	GQuark  error_quark;
	KibaDbusIcon *dbus_obj;
	KibaDesktopIcon *desktop_obj;
	GtkWidget *object = NULL;
	
	error_quark = g_quark_from_string ("kiba-dock");
	
	if (!object_id) {
		g_set_error (error, error_quark, 1, "NULL Pointer Object ID");
		return FALSE;
	}
	
	for (l = dbus->dbus_icons; l != NULL; l = l->next) {
		KibaDbusIcon *check_object = l->data;
		
		if (check_object->object_id == g_quark_from_string(object_id))
			return TRUE;
	}
	
	dbus_obj = g_new0 (KibaDbusIcon, 1);
	dbus_obj->dbus = dbus;
	dbus_obj->object_id = g_quark_from_string(object_id);
	dbus_obj->type = KIBA_DBUS_ICON_TYPE_WIDGET;
	dbus_obj->wmclass = (wmclass != NULL ? g_strdup(wmclass) : NULL);
	
	if (dbus_obj->type == KIBA_DBUS_ICON_TYPE_WIDGET) {
		KibaDrawableObject *drawable_object;
		object = kiba_icon_new_from_file (PLUGIN_NAME, icon);
		kiba_add_plugin_object(kiba, dbus->plugin, KIBA_OBJECT(object));
		dbus_obj->object.widget = object;
		gtk_widget_show (object);
		
		drawable_object = KIBA_DRAWABLE_OBJECT(object);
		kiba_object_set_requisition_factor (KIBA_OBJECT(object), dbus->scale, dbus->scale);
		drawable_object->surface_data->width  = object->allocation.width;
		drawable_object->surface_data->height = object->allocation.height;
	}
	
	kiba_object_set_title (KIBA_OBJECT(object), object_id);
	
	/* add object signal handler */
	g_signal_connect (G_OBJECT (object), "button-press-event",
			G_CALLBACK (kiba_object_button_press_cb), dbus_obj);
	g_signal_connect (G_OBJECT (object), "button-release-event",
			G_CALLBACK (kiba_object_button_release_cb), dbus_obj);
	g_signal_connect (G_OBJECT (object), "scroll-event",
			G_CALLBACK (kiba_object_scroll_event), dbus_obj);
	
	dbus->dbus_icons = g_list_append (dbus->dbus_icons, dbus_obj);
	
	return TRUE;
}

static void kiba_dbus_icon_remove_desktop (KibaDbusIcon *icon) {
	if (icon->type != KIBA_DBUS_ICON_TYPE_DESKTOP)
		return;
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(icon->object.icon->object),
		G_CALLBACK(kiba_object_button_press_cb), icon);
	g_signal_handlers_disconnect_by_func (G_OBJECT(icon->object.icon->object),
		G_CALLBACK(kiba_object_button_release_cb), icon);
	g_signal_handlers_disconnect_by_func (G_OBJECT(icon->object.icon->object),
		G_CALLBACK(kiba_object_scroll_event), icon);
	
	kiba_desktop_icon_set_file (icon->object.icon, icon->object.icon->entry.uri);
	
	{ //TODO add new object for the icon
		KibaDbus *dbus = icon->dbus;
		
		dbus->dbus_icons = g_list_remove (dbus->dbus_icons, icon);
		
		if (icon->wmclass)
			g_free (icon->wmclass);
		g_free (icon);
	}
}

static void kiba_dbus_icon_add_desktop (KibaDbusIcon    *icon,
					KibaDesktopIcon *desktop_icon) {
	if (icon->type == KIBA_DBUS_ICON_TYPE_DESKTOP)
		return;
	
	icon->type = KIBA_DBUS_ICON_TYPE_DESKTOP;
	icon->object.icon = desktop_icon;
	if (KIBA_IS_OBJECT(icon->object.widget))
		kiba_object_destroy (KIBA_OBJECT(icon->object.widget));
	
	g_signal_connect (G_OBJECT (icon->object.icon->object), "button-press-event",
			G_CALLBACK (kiba_object_button_press_cb), icon);
	g_signal_connect (G_OBJECT (icon->object.icon->object), "button-release-event",
			G_CALLBACK (kiba_object_button_release_cb), icon);
	g_signal_connect (G_OBJECT (icon->object.icon->object), "scroll-event",
			G_CALLBACK (kiba_object_scroll_event), icon);
}

/**
Class
**/

static void     kiba_dbus_class_intern_init (gpointer klass)
{
  kiba_dbus_parent_class = g_type_class_peek_parent (klass); 
  kiba_dbus_class_init ((KibaDbusClass*) klass); 
}

GType
kiba_dbus_get_type (void)
{
  return kiba_dbus_type_id;
}

static void
kiba_dbus_register_type (GTypeModule *type_module)
{
  const GTypeInfo g_define_type_info = {
    sizeof (KibaDbusClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) kiba_dbus_class_intern_init,
    (GClassFinalizeFunc) NULL,
    NULL,   /* class_data */
    sizeof (KibaDbus),
    0,      /* n_preallocs */
    (GInstanceInitFunc) kiba_dbus_init, 
    NULL    /* value_table */
  };
  kiba_dbus_type_id = g_type_module_register_type (type_module,
					      G_TYPE_OBJECT,
					      "KibaDbus",
					      &g_define_type_info, 0);
}

#include "kiba-dbus-glue.h"

static void kiba_dbus_class_init (KibaDbusClass *class) {
	GObjectClass *object_class = G_OBJECT_CLASS (class);
	
	kiba_dbus_signals[SIG_MENU_REQUEST] =
		g_signal_new ("icon_menu",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (KibaDbusClass, icon_menu),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 1,
			      G_TYPE_STRING);
	kiba_dbus_signals[SIG_ICON_RELEASED] =
		g_signal_new ("icon_released",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (KibaDbusClass, icon_released),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 1,
			      G_TYPE_STRING);
	kiba_dbus_signals[SIG_ICON_SCROLLED] =
		g_signal_new ("icon_scrolled",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (KibaDbusClass, icon_scrolled),
			      NULL, NULL,
			      _kiba_marshal_VOID__STRING_INT,
			      G_TYPE_NONE, 2,
			      G_TYPE_STRING, G_TYPE_INT);
	kiba_dbus_signals[SIG_INIT] =
		g_signal_new ("initialled",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (KibaDbusClass, initialled),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      G_TYPE_INVALID);
	kiba_dbus_signals[SIG_FINI] =
		g_signal_new ("closed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (KibaDbusClass, closed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      G_TYPE_INVALID);
	
	dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (class),
					&dbus_glib_kiba_dock_object_info);
}

static void kiba_dbus_init (KibaDbus *dbus) {
	dbus->dbus_icons = NULL;
}

static KibaDbus* kiba_dbus_new (GTypeModule *type_module) {
	KibaDbus *dbus;
	
	kiba_dbus_register_type (type_module);
	dbus = g_object_new (KIBA_TYPE_DBUS, NULL);
	dbus->plugin = KIBA_PLUGIN(type_module);
	
	return dbus;
}

/**
KibaPrefs Callbacks
**/

static void dock_object_size_cb (const char *option_id,
			         gpointer    value,
			         gpointer    data) {
	KibaDbus *dbus = data;
	GList *l;
	
	for (l = dbus->dbus_icons; l != NULL; l = l->next) {
		KibaDbusIcon *icon = l->data;
		KibaObject *object = KIBA_OBJECT(icon->object.widget);
		KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object);
		
		if (icon->type != KIBA_DBUS_ICON_TYPE_WIDGET)
			continue;
		
		drawable_object->surface_data->width  = icon->object.widget->allocation.width;
		drawable_object->surface_data->height = icon->object.widget->allocation.height;
		char *buf = KIBA_ICON(object)->file;
		KIBA_ICON(object)->file = NULL;
		kiba_icon_set_file (KIBA_ICON(object), buf);
		g_free (buf);
	}
}

static void object_size_factor_cb (const char *option_id,
			           gpointer    value,
			           gpointer    data) {
	Kiba *kiba = kiba_get();
	KibaDbus *dbus = data;
	KibaWin *win;
	KibaDock *dock;
	GList *l;
	
	win = kiba_get_plugin_win(kiba, dbus->plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	dbus->scale = g_ascii_strtod((char*) value, NULL);
	
	for (l = dbus->dbus_icons; l != NULL; l = l->next) {
		KibaDbusIcon *icon = l->data;
		KibaObject *object = KIBA_OBJECT(icon->object.widget);
		KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object);
		char *buf;
		
		if (icon->type != KIBA_DBUS_ICON_TYPE_WIDGET)
			continue;
		
		kiba_object_set_requisition_factor (object, dbus->scale, dbus->scale);
		drawable_object->surface_data->width  = icon->object.widget->allocation.width;
		drawable_object->surface_data->height = icon->object.widget->allocation.height;
		buf = KIBA_ICON(object)->file;
		KIBA_ICON(object)->file = NULL;
		kiba_icon_set_file (KIBA_ICON(object), buf);
		g_free (buf);
	}
	gtk_widget_queue_resize (GTK_WIDGET(dock));
}

static void object_alpha_cb (const char *option_id,
			     gpointer    value,
			     gpointer    data) {
	KibaDbus *dbus = data;
	GList *l;
	float alpha;
	
	alpha = g_ascii_strtod((char*) value, NULL);
	
	for (l = dbus->dbus_icons; l != NULL; l = l->next) {
		KibaDbusIcon *icon = l->data;
		KibaObject *object;
		
		if (icon->type != KIBA_DBUS_ICON_TYPE_WIDGET)
			continue;
		
		object = KIBA_OBJECT(icon->object.widget);
		
		kiba_drawable_object_set_alpha (KIBA_DRAWABLE_OBJECT(object), alpha);
	}
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaDbus *dbus;
	GList *l;
	DBusGConnection  *connection;
	DBusGProxy *proxy;
	guint32 ret;
	GError *error = NULL;
	
	dbus = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	if (!KIBA_IS_DBUS(dbus))
		return;
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_remove_callback_by_data (file, dbus);
	
	for (l = dbus->dbus_icons; l != NULL; l = dbus->dbus_icons) {
		KibaDbusIcon *icon = l->data;
		
		if (icon->type == KIBA_DBUS_ICON_TYPE_WIDGET)
			kiba_object_destroy (KIBA_OBJECT(icon->object.widget));
		
		dbus->dbus_icons = g_list_remove (dbus->dbus_icons, l->data);
	}
	g_list_free (dbus->dbus_icons);
	
	/* kill dbus scripts */
	kiba_dbus_scripts_kill ();
	
	/* release the name on the bus */
	connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (connection != NULL) {
		proxy = dbus_g_proxy_new_for_name (connection, DBUS_SERVICE_DBUS,
					DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
		org_freedesktop_DBus_release_name (proxy, KIBA_NAMESPACE, 0, &error);
	}
	
	g_signal_emit (G_OBJECT (dbus),
			kiba_dbus_signals[SIG_FINI], 0);
	
	g_object_unref (dbus);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaDbus *dbus;
	DBusGConnection *connection;
	DBusGProxy *proxy;
	guint32 ret;
	GError *error = NULL;
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");	
#endif
	
	/* Get the connection and ensure the name is not used yet */
	connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (connection == NULL) {
		DEBUG_ERROR ("Failed to make connection to session bus: %s", error->message);
		g_error_free (error);
		return FALSE;
	}
	
	proxy = dbus_g_proxy_new_for_name (connection, DBUS_SERVICE_DBUS,
					DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
	if (!org_freedesktop_DBus_request_name (proxy, KIBA_NAMESPACE, 0, &ret, &error)) {
		DEBUG_ERROR ("There was an error requesting the name: %s", error->message);
		g_error_free (error);
		return FALSE;
	}
	
	if (ret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER)
		return FALSE;
	
	/* initiall the KibaDbus Class */
	dbus = kiba_dbus_new (G_TYPE_MODULE(plugin));
	if (dbus == NULL) {
		DEBUG_ERROR ("Failed to create the KibaDbus instance");
		return FALSE;
	}
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) dbus);
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_add_callback (file, "dock_object_size",
		dock_object_size_cb, dbus, TRUE);
	kiba_prefs_add_callback (file, "dbus_size_factor",
		object_size_factor_cb, dbus, TRUE);
	kiba_prefs_add_callback (file, "dbus_alpha",
		object_alpha_cb, dbus, TRUE);
	
	/* Register the app on the bus */
	dbus_g_connection_register_g_object (connection,
		KIBA_DBUS_PATH, G_OBJECT (dbus));
	
	g_signal_emit (G_OBJECT (dbus),
			kiba_dbus_signals[SIG_INIT], 0);
	
	/* launch dbus scripts */
	kiba_dbus_scripts_launch ();
	
	return TRUE;
}

static PrefMember dbus_options[] = {
    { "dbus_alpha", "1.0", KIBA_DOUBLE },
    { "dbus_size_factor", "1.0", KIBA_DOUBLE }
};

static KibaPluginVtable vtable = {
	"DBus Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_ONCE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
};

KIBA_INIT_PLUGIN (vtable, dbus_options, NULL, NULL, NULL);

