/*
 *  Copyright (C) 2007  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
 *  Note   : based on Gnome Menu Applet
*/

#include "config.h"

#include <string.h>
#include <libintl.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <libgnomeui/libgnomeui.h>
#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-mime-handlers.h>
#define GMENU_I_KNOW_THIS_IS_UNSTABLE
#include <gmenu-tree.h>

#include <kiba.h>
#include <kiba-dock.h>
#include <kiba-icon.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-screen.h>
#include <kiba-debug.h>
#include <kiba-dialog.h>
#include <kiba-menu-items.h>

#define _(String)             dgettext (GETTEXT_PACKAGE, String)
#define PLUGIN_NAME           "gnome_menu"
#define BOOKMARKS_FILENAME    ".gtk-bookmarks"
#define MAX_ITEMS_OR_SUBMENU  5
#define GSTOCK_ICON_SIZE      GTK_ICON_SIZE_MENU

typedef struct _KibaGMenu      KibaGMenu;
typedef struct _KibaMenuItem   KibaMenuItem;

struct _KibaGMenu {
	GtkWidget        *icon;
	
	GtkWidget        *menu;
	GtkRecentManager *recent_manager;
	
	/* settings */
	float             alpha;
	float             icon_size;
};

struct _KibaMenuItem {
	GtkWidget *widget;
	GtkWidget *menu;
};

static const char *gmenu_icon_data[3] = {
	"distributor-logo",
	"gnome-logo-icon-transparent",
	"gnome-logo-icon",
};

static void popup_menu (KibaGMenu *gmenu,
			guint      n_button,
			guint32    activate_time);

static GtkWidget* populate_menu_from_directory (GtkWidget          *menu,
						GMenuTreeDirectory *directory);

static GtkWidget* create_applications_menu (const char *menu_file,
					    const char *menu_path);

static GnomeVFSVolumeMonitor *volume_monitor = NULL;
extern char **environ;

/**
Helper Functions
**/

static void activate_uri (GtkWidget *menuitem,
			  char      *path) {
	GError    *error = NULL;
	char      *url;
	char      *escaped;

	url = gnome_vfs_make_uri_from_input_with_dirs (path,
						       GNOME_VFS_MAKE_URI_DIR_HOMEDIR);
	gnome_url_show_on_screen (url, gdk_screen_get_default (), &error);

	if (error != NULL)
		g_error_free (error);
	g_free (url);
}

static char ** panel_make_environment_for_screen (GdkScreen  *screen,
						  char      **envp) {
	char **retval = NULL;
	char  *display_name;
	int    display_index = -1;
	int    i, env_len;

	g_return_val_if_fail (GDK_IS_SCREEN (screen), NULL);

	if (envp == NULL)
		envp = environ;

	for (env_len = 0; envp[env_len]; env_len++)
		if (strncmp (envp[env_len], "DISPLAY", strlen ("DISPLAY")) == 0)
			display_index = env_len;

	retval = g_new (char *, env_len + 1);
	retval[env_len] = NULL;

	display_name = gdk_screen_make_display_name (screen);

	for (i = 0; i < env_len; i++)
		if (i == display_index)
			retval[i] = g_strconcat ("DISPLAY=", display_name, NULL);
		else
			retval[i] = g_strdup (envp[i]);

	g_assert (i == env_len);

	g_free (display_name);

	return retval;
}

static gboolean show_uri (const char *uri,
			  const char *mime_type,
			  GdkScreen  *screen,
			  GError    **error) {
	char **env;
	GnomeVFSResult result;
	GnomeVFSMimeApplication *app;
	GList *uris = NULL;

	app = gnome_vfs_mime_get_default_application_for_uri (uri, mime_type);
	if (app == NULL) {
		g_set_error (error, 0, 0, _("Could not find a suitable application."));
		return FALSE;
	}

	env = panel_make_environment_for_screen (screen, NULL);

	uris = g_list_append (uris, (gpointer)uri);
	result = gnome_vfs_mime_application_launch_with_env (app, uris, env);
	g_list_free (uris);

	g_strfreev (env);
	gnome_vfs_mime_application_free (app);

	if (result != GNOME_VFS_OK) {
		g_set_error (error, 0, 0, gnome_vfs_result_to_string (result));
		return FALSE;
	}

	return TRUE;
}

static void activate_app_def (GtkWidget      *menuitem,
			      GMenuTreeEntry *entry) {
	const char       *path;

	path = gmenu_tree_entry_get_desktop_file_path (entry);
	kiba_menu_item_activate_desktop_file (menuitem, path);
}

static char* remove_extension (const char *icon) {
	char *icon_no_extension;
	char *p;

	icon_no_extension = g_strdup (icon);
	p = strrchr (icon_no_extension, '.');
	if (p &&
	    (strcmp (p, ".png") == 0 ||
	     strcmp (p, ".xpm") == 0 ||
	     strcmp (p, ".svg") == 0)) {
	    *p = 0;
	}

	return icon_no_extension;
}

static GtkWidget* create_empty_menu (void) {
	GtkWidget *retval;
	
	retval = gtk_menu_new ();
	
	/* intercept all right button clicks makes sure they don't
	   go to the object itself */
	g_signal_connect (retval, "button_press_event",
			  G_CALLBACK (kiba_menu_item_dummy_button_press_event), NULL);

	return retval;
}

static void kiba_gmenu_update_icon (KibaGMenu *gmenu) {
	KibaDrawableObject *object = KIBA_DRAWABLE_OBJECT(gmenu->icon);
	gint i;
	
	object->surface_data->width  = gmenu->icon->allocation.width;
	object->surface_data->height = gmenu->icon->allocation.height;
	
	for (i = 0; i < G_N_ELEMENTS(gmenu_icon_data); i++) {
		if (kiba_icon_set_file (KIBA_ICON(gmenu->icon), gmenu_icon_data[i]))
			break;
	}
}

/**
Items
**/

static GtkWidget* create_submenu_entry (GtkWidget          *menu,
					GMenuTreeDirectory *directory) {
	GtkWidget *menuitem;

	menuitem = gtk_image_menu_item_new ();

	kiba_load_menu_image_deferred (menuitem,
					GSTOCK_ICON_SIZE,
					NULL,
					gmenu_tree_directory_get_icon (directory),
					"gnome-fs-directory");

	kiba_menu_item_setup (menuitem,
			GSTOCK_ICON_SIZE,
			NULL,
			gmenu_tree_directory_get_name (directory),
			TRUE);

	if (gmenu_tree_directory_get_comment (directory))
		gtk_tooltips_set_tip (kiba_get_tooltips(),
			menuitem,
			gmenu_tree_directory_get_comment (directory),
			NULL);

	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);

	return menuitem;
}

static void create_header (GtkWidget       *menu,
			   GMenuTreeHeader *header) {
	GMenuTreeDirectory *directory;
	GtkWidget          *menuitem;

	directory = gmenu_tree_header_get_directory (header);
	menuitem = create_submenu_entry (menu, directory);
	gmenu_tree_item_unref (directory);

	g_object_set_data_full (G_OBJECT (menuitem),
				"panel-gmenu-tree.header",
				gmenu_tree_item_ref (header),
				(GDestroyNotify) gmenu_tree_item_unref);

	g_signal_connect (menuitem, "activate",
			  G_CALLBACK (gtk_false), NULL);
}

static void panel_menu_items_append_place_item (const char *icon_name,
					        const char *title,
					        const char *tooltip,
					        GtkWidget  *menu,
					        GCallback   callback,
					        const char *uri) {
	GtkWidget *item;
	char      *user_data;

	item = gtk_image_menu_item_new ();
	kiba_menu_item_setup_with_icon (item,
				   GSTOCK_ICON_SIZE,
				   icon_name,
				   NULL,
				   title,
				   TRUE);

	gtk_tooltips_set_tip (kiba_get_tooltips(), item, tooltip, NULL);

	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);

	user_data = g_strdup (uri);
	g_signal_connect_data (item, "activate", callback, user_data,
			       (GClosureNotify) g_free, 0);

	g_signal_connect (G_OBJECT (item), "button_press_event",
			  G_CALLBACK (kiba_menu_item_dummy_button_press_event), NULL);

	kiba_menu_item_setup_uri_drag (item, icon_name, uri);
}

static void desktop_menu_item_append_menu (GtkWidget    *menu,
					   KibaMenuItem *menuitem) {
	gboolean  add_separator;
	GList    *children;

	if (!g_object_get_data (G_OBJECT (menu), "panel-menu-needs-appending"))
		return;

	g_object_set_data (G_OBJECT (menu), "panel-menu-needs-appending", NULL);

	add_separator = FALSE;
	children = gtk_container_get_children (GTK_CONTAINER (menu));

	if (children != NULL) {
		GList *child;

		child = children;
		while (child->next != NULL)
			child = child->next;

		add_separator = !GTK_IS_SEPARATOR (GTK_WIDGET (child->data));
	}

	g_list_free (children);

	if (add_separator)
		kiba_menu_add_separator (menu);

	kiba_menu_item_create_from_desktop (menu, "yelp.desktop", NULL);
	kiba_menu_item_create_from_desktop (menu, "gnome-about.desktop", NULL);
}

static void panel_place_menu_item_append_gtk_bookmarks (GtkWidget *menu) {
	typedef struct {
		GnomeVFSURI *uri;
		char        *label;
	} PanelBookmark;

	GtkWidget   *add_menu;
	char        *filename;
	char        *contents;
	gchar      **lines;
	GHashTable  *table;
	int          i;
	GSList      *add_bookmarks, *l;
	PanelBookmark *bookmark;

	filename = g_build_filename (g_get_home_dir (),
				     BOOKMARKS_FILENAME, NULL);

	if (!g_file_get_contents (filename, &contents, NULL, NULL)) {
		g_free (filename);
		return;
	}

	g_free (filename);

	lines = g_strsplit (contents, "\n", -1);
	g_free (contents);

	table = g_hash_table_new (g_str_hash, g_str_equal);
	add_bookmarks = NULL;

	for (i = 0; lines[i]; i++) {
		if (lines[i][0] && !g_hash_table_lookup (table, lines[i])) {
			GnomeVFSURI *uri;
			char        *space;
			char        *label;

			g_hash_table_insert (table, lines[i], lines[i]);

			space = strchr (lines[i], ' ');
			if (space) {
				*space = '\0';
				label = g_strdup (space + 1);
			} else {
				label = NULL;
			}

			uri = gnome_vfs_uri_new (lines[i]);

			if (!uri ||
			    (!strcmp (gnome_vfs_uri_get_scheme (uri), "file") &&
			     !gnome_vfs_uri_exists (uri))) {
				if (label)
					g_free (label);
				if (uri)
					gnome_vfs_uri_unref (uri);
				continue;
			}

			bookmark = g_malloc (sizeof (PanelBookmark));
			bookmark->uri = uri;
			bookmark->label = label;
			add_bookmarks = g_slist_prepend (add_bookmarks, bookmark);
		}
	}

	g_hash_table_destroy (table);
	g_strfreev (lines);

	add_bookmarks = g_slist_reverse (add_bookmarks);

	if (g_slist_length (add_bookmarks) <= MAX_ITEMS_OR_SUBMENU) {
		add_menu = menu;
	} else {
		GtkWidget *item;

		item = gtk_image_menu_item_new ();
		kiba_menu_item_setup_with_icon (item, GSTOCK_ICON_SIZE,
					   "stock_bookmark", NULL,
					   _("Bookmarks"), TRUE);

		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		gtk_widget_show (item);

		add_menu = create_empty_menu ();
		gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), add_menu);
	}

	for (l = add_bookmarks; l; l = l->next) {
		char *full_uri;
		char *display_uri;
		char *tooltip;
		char *label;

		bookmark = l->data;
		full_uri = gnome_vfs_uri_to_string (bookmark->uri,
						    GNOME_VFS_URI_HIDE_NONE);

		display_uri = gnome_vfs_format_uri_for_display (full_uri);
		/* Translators: %s is a URI */
		tooltip = g_strdup_printf (_("Open '%s'"), display_uri);
		g_free (display_uri);

		label = NULL;
		if (bookmark->label) {
			label = g_strdup (g_strstrip (bookmark->label));
			if (!label [0]) {
				g_free (label);
				label = NULL;
			}
		}

		if (!label) {
			if (!strcmp (gnome_vfs_uri_get_scheme (bookmark->uri),
							       "file")) {
				char *buffer;

				buffer = gnome_vfs_get_local_path_from_uri (full_uri);
				if (!buffer) {
					g_free (tooltip);
					g_free (full_uri);
					if (bookmark->label)
						g_free (bookmark->label);
					gnome_vfs_uri_unref (bookmark->uri);
					g_free (bookmark);
					continue;
				}

				label = g_filename_display_basename (buffer);
				g_free (buffer);
			} else if (gnome_vfs_uri_is_local (bookmark->uri)) {
				/* local non-file methods, such as burn:,
				 * font:, etc. */
				//FIXME 2.16: we should do the same than in nautilus (see nautilus_file_get_display_name_nocopy() and nautilus_get_vfs_method_display_name()
				label = gnome_vfs_format_uri_for_display (full_uri);
			} else {
				const char *path;
				const char *hostname;
				char       *displayname;

				hostname = gnome_vfs_uri_get_host_name (bookmark->uri);
				path = gnome_vfs_uri_get_path (bookmark->uri);
				if (!path || !path[0])
					displayname = g_strdup ("/");
				else
					displayname = gnome_vfs_unescape_string_for_display (path);
				if (hostname) {
					/* Translators: the first string is a
					 * path and the second string is a
					 * hostname. nautilus contains the same
					 * string to translate. */
					label = g_strdup_printf (_("%1$s on %2$s"),
								 displayname,
								 hostname);
					g_free (displayname);
				} else
					label = displayname;
			}
		}

		panel_menu_items_append_place_item ("gnome-fs-directory",
						    label,
						    tooltip,
						    add_menu,
						    G_CALLBACK (activate_uri),
						    full_uri);

		g_free (tooltip);
		g_free (full_uri);
		g_free (label);
		if (bookmark->label)
			g_free (bookmark->label);
		gnome_vfs_uri_unref (bookmark->uri);
		g_free (bookmark);
	}

	g_slist_free (add_bookmarks);
}

static void panel_place_menu_item_append_volumes (GtkWidget *menu,
						  gboolean   connected_volumes) {
	GtkWidget *add_menu;
	GList     *volumes, *l;
	GSList    *add_volumes, *sl;

	volumes = gnome_vfs_volume_monitor_get_mounted_volumes (volume_monitor);
	add_volumes = NULL;

	for (l = volumes; l; l = l->next) {
		GnomeVFSVolume *volume = l->data;

		if (!gnome_vfs_volume_is_user_visible (volume) ||
		    !gnome_vfs_volume_is_mounted (volume))
			continue;

		switch (gnome_vfs_volume_get_volume_type (volume)) {
		case GNOME_VFS_VOLUME_TYPE_CONNECTED_SERVER:
			if (connected_volumes)
				add_volumes = g_slist_prepend (add_volumes,
							       volume);
			break;
		default:
			if (!connected_volumes)
				add_volumes = g_slist_prepend (add_volumes,
							       volume);
			break;
		}
	}

	add_volumes = g_slist_sort (add_volumes,
				    (GCompareFunc) gnome_vfs_volume_compare);

	if (g_slist_length (add_volumes) <= MAX_ITEMS_OR_SUBMENU) {
		add_menu = menu;
	} else {
		GtkWidget  *item;
		const char *title;
		const char *icon;

		/* FIXME: find nice icons. We don't use NULL since it will
		 * make the menu items smaller... */
		if (connected_volumes) {
			title = _("Network Places");
			icon = "network-server";
		} else {
			title = _("Removable Media");
			icon = "";
		}

		item = gtk_image_menu_item_new ();
		kiba_menu_item_setup_with_icon (item, GSTOCK_ICON_SIZE,
					   icon, NULL, title, TRUE);

		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		gtk_widget_show (item);

		add_menu = create_empty_menu ();
		gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), add_menu);
	}

	for (sl = add_volumes; sl; sl = sl->next) {
		GnomeVFSVolume *volume = sl->data;
		char           *icon;
		char           *display_name;
		char           *activation_uri;

		icon           = gnome_vfs_volume_get_icon (volume);
		display_name   = gnome_vfs_volume_get_display_name (volume);
		activation_uri = gnome_vfs_volume_get_activation_uri (volume);

		panel_menu_items_append_place_item (icon,
						    display_name,
						    display_name, //FIXME tooltip
						    add_menu,
						    G_CALLBACK (activate_uri),
						    activation_uri);

		g_free (icon);
		g_free (display_name);
		g_free (activation_uri);
	}

	g_slist_free (add_volumes);

	g_list_foreach (volumes, (GFunc) gnome_vfs_volume_unref, NULL);
	g_list_free (volumes);
}

static void recent_documents_activate_cb (GtkRecentChooser *chooser,
					  gpointer          data) {
	GtkRecentInfo *recent_info;
	const char    *uri;
	const char    *mime_type;
	GdkScreen     *screen;
	GError        *error = NULL;

	screen = gtk_widget_get_screen (GTK_WIDGET (chooser));
	
	recent_info = gtk_recent_chooser_get_current_item (chooser);
	uri = gtk_recent_info_get_uri (recent_info);
	mime_type = gtk_recent_info_get_mime_type (recent_info);
	//FIXME gtk_recent_info_get_application_info() could be useful

	if (show_uri (uri, mime_type, screen, &error) != TRUE) {
		char *uri_utf8;

		uri_utf8 = g_filename_to_utf8 (uri, -1, NULL, NULL, NULL);
		//FIXME this could fail... Maybe we want gtk_recent_info_get_display_name()

		if (error) {
			kiba_dialog_error ("Failed to open File",
				_("Could not open recently used document '%s': %s"),
				uri_utf8, error->message);
			g_error_free (error);
		} else {
			kiba_dialog_error ("Failed to open File",
				_("Could not open recently used document '%s': %s"),
				uri_utf8, _("An unknown error occurred"));
		}

		g_free (uri_utf8);
	}


	gtk_recent_info_unref (recent_info);
}

static void panel_recent_manager_changed_cb (GtkRecentManager *manager,
					     GtkWidget        *menu_item) {
	int size;

	g_object_get (manager, "size", &size, NULL);

	gtk_widget_set_sensitive (menu_item, size > 0);
}

static GtkWidget *clear_recent_dialog = NULL;

static void clear_dialog_response (GtkWidget        *widget,
				   int               response,
				   GtkRecentManager *manager) {
        if (response == GTK_RESPONSE_YES)
		gtk_recent_manager_purge_items (manager, NULL);

	gtk_widget_destroy (widget);
}

static void recent_documents_clear_cb (GtkMenuItem      *menuitem,
				       GtkRecentManager *manager) {
	gpointer tmp;
	
	if (clear_recent_dialog != NULL) {
		gtk_window_set_screen (GTK_WINDOW (clear_recent_dialog),
				       gtk_widget_get_screen (GTK_WIDGET (menuitem)));
		gtk_window_present (GTK_WINDOW (clear_recent_dialog));
		return;
	}
	
	clear_recent_dialog = kiba_dialog_confirm(_("Clear the Recent Documents list?"),
	    _("If you clear the Recent Documents list, you clear the following:\n\n"
		"\342\200\242 All items from the Places \342\206\222 Recent Documents menu item.\n"
		"\342\200\242 All items from the recent documents list in all applications."));
	
	g_signal_connect (clear_recent_dialog, "response",
			  G_CALLBACK (clear_dialog_response), manager);
	
	tmp = &clear_recent_dialog;
	g_object_add_weak_pointer (G_OBJECT (clear_recent_dialog), tmp);
}

static void panel_recent_append_documents_menu (GtkWidget *top_menu,
						KibaGMenu *gmenu) {
	GtkWidget      *recent_menu;
	GtkWidget      *menu_item;
	int             size;

	menu_item = gtk_image_menu_item_new ();
	kiba_menu_item_setup_with_icon (menu_item,
				   GSTOCK_ICON_SIZE,
				   NULL,
				   GTK_STOCK_OPEN,
				   _("Recent Documents"),
				   TRUE);
	recent_menu = gtk_recent_chooser_menu_new_for_manager (gmenu->recent_manager);
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item), recent_menu);

	g_signal_connect (G_OBJECT (recent_menu), "button_press_event",
			  G_CALLBACK (kiba_menu_item_dummy_button_press_event), NULL);

	gtk_menu_shell_append (GTK_MENU_SHELL (top_menu), menu_item);
	gtk_widget_show_all (menu_item);
	
	gtk_recent_chooser_set_local_only (GTK_RECENT_CHOOSER (recent_menu),
					   FALSE);
	gtk_recent_chooser_set_show_tips (GTK_RECENT_CHOOSER (recent_menu),
					  TRUE);
	gtk_recent_chooser_set_sort_type (GTK_RECENT_CHOOSER (recent_menu),
					  GTK_RECENT_SORT_MRU);

	g_signal_connect (GTK_RECENT_CHOOSER (recent_menu),
			  "item-activated",
			  G_CALLBACK (recent_documents_activate_cb),
			  NULL);

	//FIXME this is not possible with GtkRecent...: egg_recent_view_gtk_set_icon_size (view, panel_menu_icon_get_size ());

	g_signal_connect_object (gmenu->recent_manager, "changed",
				 G_CALLBACK (panel_recent_manager_changed_cb),
				 menu_item, 0);

	g_object_get (gmenu->recent_manager, "size", &size, NULL);
	gtk_widget_set_sensitive (menu_item, size > 0);

	kiba_menu_add_separator (recent_menu);

	menu_item = gtk_image_menu_item_new ();
	kiba_menu_item_setup_with_icon (menu_item,
				   GSTOCK_ICON_SIZE,
				   NULL,
				   GTK_STOCK_CLEAR,
				   _("Clear Recent Documents"),
				   TRUE);
	gtk_tooltips_set_tip (kiba_get_tooltips(),
			      menu_item,
			      _("Clear all items from the recent documents list"),
			      NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (recent_menu), menu_item);

	g_signal_connect (menu_item, "activate",
			  G_CALLBACK (recent_documents_clear_cb),
			  gmenu->recent_manager);
}

static GtkWidget* menu_item_create_menu (KibaGMenu    *gmenu,
					 KibaMenuItem *menuitem) {
	GtkWidget *places_menu;
	GtkWidget *item;

	places_menu = gtk_menu_new ();

	kiba_menu_item_create_from_desktop (places_menu,
					      "nautilus-home.desktop",
					      NULL);
	/* this assumes that the Desktop dir is not the Home directory */
	char *uri;

	uri = gnome_vfs_make_uri_from_input_with_dirs ("Desktop", GNOME_VFS_MAKE_URI_DIR_HOMEDIR);

	panel_menu_items_append_place_item (
			"gnome-fs-desktop",
			"Desktop Folder|Desktop",
			"Open the desktop as a folder",
			places_menu,
			G_CALLBACK (activate_uri),
			uri);

	g_free (uri);

	panel_place_menu_item_append_gtk_bookmarks (places_menu);
	kiba_menu_add_separator (places_menu);

	kiba_menu_item_create_from_desktop (places_menu,
					"nautilus-computer.desktop",
					 NULL);

	kiba_menu_item_create_from_desktop (places_menu,
					"nautilus-cd-burner.desktop",
					NULL);

	panel_place_menu_item_append_volumes (places_menu, FALSE);
	kiba_menu_add_separator (places_menu);

	kiba_menu_item_create_from_desktop (places_menu,
					"network-scheme.desktop",
					NULL);
	panel_place_menu_item_append_volumes (places_menu, TRUE);

//TODO
// 	if (panel_is_program_in_path ("nautilus-connect-server")) {
// 		item = panel_menu_items_create_action_item (PANEL_ACTION_CONNECT_SERVER);
// 		if (item != NULL)
// 			gtk_menu_shell_append (GTK_MENU_SHELL (places_menu), item);
// 	}

	kiba_menu_add_separator (places_menu);

	kiba_menu_item_create_from_desktop (places_menu,
					"gnome-search-tool.desktop",
					NULL);
	
	panel_recent_append_documents_menu (places_menu, gmenu);
	
	return places_menu;
}

static GtkWidget* desktop_menu_item_create_menu (KibaMenuItem *menuitem) {
	GtkWidget *desktop_menu;

	desktop_menu = create_applications_menu ("settings.menu", NULL);

	g_signal_connect (desktop_menu, "show",
			  G_CALLBACK (desktop_menu_item_append_menu),
			  menuitem);

	return desktop_menu;
}

static GtkWidget* menu_item_new (KibaGMenu *gmenu,
				 gboolean   use_image) {
	KibaMenuItem *menuitem;
	GtkWidget    *accel_label;
	GtkWidget    *image;
	
	menuitem = g_new0 (KibaMenuItem, 1);
	menuitem->widget = gtk_image_menu_item_new ();
	accel_label = gtk_accel_label_new (_("Places"));
	gtk_misc_set_alignment (GTK_MISC (accel_label), 0.0, 0.5);

	gtk_container_add (GTK_CONTAINER (menuitem->widget), accel_label);
	gtk_accel_label_set_accel_widget (GTK_ACCEL_LABEL (accel_label), menuitem->widget);
	gtk_widget_show (accel_label);
	
	image = gtk_image_new_from_icon_name ("gnome-fs-directory", GSTOCK_ICON_SIZE);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem->widget), image);

	menuitem->menu = menu_item_create_menu (gmenu, menuitem);
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem->widget), menuitem->menu);

	return menuitem->widget;
}

GtkWidget* desktop_menu_item_new () {
	KibaMenuItem *menuitem;
	GtkWidget   *accel_label;
	GtkWidget   *image;

	menuitem = g_new0 (KibaMenuItem, 1);
	menuitem->widget = gtk_image_menu_item_new ();
	accel_label = gtk_accel_label_new (_("System"));
	gtk_misc_set_alignment (GTK_MISC (accel_label), 0.0, 0.5);

	gtk_container_add (GTK_CONTAINER (menuitem->widget), accel_label);
	gtk_accel_label_set_accel_widget (GTK_ACCEL_LABEL (accel_label),
					  GTK_WIDGET (menuitem->widget));
	gtk_widget_show (accel_label);

	image = gtk_image_new_from_icon_name ("gnome-fs-desktop", GSTOCK_ICON_SIZE);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem->widget), image);

	menuitem->menu = desktop_menu_item_create_menu (menuitem);
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem->widget),
				   menuitem->menu);

	return GTK_WIDGET (menuitem->widget);
}

/**
Signal Handler
**/

static void submenu_to_display (GtkWidget *menu) {
	GMenuTree          *tree;
	GMenuTreeDirectory *directory;
	const char         *menu_path;

	if (!g_object_get_data (G_OBJECT (menu), "panel-menu-needs-loading"))
		return;

	g_object_set_data (G_OBJECT (menu), "panel-menu-needs-loading", NULL);

	directory = g_object_get_data (G_OBJECT (menu),
				       "panel-menu-tree-directory");
	if (!directory) {
		menu_path = g_object_get_data (G_OBJECT (menu),
					       "panel-menu-tree-path");
		if (!menu_path)
			return;

		tree = g_object_get_data (G_OBJECT (menu), "panel-menu-tree");
		if (!tree)
			return;

		directory = gmenu_tree_get_directory_from_path (tree,
								menu_path);

		g_object_set_data_full (G_OBJECT (menu),
					"panel-menu-tree-directory",
					directory,
					(GDestroyNotify) gmenu_tree_item_unref);
	}

	if (directory)
		populate_menu_from_directory (menu, directory);
}

static gboolean submenu_to_display_in_idle (gpointer data) {
	GtkWidget *menu = GTK_WIDGET (data);

	g_object_set_data (G_OBJECT (menu), "panel-menu-idle-id", NULL);

	submenu_to_display (menu);

	return FALSE;
}

static void remove_submenu_to_display_idle (gpointer data) {
	guint idle_id = GPOINTER_TO_UINT (data);

	g_source_remove (idle_id);
}

static gboolean menuitem_button_press_event (GtkWidget      *menuitem,
					     GdkEventButton *event) {
// 	if (event->button == 3)
// 		return show_item_menu (menuitem, event);
	
	return FALSE;
}

static void edit_menu (GtkWidget *button,
		       gpointer   data) {
	KibaGMenu *menu = data;
	GError    *error = NULL;
	int        workspace;
	
	workspace = kiba_get_active_workspace_number (gdk_display);
	
	if (kiba_launch_programm ("alacarte"))
		return;
	if (kiba_launch_programm ("gmenu-simple-editor"))
		return;
	
	kiba_dialog_error (_("Failed to launch Programm"),
		_("There was an error executing '%s'"),
		"gmenu-simple-editor");
}

static void drag_data_get_menu_cb (GtkWidget        *widget,
				   GdkDragContext   *context,
				   GtkSelectionData *selection_data,
				   guint             info,
				   guint             time,
				   GMenuTreeEntry   *entry) {
	const char *path;
	char       *uri;
	char       *uri_list;

	path = gmenu_tree_entry_get_desktop_file_path (entry);
	uri = gnome_vfs_get_uri_from_local_path (path);
	uri_list = g_strconcat (uri, "\r\n", NULL);
	g_free (uri);

	gtk_selection_data_set (selection_data,
				selection_data->target, 8, (guchar *)uri_list,
				strlen (uri_list));
	g_free (uri_list);
}

static void icon_theme_changed_cb (GtkIconTheme *icon_theme,
				   gpointer      user_data) {
	KibaGMenu *gmenu = user_data;
	
	kiba_gmenu_update_icon (gmenu);
}

/**
The Menu
**/

static GtkWidget* create_fake_menu (GMenuTreeDirectory *directory) {
	GtkWidget *menu;
	guint      idle_id;
	
	menu = create_empty_menu ();

	g_object_set_data_full (G_OBJECT (menu),
				"panel-menu-tree-directory",
				gmenu_tree_item_ref (directory),
				(GDestroyNotify) gmenu_tree_item_unref);
	
	g_object_set_data (G_OBJECT (menu),
			   "panel-menu-needs-loading",
			   GUINT_TO_POINTER (TRUE));

	g_object_set_data (G_OBJECT (menu),
			   "panel-menu-needs-appending",
			   GUINT_TO_POINTER (TRUE));

	g_signal_connect (menu, "show",
			  G_CALLBACK (submenu_to_display), NULL);

	idle_id = g_idle_add_full (G_PRIORITY_LOW,
				   submenu_to_display_in_idle,
				   menu,
				   NULL);
	g_object_set_data_full (G_OBJECT (menu),
				"panel-menu-idle-id",
				GUINT_TO_POINTER (idle_id),
				remove_submenu_to_display_idle);

	g_signal_connect (menu, "button_press_event",
			  G_CALLBACK (kiba_menu_item_dummy_button_press_event), NULL);

	return menu;
}

static void create_submenu (GtkWidget          *menu,
			    GMenuTreeDirectory *directory,
			    GMenuTreeDirectory *alias_directory) {
	GtkWidget *menuitem;
	GtkWidget *submenu;

	if (alias_directory)
		menuitem = create_submenu_entry (menu, alias_directory);
	else
		menuitem = create_submenu_entry (menu, directory);
	
	submenu = create_fake_menu (directory);

	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), submenu);
}

static void create_menuitem (GtkWidget          *menu,
			     GMenuTreeEntry     *entry,
			     GMenuTreeDirectory *alias_directory) {
	GtkWidget  *menuitem;
	
	menuitem = gtk_image_menu_item_new ();

	g_object_set_data_full (G_OBJECT (menuitem),
				"panel-menu-tree-entry",
				gmenu_tree_item_ref (entry),
				(GDestroyNotify) gmenu_tree_item_unref);

	if (alias_directory)
		//FIXME: we should probably use this data when we do dnd or
		//context menu for this menu item
		g_object_set_data_full (G_OBJECT (menuitem),
					"panel-menu-tree-alias-directory",
					gmenu_tree_item_ref (alias_directory),
					(GDestroyNotify) gmenu_tree_item_unref);

	kiba_load_menu_image_deferred (menuitem,
					GSTOCK_ICON_SIZE,
					NULL,
					alias_directory ? gmenu_tree_directory_get_icon (alias_directory) :
							  gmenu_tree_entry_get_icon (entry),
					NULL);

	kiba_menu_item_setup (menuitem,
			GSTOCK_ICON_SIZE,
			NULL,
			alias_directory ? gmenu_tree_directory_get_name (alias_directory) :
					  gmenu_tree_entry_get_name (entry),
			TRUE);

	if ((alias_directory &&
	     gmenu_tree_directory_get_comment (alias_directory)) ||
	    (!alias_directory &&
	     gmenu_tree_entry_get_comment (entry)))
		gtk_tooltips_set_tip (kiba_get_tooltips(),
			menuitem,
			alias_directory ? gmenu_tree_directory_get_comment (alias_directory) :
					  gmenu_tree_entry_get_comment (entry),
			NULL);

	g_signal_connect_after (menuitem, "button_press_event",
				G_CALLBACK (menuitem_button_press_event), NULL);

	char *icon_no_ext = NULL;
	if (gmenu_tree_entry_get_icon (entry) != NULL) {
		const char *icon;
		
		icon = gmenu_tree_entry_get_icon (entry);
		if (!g_path_is_absolute (icon))
			icon_no_ext = remove_extension (icon);
	}
	
	kiba_menu_item_setup_uri_drag (menuitem, NULL, icon_no_ext);
	if (icon_no_ext)
		g_free (icon_no_ext);
	
	g_signal_connect (menuitem, "drag_data_get",
			  G_CALLBACK (drag_data_get_menu_cb), entry);

	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);

	g_signal_connect (menuitem, "activate",
			  G_CALLBACK (activate_app_def), entry);
}

static GtkWidget* populate_menu_from_directory (GtkWidget          *menu,
						GMenuTreeDirectory *directory) {
	GSList   *l;
	GSList   *items;
	gboolean  add_separator;

	add_separator = (GTK_MENU_SHELL (menu)->children != NULL);

	items = gmenu_tree_directory_get_contents (directory);

	for (l = items; l; l = l->next) {
		GMenuTreeItem *item = l->data;

		if (add_separator ||
		    gmenu_tree_item_get_type (item) == GMENU_TREE_ITEM_SEPARATOR) {
			kiba_menu_add_separator (menu);
			add_separator = FALSE;
		}

		switch (gmenu_tree_item_get_type (item)) {
		case GMENU_TREE_ITEM_DIRECTORY:
			create_submenu (menu, GMENU_TREE_DIRECTORY (item), NULL);
			break;

		case GMENU_TREE_ITEM_ENTRY:
			create_menuitem (menu, GMENU_TREE_ENTRY (item), NULL);
			break;

		case GMENU_TREE_ITEM_SEPARATOR :
			/* already added */
			break;

		case GMENU_TREE_ITEM_ALIAS:
			create_menuitem_from_alias (menu, GMENU_TREE_ALIAS (item));
			break;

		case GMENU_TREE_ITEM_HEADER:
			create_header (menu, GMENU_TREE_HEADER (item));
			break;

		default:
			break;
		}

		gmenu_tree_item_unref (item);
	}

	g_slist_free (items);

	return menu;
}

static void handle_gmenu_tree_changed (GMenuTree *tree,
				       GtkWidget *menu) {
	guint idle_id;

	while (GTK_MENU_SHELL (menu)->children)
                gtk_widget_destroy (GTK_MENU_SHELL (menu)->children->data);

	g_object_set_data_full (G_OBJECT (menu),
				"panel-menu-tree-directory",
				NULL, NULL);

	g_object_set_data (G_OBJECT (menu),
			   "panel-menu-needs-loading",
			   GUINT_TO_POINTER (TRUE));

	g_object_set_data (G_OBJECT (menu),
			   "panel-menu-needs-appending",
			   GUINT_TO_POINTER (TRUE));

	idle_id = g_idle_add_full (G_PRIORITY_LOW,
				   submenu_to_display_in_idle,
				   menu,
				   NULL);
	g_object_set_data_full (G_OBJECT (menu),
				"panel-menu-idle-id",
				GUINT_TO_POINTER (idle_id),
				remove_submenu_to_display_idle);
}

static void remove_gmenu_tree_monitor (GtkWidget *menu,
				       GMenuTree *tree) {
	gmenu_tree_remove_monitor (tree,
				  (GMenuTreeChangedFunc) handle_gmenu_tree_changed,
				  menu);
}

static void main_menu_append (GtkWidget *main_menu,
			      KibaGMenu *gmenu) {
	GtkWidget *item;
	gboolean   add_separator;
	GList     *children;
	GList     *last;

	if (!g_object_get_data (G_OBJECT (main_menu), "panel-menu-needs-appending"))
		return;

	g_object_set_data (G_OBJECT (main_menu), "panel-menu-needs-appending", NULL);

	add_separator = FALSE;
	children = gtk_container_get_children (GTK_CONTAINER (main_menu));
	last = g_list_last (children);
	if (last != NULL) {
		add_separator = !GTK_IS_SEPARATOR (GTK_WIDGET (last->data));
	}
	g_list_free (children);

	if (add_separator)
		kiba_menu_add_separator (main_menu);

	item = menu_item_new (gmenu, TRUE);
	gtk_menu_shell_append (GTK_MENU_SHELL (main_menu), item);
	gtk_widget_show (item);

	item = desktop_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (main_menu), item);
	gtk_widget_show (item);
	
//TODO action items
// 	panel_menu_items_append_lock_logout (main_menu);
}

static GtkWidget* create_applications_menu (const char *menu_file,
					    const char *menu_path) {
	GMenuTree *tree;
	GtkWidget *menu;
	guint      idle_id;

	menu = create_empty_menu ();

	tree = gmenu_tree_lookup (menu_file, GMENU_TREE_FLAGS_NONE);

	g_object_set_data_full (G_OBJECT (menu),
				"panel-menu-tree",
				gmenu_tree_ref (tree),
				(GDestroyNotify) gmenu_tree_unref);

	g_object_set_data_full (G_OBJECT (menu),
				"panel-menu-tree-path",
				g_strdup (menu_path ? menu_path : "/"),
				(GDestroyNotify) g_free);
	
	g_object_set_data (G_OBJECT (menu),
			   "panel-menu-needs-loading",
			   GUINT_TO_POINTER (TRUE));

	g_object_set_data (G_OBJECT (menu),
			   "panel-menu-needs-appending",
			   GUINT_TO_POINTER (TRUE));

	g_signal_connect (menu, "show", G_CALLBACK (submenu_to_display), NULL);

	idle_id = g_idle_add_full (G_PRIORITY_LOW,
				   submenu_to_display_in_idle,
				   menu,
				   NULL);
	g_object_set_data_full (G_OBJECT (menu),
				"panel-menu-idle-id",
				GUINT_TO_POINTER (idle_id),
				remove_submenu_to_display_idle);

	g_signal_connect (menu, "button_press_event",
			  G_CALLBACK (kiba_menu_item_dummy_button_press_event), NULL);

	gmenu_tree_add_monitor (tree,
			       (GMenuTreeChangedFunc) handle_gmenu_tree_changed,
			       menu);
	g_signal_connect (menu, "destroy",
			  G_CALLBACK (remove_gmenu_tree_monitor), tree);

	gmenu_tree_unref (tree);

	return menu;
}

static GtkWidget* create_menu (void) {
	GtkWidget *main_menu;
	
	main_menu = create_applications_menu ("applications.menu", NULL);
	
	return main_menu;
}

static void position_menu (GtkMenu   *menu,
			   int       *x,
			   int       *y,
			   gboolean  *push_in,
			   KibaGMenu *gmenu) {
	GtkRequisition  req;
	GtkWidget *dock;
	const KibaLayoutData *d;
	
	dock = GTK_WIDGET(gmenu->icon)->parent;
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	gtk_widget_size_request (GTK_WIDGET (menu), &req);
	
	*x = gmenu->icon->allocation.x + dock->parent->allocation.x;
	*y = gmenu->icon->allocation.y + dock->parent->allocation.y;
	if (d->orientation == GTK_ORIENTATION_VERTICAL)
		*x -= d->halignment == 0 ? -gmenu->icon->allocation.width : req.width;
	else
		*y -= d->valignment == 0 ? -gmenu->icon->allocation.width : req.height;
	*push_in = TRUE;
}

static void popup_menu (KibaGMenu *gmenu,
			guint      n_button,
			guint32    activate_time) {
	if (!gmenu->menu) {
		gmenu->menu = create_menu ();
		g_object_set_data (G_OBJECT (gmenu->menu), "menu_panel", gmenu);
		gtk_menu_set_screen (GTK_MENU (gmenu->menu), KIBA_GDK_SCREEN);
		g_signal_connect (gmenu->menu, "show", G_CALLBACK (main_menu_append), gmenu);
	}

	gtk_menu_set_screen (GTK_MENU (gmenu->menu), KIBA_GDK_SCREEN);

	gtk_menu_popup (GTK_MENU (gmenu->menu),
			NULL,
			NULL,
			(GtkMenuPositionFunc) position_menu,
			gmenu,
			n_button,
			activate_time);
}

/**
KibaObject Callbacks
**/

static gboolean kiba_object_button_press_cb (GtkWidget      *widget,
					     GdkEventButton *event,
					     gpointer        data) {
	KibaPlugin *plugin = data;
	KibaGMenu *gmenu;
	GtkWidget *menu;
	GtkWidget *item, *image;
	
	gmenu = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	menu = gtk_menu_new ();
	
	item = gtk_menu_item_new_with_label (_("Edit Menus"));
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
	g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (edit_menu), gmenu);
	gtk_widget_show (item);
	
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
		event->button, event->time);
	
	return TRUE;
}

static gboolean kiba_object_button_release_cb (GtkWidget      *widget,
					       GdkEventButton *event,
					       gpointer        data) {
	KibaGMenu *gmenu = data;
	
	popup_menu (gmenu, event->button, event->time);
	
	return FALSE;
}

/**
KibaPrefs Callbacks
**/

static void dock_object_size_cb (const char *option_id,
			         gpointer    value,
			         gpointer    data) {
	KibaGMenu *gmenu = data;
	
	kiba_gmenu_update_icon (gmenu);
}

static void object_size_factor_cb (const char *option_id,
			           gpointer    value,
			           gpointer    data) {
	KibaGMenu *gmenu = data;
	float factor;
	
	factor = g_ascii_strtod((char*) value,NULL);
	
	kiba_object_set_requisition_factor (KIBA_OBJECT(gmenu->icon), factor, factor);
	kiba_gmenu_update_icon (gmenu);
}

static void object_alpha_cb (const char *option_id,
			     gpointer    value,
			     gpointer    data) {
	KibaGMenu *gmenu = data;
	float alpha;
	
	alpha = g_ascii_strtod((char*) value,NULL);
	
	kiba_drawable_object_set_alpha (KIBA_DRAWABLE_OBJECT(gmenu->icon), alpha);
}

/*** DBus methods ***/

static gboolean gmenu_dbus_popup_menu (KibaPlugin *plugin,
                                       GError **error) {
	KibaGMenu *gmenu;
	
	gmenu = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	if (!gmenu->menu) {
		gmenu->menu = create_menu ();
		g_object_set_data (G_OBJECT (gmenu->menu), "menu_panel", gmenu);
		gtk_menu_set_screen (GTK_MENU (gmenu->menu), KIBA_GDK_SCREEN);
		g_signal_connect (gmenu->menu, "show", G_CALLBACK (main_menu_append), gmenu);
	}
	
	gtk_menu_set_screen (GTK_MENU (gmenu->menu), KIBA_GDK_SCREEN);
	
	gtk_menu_popup (GTK_MENU (gmenu->menu),
			NULL,
			NULL,
			NULL,
			gmenu,
			3,
			CurrentTime);
	
	return TRUE;
}

/*** Plugin methods ***/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaGMenu *gmenu;
	
	gmenu = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_remove_callback_by_data (file, gmenu);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(gtk_icon_theme_get_default()),
		G_CALLBACK(icon_theme_changed_cb), gmenu);
	
	if (GTK_IS_WIDGET(gmenu->menu))
		gtk_widget_destroy (gmenu->menu);
	
	kiba_object_destroy (KIBA_OBJECT(gmenu->icon));
	
	g_free (gmenu);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaGMenu *gmenu;
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");	
#endif
	
	if (!gnome_vfs_initialized () && !gnome_vfs_init ()) {
		DEBUG_ERROR ("Can not initialize gnome-vfs.\n");
		return FALSE;
	}
	
	gmenu = g_new0 (KibaGMenu, 1);
	
	/* add icon to the dock */
	gmenu->icon = kiba_icon_new (PLUGIN_NAME);
	kiba_add_plugin_object(kiba, plugin, KIBA_OBJECT(gmenu->icon));
	gtk_widget_show (gmenu->icon);
	kiba_object_set_title (KIBA_OBJECT(gmenu->icon), _("Main Menu"));
	
	gmenu->recent_manager = gtk_recent_manager_get_default ();
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) gmenu);
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	
	/* add object signal handler */
	g_signal_connect (G_OBJECT (gmenu->icon), "button-press-event",
			G_CALLBACK (kiba_object_button_press_cb), plugin);
	g_signal_connect (G_OBJECT (gmenu->icon), "button-release-event",
			G_CALLBACK (kiba_object_button_release_cb), gmenu);
	
	g_signal_connect (G_OBJECT (gtk_icon_theme_get_default()), "changed",
			G_CALLBACK (icon_theme_changed_cb), gmenu);
	
	kiba_prefs_add_callback (file, "dock_object_size",
		dock_object_size_cb, gmenu, FALSE);
	kiba_prefs_add_callback (file, "gmenu_size_factor",
		object_size_factor_cb, gmenu, TRUE);
	kiba_prefs_add_callback (file, "gmenu_alpha",
		object_alpha_cb, gmenu, TRUE);
	
	if (!volume_monitor)
		volume_monitor = gnome_vfs_get_volume_monitor ();
	
	return TRUE;
}

static PrefMember gmenu_options[] = {
	{ "gmenu_alpha", "1.0", KIBA_DOUBLE },
	{ "gmenu_size_factor", "1.0", KIBA_DOUBLE }
};

static KibaPluginVtable vtable = {
	"GNOME-Menu Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
};

#include "gmenu-dbus-glue.h"

KIBA_INIT_PLUGIN (vtable, gmenu_options, &dbus_glib_gmenu_dbus_object_info, NULL, NULL);
