/*
 *  Copyright (C) 2008 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <string.h>
#include <gtk/gtk.h>
#include <libintl.h>

#include <gconf/gconf-client.h>
#include <bonobo-activation/bonobo-activation.h>
#include <bonobo/bonobo-ui-component.h>
#include <bonobo/bonobo-exception.h>
#include <bonobo/bonobo-control-frame.h>
#include <bonobo/bonobo-widget.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-container-object.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-dialog.h>

#include "gnome-applet-chooser.h"

#define _(string)  dgettext (GETTEXT_PACKAGE, string)

#define KIBA_GNOME_APPLET_GCONF_DIR "/apps/kiba-gnome-applet"

/* Gnome panel sizes. */
#define GNOME_PANEL_SIZE_XX_SMALL       12
#define GNOME_PANEL_SIZE_X_SMALL        24
#define GNOME_PANEL_SIZE_SMALL          36
#define GNOME_PANEL_SIZE_MEDIUM         48
#define GNOME_PANEL_SIZE_LARGE          64
#define GNOME_PANEL_SIZE_X_LARGE        80
#define GNOME_PANEL_SIZE_XX_LARGE       128

/* Gnome applet orientations */
#define GNOME_APPLET_ORIENT_UP		0
#define GNOME_APPLET_ORIENT_DOWN	1
#define GNOME_APPLET_ORIENT_LEFT	2
#define GNOME_APPLET_ORIENT_RIGHT	3

#define APPLET_SIZE 28

typedef struct {
	char *name;
	char *iid;
}GnomeAppletData;

typedef struct {
	KibaPlugin *plugin;
	
	GList *applets;
	
	KibaAppletChooserDialog *chooser;
}KibaGnomeAppletManager;

typedef struct {
	KibaGnomeAppletManager *manager;
	gchar *kid;
	gchar *iid;
	gchar *name;
	gchar *gconf_key;
	CORBA_Object object;
	BonoboUIComponent *uic;
	Bonobo_PropertyBag prop_bag;
	KibaObject *kiba_object;
	GtkWidget *applet_widget;
}KibaGnomeApplet;

static PrefMember __hidden_options[] = {
	{ "gnome_applets_enabled_list", "[]", KIBA_LIST },
};

static gchar* __find_unique_key (gchar *applet_name) {
	GConfClient	*client;
	gchar		*key = NULL;
	gchar		*name;
	gchar		*gconfName;
	int		 i = 0;

	client = gconf_client_get_default ();

	do {
		g_free (key);
		key = g_strdup_printf (KIBA_GNOME_APPLET_GCONF_DIR "/applet_%d", i++);
		
		name = g_strdup_printf ("%s/name", key);
		gconfName = gconf_client_get_string (client, name, NULL);
		g_free (name);
		if (!gconfName) continue;
		
		if (strcmp(gconfName, applet_name) == 0) {
			g_object_unref (client);
			g_free(gconfName);
			return key;
		}
	} while (gconf_client_dir_exists (client, key, NULL));

	name = g_strdup_printf ("%s/name", key);
	gconf_client_set_string (client, name, applet_name, NULL);
	g_free (name);

	g_object_unref (client);

	return key;
}

static gchar* __find_unique_kid (KibaGnomeApplet *applet) {
	gchar *file = NULL;
	gchar *homedir;
	gint i = 0;
	
	homedir = g_strdup_printf ("%s/.kiba-dock/gnome-applets", g_get_home_dir ());
	
	do {
		g_free (file);
		file = g_strdup_printf ("%s/applet_%d", homedir, i++);
	} while (g_file_test (file, G_FILE_TEST_EXISTS));
	
	return file;
}

static void __applet_to_key_file (KibaGnomeApplet *applet) {
	GKeyFile *keyfile = NULL;
	static const char desktop_entry_group[] = "Desktop Entry";
	char *filePath;
	char *data;
	gsize length;
	gssize slength;
	
	if (applet->kid) {
		filePath = g_strdup_printf ("%s/.kiba-dock/gnome-applets/%s", g_get_home_dir (), applet->kid);
		keyfile = g_key_file_new();
		g_key_file_load_from_file (keyfile, filePath, G_KEY_FILE_KEEP_COMMENTS, NULL);
	} else {
		filePath = __find_unique_kid(applet);
		applet->kid = g_strdup(g_strrstr (filePath, "/")+1);
		keyfile = g_key_file_new();
	}
	
	g_key_file_set_string (keyfile, desktop_entry_group, "Name", applet->name);
	g_key_file_set_string (keyfile, desktop_entry_group, "IID", applet->iid);
	g_key_file_set_string (keyfile, desktop_entry_group, "GconfPath", applet->gconf_key);
	
	data = g_key_file_to_data (keyfile, &length, NULL);
	g_key_file_free(keyfile);
	if (data) {
		slength = (gssize) length;
		g_file_set_contents (filePath, data, slength, NULL);
		g_free(data);
	}
	g_free(filePath);
}

static char* __applet_list_to_string(GList *list) {
	KibaGnomeApplet *applet;
	GList *l;
	char *ret;
	char *buf;
	
	if (!list) return g_strdup("[]");
	
	applet = list->data;
	ret = g_strdup_printf("[%s", applet->kid);
	for (l = list->next; l != NULL; l = l->next) {
		applet = l->data;
		
		buf = ret;
		ret = g_strdup_printf("%s,%s", buf, applet->kid);
		g_free (buf);
	}
	buf = ret;
	ret = g_strdup_printf("%s]", buf);
	g_free (buf);
	
	return ret;
}

static void __save_kiba_prefs (KibaGnomeAppletManager *manager) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	char *appletList;
	
	file = kiba_get_plugin_config_file(kiba, manager->plugin);
	
	appletList = __applet_list_to_string(manager->applets);
	kiba_prefs_write_to_config_file(file, "gnome_applets_enabled_list", "list", appletList);
	g_free(appletList);
}

static const gchar* __get_orient_string (KibaGnomeApplet *applet) {
	const gchar *ret = NULL;
	Kiba *kiba = kiba_get();
	KibaWin *win;
	KibaDock *dock;
	
	win = kiba_get_plugin_win(kiba, applet->manager->plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	switch (dock->d.position) {
		case KIBA_POSITION_TOP_EDGE:
			ret = "up";
			break;
		case KIBA_POSITION_BOTTOM_EDGE:
			ret = "down";
			break;
		case KIBA_POSITION_RIGHT_EDGE:
			ret = "right";
			break;
		case KIBA_POSITION_LEFT_EDGE:
			ret = "left";
			break;
		default:
			g_assert_not_reached ();
	}
	
	return ret;
}

static unsigned short __size_to_gnome_size (gint size) {
	unsigned short sz;
	
	sz = size <= GNOME_PANEL_SIZE_XX_SMALL ? GNOME_PANEL_SIZE_XX_SMALL :
	     size <= GNOME_PANEL_SIZE_X_SMALL  ? GNOME_PANEL_SIZE_X_SMALL  :
	     size <= GNOME_PANEL_SIZE_SMALL    ? GNOME_PANEL_SIZE_SMALL    :
	     size <= GNOME_PANEL_SIZE_MEDIUM   ? GNOME_PANEL_SIZE_MEDIUM   :
	     size <= GNOME_PANEL_SIZE_LARGE    ? GNOME_PANEL_SIZE_LARGE    :
	     size <= GNOME_PANEL_SIZE_X_LARGE  ? GNOME_PANEL_SIZE_X_LARGE  : GNOME_PANEL_SIZE_XX_LARGE;

	return sz;
}

static const gchar* __get_size_string (KibaGnomeApplet *applet) {
	gint size;
	const gchar	*retval;
	Kiba *kiba = kiba_get();
	KibaWin *win;
	KibaDock *dock;
	
	win = kiba_get_plugin_win(kiba, applet->manager->plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	size = APPLET_SIZE;
	size = __size_to_gnome_size (size);

	switch (size) {
		case GNOME_PANEL_SIZE_XX_SMALL:
			retval = "xx-small";
			break;
		case GNOME_PANEL_SIZE_X_SMALL:
			retval = "x-small";
			break;
		case GNOME_PANEL_SIZE_SMALL:
			retval = "small";
			break;
		case GNOME_PANEL_SIZE_MEDIUM:
			retval = "medium";
			break;
		case GNOME_PANEL_SIZE_LARGE:
			retval = "large";
			break;
		case GNOME_PANEL_SIZE_X_LARGE:
			retval = "x-large";
			break;
		default:
			retval = "xx-large";
			break;
	}
	
	return retval;
}

static gchar* __construct_moniker (KibaGnomeApplet *applet) {
	gchar *moniker;
	
	/* We (still) do not support "background" and "locked_down" options. */
	moniker = g_strdup_printf ("%s!prefs_key=%s/prefs;orient=%s;size=%s",
				   applet->iid, applet->gconf_key,
				   __get_orient_string (applet),
				   __get_size_string (applet));
	
	return moniker;
}

static unsigned short __screen_position_to_gnome_applet_orientation (KibaPosition position) {
	switch (position) {
		case KIBA_POSITION_TOP_EDGE:   /* top */
			return GNOME_APPLET_ORIENT_DOWN;
		case KIBA_POSITION_BOTTOM_EDGE:   /* bottom */
			return GNOME_APPLET_ORIENT_UP;
		case KIBA_POSITION_RIGHT_EDGE:
			return GNOME_APPLET_ORIENT_RIGHT;
		case KIBA_POSITION_LEFT_EDGE:
			return GNOME_APPLET_ORIENT_LEFT;
	}
	
	return GNOME_APPLET_ORIENT_UP;
}

static gboolean __update_applet_size (KibaGnomeApplet *applet, int size) {
	Kiba *kiba = kiba_get();
	KibaDock *dock;
	KibaWin *win;
	
	win = kiba_get_plugin_win(kiba, applet->manager->plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	gint sz = __size_to_gnome_size (size);
	bonobo_pbclient_set_short (applet->prop_bag, "panel-applet-size", sz, NULL);
	
	switch (dock->d.position) {
		case KIBA_POSITION_TOP_EDGE:
		case KIBA_POSITION_BOTTOM_EDGE:
			gtk_widget_set_size_request (GTK_WIDGET (applet->applet_widget), -1, size);
			break;
		case KIBA_POSITION_RIGHT_EDGE:
		case KIBA_POSITION_LEFT_EDGE:
			gtk_widget_set_size_request (GTK_WIDGET (applet->applet_widget), size, -1);
			break;
		default:
			g_assert_not_reached ();
	}
}

static void __dock_position_changed (KibaDock *dock,
                                     KibaPosition position,
                                     gpointer data) {
	unsigned short orientation;
	KibaGnomeAppletManager *manager = (KibaGnomeAppletManager*) data;
	GList *l;
	
	orientation = __screen_position_to_gnome_applet_orientation (position);
	
	for (l = manager->applets; l != NULL; l = l->next) {
		KibaGnomeApplet *applet = l->data;
		
		/*
		 * Workaround: Xfce panel makes all items in all panels insensitive when
		 * moving items between panels. Some Gnome applets (noticeably the Mixer
		 * applet) misbehave when the following happens:
		 * 1. Applet is set insensitive;
		 * 2. Applet changes orientation;
		 * 3. Applet is set sensitive again.
		 * The following makes sure the applet is set sensitive before it
		 * changes orientation.
		 */
// 		gtk_widget_set_sensitive (GTK_WIDGET(), TRUE);
		
		bonobo_pbclient_set_short (applet->prop_bag, "panel-applet-orient", orientation, NULL);
		gtk_widget_queue_resize (GTK_WIDGET (applet->kiba_object));
		__update_applet_size(applet, APPLET_SIZE);
	}
}

static void __free_applet (KibaGnomeApplet *applet) {
	applet->manager->applets = g_list_remove(applet->manager->applets, applet);
	g_free(applet->iid);
	if (applet->kid)
		g_free(applet->kid);
	g_free(applet->name);
	g_free(applet->gconf_key);
	g_free(applet);
}

static void __remove_applet_configuration (KibaGnomeApplet *applet) {
	GConfClient *client;
	char *filePath;
	
	if (applet->kid) {
		filePath = g_strdup_printf ("%s/.kiba-dock/gnome-applets/%s", g_get_home_dir (), applet->kid);
		if (g_file_test(filePath, G_FILE_TEST_EXISTS)) {
			g_remove(filePath);
		}
		g_free(filePath);
	}
	
	if (applet->gconf_key) {
		client = gconf_client_get_default ();
		gconf_client_recursive_unset (client, applet->gconf_key, GCONF_UNSET_INCLUDING_SCHEMA_NAMES, NULL);
		g_object_unref (client);
	}
}

static void __connection_broken (ORBitConnection *conn,
                                 KibaGnomeApplet *applet) {
	KibaGnomeAppletManager *manager = applet->manager;
	char *title;
	
	title = g_strdup_printf(_("'%s' has quit unexpectedly."), applet->name);
	kiba_dialog_error(title, _("This applet will be removed from the plugin."));
	g_free(title);
	
	__remove_applet_configuration(applet);
	__free_applet(applet);
	__save_kiba_prefs(manager);
}

static void __menu_item_activated (BonoboUIComponent *uic, gpointer mi, const char *cname) {
	gtk_menu_item_activate (GTK_MENU_ITEM(mi));
}

static void __remove_item_activated (GtkWidget *widget,
                                     KibaGnomeApplet *applet) {
	KibaGnomeAppletManager *manager = applet->manager;
	
	__remove_applet_configuration(applet);
	
	if (applet->prop_bag) {
		bonobo_object_release_unref (applet->prop_bag, NULL);
		applet->prop_bag = CORBA_OBJECT_NIL;
	}
	
	if (applet->uic) {
		bonobo_object_unref (BONOBO_OBJECT (applet->uic));
		applet->uic = NULL;
	}
	
	if (applet->object) {
		ORBit_small_unlisten_for_broken (applet->object, G_CALLBACK (__connection_broken));
		CORBA_Object_release (applet->object, NULL);
		applet->object = CORBA_OBJECT_NIL;
	}
	
	if (KIBA_IS_OBJECT(applet->kiba_object))
		kiba_object_destroy(applet->kiba_object);
	
	__free_applet(applet);
	__save_kiba_prefs(manager);
}

static void __setup_menu_items (KibaGnomeApplet *applet, BonoboUIComponent *uic) {
	GtkWidget *item;
	
	item = gtk_image_menu_item_new_from_stock(GTK_STOCK_REMOVE, NULL);
	g_signal_connect(G_OBJECT(item), "activate",
		G_CALLBACK(__remove_item_activated), applet);
	bonobo_ui_component_add_verb (uic, "Remove", __menu_item_activated, item);
}

static void __applet_activated (Bonobo_Unknown object,
                                CORBA_Environment *ev,
                                gpointer data) {
	GtkWidget *container_object;
	GtkWidget *bw, *child = NULL;
	CORBA_Object control;
	CORBA_Environment corba_ev;
	BonoboControlFrame *frame;
	BonoboUIComponent *uic;
	Bonobo_PropertyBag  prop_bag;
	KibaGnomeApplet *applet = (KibaGnomeApplet*) data;
	Kiba *kiba = kiba_get();
	KibaWin *win;
	KibaDock *dock;
	gchar *error;
	
	win = kiba_get_plugin_win(kiba, applet->manager->plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	if (BONOBO_EX (ev) || object == CORBA_OBJECT_NIL) {
		error = bonobo_exception_get_text (ev);
		CORBA_exception_free (ev);
		g_warning ("Failed to load applet '%s' (can't get CORBA object): %s\n", applet->iid, error);
		__free_applet(applet);
		g_free (error);
		return;
	}
	
	control = CORBA_Object_duplicate (object, NULL);
	bw = bonobo_widget_new_control_from_objref (object, CORBA_OBJECT_NIL);
	bonobo_object_release_unref (object, NULL);
	if (!bw) {
		g_warning ("Failed to load applet '%s' (can't get bonobo widget)\n", applet->iid);
		__free_applet (applet);
		return;
	}
	
	frame = bonobo_widget_get_control_frame (BONOBO_WIDGET (bw));
	if (!frame) {
		g_warning ("Failed to load applet '%s' (can't get control frame)\n", applet->iid);
		gtk_object_sink (GTK_OBJECT (bw));
		__free_applet (applet);
		return;
	}
	
	CORBA_exception_init (&corba_ev);
	prop_bag = bonobo_control_frame_get_control_property_bag (frame, &corba_ev);
	if (prop_bag == NULL || BONOBO_EX (&corba_ev)) {
		error = bonobo_exception_get_text (&corba_ev);
		CORBA_exception_free (&corba_ev);
		g_warning ("Failed to load applet '%s' (can't get property bag): %s\n", applet->iid, error);
		gtk_object_sink (GTK_OBJECT (bw));
		__free_applet (applet);
		g_free (error);
		return;
	}
	
	uic = bonobo_control_frame_get_popup_component (frame, &corba_ev);
	if (uic == NULL || BONOBO_EX (&corba_ev)) {
		error = bonobo_exception_get_text (&corba_ev);
		CORBA_exception_free (&corba_ev);
		g_warning ("Failed to load applet '%s' (can't get popup component): %s\n", applet->iid, error);
		gtk_object_sink (GTK_OBJECT (bw));
		__free_applet (applet);
		g_free (error);
		return;
	}
	
	/* load custom menu item(s) */
	bonobo_ui_component_freeze (uic, CORBA_OBJECT_NIL);
	bonobo_ui_util_set_ui (uic, DATADIR "/kiba-dock/ui", "KIBA_Panel_Popup.xml", "kiba-gnome-applet-plugin", &corba_ev);
	if (BONOBO_EX (&corba_ev)) {
		error = bonobo_exception_get_text (&corba_ev);
		CORBA_exception_free (&corba_ev);
		g_warning ("Failed to load applet '%s' (can't set ui): %s\n", applet->iid, error);
		gtk_object_sink (GTK_OBJECT (bw));
		__free_applet (applet);
		g_free (error);
		return;
	}
	__setup_menu_items (applet, uic);
	bonobo_ui_component_thaw (uic, CORBA_OBJECT_NIL);

	gtk_widget_show (bw);
	
	applet->object = control;
	applet->uic = uic;
	applet->prop_bag = prop_bag;
	applet->applet_widget = bw;
	ORBit_small_listen_for_broken (object, G_CALLBACK (__connection_broken), applet);
	
	__update_applet_size(applet, APPLET_SIZE);
	
	/* set up the object on the dock */
	container_object = kiba_container_object_new(applet->name);
	gtk_container_add (GTK_CONTAINER(container_object), bw);
	kiba_add_plugin_object(kiba, applet->manager->plugin, KIBA_OBJECT(container_object));
	gtk_widget_show(container_object);
	
	applet->kiba_object = KIBA_OBJECT(container_object);
}

static KibaGnomeApplet* __setup_applet_object (KibaGnomeAppletManager *manager,
                                               char *iid,
                                               char *name,
                                               char *gconf_key) {
	KibaGnomeApplet *applet;
	CORBA_Environment ev;
	gchar *moniker;
	
	applet = g_new0 (KibaGnomeApplet, 1);
	applet->manager = manager;
	applet->kid = NULL;
	applet->iid = iid;
	applet->name = name;
	applet->gconf_key = gconf_key;
	
	CORBA_exception_init (&ev);
	moniker = __construct_moniker (applet);
	bonobo_get_object_async (moniker, "IDL:Bonobo/Control:1.0", &ev, __applet_activated, applet);
	g_free (moniker);
	
	manager->applets = g_list_append(manager->applets, applet);
	
	return applet;
}

static KibaGnomeApplet* __setup_applet_object_from_applet_info (KibaGnomeAppletManager *manager,
                                                                GnomeAppletInfo *applet,
                                                                char *gconf_key) {
	return __setup_applet_object(manager, g_strdup(applet->iid), g_strdup(applet->name), gconf_key);
}

static gboolean __populate_dialog_response_cb (GtkWidget *dialog,
                                               gint response,
                                               gpointer data) {
	KibaGnomeAppletManager *manager = data;
	GtkTreeSelection *sel;
	GtkTreeModel *model;
	GtkTreeIter iter;
	GnomeAppletInfo *applet;
	
	if (response == GTK_RESPONSE_OK) {
		sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (manager->chooser->tv));
		if (gtk_tree_selection_get_selected (sel, &model, &iter)) {
			KibaGnomeApplet *kiba_applet;
			gchar *gconf_key;
			
			gtk_tree_model_get (model, &iter, 0, &applet, -1);
			
			gconf_key = __find_unique_key (applet->name);
			kiba_applet = __setup_applet_object_from_applet_info(manager, applet, gconf_key);
			/* save the applet keyfile */
			__applet_to_key_file(kiba_applet);
			/* save new applet to kiba-pres */
			__save_kiba_prefs(manager);
		}
		
		gtk_widget_destroy(dialog);
		return TRUE;
	}
	
	return FALSE;
}

static void __populate_dialog_destroy_cb (GtkWidget *widget,
                                          gpointer data) {
	KibaGnomeAppletManager *manager = data;
	manager->chooser = NULL;
}

static void __populate_dock (KibaDock *dock,
                             gpointer  data) {
	KibaGnomeAppletManager *manager = data;
	GtkWidget *dialog;
	char *argv[] = { "kiba-dock", };
	int argc = 1;
	
	if (manager->chooser) {
		gtk_window_present(GTK_WINDOW(manager->chooser->dialog));
		return;
	}
	
	bonobo_ui_init (argv[0], VERSION, &argc, argv);
	
	manager->chooser = kiba_applet_chooser_dialog_new ();
	gtk_widget_show_all (manager->chooser->dialog);
	
	g_signal_connect (G_OBJECT(manager->chooser->dialog), "response",
		G_CALLBACK (__populate_dialog_response_cb), manager);
	g_signal_connect (G_OBJECT(manager->chooser->dialog), "destroy",
		G_CALLBACK (__populate_dialog_destroy_cb), manager);
}

/*** Plugin methods ***/

static void plugin_unload (KibaPlugin *plugin) {
	KibaGnomeAppletManager *manager;
	Kiba *kiba = kiba_get();
	KibaWin *win;
	KibaDock *dock;
	GList *l;
	GSList *sl;
	
	manager = g_object_get_data(G_OBJECT(plugin), "gnome-applets");
	g_object_set_data(G_OBJECT(plugin), "gnome-applets", NULL);
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	kiba_dock_remove_populate_func (dock,
		(KibaDockPopulateFunc) __populate_dock);
	
	if (manager->chooser) {
		gtk_widget_destroy(manager->chooser->dialog);
	}
	
	/* unload applets */
	for (l = manager->applets; l != NULL; l = manager->applets) {
		KibaGnomeApplet *applet = l->data;
		
		if (applet->prop_bag) {
			bonobo_object_release_unref (applet->prop_bag, NULL);
			applet->prop_bag = CORBA_OBJECT_NIL;
		}
	
		if (applet->uic) {
			bonobo_object_unref (BONOBO_OBJECT (applet->uic));
			applet->uic = NULL;
		}
		
		if (applet->object) {
			ORBit_small_unlisten_for_broken (applet->object, G_CALLBACK (__connection_broken));
			CORBA_Object_release (applet->object, NULL);
			applet->object = CORBA_OBJECT_NIL;
		}
		
		if (KIBA_IS_OBJECT(applet->kiba_object))
			kiba_object_destroy(applet->kiba_object);
		
		__free_applet(applet);
	}
	
	g_free(manager);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	static const char desktop_entry_group[] = "Desktop Entry";
	KibaGnomeAppletManager *manager;
	Kiba *kiba = kiba_get();
	KibaWin *win;
	KibaDock *dock;
	char *homedir;
	GSList *availableApplets, *l;
	GSList *enabledApplets;
	KibaPrefsFile *file;
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
#endif
	
	homedir = g_strdup_printf ("%s/.kiba-dock/gnome-applets", g_get_home_dir());
	if (!g_file_test(homedir, G_FILE_TEST_EXISTS)) {
		if (g_mkdir (homedir, 7664) != 0) {
			DEBUG_ERROR ("Failed to create applet directory at '%s'!\n", homedir);
			return FALSE;
		}
	}
	
	manager = g_new0(KibaGnomeAppletManager, 1);
	manager->applets = NULL;
	manager->chooser = NULL;
	manager->plugin = plugin;
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	file = kiba_get_plugin_config_file(kiba, plugin);
	
	kiba_prefs_add_options (file, __hidden_options, G_N_ELEMENTS(__hidden_options));
	
	/* loading enabled applets */
	enabledApplets = kiba_prefs_get_list (file, "gnome_applets_enabled_list", NULL);
	if (enabledApplets != NULL) {
		char *argv[] = { "kiba-dock", };
		int argc = 1;
		bonobo_ui_init (argv[0], VERSION, &argc, argv);
	}
	for (l = enabledApplets; l != NULL; l = l->next) {
		KibaGnomeApplet *applet;
		GKeyFile *keyfile = NULL;
		const gchar *kid = l->data;
		gchar *filePath;
		gchar *name, *iid, *gconf_key;
		
		filePath = g_strdup_printf ("%s/.kiba-dock/gnome-applets/%s", g_get_home_dir (), kid);
		if (!g_file_test(filePath, G_FILE_TEST_EXISTS)) {
			g_free(filePath);
			continue;
		}
		
		keyfile = g_key_file_new();
		g_key_file_load_from_file (keyfile, filePath, G_KEY_FILE_NONE, NULL);
		if (!keyfile) continue;
		
		name = g_key_file_get_string (keyfile, desktop_entry_group, "Name", NULL);
		if (!name) {
			g_key_file_free(keyfile);
			continue;
		}
		iid = g_key_file_get_string (keyfile, desktop_entry_group, "IID", NULL);
		if (!iid) {
			g_free(name);
			g_key_file_free(keyfile);
			continue;
		}
		gconf_key = g_key_file_get_string (keyfile, desktop_entry_group, "GconfPath", NULL);
		if (!gconf_key) {
			g_free(iid);
			g_free(name);
			g_key_file_free(keyfile);
			continue;
		}
		
		//TODO check if the gconf path is valid
		
		applet = __setup_applet_object (manager, iid, name, gconf_key);
		applet->kid = g_strdup(kid);
	}
	
	/* saving the kid list to kiba prefs, it may changed */
	__save_kiba_prefs(manager);
	
	g_signal_connect(G_OBJECT(dock), "position-changed",
		G_CALLBACK(__dock_position_changed), manager);
	
	kiba_dock_add_populate_func (dock,
		(KibaDockPopulateFunc) __populate_dock,
		"gnome-logo-icon-transparent",
		_("Add gnome applet"),
		(gpointer) manager);
	
	g_object_set_data(G_OBJECT(plugin), "gnome-applets", manager);
	
	return TRUE;
}

static PrefMember plugin_options[] = {
};

static KibaPluginVtable vtable = {
	"Display Gnome applets in kiba-dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
};

KIBA_INIT_PLUGIN (vtable, plugin_options, NULL, NULL, NULL);
