/*
 *  Copyright (C) 2007  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <libintl.h>
#include <math.h>
#include <gtk/gtk.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-drawable-object.h>
#include <kiba-fixed-object.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>

#define _(string)   gettext (string)
#define PLUGIN_NAME "pulse"

#define OBJECT_SIZE(obj) MAX(GTK_WIDGET(obj)->allocation.width, GTK_WIDGET(obj)->allocation.height)

typedef struct _KibaPulse       KibaPulse;
typedef struct _KibaPulseObject KibaPulseObject;

struct _KibaPulseObject {
	KibaPulse  *pulse;
	GtkWidget  *widget;
	float      factor;
	float      max_pulse_factor;
	gboolean    entered;
	gboolean    urgent;
	gint        pulse_id;
};

typedef enum {
	KIBA_PULSE_URGENT = 0,
	KIBA_PULSE_HOVERING,
	KIBA_PULSE_TRIGGERING
}KibaPulseFlags;

struct _KibaPulse {
	GList *objects;
	KibaPulseFlags flags;
	
	float factor;
	float exec_factor;
	gint speed;
};

static KibaPulseObject* kiba_pulse_get_object_for_widget (KibaPulse  *pulse,
							GtkWidget *widget) {
	GList *l;
	
	for (l = pulse->objects; l != NULL; l = l->next) {
		KibaPulseObject *object = l->data;
		
		if (object->widget == widget)
			return object;
	}
	
	return NULL;
}

static void kiba_pulse_object_queue_draw (KibaPulseObject *object) {
	GdkRectangle area;
	KibaWin *win;
	GtkWidget *dock;
	const KibaLayoutData *d;
	gint width, height;
	GList *l;
	KibaObjectCustomOffsetData *data;
	float offsetX = 0.0, offsetY = 0.0;
	float rotateAngle = 0.0;
	
	dock = object->widget->parent;
	win = KIBA_WIN(dock->parent);
	
	d = &kiba_layout_data[KIBA_DOCK(KIBA_WIN_CHILD(win))->d.position];
	
	for (l = KIBA_OBJECT(object->widget)->custom_offset_data; l != NULL; l = l->next) {
		float customX = 0.0, customY = 0.0;
		data = l->data;
		
		data->place_func (KIBA_OBJECT(object->widget), &customX, &customY, data->data);
		offsetX += customX;
		offsetY += customY;
	}
	
	area.width = object->factor*object->widget->allocation.width;
	area.height = object->factor*object->widget->allocation.height;
	for (l = KIBA_OBJECT(object->widget)->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		float custom_factor_x = 1.0;
		float custom_factor_y = 1.0;
		
		if (data->scale_widget)
			continue;
		
		data->scale_func (KIBA_OBJECT(object->widget),
			&custom_factor_x,
			&custom_factor_y,
			data->data);
		
		area.width *= custom_factor_x;
		area.height *= custom_factor_y;
	}
	
	area.x = object->widget->allocation.x - (area.width-object->widget->allocation.width)/2;
	area.y = object->widget->allocation.y - (area.height-object->widget->allocation.height)/2 ;
	
	area.y -= (d->valignment-1)*(offsetY+1);
	area.x -= (d->halignment-1)*(offsetX+1);
	
	for (l = KIBA_DRAWABLE_OBJECT(object->widget)->custom_rotate_funcs; l != NULL; l = l->next) {
		float customAngle = 0.0;
		KibaDrawableObjectCustomRotateData *data = l->data;
		
		data->func (KIBA_DRAWABLE_OBJECT(object->widget), &customAngle, data->data);
		rotateAngle += customAngle;
	}
	
	width = area.width;
	height = area.height;
	if (rotateAngle != 0.0) {
		area.height = sqrt (height * height + width  * width);
		area.width = area.height;
		area.y -= (area.height - height)/2;
		area.x -= (area.width - width)/2;
	}
	
	kiba_win_invalidate_area (GTK_WIDGET(win), area);
}

static gint pulse_object_timeout (gpointer data) {
	KibaPulseObject *object = data;
	
	object->factor += 0.02 + 0.02*object->pulse->speed/50.0;
	
	kiba_pulse_object_queue_draw (object);
	
	if (object->factor >= object->max_pulse_factor) {
		gboolean stop_pulse = FALSE;
		gboolean urgent;
		gboolean hovering;
		
		hovering = (GET_FLAG(object->pulse->flags, KIBA_PULSE_HOVERING) && object->entered);
		if (!hovering) {
			urgent = (GET_FLAG(object->pulse->flags, KIBA_PULSE_URGENT) && object->urgent);
			if (!urgent) {
				stop_pulse = TRUE;
			}
		}
		
		object->factor = 1.0;
		
		if (object->max_pulse_factor == object->pulse->exec_factor)
			object->max_pulse_factor = object->pulse->factor;
		
		if (stop_pulse || GET_FLAG(KIBA_OBJECT(object->widget)->state, KIBA_OBJECT_DRAGGING)) {
			object->pulse_id = 0;
			return FALSE;
		}
	}
	
	return TRUE;
}

void kiba_object_start_pulsing (KibaPulseObject *object,
				float           factor) {
	if (factor <= 1.0)
		return;
	
	if (GET_FLAG(KIBA_OBJECT(object->widget)->state, KIBA_OBJECT_DRAGGING))
		return;
	
	object->max_pulse_factor = factor;
	if (!object->entered)
		object->entered = TRUE;
	
	if (!object->pulse_id)
		object->pulse_id = g_timeout_add (20, pulse_object_timeout, object);
}

void kiba_object_stop_pulsing (KibaPulseObject *object) {
	if (object->entered)
		object->entered = FALSE;
}

static void kiba_prefs_hovering_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaPulse *pulse = data;
	gboolean hovering = GPOINTER_TO_INT(value);
	
	if (hovering) {
		if (!GET_FLAG(pulse->flags, KIBA_PULSE_HOVERING))
			SET_FLAG(pulse->flags, KIBA_PULSE_HOVERING);
	}
	else if (GET_FLAG(pulse->flags, KIBA_PULSE_HOVERING))
		DEL_FLAG(pulse->flags, KIBA_PULSE_HOVERING);
}

static void kiba_prefs_triggering_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaPulse *pulse = data;
	gboolean triggering = GPOINTER_TO_INT(value);
	
	if (triggering) {
		if (!GET_FLAG(pulse->flags, KIBA_PULSE_TRIGGERING))
			SET_FLAG(pulse->flags, KIBA_PULSE_TRIGGERING);
	}
	else if (GET_FLAG(pulse->flags, KIBA_PULSE_TRIGGERING))
		DEL_FLAG(pulse->flags, KIBA_PULSE_TRIGGERING);
}

static void kiba_prefs_urgent_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaPulse *pulse = data;
	gboolean urgent = GPOINTER_TO_INT(value);
	
	if (urgent) {
		if (!GET_FLAG(pulse->flags, KIBA_PULSE_URGENT))
			SET_FLAG(pulse->flags, KIBA_PULSE_URGENT);
	}
	else if (GET_FLAG(pulse->flags, KIBA_PULSE_URGENT))
		DEL_FLAG(pulse->flags, KIBA_PULSE_URGENT);
}

static void kiba_prefs_pulse_factor_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaPulse *pulse = data;
	
	pulse->factor = g_ascii_strtod((char*) value, NULL);
}

static void kiba_prefs_pulse_exec_factor_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaPulse *pulse = data;
	
	pulse->exec_factor = g_ascii_strtod((char*) value, NULL);
}

static void kiba_prefs_pulse_step_cb (const char *option_id,
				     gpointer    value,
				     gpointer    data) {
	KibaPulse *pulse = data;
	
	pulse->speed = GPOINTER_TO_INT(value);
}

static void object_urgent_event_cb (KibaObject *object,
				    gboolean    urgent,
				    gpointer    data) {
	KibaPulseObject *pulse = data;
	
	if(urgent) {
		pulse->urgent = TRUE;
		if(GET_FLAG(pulse->pulse->flags, KIBA_PULSE_URGENT))
			kiba_object_start_pulsing(pulse, pulse->pulse->factor);
	} else {
		pulse->urgent = FALSE;
	}
}

static gboolean object_release_event_cb (GtkWidget      *widget,
					 GdkEventButton *event,
					 gpointer        data) {
	KibaPulseObject *object = data;
	
	if (GET_FLAG(object->pulse->flags, KIBA_PULSE_TRIGGERING))
		kiba_object_start_pulsing(object, object->pulse->exec_factor);
	
	return FALSE;
}

static gboolean object_enter_event_cb  (GtkWidget        *widget,
					GdkEventCrossing *event,
					gpointer          data) {
	KibaPulseObject *object = data;
	
	if (GET_FLAG(object->pulse->flags, KIBA_PULSE_HOVERING))
		kiba_object_start_pulsing(object, object->pulse->factor);
	
	return FALSE;
}

static gboolean object_leave_event_cb  (GtkWidget        *widget,
					GdkEventCrossing *event,
					gpointer          data) {
	KibaPulseObject *object = data;
	
	object->entered = FALSE;
	
	return FALSE;
}

void kiba_pulse_paint_object (KibaDrawableObject *drawable_object,
			      cairo_t    *cr,
			      gpointer    data) {
	KibaPulseObject *pulse_object = data;
	KibaObject *object = KIBA_OBJECT(drawable_object);
	cairo_surface_t *surface;
	float fade_factor;
	gint unscaled_width;
	gint unscaled_height;
	
	if (pulse_object->factor <= 1.0)
		return;
	
	unscaled_width  = (int) ((object->size.width) * object->scale_x);
	unscaled_height = (int) ((object->size.height) * object->scale_y);
	
	if (drawable_object->switch_surface)
		surface = drawable_object->switch_surface;
	else
		surface = drawable_object->surface_data->surface;
	
	g_return_if_fail (surface != NULL);
	
	fade_factor = 1.0 - (OBJECT_SIZE(object)*(pulse_object->factor-1.0)) /
		((pulse_object->max_pulse_factor * OBJECT_SIZE(object)) -
		OBJECT_SIZE(object) + 1); //+1 is to avoid division with 0
	
	cairo_translate (cr,
		-(unscaled_width*(pulse_object->factor-1.0))/2,
		-(unscaled_height*(pulse_object->factor-1.0))/2);
	cairo_scale (cr, pulse_object->factor, pulse_object->factor);
	cairo_set_source_surface (cr, surface, 0, 0);
	cairo_paint_with_alpha (cr, object->destroy_factor*drawable_object->alpha*fade_factor*drawable_object->scroll_alpha);
}

static void dock_object_add_cb (GtkContainer *container,
				GtkWidget    *widget,
				gpointer      data) {
	KibaPulse *pulse = data;
	KibaPulseObject *object;
	
	if (!KIBA_IS_DRAWABLE_OBJECT(widget))
		return;
	if (kiba_pulse_get_object_for_widget (pulse, widget) != NULL)
		return;
	
	object = g_new0 (KibaPulseObject, 1);
	object->widget = widget;
	object->factor = 1.0;
	object->pulse = pulse;
	object->pulse_id = 0;
	
	pulse->objects = g_list_append (pulse->objects, object);
	
	kiba_drawable_object_add_custom_paint_function (KIBA_DRAWABLE_OBJECT(widget),
		kiba_pulse_paint_object, FALSE, object);
	
	g_signal_connect (G_OBJECT(widget), "enter-notify-event",
		G_CALLBACK(object_enter_event_cb), object);
	g_signal_connect (G_OBJECT(widget), "leave-notify-event",
		G_CALLBACK(object_leave_event_cb), object);
	g_signal_connect (G_OBJECT(widget), "button-release-event",
		G_CALLBACK(object_release_event_cb), object);
	g_signal_connect (G_OBJECT(widget), "urgent-toggled",
		G_CALLBACK(object_urgent_event_cb), object);
}

static void dock_object_remove_cb (GtkContainer *container,
				   GtkWidget    *widget,
				   gpointer      data) {
	KibaPulse *pulse = data;
	KibaPulseObject *object;
	
	object = kiba_pulse_get_object_for_widget (pulse, widget);
	
	if (!object)
		return;
	
	kiba_drawable_object_remove_custom_paint_function_by_data (KIBA_DRAWABLE_OBJECT(widget), object);
	
	if (object->pulse_id)
		g_source_remove (object->pulse_id);
	
	pulse->objects = g_list_remove (pulse->objects, object);
	
	g_free (object);
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPulse *pulse;
	KibaDock *dock;
	KibaWin *win;
	GList *l;
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	pulse = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	for (l = pulse->objects; l != NULL; l = pulse->objects) {
		KibaPulseObject *object = l->data;
		
		if (object->pulse_id)
			g_source_remove (object->pulse_id);
		
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_enter_event_cb), object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_leave_event_cb), object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_release_event_cb), object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_urgent_event_cb), object);
		
		kiba_drawable_object_remove_custom_paint_function_by_data (KIBA_DRAWABLE_OBJECT(object->widget), object);
		
		g_free (object);
		pulse->objects = g_list_remove (pulse->objects, l->data);
	}
	g_list_free (pulse->objects);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_add_cb), pulse);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_remove_cb), pulse);
	
	g_free (pulse);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaPulse *pulse;
	KibaWin *win;
	KibaDock *dock;
	GtkContainer *container;
	GList *l;
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	container = GTK_CONTAINER(dock);
	
	pulse = g_new0(KibaPulse, 1);
	pulse->factor = 2.0;
	pulse->exec_factor = 1.0;
	pulse->speed = 30;
	pulse->objects = NULL;
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_add_callback (file, "pulse_hovering",
		kiba_prefs_hovering_cb, pulse, TRUE);
	kiba_prefs_add_callback (file, "pulse_triggering",
		kiba_prefs_triggering_cb, pulse, TRUE);
	kiba_prefs_add_callback (file, "pulse_urgent",
		kiba_prefs_urgent_cb, pulse, TRUE);
	kiba_prefs_add_callback (file, "pulse_factor",
		kiba_prefs_pulse_factor_cb, pulse, TRUE);
	kiba_prefs_add_callback (file, "pulse_exec_factor",
		kiba_prefs_pulse_exec_factor_cb, pulse, TRUE);
	kiba_prefs_add_callback (file, "pulse_speed",
		kiba_prefs_pulse_step_cb, pulse, TRUE);
	
	g_signal_connect (G_OBJECT(dock), "remove",
		G_CALLBACK(dock_object_remove_cb), pulse);
	g_signal_connect_after (G_OBJECT(dock), "add",
		G_CALLBACK(dock_object_add_cb), pulse);
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		
		dock_object_add_cb (container, child, pulse);
	}
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) pulse);
	
	return TRUE;
}

static PrefMember pulse_options[] = {
    { "pulse_hovering", "TRUE", KIBA_BOOLEAN },
    { "pulse_triggering", "TRUE", KIBA_BOOLEAN },
    { "pulse_urgent", "TRUE", KIBA_BOOLEAN },
    { "pulse_factor", "2.0", KIBA_DOUBLE },
    { "pulse_exec_factor", "2.5", KIBA_DOUBLE },
    { "pulse_speed", "30", KIBA_INT }
};

static KibaPluginVtable vtable = {
	"Pulse Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
};

KIBA_INIT_PLUGIN (vtable, pulse_options, NULL, NULL, NULL);
