/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */

/*
   kiba-mime-extensions.c: MIME database manipulation
 
   Copyright (C) 2004 Novell, Inc.
 
   The Gnome Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The Gnome Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the Gnome Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Authors: Dave Camp <dave@novell.com>

   Adapted by Daniel Bessler 2008 for kiba-dock
*/

#include <config.h>
#include "kiba-mime-extensions.h"

#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include <gtk/gtk.h>

#include <kiba-utils.h>

static char *
get_user_dir (const char *suffix)
{
	const char *xdg_data_home;
	char *user_dir;
	
	xdg_data_home = g_getenv ("XDG_DATA_HOME");
	
	if (xdg_data_home) {
		user_dir = g_build_filename (xdg_data_home, suffix, NULL);
	} else {
		const char *home;
		
		home = g_getenv ("HOME");
		
		if (home) {
			user_dir = g_build_filename (home, "/.local/share/", suffix, NULL);
		} else {
			return NULL;
		}
	}
	return user_dir;
}

static void
mime_update_program_done (GPid     pid,
			   gint     status,
			   gpointer data)
{
	/* Did the application exit correctly */
	if (WIFEXITED (status) &&
	    WEXITSTATUS (status) == 0) {
		/* FIXME: what is the right thing to do here??  We're going to
		 * have to rethink the gnome-vfs APIs.
		 */
		gnome_vfs_mime_reload ();
	}
}

static void
run_update_command (char *command,
		    char *subdir)
{
	char *argv[3] = {
		NULL,
		NULL
	};
	GPid pid = 0;
	GError *error = NULL;

	argv[0] = command;
	argv[1] = get_user_dir (subdir);

	if (g_spawn_async ("/", argv,
			   NULL,       /* envp */
			   G_SPAWN_SEARCH_PATH |
			   G_SPAWN_STDOUT_TO_DEV_NULL |
			   G_SPAWN_STDERR_TO_DEV_NULL |
			   G_SPAWN_DO_NOT_REAP_CHILD,
			   NULL, NULL, /* No setup function */
			   &pid,
			   &error)) {
		g_child_watch_add (pid, mime_update_program_done, NULL);
	}
	/* If we get an error at this point, it's quite likely the user doesn't
	 * have an installed copy of either 'update-mime-database' or
	 * 'update-desktop-database'.  I don't think we want to popup an error
	 * dialog at this point, so we just do a g_warning to give the user a
	 * chance of debugging it.
	 */
	if (error) {
		g_warning ("%s", error->message);
		g_error_free (error);
	}
	g_free (argv[1]);
}

static gboolean 
ensure_application_dir (void)
{
	char *user_dir = get_user_dir ("applications");
	
	if (!kiba_utils_recursive_mkdir (user_dir)) {
		g_free (user_dir);
		return FALSE;
	}
	
	g_free (user_dir);
	return TRUE;
}

/* stole this from update-desktop-database */
#define TEMP_CACHE_FILENAME_PREFIX ".defaults.list"
#define TEMP_CACHE_FILENAME_MAX_LENGTH 64

static int
open_temp_cache_file (const char *dir, char **filename, GError **error)
{
	GRand *rand;
	GString *candidate_filename;
	int fd;
	
	enum {
		CHOICE_UPPER_CASE = 0, 
		CHOICE_LOWER_CASE, 
		CHOICE_DIGIT, 
		NUM_CHOICES
	} choice;
	
	candidate_filename = g_string_new (TEMP_CACHE_FILENAME_PREFIX);
	
	/* Generate a unique candidate_filename and open it for writing
	*/
	rand = g_rand_new ();
	fd = -1;
	while (fd < 0) {
		char *full_path;
	
		if (candidate_filename->len > TEMP_CACHE_FILENAME_MAX_LENGTH) 
			g_string_assign (candidate_filename, TEMP_CACHE_FILENAME_PREFIX);
	
		choice = g_rand_int_range (rand, 0, NUM_CHOICES);
	
		switch (choice) {
			case CHOICE_UPPER_CASE:
				g_string_append_c (candidate_filename,
									g_rand_int_range (rand, 'A' , 'Z' + 1));
				break;
			case CHOICE_LOWER_CASE:
				g_string_append_c (candidate_filename,
									g_rand_int_range (rand, 'a' , 'z' + 1));
				break;
			case CHOICE_DIGIT:
				g_string_append_c (candidate_filename,
									g_rand_int_range (rand, '0' , '9' + 1));
				break;
			default:
				g_assert_not_reached ();
				break;
		}
	
		full_path = g_build_filename (dir, candidate_filename->str, NULL);
		fd = open (full_path, O_CREAT | O_WRONLY | O_EXCL, 0644); 
	
		if (fd < 0) {
			if (errno != EEXIST)
				{
				g_set_error (error, G_FILE_ERROR,
							g_file_error_from_errno (errno),
							"%s", g_strerror (errno));
			*filename = NULL;
				break;
				}
		}
		else if (fd >= 0 && filename != NULL) {
			*filename = full_path;
			break;
		}
	
		g_free (full_path);
	}
	
	g_rand_free (rand);
	
	g_string_free (candidate_filename, TRUE);
	
	return fd;
}

static gboolean
line_is_for_mime_type (const char *line,
		       const char *mime_type)
{
	if (!strncmp (line, mime_type, strlen (mime_type))) {
		const char *p;
		
		p = line + strlen (mime_type);
		while (g_ascii_isspace (*p)) {
			p++;
		}
		if (*p == '=') {
			return TRUE;
		}
	}

	return FALSE;
}

/*** public methods ***/

gboolean kiba_mime_application_is_user_owned (const char *id) {
	gboolean ret;
	char *user_dir;
	char *path;
	
	user_dir = get_user_dir ("applications");
	path = g_build_filename (user_dir, id, NULL);

	ret = g_file_test (path, G_FILE_TEST_EXISTS);

	g_free (path);
	g_free (user_dir);

	return ret;
}

// TODO
// gboolean kiba_mime_application_add () {
// }

void kiba_mime_application_remove (const char *id) {
	char *user_dir;
	char *path;
	
	user_dir = get_user_dir ("applications");
	path = g_build_filename (user_dir, id, NULL);

	unlink (path);
	
	g_free (path);
	g_free (user_dir);
	
	run_update_command ("update-desktop-database", "applications");
}

gboolean kiba_mime_set_default_application (const char *mime_type,
                                            const char *id) {
	char *applications_dir;
	char *list_filename;
	char *temp_filename;
	char *new_line;
	int list_file_fd;
	int temp_file_fd;
	GIOChannel *temp;
	GError *error = NULL;
	gboolean ret;

	if (!ensure_application_dir ()) {
		return FALSE;
	}
	
	applications_dir = get_user_dir ("applications");
	list_filename = g_build_filename (applications_dir, "defaults.list", NULL);

	temp_file_fd = open_temp_cache_file (applications_dir, 
					     &temp_filename, 
					     &error);

	g_free (applications_dir);
	if (error) {
		g_free (list_filename);
		return FALSE;
	}

	temp = g_io_channel_unix_new (temp_file_fd);
	g_io_channel_set_close_on_unref (temp, TRUE);

	ret = TRUE;
	
	list_file_fd = open (list_filename, O_RDONLY);
	if (list_file_fd > -1) {
		GIOChannel *list;
		char *line;
		gsize terminator_pos;
		gboolean have_newline;
		have_newline = FALSE;

		list = g_io_channel_unix_new (list_file_fd);
		g_io_channel_set_close_on_unref (list, TRUE);
		
		while (ret && 
		       g_io_channel_read_line (list, 
					       &line, 
					       NULL, 
					       &terminator_pos, 
					       NULL) == G_IO_STATUS_NORMAL) {
			if (line_is_for_mime_type (line, mime_type)) {
				g_free (line);
				continue;
			}
			
			ret = (g_io_channel_write_chars (temp, line, 
							 strlen (line), 
							 NULL, NULL) == G_IO_STATUS_NORMAL);
			have_newline = (line[terminator_pos] == '\n');

			g_free (line);
		}

		if (ret && !have_newline) {
			ret = (g_io_channel_write_chars (temp, "\n", 1, NULL, NULL) == G_IO_STATUS_NORMAL);
		}
		
		g_io_channel_unref (list);
	} else {
		const char *content;
		
		content = "[Default Applications]\n";

		ret = (g_io_channel_write_chars (temp, content, strlen (content), NULL, NULL) == G_IO_STATUS_NORMAL);
	}

	new_line = g_strdup_printf ("%s=%s\n", mime_type, id);
	ret = (g_io_channel_write_chars (temp, new_line, strlen (new_line), NULL, NULL) == G_IO_STATUS_NORMAL);
	g_free (new_line);

	g_io_channel_unref (temp);
	
	if (ret != FALSE) {
		if (rename (temp_filename, list_filename) < 0) {
			ret = FALSE;
			unlink (temp_filename);
		}
	} else {
		unlink (temp_filename);
	}
	
	g_free (temp_filename);
	g_free (list_filename);

	gnome_vfs_mime_reload ();

	return ret;
}

GList* kiba_mime_get_available_mime_types (const char *mime_type) {
	GList *res;
	
	res = gnome_vfs_mime_get_all_applications (mime_type);
	
	return res;
}

GdkPixbuf* kiba_mime_get_icon (KibaMimeApplication *app) {
	GdkPixbuf *pixbuf = NULL;
	const char *icon;
	
	icon = gnome_vfs_mime_application_get_icon (app);
	
	if (icon != NULL) {
		if (g_path_is_absolute (icon)) {
			pixbuf = gdk_pixbuf_new_from_file_at_size (icon, 24, 24, NULL);
		}
		if (pixbuf == NULL) {
			pixbuf = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (), icon, 24, 0, NULL);
		}
		if (pixbuf == NULL && strrchr (icon, '.') != NULL) {
			char *name;
			name = g_strndup (icon, strlen (icon) - strlen (strrchr (icon, '.')));
			pixbuf = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (), name, 24, 0, NULL);
			g_free (name);
		}
		
		if (pixbuf == NULL) {
			char *filename;
			filename = g_strdup_printf (DATADIR"/pixmaps/%s", icon);
			pixbuf = gdk_pixbuf_new_from_file_at_size (filename, 24, 24, NULL);
			g_free (filename);
		}
	}
	
	return pixbuf;
}

void kiba_mime_application_free (KibaMimeApplication *app) {
	gnome_vfs_mime_application_free (app);
}
