/*
 *  Copyright (C) 2008 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <string.h>
#include <sys/stat.h>
#include <libintl.h>
#include <glib.h>

#include <libgnomevfs/gnome-vfs-utils.h>
#include <libgnomevfs/gnome-vfs-file-info.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>

#include "kiba-mime-application-chooser.h"

#include "stack-file-property-window.h"


#define _(string)  dgettext (GETTEXT_PACKAGE, string)

typedef struct {
	char *path;
	GtkWidget *window;
	GtkWidget *notebook;
}KibaFilePropertyWindow;

/*** Private methods ***/

static void __window_destroy_cb (GtkWidget *win, gpointer data) {
	KibaFilePropertyWindow *window = (KibaFilePropertyWindow*) data;
	g_free(window->path);
	g_free(window);
}

static void __destroy_cb (GtkButton *button, gpointer data) {
	GtkWidget *window = (GtkWidget*) data;
	gtk_widget_destroy(window);
}

static void create_emblems_page (KibaFilePropertyWindow *window) {
// 	GtkWidget *emblems_table, *scroller;
// 	
// 	/* The emblems wrapped table */
// 	scroller = eel_scrolled_wrap_table_new (TRUE, &emblems_table);
// 	gtk_container_set_border_width (GTK_CONTAINER (emblems_table), 12);
// 	gtk_widget_show (scroller);
// 	gtk_notebook_append_page (GTK_NOTEBOOK(window->notebook), scroller, gtk_label_new (_("Emblems")));
// 	//TODO fill emblems table
// 	gtk_widget_show_all (emblems_table);
}

static void create_open_with_page (KibaFilePropertyWindow *window) {
	GtkWidget *mime_chooser;
	char *mime_type;
	
	mime_type =  gnome_vfs_get_mime_type(window->path);
	mime_chooser = kiba_mime_application_chooser_new (window->path, mime_type);
	gtk_widget_show (mime_chooser);
	g_free (mime_type);

	gtk_notebook_append_page (GTK_NOTEBOOK(window->notebook), mime_chooser, gtk_label_new (_("Open With")));
}

static void create_basic_page (KibaFilePropertyWindow *window) {
	GtkWidget *table;
	GnomeVFSFileInfo *info;
	float file_size;
	char *size_string;
	const char *fileName;
	char *accessed, *modified;
	gint i;
	
	//TODO icon
	
	char *buf = g_strrstr(window->path, "/");
	if (buf) fileName = buf+1;
	else fileName = window->path;
	
	info = gnome_vfs_file_info_new ();
	gnome_vfs_get_file_info (window->path, info, GNOME_VFS_FILE_INFO_DEFAULT | GNOME_VFS_FILE_INFO_GET_MIME_TYPE);
	size_string = gnome_vfs_format_file_size_for_display (info->size);
	
	if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_ATIME) {
		accessed = ctime(&(info->atime));
		accessed[strlen(accessed)-1] = '\0';
	}
	else accessed = " ";
	if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_MTIME) {
		modified = ctime(&(info->mtime));
		modified[strlen(modified)-1] = '\0';
	}
	else modified = " ";
	
	const char* labelData[] = {
		_("Name"), fileName,
// 		_("Type"), type, /*TODO*/
		_("Size"), size_string,
		_("Path"), window->path,
		_("Mime-Type"), gnome_vfs_file_info_get_mime_type (info),
		" ", " ",
		_("Modified"), modified,
		_("Accessed"), accessed,
	};
	
	table = gtk_table_new (G_N_ELEMENTS(labelData), 2, FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(table), 5);
	gtk_table_set_row_spacings (GTK_TABLE(table), 5);
	gtk_table_set_col_spacings (GTK_TABLE(table), 5);
	
	for (i = 0; i < G_N_ELEMENTS(labelData); i++) {
		const char *title = labelData[i++];
		const char *value = labelData[i];
		GtkWidget *label;
		char *markup;
		
		label = gtk_label_new(title);
		markup = g_markup_printf_escaped ("<span weight=\"bold\">%s</span>", GTK_LABEL (label)->text);
		gtk_label_set_markup (GTK_LABEL (label), markup);
		g_free (markup);
		gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
		gtk_widget_show (label);
		gtk_table_attach (GTK_TABLE(table), label, 0, 1, i, i+1, GTK_FILL, GTK_FILL, 0, 0);
		gtk_widget_show(label);
		
		label = gtk_label_new(value);
		markup = g_markup_printf_escaped ("<span style=\"italic\">%s</span>", GTK_LABEL (label)->text);
		gtk_label_set_markup (GTK_LABEL (label), markup);
		g_free (markup);
		gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
		gtk_widget_show (label);
		gtk_table_attach (GTK_TABLE(table), label, 1, 2, i, i+1, GTK_FILL, GTK_FILL, 0, 0);
		gtk_widget_show(label);
	}
	
	gtk_notebook_append_page (GTK_NOTEBOOK(window->notebook), table, gtk_label_new (_("Basic")));
	gtk_widget_show(table);
	
	gnome_vfs_file_info_unref(info);
}

/*** Public methods ***/

GtkWidget* kiba_stack_file_property_window_new (const char *filePath) {
	KibaFilePropertyWindow *window;
	GtkWidget *mainbox, *buttonbox;
	GtkWidget *button;
	char *mime_type;
	const char *fileName;
	char *title;
	
	window = g_new0(KibaFilePropertyWindow, 1);
	window->path = g_strdup(filePath);
	
	window->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	
	gtk_window_set_type_hint (GTK_WINDOW(window->window), GDK_WINDOW_TYPE_HINT_DIALOG);
	gtk_window_set_default_size (GTK_WINDOW(window->window), 300, 400);
	
	char *buf = g_strrstr(filePath, "/");
	if (buf)
		fileName = buf+1;
	else
		fileName = filePath;
	title = g_strdup_printf(_("Properties of %s"), fileName);
	gtk_window_set_title (GTK_WINDOW(window->window), title);
	g_free(title);
	
	mainbox = gtk_vbox_new(FALSE, 0);
	
	window->notebook = gtk_notebook_new();
	gtk_box_pack_start (GTK_BOX (mainbox),
			window->notebook,
			TRUE, TRUE, 5);
	gtk_widget_show(window->notebook);
	
	/* adding notebook pages */
	create_basic_page(window);
// 	create_emblems_page(window);
// 	create_permission_page(window);
	create_open_with_page(window);
	
	gtk_notebook_set_current_page(GTK_NOTEBOOK(window->notebook), 0);
	
	buttonbox = gtk_hbox_new(FALSE, 0);
	
	button = gtk_button_new_from_stock (GTK_STOCK_CLOSE);
	g_signal_connect(G_OBJECT(button), "released", G_CALLBACK(__destroy_cb), window->window);
	gtk_box_pack_end (GTK_BOX (buttonbox),
			button,
			FALSE, FALSE, 5);
	gtk_widget_show(button);
	gtk_box_pack_start (GTK_BOX (mainbox),
			buttonbox,
			FALSE, FALSE, 5);
	gtk_widget_show(buttonbox);
	
	gtk_container_add(GTK_CONTAINER(window->window), mainbox);
	gtk_widget_show(mainbox);
	
	g_signal_connect(G_OBJECT(window->window), "destroy",
		G_CALLBACK(__window_destroy_cb), window);
	
	return window->window;
}
