/*
 *  Copyright (C) 2007-2008 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <string.h>

#include <kiba.h>
#include <kiba-utils.h>
#include <kiba-image.h>
#include <kiba-screen.h>
#include <kiba-surface-factory.h>
#include <kiba-cairo-utils.h>
#include <cairo.h>

#include <libgnomevfs/gnome-vfs-utils.h>
#include <libgnomevfs/gnome-vfs-xfer.h>
#include <libgnomevfs/gnome-vfs-async-ops.h>

#include "stack.h"
#include "stack-icon.h"
#include "stack-icon-view.h"

static KibaStackIcon *stack_dnd_icon = NULL;
static KibaStackIconView *stack_icon_view = NULL;

typedef struct {
	KibaStackIconView *view;
	GnomeVFSAsyncHandle *handle_return;
	GdkDragContext *context;
	GtkWidget *box;
	char *url;
}KibaStackXferData;

/*** Private methods ***/

/**
 * Sort helper function, to sort the icons by their name
 */
static gint __sort_icons_by_name (gconstpointer apointer, gconstpointer bpointer) {
	const KibaStackIcon *aicon = apointer;
	const KibaStackIcon *bicon = bpointer;
	
	return kiba_utils_sort_string_list (aicon->name, bicon->name);
}

/*** Drag window ***/

static gboolean kiba_icon_view_dnd_expose_transparent (GtkWidget      *widget,
						       GdkEventExpose *expose) {
	cairo_t *cr = NULL;
	cairo_pattern_t *pattern = NULL;
	GtkWidget *child = NULL;
	
	child = gtk_bin_get_child (GTK_BIN (widget));
	
	if (!GTK_WIDGET_REALIZED (widget))
		return FALSE;
	
	cr = gdk_cairo_create (widget->window);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;
	
	cairo_rectangle (cr,
		expose->area.x, expose->area.y,
		expose->area.width, expose->area.height);
	cairo_clip (cr);
	
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_set_source_rgba (cr, 0, 0, 0, 0);
	cairo_paint (cr);
	cairo_destroy (cr);
	
	if (child)
		gtk_container_propagate_expose (GTK_CONTAINER (widget), child,  expose);
	
	return FALSE;
}

static void kiba_icon_view_dnd_image_screen_changed (GtkWidget *widget,
						     GdkScreen *previous_screen) {
	GdkScreen *screen;
	GdkColormap *colormap = NULL;
	GdkVisual *visual = NULL;
	
	screen = gtk_widget_get_screen (widget);
	
	colormap = gdk_screen_get_rgba_colormap (screen);
	visual   = gdk_screen_get_rgba_visual (screen);
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgb_colormap (screen);
		visual   = gdk_screen_get_rgb_visual (screen);
	}
	
	gtk_widget_set_visual (widget, visual);
	gtk_widget_set_colormap (widget, colormap);
}

static void kiba_icon_view_dnd_image_drag_end_cb (GtkWidget      *widget,
						  GdkDragContext *drag_context,
						  gpointer        user_data) {
	GtkWidget *image = user_data;
	
	gtk_widget_destroy (image);
}

static void unset_dnd_icon (GtkWidget      *widget,
			    GdkDragContext *drag_context) {
	stack_dnd_icon = NULL;
}

static gint __sort_icons (gconstpointer apointer,
                          gconstpointer bpointer) {
	const KibaIconViewChild *achild = apointer;
	const KibaIconViewChild *bchild = bpointer;
	const KibaStackIcon *aicon = g_object_get_data (G_OBJECT(achild->widget), "stack-icon");
	const KibaStackIcon *bicon = g_object_get_data (G_OBJECT(bchild->widget), "stack-icon");
	if (!aicon) return -1;
	if (!bicon) return 1;
	
	/* keeping the "..." item first */
	if (g_str_has_prefix(aicon->url, bicon->url))
		return 1;
	else if (g_str_has_prefix(bicon->url, aicon->url))
		return -1;
	
	/* keeping directories before file items */
	if (aicon->type == KIBA_STACK_ICON_TYPE_DIRECTORY && bicon->type != KIBA_STACK_ICON_TYPE_DIRECTORY)
		return -1;
	else if (bicon->type == KIBA_STACK_ICON_TYPE_DIRECTORY && aicon->type != KIBA_STACK_ICON_TYPE_DIRECTORY)
		return 1;
	
	/* sort alphabetic */
	return kiba_utils_sort_string_list(aicon->name, bicon->name);
}

/*** Image box ***/

/**
 * set markup for the icon-view item titles
 */
static void __set_image_box_markup (KibaStackIconView *icon_view,
                                    GtkWidget *box,
                                    GtkWidget *label) {
	KibaIconView *view;
	char *markup;
	
	if (!icon_view->win) return;
	
	view = KIBA_ICON_VIEW(icon_view->win->icon_view);
	
	if (box == kiba_icon_view_get_selected_child(view)) {
		markup = g_markup_printf_escaped (
			"<span size=\"small\" weight=\"bold\">%s</span>",
			GTK_LABEL (label)->text);
	}
	else {
		markup = g_markup_printf_escaped (
			"<span size=\"small\">%s</span>",
			 GTK_LABEL (label)->text);
	}
	gtk_label_set_markup (GTK_LABEL (label), markup);
	g_free (markup);
}

static void __icon_loaded (cairo_surface_t *surface,
                           gint size,
                           gpointer user_data) {
	GtkWidget *image = user_data;
	
	kiba_image_set_surface (KIBA_IMAGE(image),
		surface,
		size, size);
}

static gboolean __image_button_press_cb (GtkWidget      *widget,
				         GdkEventButton *event,
				         gpointer        data) {
	KibaStackIcon *icon = data;
	
	if (event->button == 3) {
		GtkWidget *menu;
		
		menu = kiba_stack_create_context_menu_for_icon (icon);
		if (!menu)
			return FALSE;
		
		gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
			event->button, gtk_get_current_event_time());
	}
	
	return FALSE;
}

static void __image_drag_begin_cb (GtkWidget      *widget,
				   GdkDragContext *drag_context,
				   gpointer        data) {
	KibaStackIcon *icon = data;
	GtkWidget *image;
	GtkWidget *win;
	GtkWidgetClass *widget_class;
	
	win = gtk_window_new (GTK_WINDOW_POPUP);
	widget_class = GTK_WIDGET_GET_CLASS(win);
	widget_class->expose_event = kiba_icon_view_dnd_expose_transparent;
	widget_class->screen_changed = kiba_icon_view_dnd_image_screen_changed;
	kiba_icon_view_dnd_image_screen_changed (win, NULL);
	gtk_window_set_type_hint (GTK_WINDOW(win), GDK_WINDOW_TYPE_HINT_DND);
	
	image = kiba_image_new ();
	kiba_image_set_surface (KIBA_IMAGE(image),
		icon->surface,
		widget->allocation.width,
		widget->allocation.height);
	kiba_image_set_size (KIBA_IMAGE(image),
		widget->allocation.width,
		widget->allocation.height);
	gtk_widget_show (image);
	
	gtk_container_add (GTK_CONTAINER(win), image);
	
	gtk_drag_set_icon_widget (drag_context, win, 0, 0);
	
	stack_dnd_icon = icon;
	
	g_signal_connect (G_OBJECT(widget), "drag-end", G_CALLBACK(unset_dnd_icon), NULL);
}

/**
 * append the icon uri to the drag path
 */
static void __image_drag_data_get_cb (GtkWidget        *widget,
				      GdkDragContext   *drag_context,
				      GtkSelectionData *data,
				      guint             info,
				      guint             time,
				      gpointer          user_data) {
	KibaStackIcon *icon = user_data;
	GString *result;
	char *string;
	
	result = g_string_new (NULL);
	
	string = gnome_vfs_make_uri_from_shell_arg (icon->url);
	g_string_append (result, string);
	g_free (string);
	g_string_append (result, "\r\n");
	
	gtk_selection_data_set (data, data->target, 8, result->str, result->len);
}

static void __image_destroy_cb (GtkWidget *widget) {
	KibaImage *image = KIBA_IMAGE(widget);
	kiba_surface_factory_remove_surface(image->surface);
}

/**
 * create a icon-view item for a KibaStackIcon,
 * using KibaImage for the icon and GtkLabel for the text
 */
static GtkWidget* __create_image_box_for_icon (KibaStackIconView *icon_view, KibaStackIcon *icon, gint size) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	KibaObject *object = KIBA_OBJECT(icon_view->stack->widget);
	GtkWidget *image;
	GtkWidget *box;
	GtkWidget *label;
	char *markup;
	const char *iconPath;
	
	if (icon->custom_icon) iconPath = icon->custom_icon;
	else iconPath = icon->mime_icon;
	
	box = gtk_vbox_new(FALSE, 0);
	
	image = kiba_image_new ();
	kiba_image_set_size (KIBA_IMAGE(image), size, size);
	kiba_surface_factory_load_icon (iconPath, size, __icon_loaded, image);
	/*if (!KIBA_IMAGE(image)->surface) {
		GtkIconInfo *info;
		info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default(), GTK_STOCK_REFRESH, size, 0);
		if (info != NULL) {
			kiba_surface_factory_load_icon (KIBA_SURFACE_FACTORY_LOAD_IMMEDIATELY,
				gtk_icon_info_get_filename (info), size, __icon_loaded, image);
			gtk_icon_info_free (info);
		}
	}*/
	gtk_box_pack_start (GTK_BOX (box), image, FALSE, FALSE, 0);
	gtk_widget_show (image);
	
	label = gtk_label_new(icon->name);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
	gtk_label_set_ellipsize (GTK_LABEL(label), PANGO_ELLIPSIZE_END);
	__set_image_box_markup (icon_view, box, label);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (box), label, TRUE, TRUE, 0);
	
	gtk_drag_source_set (image,
		GDK_BUTTON1_MASK | GDK_BUTTON2_MASK,
		targets, G_N_ELEMENTS (targets),
		GDK_ACTION_MOVE | GDK_ACTION_COPY);
	
	g_signal_connect (GTK_OBJECT(image), "drag-data-get",
		G_CALLBACK(__image_drag_data_get_cb), icon);
	g_signal_connect (GTK_OBJECT(image), "drag-begin",
		G_CALLBACK(__image_drag_begin_cb), icon);
	g_signal_connect (GTK_OBJECT(image), "button-press-event",
		G_CALLBACK(__image_button_press_cb), icon);
	g_signal_connect (GTK_OBJECT(image), "destroy",
		G_CALLBACK(__image_destroy_cb), NULL);
	
	g_object_set_data (G_OBJECT(box), "stack-image", image);
	g_object_set_data (G_OBJECT(box), "stack-icon", icon);
	
	gtk_widget_set_usize (box, size + 30, -1);
	
	return box;
}

static void kiba_stack_dir_view_destroyed_cb (GtkWidget *widget,
					      gpointer   data) {
	KibaStackIcon *icon = data;
	kiba_stack_icon_destroy (icon);
}

/**
create a KibaStackIcon for the given uri and put it in the view win
**/
static GtkWidget* create_icon_for_uri (KibaStackIconView *icon_view,
				       const char  *uri) {
	GtkWidget *ret;
	KibaStackIcon *icon;
	
	icon = kiba_stack_icon_new (icon_view->stack, uri);
	if (!icon) return NULL;
	kiba_stack_icon_update_surface (icon);
	
	ret = __create_image_box_for_icon (icon_view, icon, icon_view->stack->stack_plugin->view_icon_size);
	kiba_icon_view_win_add (icon_view->win, ret);
	
	g_signal_connect (G_OBJECT(ret), "destroy",
		G_CALLBACK(kiba_stack_dir_view_destroyed_cb), icon);
	
	return ret;
}

/*** icon view ***/

/**
 * setting the drag action the view wanna take
 */
static gboolean kiba_icon_view_drag_motion (GtkWidget      *widget,
					    GdkDragContext *dc,
					    gint            x,
					    gint            y,
					    guint           t,
					    gpointer        data) {
	KibaStackIconView *icon_view = data;
	GdkDragAction suggested_action;
	GtkWidget *src_widget;
	
	/* Get source widget */
	src_widget = gtk_drag_get_source_widget(dc);
	
	if (src_widget == widget) /* source == target, skip this one */
		suggested_action = 0;
	/* expect moving for directories */
	else if (icon_view->type == KIBA_STACK_TYPE_DIRECTORY)
		suggested_action = GDK_ACTION_MOVE;
	else
		suggested_action = GDK_ACTION_COPY;
	
	/* set the drag status */
	if(dc->actions == GDK_ACTION_MOVE) {
		gdk_drag_status(dc, GDK_ACTION_MOVE, t);
	}
	else if(dc->actions == GDK_ACTION_COPY) {
		gdk_drag_status(dc, GDK_ACTION_COPY, t);
	}
	else if(dc->actions == GDK_ACTION_LINK) {
		gdk_drag_status(dc, GDK_ACTION_LINK, t);
	}
	else if(dc->actions & suggested_action) {
		gdk_drag_status(dc, suggested_action, t);
	}
	else {
		gdk_drag_status(dc, 0, t);
	}
	
	return FALSE;
}

static gint  kiba_stack_xfer_progress_callback (GnomeVFSAsyncHandle      *handle,
						GnomeVFSXferProgressInfo *info,
						gpointer                  user_data) {
	KibaStackXferData *data = user_data;
	
	if (info->phase >= GNOME_VFS_XFER_PHASE_INITIAL && !data->box) {
		GtkWidget *image;
		GtkWidget *label;
		
		data->box = gtk_vbox_new (FALSE, 0);
		
		image = gtk_image_new_from_stock (GTK_STOCK_REFRESH, GTK_ICON_SIZE_DIALOG);
		gtk_widget_set_usize (image,
			data->view->stack->stack_plugin->view_icon_size,
			data->view->stack->stack_plugin->view_icon_size);
		gtk_widget_show (image);
		gtk_box_pack_start (GTK_BOX (data->box), image, FALSE, FALSE, 0);
		
		label = gtk_label_new (g_strrstr(data->url, "/")+1);
		gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
		gtk_label_set_ellipsize (GTK_LABEL(label), PANGO_ELLIPSIZE_END);
		__set_image_box_markup (data->view, data->box, label);
		gtk_widget_show (label);
		gtk_box_pack_start (GTK_BOX (data->box), label, FALSE, FALSE, 0);
		
		kiba_icon_view_win_add (data->view->win, data->box);
	}
	
	if (info->phase == GNOME_VFS_XFER_PHASE_COMPLETED) {
		gtk_widget_destroy (data->box);
		create_icon_for_uri (data->view, data->url);
		g_free (data->url);
		g_free (data);
	}
	
	return 1;
}

/**
 * a directory icon-view received a drag,
 * we gonna copy/link or move the file to the directory
 */
static gboolean kiba_stack_extern_drag_receive_cb (KibaStackIconView *view,
						   GdkDragContext *drag_context,
						   const char *filepath) {
	KibaIconView *icon_view;
	GnomeVFSXferOptions xfer_options;
	GList *l;
	char *new_path;
	char *file_name;
	gboolean async_xfer_callback;
	
	icon_view = KIBA_ICON_VIEW(view->win->icon_view);
	
	if (!g_file_test(filepath, G_FILE_TEST_IS_DIR))
		file_name = g_strrstr(filepath, "/")+1;
	else
		file_name = "";
	new_path = g_strdup_printf ("%s/%s", view->path, file_name);
	
	for (l = gtk_container_get_children(GTK_CONTAINER(icon_view)); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		KibaStackIcon *icon = g_object_get_data (G_OBJECT(child), "stack-icon");
		
		if (g_strrstr (icon->url, new_path)) {
			g_free (new_path);
			return 0;
		}
	}
	
	async_xfer_callback = FALSE;
	switch (drag_context->action) {
		case GDK_ACTION_MOVE: {
			async_xfer_callback = TRUE;
			xfer_options = GNOME_VFS_XFER_REMOVESOURCE;
			break;
		}
		case GDK_ACTION_COPY: {
			xfer_options = GNOME_VFS_XFER_DEFAULT;
			async_xfer_callback = TRUE;
			break;
		}
		case GDK_ACTION_LINK: {
			link (filepath, new_path);
			create_icon_for_uri (view, new_path);
			break;
		}
	}
	
	if (async_xfer_callback) {
		KibaStackXferData *xfer_data;
		GnomeVFSURI *source_uri;
		GnomeVFSURI *target_uri;
		GList *src_list;
		GList *dest_list;
		char *escaped_uri;
		
		escaped_uri = gnome_vfs_make_uri_from_input (filepath);
		source_uri = gnome_vfs_uri_new (escaped_uri);
		g_free (escaped_uri);
		escaped_uri = gnome_vfs_make_uri_from_input (new_path);
		target_uri = gnome_vfs_uri_new (escaped_uri);
		g_free (escaped_uri);
		
		src_list  = g_list_append (NULL, source_uri);
		dest_list = g_list_append (NULL, target_uri);
		
		xfer_data = g_new0 (KibaStackXferData, 1);
		xfer_data->view = view;
		xfer_data->context = drag_context;
		xfer_data->box = NULL;
		xfer_data->url = g_strdup (new_path);
		xfer_data->handle_return = NULL;
		
		gnome_vfs_async_xfer (&xfer_data->handle_return,
				      src_list,
				      dest_list,
				      xfer_options,
				      GNOME_VFS_XFER_ERROR_MODE_ABORT,
				      GNOME_VFS_XFER_OVERWRITE_MODE_ABORT,
				      GNOME_VFS_PRIORITY_DEFAULT,
				      kiba_stack_xfer_progress_callback,
				      xfer_data,
				      NULL,
				      NULL);
	}
	
	g_free (new_path);
	
	return 1;
}

/**
 * a icon-view received a drag
 */
static void kiba_icon_view_drag_received_cb (GtkWidget        *widget,
					     GdkDragContext   *drag_context,
					     gint              x,
					     gint              y,
					     GtkSelectionData *selection_data,
					     guint             info,
					     guint             time,
					     gpointer          user_data) {
	static const char *file_prefix = "file://";
	
	KibaStackIconView *view = user_data;
	
	if (view->type == KIBA_STACK_TYPE_CUSTOM) {
		if (kiba_stack_drag_receive_cb (view->stack, drag_context, selection_data))
			gtk_drag_finish (drag_context,
				TRUE,
				FALSE, time);
		else {
			gdk_drag_status (drag_context, 0, time);
		}
	}
	/* the icon-view is a directory */
	else {
		char *filepath;
		char *string;
		char *unescaped;
		gint new_icons;
		
		char data[strlen((char*) selection_data->data)];
		sprintf(data, "%s", (char*) selection_data->data);
		g_strstrip (data);
		
		new_icons = 0;
		while ((string = g_strrstr ((char*) data, "\n")) != NULL) {
			g_strstrip (string);
			
			if (g_str_has_prefix(string, file_prefix)) {
				unescaped = gnome_vfs_unescape_string (string + strlen(file_prefix), NULL);
			}
			else {
				unescaped = gnome_vfs_unescape_string (string, NULL);
			}
			filepath = unescaped;
			
			if (g_file_test(filepath, G_FILE_TEST_EXISTS)) {
				new_icons += kiba_stack_extern_drag_receive_cb (view, drag_context, filepath);
			}
			
			string[0] = '\0';
		}
		g_strstrip (data);
		if (g_str_has_prefix(data, file_prefix)) {
			unescaped = gnome_vfs_unescape_string (data + strlen(file_prefix), NULL);
		}
		else {
			unescaped = gnome_vfs_unescape_string (data, NULL);
		}
		filepath = unescaped;
		
		if (g_file_test(filepath, G_FILE_TEST_EXISTS)) {
			new_icons += kiba_stack_extern_drag_receive_cb (view, drag_context, filepath);
		}
		g_free (unescaped);
		
		if (new_icons > 0)
			gtk_drag_finish (drag_context,
				TRUE,
				FALSE, /* currently we remove the file on our own,
					  but we should involve the drag-data-delete event */
				time);
		else
			gdk_drag_status (drag_context, 0, time);
	}
}

/**
 * Place the icon view on the object (below for a top-dock, above for a bottom-dock ....)
 */
static void kiba_stack_place_icon_view (KibaStackIconView *icon_view) {
	Kiba *kiba = kiba_get();
	KibaWin *win;
	KibaObject *object;
	KibaScreen *screen;
	gint w, h, x, y;
	const KibaLayoutData *d;
	GtkWidget *widget = icon_view->stack->widget;
	
	screen = kiba_screen_get();
	win = kiba_get_plugin_win(kiba, icon_view->stack->stack_plugin->plugin);
	object = KIBA_OBJECT(widget);
	
	if (!KIBA_IS_WIN(win))
		return;
	if (!KIBA_IS_ICON_VIEW_WIN(icon_view->win))
		return;
	if (!GTK_WIDGET_REALIZED(icon_view->win->win))
		return;
	
	d = &kiba_layout_data[KIBA_DOCK(KIBA_WIN_CHILD(win))->d.position];
	
	struct {
		gint  *x, *y;
		gint  *w, *h;
		gint  alignment;
	} magic[2] = {
		{ &x, &y , &w, &h, d->valignment},
		{ &y, &x , &h, &w, d->halignment}
	};
	w = icon_view->win->win->allocation.width;
	h = icon_view->win->win->allocation.height;
	x = widget->allocation.x + GTK_WIDGET(win)->allocation.x;
	y = widget->allocation.y + GTK_WIDGET(win)->allocation.y;
	
	*magic[d->orientation].x -= (*magic[d->orientation].w - widget->allocation.width)/2;
	*magic[d->orientation].y -= (magic[d->orientation].alignment == 0 ? - widget->allocation.height : *magic[d->orientation].h);
	
	if (x < screen->geometry.x)
		x = screen->geometry.x;
	else if (x + icon_view->win->win->allocation.width > screen->geometry.x + screen->geometry.width)
		x = screen->geometry.x + screen->geometry.width - icon_view->win->win->allocation.width;
	
	if (y < screen->geometry.y)
		y = screen->geometry.y;
	else if (y + icon_view->win->win->allocation.height > screen->geometry.y + screen->geometry.height)
		y = screen->geometry.y + screen->geometry.height - icon_view->win->win->allocation.height;
	
	gdk_window_move (icon_view->win->win->window, x, y);
}

static void kiba_stack_icon_view_size_allocate_cb (GtkWidget     *widget,
						   GtkAllocation *allocation,
						   gpointer       data) {
	KibaStackIconView *icon_view = data;
	kiba_stack_place_icon_view (icon_view);
}

/**
 * clean up invalid pointers after the icon view win was destroyed
 */
static void kiba_stack_win_destroyed_cb (GtkWidget *widget,
					 gpointer   data) {
	KibaStackIconView *icon_view = data;
	
	g_signal_handlers_disconnect_by_func (
	     G_OBJECT(icon_view->stack->widget),
	     G_CALLBACK(kiba_stack_icon_view_size_allocate_cb),
	     icon_view);
	
	icon_view->icon = NULL;
	icon_view->win = NULL;
	icon_view->type = KIBA_STACK_TYPE_NONE;
	icon_view->stack = NULL;
	if (icon_view->path)
		g_free (icon_view->path);
	icon_view->path = NULL;
}

static void kiba_icon_view_child_unselected_cb (KibaIconView *icon_view,
					        GtkWidget    *child,
					        gpointer      user_data) {
	KibaStackIconView *view = user_data;
	GList *l;
	
	for (l = gtk_container_get_children(GTK_CONTAINER(child)); l != NULL; l = l->next) {
		GtkWidget *widget = l->data;
		
		if (!GTK_IS_LABEL(widget))
			continue;
		
		__set_image_box_markup(view, child, widget);
	}
}

static void kiba_icon_view_child_selected_cb (KibaIconView *icon_view,
					      GtkWidget    *child,
					      gpointer      user_data) {
	KibaStackIconView *view = user_data;
	KibaStackIcon *selected;
	GList *l;
	
	selected = g_object_get_data (G_OBJECT(child), "stack-icon");
	
	for (l = gtk_container_get_children(GTK_CONTAINER(child)); l != NULL; l = l->next) {
		GtkWidget *widget = l->data;
		
		if (!GTK_IS_LABEL(widget))
			continue;
		
		__set_image_box_markup(view, child, widget);
	}
	
	if (view->type == KIBA_STACK_TYPE_CUSTOM)
		kiba_stack_select_icon (selected->stack, selected);
}

static void kiba_icon_view_child_activated_cb (KibaIconView *icon_view,
					       GtkWidget    *child,
					       gpointer      user_data) {
	kiba_stack_activate_icon (g_object_get_data (G_OBJECT(child), "stack-icon"));
}

/**
 * create a icon-view dialog window
 */
static KibaIconViewWin* __create_icon_view (KibaStackIconView *icon_view) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	KibaIconViewWin *view_win;
	KibaInfoWin *info_win;
	KibaIconView *view;
	
	view_win = kiba_icon_view_win_new(KIBA_ICON_VIEW_BOX_POSITION_RIGHT,
			icon_view->stack->stack_plugin->view_columns,
			icon_view->stack->stack_plugin->view_rows, TRUE);
// 	gtk_window_set_type_hint (GTK_WINDOW(view_win->win), GDK_WINDOW_TYPE_HINT_DOCK);
	gtk_window_set_type_hint (GTK_WINDOW(view_win->win), GDK_WINDOW_TYPE_HINT_DIALOG);
	gtk_window_stick (GTK_WINDOW(view_win->win));
	
	info_win = KIBA_INFO_WIN(view_win->win);
	view = KIBA_ICON_VIEW(view_win->icon_view);
	gtk_window_set_role (GTK_WINDOW(view_win->win), "kiba-stack");
	
	kiba_icon_view_set_sort_method (view, __sort_icons);
	
	gtk_drag_dest_set (view_win->icon_view,
		GTK_DEST_DEFAULT_ALL,
		targets, G_N_ELEMENTS (targets),
		GDK_ACTION_COPY | GDK_ACTION_LINK | GDK_ACTION_MOVE);
	gtk_drag_dest_add_uri_targets (view_win->icon_view);
	
	/* adding signal handlers */
	g_signal_connect(G_OBJECT (icon_view->stack->widget), "size-allocate",
		G_CALLBACK(kiba_stack_icon_view_size_allocate_cb), icon_view);
	g_signal_connect_after(G_OBJECT (view_win->win), "size-allocate",
		G_CALLBACK(kiba_stack_icon_view_size_allocate_cb), icon_view);
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "drag-motion",
		G_CALLBACK(kiba_icon_view_drag_motion), icon_view);
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "drag-data-received",
		G_CALLBACK(kiba_icon_view_drag_received_cb), icon_view);
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "child-selected",
		G_CALLBACK(kiba_icon_view_child_selected_cb), icon_view);
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "child-unselected",
		G_CALLBACK(kiba_icon_view_child_unselected_cb), icon_view);
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "child-activated",
		G_CALLBACK(kiba_icon_view_child_activated_cb), NULL);
	g_signal_connect (G_OBJECT(view_win->win), "destroy",
		G_CALLBACK(kiba_stack_win_destroyed_cb), icon_view);
	
	icon_view->win = view_win;
	
	return view_win;
}

/**
 * create a custom icon view win, containing all files/dirs in the stack
 */
static KibaIconViewWin* __create_icon_view_win (KibaStackIconView *icon_view) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	KibaIconViewWin *view_win;
	GtkWidget *box;
	GList *l;
	GList *sorted;
	
	sorted = g_list_sort (g_list_copy(icon_view->stack->icons), __sort_icons_by_name);
	view_win = __create_icon_view (icon_view);
	for (l = sorted; l != NULL; l = l->next) {
		KibaStackIcon *icon = l->data;
		
		box = __create_image_box_for_icon (icon_view, icon, icon_view->stack->stack_plugin->view_icon_size);
		kiba_icon_view_win_add (view_win, box);
		if (icon == icon_view->stack->icons->data) {
			// TODO: the currently selected icon is maybe unshown
			// kiba_icon_view_win_show_child (view_win->icon_view, box);
			kiba_icon_view_select_child (
				KIBA_ICON_VIEW(view_win->icon_view),
				box);
		}
	}
	g_list_free (sorted);
	
	icon_view->type = KIBA_STACK_TYPE_CUSTOM;
	if (icon_view->path)
		g_free (icon_view->path);
	icon_view->path = g_strdup (icon_view->stack->uri);
	
	return view_win;
}

/*** Public methods ***/

void kiba_stack_icon_view_add_icon (KibaStackIconView *icon_view, KibaStackIcon *icon) {
	GtkWidget *box;
	box = __create_image_box_for_icon(icon_view, icon, icon_view->stack->stack_plugin->view_icon_size);
	kiba_icon_view_win_add (icon_view->win, box);
}

/**
create a directory icon view, containing all files of the directory and a "..." item
**/
gboolean kiba_icon_view_activate_directory_icon (KibaStackIcon *directory_icon) {
	KibaStackIconView *icon_view = kiba_stack_icon_view_get();
	GtkWidget *box = NULL;
	GError *error = NULL;
	char *file_name;
	char *parent_dir;
	char *parent_nedle;
	GDir *dir;
	GList *l;
	GList *pathes = NULL;
	
	dir = g_dir_open (directory_icon->url, 0, &error);
	if (error != NULL) {
		g_error_free (error);
		return FALSE;
	}
	
	if (!KIBA_IS_ICON_VIEW_WIN(icon_view->win)) {
		__create_icon_view (icon_view);
	}
	else {
		KibaIconView *view;
		view = KIBA_ICON_VIEW(icon_view->win->icon_view);
		
		for (l = gtk_container_get_children(GTK_CONTAINER(view));
		     l != NULL; l = gtk_container_get_children(GTK_CONTAINER(view))) {
			kiba_icon_view_win_remove (icon_view->win, GTK_WIDGET(l->data));
		}
	}
	icon_view->type = KIBA_STACK_TYPE_DIRECTORY;
	icon_view->stack = directory_icon->stack;
	if (icon_view->path)
		g_free (icon_view->path);
	icon_view->path = g_strdup (directory_icon->url);
	
	parent_nedle = g_strrstr (directory_icon->url, "/");
	if (parent_nedle) {
		if (parent_nedle != directory_icon->url) {
			parent_dir = g_strndup (directory_icon->url, strlen(directory_icon->url) - strlen(parent_nedle));
			box = create_icon_for_uri (icon_view, parent_dir);
			g_free (parent_dir);
		}
		else if (strcmp (parent_nedle, "/") != 0)
			box = create_icon_for_uri (icon_view, "/");
		
		if (box) {
			for (l = gtk_container_get_children(GTK_CONTAINER(box)); l != NULL; l = l->next) {
				GtkWidget *child = l->data;
				char *markup;
				
				if (!GTK_IS_LABEL(child))
					continue;
				
				gtk_label_set_text (GTK_LABEL(child), "...");
				markup = g_markup_printf_escaped (
					"<span size=\"small\">%s</span>",
					GTK_LABEL (child)->text);
				gtk_label_set_markup (GTK_LABEL (child), markup);
				g_free (markup);
			}
		}
	}
	
	while ((file_name = (char*) g_dir_read_name (dir)) != NULL) {
		char *uri;
		
		/* skipping hidden files */
		if (file_name[0] == '.')
			continue;
		
		if (strcmp (directory_icon->url, "/") == 0)
			uri = g_strdup_printf ("/%s", file_name);
		else
			uri = g_strdup_printf ("%s/%s", directory_icon->url, file_name);
		
		pathes = g_list_append (pathes, uri);
	}
	g_dir_close (dir);
	
	pathes = g_list_sort (pathes, kiba_utils_sort_string_list);
	
	for (l = pathes; l != NULL; l = pathes) {
		char *path = l->data;
		
		box = create_icon_for_uri (icon_view, path);
		g_free (path);
		pathes = g_list_remove (pathes, l->data);
	}
	g_list_free (pathes);
	
	gtk_window_present (GTK_WINDOW(icon_view->win->win));
	gdk_window_set_keep_above (icon_view->win->win->window, TRUE);
	kiba_stack_place_icon_view (icon_view);
	
	return TRUE;
}

KibaStackIconView* kiba_stack_icon_view_get () {
	if (!stack_icon_view) {
		stack_icon_view = g_new0(KibaStackIconView, 1);
		stack_icon_view->type = KIBA_STACK_TYPE_NONE;
		stack_icon_view->win = NULL;
		stack_icon_view->stack = NULL;
		stack_icon_view->icon = NULL;
		stack_icon_view->path = NULL;
	}
	
	return stack_icon_view;
}

void kiba_stack_icon_view_toggle (KibaStack *stack,
                                  KibaStackIcon *icon) {
	KibaStackIconView *icon_view = kiba_stack_icon_view_get();
	
	if (KIBA_IS_ICON_VIEW_WIN(icon_view->win)) {
		gboolean toggling;
		
		toggling = (icon_view->stack == stack);
		
		gtk_widget_destroy (GTK_WIDGET(icon_view->win->win));
		icon_view->type = KIBA_STACK_TYPE_NONE;
		if (icon_view->path)
			g_free (icon_view->path);
		icon_view->path = NULL;
		
		if (toggling) {
			icon_view->stack = NULL;
			return;
		}
	}
	
	icon_view->stack = stack;
	icon_view->icon = icon;
	
	icon_view->win = __create_icon_view_win (icon_view);
	gtk_widget_show (icon_view->win->win);
	gdk_window_set_keep_above (icon_view->win->win->window, TRUE);
	kiba_stack_place_icon_view (icon_view);
}

void kiba_stack_icon_view_destroy () {
	g_free(stack_icon_view);
	if (KIBA_IS_ICON_VIEW_WIN(stack_icon_view->win))
		gtk_widget_destroy (stack_icon_view->win->win);
	stack_icon_view = NULL;
}
