/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <string.h>
#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>
#include <libgnomeui/libgnomeui.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-debug.h>
#include <kiba-prefs.h>
#include <kiba-image.h>
#include <kiba-object.h>
#include <kiba-globals.h>
#include <kiba-utils.h>
#include <kiba-surface-factory.h>

#include "stack-icon.h"
#include "stack-icon-view.h"

/**
Class Info:
  Each KibaStackIcon instance represents one File in the Filesystem. it tries to find the standart application for this file, but also allows setting of custom applications.
  The struct contains a cairo_surface_t surface buffer displaying a custom icon, thumbnail, mime icon or the application icon.
  KibaStackIcon's are usually grouped together in a KibaStack instance.
**/

static GList *dirty_icon_list = NULL;

/**
Helper Functions
**/

static void __icon_loaded (cairo_surface_t *surface,
                           gint size,
                           gpointer user_data) {
	KibaStackIcon *icon = user_data;
	
	if (icon->surface) {
		kiba_surface_factory_remove_surface(icon->surface);
		cairo_surface_destroy(icon->surface);
	}
	icon->surface = cairo_surface_reference(surface);
	
	if (icon->stack->icons->data == icon) {
		KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(icon->stack->widget);
		if (drawable_object->surface_data->surface)
			cairo_surface_destroy(drawable_object->surface_data->surface);
		drawable_object->surface_data->surface = cairo_surface_reference (surface);
		kiba_drawable_object_queue_draw (drawable_object);
	}
}

static void __thumbnail_loaded_cb (const char *thumbPath, gpointer data) {
	KibaStackIcon *icon = data;
	KibaStackIconView *icon_view;
	gint size;
	
	if (!g_list_find(dirty_icon_list, icon))
		return;
	
	size = icon->stack->stack_plugin->view_icon_size;
	
	icon->type = KIBA_STACK_ICON_TYPE_THUMB_FILE;
	if (icon->mime_icon)
		g_free(icon->mime_icon);
	icon->mime_icon = g_strdup(thumbPath);
	
	kiba_stack_icon_update_surface(icon);
	
	/* change surface of the icon view image, if there is one */
	icon_view = kiba_stack_icon_view_get();
	if (KIBA_IS_ICON_VIEW_WIN(icon_view->win) && icon_view->stack == icon->stack) {
		KibaIconView *view;
		GList *l;
		
		view = KIBA_ICON_VIEW(icon_view->win->icon_view);
		
		for (l = gtk_container_get_children(GTK_CONTAINER(view)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			KibaStackIcon *check_icon;
			
			check_icon = g_object_get_data (G_OBJECT(child), "stack-icon");
			if (check_icon == icon) {
				GtkWidget *image;
				
				image = g_object_get_data (G_OBJECT(child), "stack-image");
				kiba_image_set_surface (KIBA_IMAGE(image),
					icon->surface,
					size, size);
				break;
			}
		}
	}
}

/**
 * __load:
 *
 * set all url related variables of stack_icon (mime_type, icon, exec, name and type)
 */
static gboolean __load (KibaStackIcon *stack_icon, GError **error) {
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	GnomeThumbnailFactory *thumb_factory = NULL;
	GQuark  error_quark;
	GnomeVFSResult result;
	GnomeVFSFileInfo info;
	char *utf8_uri = NULL;
	char *mime_type;
	
	kiba_plugin_loader_invoke_service (loader,
		stack_icon->stack->stack_plugin->factoryGet,
		kiba_utils_value_array_init(1, G_TYPE_POINTER, &thumb_factory));
	
	error_quark = g_quark_from_string ("kiba-dock");
	
	mime_type = gnome_vfs_get_mime_type (stack_icon->url);
	if (!mime_type) {
		g_set_error (error, error_quark, 1, "Failed to get MIME Type from url '%s'", stack_icon->url);
		return FALSE;
	}
	
	if (!g_utf8_validate (stack_icon->url, -1, NULL)) {
		utf8_uri = g_locale_to_utf8 (stack_icon->url, -1, NULL, NULL, NULL);
	}
	if (!utf8_uri) {
		utf8_uri = g_strdup(stack_icon->url);
	}
	
	stack_icon->type = -1;
	stack_icon->name = g_strrstr(utf8_uri, "/")+1;
	
	/* its a desktop file, we gonna create a special launcher icon */
	if (g_str_has_suffix (stack_icon->url, ".desktop")) {
		static const char desktop_entry_group[] = "Desktop Entry";
		
		GKeyFile *file;
		GError *error = NULL;
		
		file = g_key_file_new ();
		if (g_key_file_load_from_file (file, stack_icon->url, 0, &error)) {
			stack_icon->mime_icon = g_key_file_get_value (file, desktop_entry_group, "Icon", &error);
			g_clear_error (&error);
			error = NULL;
			g_key_file_free (file);
		}
		
		stack_icon->type = KIBA_STACK_ICON_TYPE_DESKTOP;
	}
	/* get the icon from the file info */
	else if (gnome_vfs_get_file_info (stack_icon->url, &info, GNOME_VFS_FILE_INFO_DEFAULT) == GNOME_VFS_OK) {
		GnomeVFSURI *vfs_uri;
		char *uri_string = NULL;
		char *escaped_uri;
		
		escaped_uri = gnome_vfs_make_uri_from_input (stack_icon->url);
		vfs_uri = gnome_vfs_uri_new (escaped_uri);
		g_free (escaped_uri);
		if (vfs_uri != NULL) {
			uri_string = gnome_vfs_uri_to_string (vfs_uri, GNOME_VFS_URI_HIDE_NONE);
			gnome_vfs_uri_unref (vfs_uri);
		}
		
		/* try to find a thumbnail */
		if (uri_string && gnome_thumbnail_factory_can_thumbnail (thumb_factory,
						   uri_string,
						   mime_type,
						   info.mtime)) {
			stack_icon->mime_icon = gnome_thumbnail_factory_lookup (thumb_factory, uri_string, info.mtime);
			if (stack_icon->mime_icon)
				stack_icon->type = KIBA_STACK_ICON_TYPE_THUMB_FILE;
			if (!stack_icon->mime_icon) {
				kiba_plugin_loader_invoke_service (loader,
					stack_icon->stack->stack_plugin->loadThumb,
					kiba_utils_value_array_init(3,
						G_TYPE_STRING, uri_string,
						G_TYPE_POINTER, __thumbnail_loaded_cb,
						G_TYPE_POINTER, stack_icon));
			}
		}
		
		if (!stack_icon->mime_icon) {
			stack_icon->type = KIBA_STACK_ICON_TYPE_MIME_FILE;
			stack_icon->mime_icon = gnome_icon_lookup (gtk_icon_theme_get_default(),
							NULL, stack_icon->url, NULL,
							&info, mime_type,
							GNOME_ICON_LOOKUP_FLAGS_NONE, NULL);
		}
	}
	if (g_file_test(stack_icon->url, G_FILE_TEST_IS_DIR)) {
		stack_icon->type = KIBA_STACK_ICON_TYPE_DIRECTORY;
	}
	
	if (!stack_icon->mime_icon) {
		GnomeVFSMimeApplication *app;
		const char *app_icon;
		
		app  = gnome_vfs_mime_get_default_application_for_uri (stack_icon->url, mime_type);
		app_icon = gnome_vfs_mime_application_get_icon (app);
		if (app_icon)
			stack_icon->mime_icon = g_strdup(app_icon);
	}
	
	return TRUE;
}

void __file_monitor_callback (GnomeVFSMonitorHandle *handle,
                              const gchar *monitor_uri,
                              const gchar *info_uri,
                              GnomeVFSMonitorEventType event_type,
                              gpointer user_data) {
	KibaStackIconView *icon_view = kiba_stack_icon_view_get();
	KibaStackIcon *icon = user_data;
	
	switch (event_type) {
		case GNOME_VFS_MONITOR_EVENT_DELETED: {
			if (icon->stack && g_list_find(icon->stack->icons, icon)) {
				kiba_stack_remove(icon->stack, icon);
			}
			if (icon_view->stack == icon->stack) {
				KibaIconView *view;
				GList *l;
				
				view = KIBA_ICON_VIEW(icon_view->win->icon_view);
				
				for (l = gtk_container_get_children(GTK_CONTAINER(view)); l != NULL; l = l->next) {
					GtkWidget *child = l->data;
					KibaStackIcon *check_icon;
					
					check_icon = g_object_get_data (G_OBJECT(child), "stack-icon");
					if (check_icon == icon) {
						kiba_icon_view_win_remove (icon_view->win, child);
						return; /* kiba_icon_view_win_remove calls kiba_stack_icon_destroy, so we return here */
					}
				}
			}
			kiba_stack_icon_destroy(icon);
			break;
		}
		//TODO do something fancy with the events ;)
		case GNOME_VFS_MONITOR_EVENT_CHANGED:
		case GNOME_VFS_MONITOR_EVENT_STARTEXECUTING:
		case GNOME_VFS_MONITOR_EVENT_STOPEXECUTING:
		case GNOME_VFS_MONITOR_EVENT_CREATED:
		case GNOME_VFS_MONITOR_EVENT_METADATA_CHANGED:
			break;
	}
}

/*** Public methods ***/

/**
 * Returns command line of @icon.
 */
char* kiba_stack_icon_get_command (KibaStackIcon *icon) {
	GnomeVFSMimeApplication *app;
	char *utf8_uri = NULL;
	char *exec = NULL;
	char *mime_type;
	
	mime_type = gnome_vfs_get_mime_type (icon->url);
	
	if (!g_utf8_validate (icon->url, -1, NULL)) {
		utf8_uri = g_locale_to_utf8 (icon->url, -1, NULL, NULL, NULL);
	}
	if (!utf8_uri) {
		utf8_uri = g_strdup(icon->url);
	}
	
	if (icon->type == KIBA_STACK_ICON_TYPE_DESKTOP) {
		static const char desktop_entry_group[] = "Desktop Entry";
		
		GKeyFile *file;
		GError *error = NULL;
		
		file = g_key_file_new ();
		if (g_key_file_load_from_file (file, icon->url, 0, &error)) {
			exec = g_key_file_get_value (file, desktop_entry_group, "Exec", &error);
			g_clear_error (&error);
			error = NULL;
			if (!exec) {
				char *buf_key;
				
				buf_key = g_key_file_get_value (file, desktop_entry_group, "URL", &error);
				g_clear_error (&error);
				error = NULL;
				if (!buf_key) {
					buf_key = g_key_file_get_value (file, desktop_entry_group, "Path", &error);
					g_clear_error (&error);
					error = NULL;
				}
				if (buf_key) {
					exec = g_strdup_printf ("%s %s", file_manager_global, buf_key);
					g_free (buf_key);
				}
			}
			g_key_file_free (file);
		}
	} else {
		char *vfs_uri = NULL;
		char *validate;
		
		app  = gnome_vfs_mime_get_default_application_for_uri (icon->url, mime_type);
		vfs_uri = gnome_vfs_make_uri_from_shell_arg (utf8_uri);
		
		/* this character is not escaped by the gnome vfs escape function :/ */
		if (g_strrstr (vfs_uri, "'")) {
			char *buf = vfs_uri;
			
			vfs_uri = gnome_vfs_escape_set (buf, "'");
			g_free (buf);
		}
		if (app) {
			const char *command_line;
			
			command_line = gnome_vfs_mime_application_get_exec (app);
			if (command_line) exec = g_strdup_printf ("%s %s", command_line, vfs_uri);
		}
		
		g_free (vfs_uri);
	}
	
	return exec;
}

/**
 * kiba_stack_icon_update_surface:
 *
 * (re)load the surface buffer for the KibaStackIcon instance
 */
void kiba_stack_icon_update_surface (KibaStackIcon *icon) {
	KibaObject *object;
	char *iconPath;
	char *uri = NULL;
	gint size;
	
	object = KIBA_OBJECT(icon->stack->widget);
	if (g_list_find(icon->stack->icons, icon))
		size = object->size.width*object->scale_x; /* we can ignor height here, because stack icons are squared */
	else
		size = icon->stack->stack_plugin->view_icon_size;
	
	/* first try the custom icon */
	if (icon->custom_icon)
		uri = kiba_get_uri_for_icon(icon->custom_icon, size);
	if (!uri && icon->mime_icon)
		uri = kiba_get_uri_for_icon(icon->mime_icon, size);
	if (!uri) {
		GnomeVFSMimeApplication *app;
		const char *appIcon;
		char *mime_type;
		
		mime_type = gnome_vfs_get_mime_type (icon->url);
		app  = gnome_vfs_mime_get_default_application_for_uri (icon->url, mime_type);
		g_free(mime_type);
		if (app)
			appIcon = gnome_vfs_mime_application_get_icon (app);
		else
			appIcon = NULL;
		
		if (appIcon)
			uri = kiba_get_uri_for_icon(appIcon, size);
	}
	if (!uri)
		uri = kiba_get_uri_for_icon(GTK_STOCK_MISSING_IMAGE, size);
	
	kiba_surface_factory_load_icon (uri, size, __icon_loaded, icon);
	g_free(uri);
}

/*** Class Functions ***/

KibaStackIcon* kiba_stack_icon_new (KibaStack *stack, const char *uri) {
	KibaStackIcon *icon;
	GError *error = NULL;
	
	g_return_if_fail (stack != NULL);
	g_return_val_if_fail(uri != NULL, NULL);
	
	icon = g_new0 (KibaStackIcon, 1);
	icon->url = g_strdup(uri);
	icon->mime_icon = NULL;
	icon->custom_icon = NULL;
	icon->surface = NULL;
	icon->stack = stack;
	icon->monitor = NULL;
	
	if (!__load(icon, &error)) {
		DEBUG_ERROR ("Failed to get KibaStackIcon from url '%s'\n", uri);
		if (error != NULL) {
			DEBUG_ERROR ("%s\n", error->message);
			g_free (error);
		}
		g_free(icon->url);
		g_free(icon);
		return NULL;
	}
	
	gnome_vfs_monitor_add (&icon->monitor, uri, GNOME_VFS_MONITOR_FILE, __file_monitor_callback, icon);
	
	dirty_icon_list = g_list_append(dirty_icon_list, icon);
	
	return icon;
}

void kiba_stack_icon_destroy (KibaStackIcon *icon) {
	g_return_if_fail (icon != NULL);
	
	if (icon->url) g_free (icon->url);
	
	if (icon->surface) {
		cairo_surface_destroy (icon->surface);
		kiba_surface_factory_remove_surface(icon->surface);
	}
	
	if (icon->monitor)
		gnome_vfs_monitor_cancel (icon->monitor);
	
	//FIXME: maybe a thumbnail is about to be loaded for this icon
	dirty_icon_list = g_list_remove(dirty_icon_list, icon);
	
	g_free (icon);
}
