/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <string.h>
#include <libintl.h>
#include <libgnomevfs/gnome-vfs-utils.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-icon.h>
#include <kiba-object.h>
#include <kiba-menu-items.h>
#include <kiba-info-win.h>
#include <kiba-icon-view.h>
#include <kiba-image.h>
#include <kiba-screen.h>
#include <kiba-utils.h>
#include <kiba-dialog.h>

#include "stack.h"
#include "stack-icon.h"
#include "stack-icon-view.h"
#include "stack-file-property-window.h"

#define _(string)  dgettext (GETTEXT_PACKAGE, string)

#define MAX_OBJECTS VIEW_COLUMNS*VIEW_MAX_ROWS
#define STACK_APPLICATION_DIR "/usr/share/applications"

static void kiba_stack_add (KibaStack *stack, KibaStackIcon *icon);

/**
Class Info:
	Each stack instance represents one Dock object, it contains multiple KibaStackIcon's.
	The currently selected KibaStackIcon provides the icon for the object on the dock.
	KibaStack uses a KibaIconViewWin to show the user a grid containing all KibaStackIcon and to show the content of the activated directory icon.
	a new KibaStack can be created by dragging files to the dock,
	the stack can be filled by dragging files or directories to the dock object or to the icon view.
**/

/**
Helper Functions
**/

static void __update_object_surface (KibaStack *stack) {
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(stack->widget);
	KibaStackIcon *icon = stack->icons->data;
	
	if (drawable_object->surface_data->surface)
		cairo_surface_destroy(drawable_object->surface_data->surface);
	drawable_object->surface_data->surface = cairo_surface_reference (icon->surface);
	kiba_drawable_object_queue_draw (drawable_object);
}

/**
Returns wether a file was allready added to the stack
**/
static gboolean kiba_stack_contains_file (KibaStack  *stack,
					  const char *path) {
	GList *l;
	
	for (l = stack->icons; l != NULL; l = l->next) {
		KibaStackIcon *icon = l->data;
		
		if (strcmp (icon->url, path) == 0)
			return TRUE;
	}
	
	return FALSE;
}

static void kiba_stack_file_properties_cb (GtkWidget *widget,
					   gpointer   user_data) {
	KibaStackIcon *icon = user_data;
	static GtkWidget *file_property_win = NULL;
	
	if (GTK_IS_WIDGET(file_property_win))
		gtk_widget_destroy(file_property_win);
	
	file_property_win = kiba_stack_file_property_window_new(icon->url);
	gtk_widget_show_all(file_property_win);
	gtk_window_present(GTK_WINDOW(file_property_win));
}

static gboolean kiba_stack_remove_icon_cb (GtkWidget *widget,
					   gpointer   data) {
	KibaStackIcon *icon = data;
	KibaStack *stack = icon->stack;
	
	if (g_list_find(icon->stack->icons, icon))
		kiba_stack_remove (stack, icon);
	kiba_stack_icon_destroy (icon);
	
	return TRUE;
}

static gboolean kiba_stack_remove_cb (GtkWidget *widget,
				      gpointer   data) {
	KibaStack *stack = data;
	GList *l;
	
	for (l = stack->icons; l != NULL; l = stack->icons) {
		kiba_stack_remove_icon_cb (widget, l->data);
	}
}

static gboolean kiba_stack_button_press_cb (GtkWidget      *widget,
					    GdkEventButton *event,
					    gpointer        data) {
	KibaStack *stack = data;
	
	if (event->button == 3) {
		GtkWidget *menu;
		GtkWidget *item;
		GtkWidget *image;
		
		menu = kiba_stack_create_context_menu_for_icon (stack->icons->data);
		if (!menu) return FALSE;
		
		item = gtk_image_menu_item_new_with_label (_("Remove Stack"));
		image = gtk_image_new_from_icon_name (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (kiba_stack_remove_cb), stack);
		gtk_widget_show (item);
	
		gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
			event->button, gtk_get_current_event_time());
		
		return TRUE;
	}
	//TODO item for directory icons to show the folder in nautilus
	
	return FALSE;
}

/**
a drag was received on the custom icon view or on the dock object,
we gonna add new KibaStackIcon's to the stack for the dragged files
**/
gboolean kiba_stack_drag_receive_cb (KibaStack        *stack,
					    GdkDragContext   *drag_context,
					    GtkSelectionData *selection_data) {
	static const char *file_prefix = "file://";
	
	KibaStackPlugin *plugin = stack->stack_plugin;
	KibaStackIcon *icon = NULL;
	KibaStackIconView *icon_view;
	char *string;
	const char *homedir;
	char *filename;
	char *unescaped;
	gboolean file_added;
	GList *l;
	
	homedir = g_get_home_dir ();
	
	char stack_dir[strlen(homedir)+strlen(STACK_DIR)];
	sprintf (stack_dir, "%s" STACK_DIR, homedir);
	
	char data[strlen((char*) selection_data->data)];
	sprintf(data, "%s", (char*) selection_data->data);
	g_strstrip (data);
	
	file_added = FALSE;
	
	/* walk through the drag data (the items are seperated with "\n") */
	while ((string = g_strrstr (data, "\n")) != NULL) {
		g_strstrip (string);
		
		/* get the filename */
		if (g_str_has_prefix(string, file_prefix)) {
			unescaped = gnome_vfs_unescape_string (string + strlen(file_prefix), NULL);
		}
		else {
			unescaped = gnome_vfs_unescape_string (string, NULL);
		}
		filename = unescaped;
		
		/* check if its a valid file */
		if (!g_file_test(filename, G_FILE_TEST_EXISTS)) {
			string[0] = '\0';
			continue;
		}
		
		/* skip if the file was allready added */
		if (kiba_stack_contains_file (stack, filename)) {
			string[0] = '\0';
			continue;
		}
		
		/* create, add and select a new KibaStackIcon instance */
		icon = kiba_stack_icon_new(stack, filename);
		g_free (unescaped);
		if (!icon) continue;
		
		kiba_stack_add (stack, icon);
		kiba_stack_select_icon (stack, icon);
		
		file_added = TRUE;
		
		string[0] = '\0';
	}
	g_strstrip (data);
	/* same for the last item */
	if (g_str_has_prefix(data, file_prefix)) {
		unescaped = gnome_vfs_unescape_string (data + strlen(file_prefix), NULL);
	}
	else {
		unescaped = gnome_vfs_unescape_string (data, NULL);
	}
	filename = unescaped;
	
	if (g_file_test(filename, G_FILE_TEST_EXISTS) && !kiba_stack_contains_file(stack, filename)) {
		icon = kiba_stack_icon_new(stack, filename);
		if (icon) {
			kiba_stack_add (stack, icon);
			kiba_stack_select_icon (stack, icon);
			file_added = TRUE;
		}
	}
	g_free (unescaped);
	
	/* select the icon-view item of the last added KibaStackIcon */
	icon_view = kiba_stack_icon_view_get();
	if (KIBA_IS_ICON_VIEW_WIN(icon_view->win) && icon_view->stack == stack && icon) {
		KibaIconView *view;
		
		view = KIBA_ICON_VIEW(icon_view->win->icon_view);
		
		for (l = gtk_container_get_children(GTK_CONTAINER(view)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			KibaStackIcon *check_icon;
			
			check_icon = g_object_get_data (G_OBJECT(child), "stack-icon");
			if (check_icon == icon) {
				kiba_icon_view_select_child (view, child);
				break;
			}
		}
	}
	
	return file_added;
}

static void unset_custom_icon (GtkWidget     *widget,
			       KibaStackIcon *icon) {
	g_free (icon->custom_icon);
	icon->custom_icon = NULL;
	kiba_stack_icon_update_surface (icon);
}

static gboolean kiba_stack_button_release_cb (GtkWidget      *widget,
					      GdkEventButton *event,
					      gpointer        data) {
	KibaStack *stack = data;
	
	if (event->button == 3)
		return TRUE;
	
	if (event->button == 2) {
		kiba_stack_activate_icon ((KibaStackIcon*) stack->icons->data);
	} else {
		kiba_stack_icon_view_toggle (stack, (KibaStackIcon*) stack->icons->data);
	}
	
	return TRUE;
}

static void kiba_object_drag_received_cb (GtkWidget        *widget,
					  GdkDragContext   *drag_context,
					  gint              x,
					  gint              y,
					  GtkSelectionData *selection_data,
					  guint             info,
					  guint             time,
					  gpointer          user_data) {
	KibaStack *stack = user_data;
	
	gdk_drag_status (drag_context, GDK_ACTION_COPY, time);
	kiba_stack_drag_receive_cb (stack, drag_context, selection_data);
	gtk_drag_finish (drag_context,
		TRUE,
		FALSE, time);
}

static void kiba_stack_scroll_cb (GtkWidget      *widget,
				  GdkEventScroll *event,
				  gpointer        data) {
	KibaStackIconView *icon_view = kiba_stack_icon_view_get();
	KibaStackIcon *newActive;
	KibaStack *stack = (KibaStack*) data;
	
	if (g_list_length(stack->icons) <= 1)
		return;
	
	kiba_drawable_object_switch_surface (KIBA_DRAWABLE_OBJECT(widget));
	kiba_utils_glist_scroll (&stack->icons, event->direction);
	__update_object_surface(stack);
	newActive = stack->icons->data;
	kiba_object_set_title (KIBA_OBJECT(stack->widget), newActive->name);
	
	if (icon_view->stack == stack) {
		KibaIconView *view;
		GList *l;
		
		view = KIBA_ICON_VIEW(icon_view->win->icon_view);
		
		for (l = gtk_container_get_children(GTK_CONTAINER(view)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			KibaStackIcon *check_icon;
			
			check_icon = g_object_get_data (G_OBJECT(child), "stack-icon");
			if (check_icon == newActive) {
				kiba_icon_view_select_child (view, child);
				break;
			}
		}
	}
}

static void kiba_stack_add (KibaStack *stack,
                            KibaStackIcon *icon) {
	KibaStackIconView *icon_view = kiba_stack_icon_view_get();
	Kiba *kiba = kiba_get();
	
	stack->icons = g_list_append (stack->icons, icon);
	
	icon->stack = stack;
	
	if (g_list_length(stack->icons) == 1) {
		KibaDrawableObject *drawable_object;
		stack->widget = kiba_icon_new(STACK_PLUGIN_NAME);
		kiba_add_plugin_object(kiba, stack->stack_plugin->plugin, KIBA_OBJECT(stack->widget));
		gtk_widget_show (stack->widget);
		
		kiba_object_set_title (KIBA_OBJECT(stack->widget), icon->name);
		
		drawable_object = KIBA_DRAWABLE_OBJECT(stack->widget);
		drawable_object->surface_data->width  = KIBA_OBJECT(stack->widget)->size.width;
		drawable_object->surface_data->height = KIBA_OBJECT(stack->widget)->size.height;
		
		g_signal_connect_after (G_OBJECT (stack->widget), "button-press-event",
				G_CALLBACK (kiba_stack_button_press_cb), stack);
		g_signal_connect_after (G_OBJECT (stack->widget), "button-release-event",
				G_CALLBACK (kiba_stack_button_release_cb), stack);
		g_signal_connect (G_OBJECT (stack->widget), "scroll-event",
				G_CALLBACK (kiba_stack_scroll_cb), stack);
		g_signal_connect (G_OBJECT (stack->widget), "drag-data-received",
				G_CALLBACK (kiba_object_drag_received_cb), stack);
	}
	kiba_stack_icon_update_surface (icon);
	
	if (icon_view->stack == stack && icon_view->type == KIBA_STACK_TYPE_CUSTOM) {
		kiba_stack_icon_view_add_icon(icon_view, icon);
	}
}

void kiba_stack_remove (KibaStack *stack,
                        KibaStackIcon *icon) {
	KibaStackIconView *icon_view = kiba_stack_icon_view_get();
	GtkWidget *box;
	GList *l;
	
	g_return_if_fail (g_list_find (stack->icons, icon) != NULL);
	
	stack->icons = g_list_remove (stack->icons, icon);
	
	if (icon_view->stack == stack) {
		KibaIconView *view;
		view = KIBA_ICON_VIEW(icon_view->win->icon_view);
		
		for (l = gtk_container_get_children(GTK_CONTAINER(view)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			KibaStackIcon *check_icon;
			
			check_icon = g_object_get_data (G_OBJECT(child), "stack-icon");
			if (check_icon == icon) {
				kiba_icon_view_win_remove (icon_view->win, child);
				break;
			}
		}
	}
	
	/* the removed object was the last one in the stack */
	if (g_list_length(stack->icons) == 0) {
		if (stack->uri) {
			g_remove (stack->uri);
		}
		stack->stack_plugin->stacks = g_list_remove (stack->stack_plugin->stacks, stack);
		kiba_stack_destroy (stack);
	}
	else {
		kiba_drawable_object_switch_surface (KIBA_DRAWABLE_OBJECT(stack->widget));
		__update_object_surface(stack);
		kiba_object_set_title (KIBA_OBJECT(stack->widget), icon->name);
	}
}

/*** Public Functions ***/

/**
Select the given KibaStackIcon (switching dock object surface)
**/
void kiba_stack_select_icon (KibaStack *stack,
			     KibaStackIcon *icon) {
	kiba_drawable_object_switch_surface (KIBA_DRAWABLE_OBJECT(stack->widget));
	while (stack->icons->data != icon)
		kiba_utils_glist_scroll (&stack->icons, GDK_SCROLL_DOWN);
	__update_object_surface(stack);
	kiba_object_set_title (KIBA_OBJECT(stack->widget), icon->name);
}

GtkWidget* kiba_stack_create_context_menu_for_icon (KibaStackIcon *icon) {
	GtkWidget *menu = NULL;
	GtkWidget *item;
	GtkWidget *image;
	
	menu = gtk_menu_new ();
	
	item = gtk_image_menu_item_new_with_label (_("Properties"));
	image = gtk_image_new_from_icon_name (GTK_STOCK_PROPERTIES, GTK_ICON_SIZE_MENU);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
	g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (kiba_stack_file_properties_cb), icon);
	gtk_widget_show (item);
	
	if (g_list_find(icon->stack->icons, icon)) {
		kiba_menu_add_separator (GTK_WIDGET (menu));
		
		item = gtk_image_menu_item_new_with_label (_("Remove Stack Item"));
		image = gtk_image_new_from_icon_name (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (kiba_stack_remove_icon_cb), icon);
		gtk_widget_show (item);
	}
	
	return menu;
}

/**
parse the stack text file at stack->uri
and add a KibaStackIcon for each file
**/
static void kiba_stack_parse_file (KibaStack *stack) {
	FILE *file;
	char icon_uri[400];
	
	file = fopen (stack->uri, "r");
	if (file == NULL)
		return;
	
	while (fgets(icon_uri, 400, file) != NULL) {
		KibaStackIcon *stack_icon;
		
		g_strstrip (icon_uri);
		
		if (!g_file_test(icon_uri, G_FILE_TEST_EXISTS))
			continue;
		
		stack_icon = kiba_stack_icon_new(stack, icon_uri);
		if (stack_icon)
			kiba_stack_add (stack, stack_icon);
	}
	
	fclose (file);
}

static void __execute_file_response (GtkDialog *dialog,
                                     gint response_id,
                                     gpointer data) {
	KibaStackIcon *icon = data;
	
	switch (response_id) {
		case GTK_RESPONSE_YES:
			kiba_object_launch (KIBA_OBJECT(icon->stack->widget), icon->url);
			gtk_widget_destroy(GTK_WIDGET(dialog));
			break;
		case GTK_RESPONSE_CANCEL:
			gtk_widget_destroy(GTK_WIDGET(dialog));
			break;
	}
}

/**
activate a stack icon, launch the mime app for files and open the directory view for directory icons
**/
void kiba_stack_activate_icon (KibaStackIcon *icon) {
	KibaStackIconView *icon_view = kiba_stack_icon_view_get();
	char *command;
	gboolean custom_stack_icon;
	
	command = kiba_stack_icon_get_command(icon);
	
	if (g_list_find (icon->stack->icons, icon))
		custom_stack_icon = TRUE;
	else
		custom_stack_icon = FALSE;
	
	/* the icon is a directory */
	if (icon->type == KIBA_STACK_ICON_TYPE_DIRECTORY) {
		if (icon_view->win && icon->stack != icon_view->stack) {
			kiba_stack_icon_view_toggle (icon->stack, icon);
		}
		icon_view->stack = icon->stack;
		if (custom_stack_icon) {
			if (icon_view->icon) {
				unset_custom_icon (NULL, icon_view->icon);
				if (icon_view->win->win)
					g_signal_handlers_disconnect_by_func (G_OBJECT(icon_view->win->win),
						G_CALLBACK(unset_custom_icon), icon_view->icon);
			}
			icon_view->icon = icon;
		}
		kiba_icon_view_activate_directory_icon (icon);
		if (icon->stack->stack_plugin->open_dirs && custom_stack_icon) {
			if (icon->custom_icon)
				g_free (icon->custom_icon);
			icon->custom_icon = g_strdup("folder-open");
			kiba_stack_icon_update_surface (icon);
			g_signal_connect (G_OBJECT(icon_view->win->win), "destroy",
				G_CALLBACK(unset_custom_icon), icon);
		}
	}
	/* a mime app was set, we can launch the file */
	else if (command) {
		//TODO: if the file is executable, popup a dialog offering to display or execute the file
		kiba_object_launch (KIBA_OBJECT(icon->stack->widget), command);
	}
	/* no mime app was associated previously, we gonna create a app-chooser */
	else {
		if (g_file_test(icon->url, G_FILE_TEST_IS_EXECUTABLE)) {
			char *title;
			char *message;
			GtkWidget *dialog;
			title = g_strdup_printf (_("Launching '%s'?"), icon->url);
			message = g_strdup_printf("%s",
				_("Are you sure you want to launch this executable?"));
			dialog = kiba_dialog_confirm(title, message);
			g_free(title);
			g_free(message);
			
			g_signal_connect(G_OBJECT(dialog), "response",
				G_CALLBACK(__execute_file_response), icon);
		} else {
			char *title;
			char *message;
			title = g_strdup_printf (_("Cannot display '%s'!"), icon->url);
			message = g_strdup_printf("%s\n%s",
				_("No associated MIME application!"),
				_("Please select an application from the file properties dialog."));
			kiba_dialog_error(title, message);
			g_free(title);
			g_free(message);
		}
	}
	
	g_free(command);
}

/**
Class Functions
**/

KibaStack* kiba_stack_new (KibaStackPlugin *plugin,
			   const char      *uri) {
	Kiba *kiba = kiba_get();
	KibaStack *stack;
	KibaWin *win;
	
	win = kiba_get_plugin_win(kiba, plugin->plugin);
	
	stack = g_new0 (KibaStack, 1);
	stack->stack_plugin = plugin;
	stack->uri = g_strdup (uri);
	stack->widget = NULL;
	stack->icons  = NULL;
	
	kiba_stack_parse_file(stack);
	
	return stack;
}

void kiba_stack_destroy (KibaStack *stack) {
	KibaStackIconView *icon_view = kiba_stack_icon_view_get();
	GList *l;
	
	if (icon_view && icon_view->stack == stack) {
		gtk_widget_destroy (icon_view->win->win);
	}
	
	for (l = stack->icons; l != NULL; l = stack->icons) {
		KibaStackIcon *icon = l->data;
		
		stack->icons = g_list_remove (stack->icons, l->data);
		
		kiba_stack_icon_destroy (icon);
	}
	g_list_free (stack->icons);
	
	if (KIBA_IS_OBJECT(stack->widget))
		kiba_object_destroy (KIBA_OBJECT(stack->widget));
	
	g_free (stack->uri);
	g_free (stack);
}
