/*
 *  Copyright (C) 2007  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <libintl.h>
#include <math.h>
#include <gtk/gtk.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-object.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>

#define _(string)   gettext (string)
#define PLUGIN_NAME "switch"

typedef struct _KibaSwitch       KibaSwitch;
typedef struct _KibaSwitchObject KibaSwitchObject;

struct _KibaSwitchObject {
	KibaSwitch *kiba_switch;
	GtkWidget *widget;
	float factor;
	gint switch_id;
};

struct _KibaSwitch {
	GList *objects;
	
	float step;
	gboolean object3d;
	gint orientation;
};

//TODO remove _KibaSwitchObject's after switching

//must be %2=0
#define OBJECT_DEPTH 8
#define SURFACE_SPACE 0.75f

/**
Helper Functions
**/

static KibaSwitchObject* kiba_switch_get_object_for_widget (KibaSwitch  *kiba_switch,
							    GtkWidget *widget) {
	GList *l;
	
	for (l = kiba_switch->objects; l != NULL; l = l->next) {
		KibaSwitchObject *object = l->data;
		
		if (object->widget == widget)
			return object;
	}
	
	return NULL;
}

static void __queue_draw(KibaSwitchObject *switch_obj) {
	KibaObject *object = KIBA_OBJECT(switch_obj->widget);
	KibaWin *win = KIBA_WIN(switch_obj->widget->parent->parent);
	GdkRectangle area;
	GtkAllocation allocation = switch_obj->widget->allocation;
	float scale_x = 1.0f, scale_y = 1.0f;
	float offsetX = 0.0f, offsetY = 0.0f;
	GList *l;
	const KibaLayoutData *d;
	
	d = &kiba_layout_data[KIBA_DOCK(KIBA_WIN_CHILD(win))->d.position];
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		float customX = 0.0f, customY = 0.0f;
		KibaObjectCustomScaleData *data = l->data;
		
		if (data->scale_widget)
			continue;
		
		data->scale_func (object, &customX, &customY, data->data);
		if (customX > 1.0f) scale_x *= customX;
		if (customY > 1.0f) scale_y *= customY;
	}
	
	for (l = object->custom_offset_data; l != NULL; l = l->next) {
		float customX = 0.0f, customY = 0.0f;
		KibaObjectCustomOffsetData *data = l->data;
		
		data->place_func (object, &customX, &customY, data->data);
		offsetX += customX;
		offsetY += customY;
	}
	
	area.width = allocation.width*scale_x;
	area.height = allocation.height*scale_y;
	area.x = allocation.x - (area.width - allocation.width)/2 - 8;
	area.y = allocation.y - (area.height - allocation.height)/2 - 8;
	area.height += 16;
	area.width  += 16;
	
	area.y -= (d->valignment-1)*(offsetY);
	area.x -= (d->halignment-1)*(offsetX);
	
	kiba_win_invalidate_area (GTK_WIDGET(win), area);
}

static gint kiba_switch_object_timeout (gpointer data) {
	KibaSwitchObject *object = data;
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object->widget);
	
	kiba_drawable_object_queue_draw (drawable_object);
	
	if (drawable_object->switch_surface)
		object->factor -= object->kiba_switch->step;
	else
		object->factor += object->kiba_switch->step;
	
	if (object->factor <= 0.05) {
		object->factor = 0.05; /* scaling down to 0.0 produces a mysterios white rectangle */
		
		cairo_surface_destroy (drawable_object->switch_surface);
		drawable_object->switch_surface = NULL;
	}
	else if (object->factor >= 1.0) {
		object->factor = 1.0;
		kiba_object_update_scale_factor (KIBA_OBJECT(object->widget));
		
		object->switch_id = 0;
		
		return FALSE;
	}
	
	kiba_object_update_scale_factor (KIBA_OBJECT(object->widget));
	if (object->kiba_switch->object3d)
		__queue_draw(object);
	
	return TRUE;
}

static void kiba_object_scale_func (KibaObject *object,
				    float     *scale_x,
				    float     *scale_y,
				    gpointer    data) {
	KibaSwitchObject *switch_object = data;
	
	if (GET_FLAG(switch_object->kiba_switch->orientation, GTK_ORIENTATION_VERTICAL)) {
		*scale_y = switch_object->factor;
		if (GET_FLAG(switch_object->kiba_switch->orientation, GTK_ORIENTATION_HORIZONTAL))
			*scale_x = switch_object->factor;
		else
			*scale_x = 1.0;
	}
	else {
		*scale_x = switch_object->factor;
		*scale_y = 1.0;
	}
}

static void kiba_object_max_scale_func (KibaObject *object,
				        float     *scale_x,
				        float     *scale_y,
				        gpointer    data) {
	KibaSwitchObject *switch_object = data;
	
	*scale_x = 1.0;
	*scale_y = 1.0;
}

static void __paint_3d_object_before (KibaDrawableObject *drawable_object,
                               cairo_t *cr,
                               gpointer data) {
	KibaSwitchObject *switch_obj = data;
	gint i, depth;
	gint factor;
	
	if (switch_obj->factor >= 1.0f) return;
	
	if (drawable_object->switch_surface)
		factor = -1;
	else
		factor = 1;
	
	for (i = OBJECT_DEPTH/2; i > 0; i--) {
		cairo_save(cr);
			cairo_scale(cr, 1.0f/switch_obj->factor, 1.0f/switch_obj->factor);
			if (GET_FLAG(switch_obj->kiba_switch->orientation, GTK_ORIENTATION_HORIZONTAL))
				cairo_translate(cr, factor*i*SURFACE_SPACE*(1.0f - switch_obj->factor), 0);
			else
				cairo_translate(cr, 0, factor*i*SURFACE_SPACE*(1.0f - switch_obj->factor));
			cairo_scale(cr, switch_obj->factor, switch_obj->factor);
			if (!drawable_object->paint)
				kiba_drawable_object_paint (drawable_object, cr, NULL);
			else
				drawable_object->paint (drawable_object, cr, drawable_object->paint_data);
		cairo_restore (cr);
	}
}

static void __paint_3d_object_after (KibaDrawableObject *drawable_object,
                               cairo_t *cr,
                               gpointer data) {
	KibaSwitchObject *switch_obj = data;
	gint i, depth, factor;
	
	if (switch_obj->factor >= 1.0f) return;
	
	if (drawable_object->switch_surface)
		factor = 1;
	else
		factor = -1;
	
	for (i = 1; i < OBJECT_DEPTH/2 + 1; i++) {
		cairo_save(cr);
			cairo_scale(cr, 1.0f/switch_obj->factor, 1.0f/switch_obj->factor);
			if (GET_FLAG(switch_obj->kiba_switch->orientation, GTK_ORIENTATION_HORIZONTAL))
				cairo_translate(cr, factor*i*SURFACE_SPACE*(1.0f - switch_obj->factor), 0);
			else
				cairo_translate(cr, 0, factor*i*SURFACE_SPACE*(1.0f - switch_obj->factor));
			cairo_scale(cr, switch_obj->factor, switch_obj->factor);
			if (!drawable_object->paint)
				kiba_drawable_object_paint (drawable_object, cr, NULL);
			else
				drawable_object->paint (drawable_object, cr, drawable_object->paint_data);
		cairo_restore (cr);
	}
}

void kiba_switch_object (KibaDrawableObject *drawable_object,
			 gpointer    data) {
	KibaSwitch *kiba_switch = data;
	KibaSwitchObject *switch_obj;
	KibaObject *object = KIBA_OBJECT(drawable_object);
	
	switch_obj = kiba_switch_get_object_for_widget (kiba_switch, GTK_WIDGET(object));
	
	if (!switch_obj) {
		switch_obj = g_new0(KibaSwitchObject, 1);
		switch_obj->factor = 1.0;
		switch_obj->kiba_switch = kiba_switch;
		switch_obj->switch_id = 0;
		switch_obj->widget = GTK_WIDGET(object);
		
		kiba_switch->objects = g_list_append (kiba_switch->objects, switch_obj);
		
		kiba_object_add_custom_scale_function (object,
			kiba_object_scale_func,
			kiba_object_max_scale_func,
			FALSE,
			switch_obj);
		if (kiba_switch->object3d) {
			kiba_drawable_object_add_custom_paint_function (drawable_object,
				__paint_3d_object_before,
				TRUE,
				switch_obj);
			kiba_drawable_object_add_custom_paint_function (drawable_object,
				__paint_3d_object_after,
				FALSE,
				switch_obj);
		}
	}
	
	if (!drawable_object->switch_surface) {
		drawable_object->switch_surface = cairo_surface_reference (drawable_object->surface_data->surface);
		drawable_object->surface_data->surface = NULL;
	}
	
	if (switch_obj->switch_id)
		return;
	
	switch_obj->switch_id = g_timeout_add (20, kiba_switch_object_timeout, switch_obj);
}

/**
Signal Handler
**/

static void kiba_prefs_switch_3d_cb (const char *option_id,
				     gpointer    value,
				     gpointer    data) {
	KibaSwitch *kiba_switch = data;
	GList *l;
	
	kiba_switch->object3d = GPOINTER_TO_INT(value);
	
	if (!kiba_switch->object3d) {
		for (l = kiba_switch->objects; l != NULL; l = l->next) {
			KibaSwitchObject *object = l->data;
			kiba_drawable_object_remove_custom_paint_function_by_data (KIBA_DRAWABLE_OBJECT(object->widget),
				object);
		}
	}
	else {
		for (l = kiba_switch->objects; l != NULL; l = l->next) {
			KibaSwitchObject *object = l->data;
			kiba_drawable_object_add_custom_paint_function (KIBA_DRAWABLE_OBJECT(object->widget),
				__paint_3d_object_before,
				TRUE,
				object);
			kiba_drawable_object_add_custom_paint_function (KIBA_DRAWABLE_OBJECT(object->widget),
				__paint_3d_object_after,
				FALSE,
				object);
		}
	}
}

static void kiba_prefs_switch_step_cb (const char *option_id,
				     gpointer    value,
				     gpointer    data) {
	KibaSwitch *kiba_switch = data;
	
	kiba_switch->step = g_ascii_strtod((char*) value, NULL);
}

static void kiba_prefs_orientation_cb (const char *option_id,
					      gpointer    value,
					      gpointer    data) {
	KibaSwitch *kiba_switch = data;
	char *orient = (char*) value;
	gboolean horizontal = FALSE, vertical = FALSE;
	
	if (strcmp (orient, "horizontal") == 0)
		horizontal = TRUE;
	else if (strcmp (orient, "vertical") == 0)
		vertical = TRUE;
	else if (strcmp (orient, "both") == 0) {
		vertical = TRUE;
		horizontal = TRUE;
	}
	
	if (horizontal) {
		if (!GET_FLAG(kiba_switch->orientation, GTK_ORIENTATION_HORIZONTAL))
			SET_FLAG(kiba_switch->orientation, GTK_ORIENTATION_HORIZONTAL);
	}
	else if (GET_FLAG(kiba_switch->orientation, GTK_ORIENTATION_HORIZONTAL))
		DEL_FLAG(kiba_switch->orientation, GTK_ORIENTATION_HORIZONTAL);
	if (vertical) {
		if (!GET_FLAG(kiba_switch->orientation, GTK_ORIENTATION_VERTICAL))
			SET_FLAG(kiba_switch->orientation, GTK_ORIENTATION_VERTICAL);
	}
	else if (GET_FLAG(kiba_switch->orientation, GTK_ORIENTATION_VERTICAL))
		DEL_FLAG(kiba_switch->orientation, GTK_ORIENTATION_VERTICAL);
}

static void dock_object_add_cb (GtkContainer *container,
				GtkWidget    *widget,
				gpointer      data) {
	if (KIBA_IS_DRAWABLE_OBJECT(widget))
		g_signal_connect (G_OBJECT(widget), "switch-event",
			G_CALLBACK(kiba_switch_object), data);
}

static void dock_object_remove_cb (GtkContainer *container,
				   GtkWidget    *widget,
				   gpointer      data) {
	
	KibaSwitchObject *object = NULL;
	
	object = kiba_switch_get_object_for_widget (data, widget);
	
	if (!object)
		return;
	
	if (object->switch_id)
		g_source_remove (object->switch_id);
	
	if (object->factor != 1.0) {
		object->factor = 1.0;
		kiba_object_update_scale_factor (KIBA_OBJECT(object->widget));
	}
	
	kiba_object_remove_custom_scale_function_by_data (KIBA_OBJECT(widget),
		object);
	if (object->kiba_switch->object3d)
		kiba_drawable_object_remove_custom_paint_function_by_data (KIBA_DRAWABLE_OBJECT(object->widget),
			object);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(kiba_switch_object), object->kiba_switch);
	
	object->kiba_switch->objects = g_list_remove (object->kiba_switch->objects, object);
	g_free (object);
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaSwitch *kiba_switch;
	KibaWin *win;
	KibaDock *dock;
	GList *l;
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	kiba_switch = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	for (l = kiba_switch->objects; l != NULL; l = kiba_switch->objects) {
		KibaSwitchObject *object = l->data;
		
		if (object->switch_id)
			g_source_remove (object->switch_id);
		
		if (object->factor != 1.0) {
			object->factor = 1.0;
			kiba_object_update_scale_factor (KIBA_OBJECT(object->widget));
		}
		
		kiba_object_remove_custom_scale_function_by_data (KIBA_OBJECT(object->widget),
			object);
		if (kiba_switch->object3d)
			kiba_drawable_object_remove_custom_paint_function_by_data (KIBA_DRAWABLE_OBJECT(object->widget),
				object);
		
		g_free (object);
		kiba_switch->objects = g_list_remove (kiba_switch->objects, l->data);
	}
	g_list_free (kiba_switch->objects);
	
	for (l = gtk_container_get_children(GTK_CONTAINER(dock)); l != NULL; l = l->next) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(l->data),
			G_CALLBACK(kiba_switch_object), kiba_switch);
	}
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_add_cb), kiba_switch);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_remove_cb), kiba_switch);
	
	g_free (kiba_switch);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaSwitch *kiba_switch;
	KibaWin *win;
	KibaDock *dock;
	GtkContainer *container;
	KibaPrefsFile *file;
	GList *l;
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	container = GTK_CONTAINER(dock);
	
	kiba_switch = g_new0(KibaSwitch, 1);
	kiba_switch->step = 0.5;
	kiba_switch->orientation = GTK_ORIENTATION_VERTICAL;
	kiba_switch->objects = NULL;
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_add_callback (file, "switch_step",
		kiba_prefs_switch_step_cb, kiba_switch, TRUE);
	kiba_prefs_add_callback (file, "switch_3d_objects",
		kiba_prefs_switch_3d_cb, kiba_switch, TRUE);
	kiba_prefs_add_callback (file, "switch_orientation",
		kiba_prefs_orientation_cb, kiba_switch, TRUE);
	
	g_signal_connect_after (G_OBJECT(dock), "add",
		G_CALLBACK(dock_object_add_cb), kiba_switch);
	g_signal_connect_after (G_OBJECT(dock), "remove",
		G_CALLBACK(dock_object_remove_cb), kiba_switch);

	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		
		dock_object_add_cb (container, child, kiba_switch);
	}
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) kiba_switch);
	
	return TRUE;
}

static PrefMember switch_options[] = {
    { "switch_step", "0.5", KIBA_DOUBLE },
    { "switch_3d_objects", "FALSE", KIBA_BOOLEAN },
    { "switch_orientation", "horizontal", KIBA_STRING },
};

static KibaPluginVtable vtable = {
	"Switch Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
};

KIBA_INIT_PLUGIN (vtable, switch_options, NULL, NULL, NULL);
