/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <math.h>
#include <libintl.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <glibtop.h>
#include <glibtop/cpu.h>
#include <glibtop/mem.h>
#include <glibtop/swap.h>
#include <glibtop/mountlist.h>
#include <glibtop/fsusage.h>
#include <glibtop/netlist.h>
#include <glibtop/netload.h>
#include <glibtop/procstate.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-drawable-object.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-screen.h>
#include <kiba-title.h>
#include <kiba-debug.h>
#include <kiba-dialog.h>
#include <kiba-cairo-utils.h>
#include <kiba-utils.h>
#include <kiba-macros.h>
#include <kiba-svg-macro.h>
#ifdef USE_SVG
  #include <librsvg/rsvg.h>
#endif

#define _(String)    dgettext (GETTEXT_PACKAGE, String)
#define PLUGIN_NAME  "sysinfo"
#define SYSINFO_DIR  "/.kiba-dock/"PLUGIN_NAME
#define FINGER_STEP  0.01

typedef struct _KibaSysinfo KibaSysinfo;

enum {
	KIBA_SYS_CPU = 0,
	KIBA_SYS_MEM,
	KIBA_SYS_SWAP,
	KIBA_SYS_NET,
	KIBA_SYS_ROOT,
};

static const gchar* sysinfo_device_data[] = {
	"cpu",      /**< KIBA_SYS_CPU.    */
	"memory",   /**< KIBA_SYS_MEM.    */
	"swap",     /**< KIBA_SYS_SWAP.   */
	"net",      /**< KIBA_SYS_NET.    */
	"root"      /**< KIBA_SYS_ROOT.   */
};

struct _KibaSysinfo {
	GtkWidget       *object;
	gint             flag;
	
	float           sysinfo_used_percent;
	float           sysinfo_last_percent;
	
	struct {
		unsigned int         frequency;
		struct {
			unsigned int total, idle;
		}last[GLIBTOP_NCPU];
	}cpu;
	
	struct {
		unsigned int   last_in;
	}mem;
	
	gboolean           device_changed;
	
	CairoSurfaceData   foreground_surface;
	CairoSurfaceData   hand_surface;
	PangoLayout       *pango_layout;
	KibaRGBA           text_rgba;
	
	gint               timeout_id;
	gint               finger_animate_id;
	
	char              *theme;
	gint               max_netload;
};

typedef enum {
	KIBA_SYSINFO_HAND_SURFACE = 0,
	KIBA_SYSINFO_FOREGROUND,
	KIBA_SYSINFO_BACKGROUND
}KibaSysinfoIconLevel;

static const struct { const char *name; KibaSysinfoIconLevel level; } sysinfo_theme_data[] = {
	"sysinfo-hand", KIBA_SYSINFO_HAND_SURFACE,
	"sysinfo-frame", KIBA_SYSINFO_FOREGROUND,
	"sysinfo-face", KIBA_SYSINFO_BACKGROUND,
};

/**
Helper Functions
**/

#ifdef USE_SVG

static char* kiba_sysinfo_get_theme_path (KibaSysinfo *sysinfo) {
	char *dir;
	
	if (!sysinfo->theme || strcmp (sysinfo->theme, "none") == 0)
		sysinfo->theme = "default";
	
	if (g_strrstr (sysinfo->theme, "/") &&
	    g_file_test (sysinfo->theme, G_FILE_TEST_IS_DIR))
		return g_strdup (sysinfo->theme);
	
	dir = g_strdup_printf ("%s" SYSINFO_DIR "/themes/%s", g_get_home_dir (), sysinfo->theme);
	if (g_file_test (dir, G_FILE_TEST_IS_DIR))
		return dir;
	g_free (dir);
	dir = NULL;
	
	dir = g_strdup_printf (KIBA_SYSINFO_THEME_DIR "/%s", sysinfo->theme);
	if (g_file_test (dir, G_FILE_TEST_IS_DIR))
		return dir;
	g_free (dir);
	dir = NULL;
	
	dir = g_strdup_printf (KIBA_PLUGINS_TOP_SRCDIR "/files/sysinfo_themes/%s", sysinfo->theme);
	if (g_file_test (dir, G_FILE_TEST_IS_DIR))
		return dir;
	g_free (dir);
	dir = NULL;
	
	return NULL;
}

static void kiba_sysinfo_set_surface_size (KibaSysinfo *sysinfo) {
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(sysinfo->object);
	gint i;
	
	drawable_object->surface_data->width  = sysinfo->object->allocation.width;
	drawable_object->surface_data->height = sysinfo->object->allocation.width;
	
	CairoSurfaceData *surface_data[KIBA_SYSINFO_FOREGROUND+1] = {
		&sysinfo->hand_surface,
		&sysinfo->foreground_surface
	};
	
	for (i = 0; i < KIBA_SYSINFO_FOREGROUND+1; i++) {
		surface_data[i]->width  = sysinfo->object->allocation.width;
		surface_data[i]->height = sysinfo->object->allocation.width;
	}
}

static float kiba_sysinfo_get_rotate_angle (KibaSysinfo *sysinfo) {
	if (sysinfo->sysinfo_used_percent > sysinfo->sysinfo_last_percent + FINGER_STEP)
		return sysinfo->sysinfo_last_percent + FINGER_STEP;
	else if (sysinfo->sysinfo_used_percent < sysinfo->sysinfo_last_percent - FINGER_STEP)
		return sysinfo->sysinfo_last_percent - FINGER_STEP;
	
	return sysinfo->sysinfo_used_percent;
}

/* run the full-scale system process monitor */
static void start_procman (KibaSysinfo *sysinfo) {
	GError *error = NULL;
	
	gdk_spawn_command_line_on_screen (
			KIBA_GDK_SCREEN,
			"gnome-system-monitor", &error);
	if (error) {
		kiba_dialog_error (_("Failed to launch Programm"),
			_("There was an error executing '%s': %s"),
			"gnome-system-monitor", error->message);
		
		g_error_free (error);
	}
}

static void kiba_sysinfo_set_device (KibaSysinfo *sysinfo,
				     const char  *device) {
	if (strcmp (device, "swap") == 0)
		sysinfo->flag = KIBA_SYS_SWAP;
	else if (strcmp (device, "root") == 0)
		sysinfo->flag = KIBA_SYS_ROOT;
	else if (strcmp (device, "net") == 0)
		sysinfo->flag = KIBA_SYS_NET;
	else if (strcmp (device, "memory") == 0)
		sysinfo->flag = KIBA_SYS_MEM;
	else
		sysinfo->flag = KIBA_SYS_CPU;
	
	sysinfo->device_changed = TRUE;
}

static gint sysinfo_animate_finger_step (gpointer data) {
	KibaSysinfo *sysinfo = data;
	float load_percent;
	
	load_percent = kiba_sysinfo_get_rotate_angle (sysinfo);
	
	sysinfo->sysinfo_last_percent = load_percent;
	
	if (sysinfo->sysinfo_last_percent == sysinfo->sysinfo_used_percent) {
		sysinfo->finger_animate_id = 0;
		return FALSE;
	}
	
	kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(sysinfo->object));
	
	return TRUE;
}

static gboolean kiba_sysinfo_time_handler (KibaSysinfo *sysinfo) {
	float current_percent = sysinfo->sysinfo_used_percent;
	KibaWin *win;
	GtkWidget *dock;
	cairo_t *cr;
	char *icon_text;
	
	dock = GTK_WIDGET(sysinfo->object)->parent;
	win = KIBA_WIN(dock->parent);
	
	switch (sysinfo->flag) {
		case KIBA_SYS_SWAP: {
			glibtop_swap swap;
			unsigned long used;
			
			glibtop_get_swap (&swap);
			
			used = (unsigned long) (swap.used/(1024*1024));
			current_percent = ((float) used * 1024.0 * 1024.0) / (float) swap.total;
			
			if (sysinfo->device_changed || current_percent != sysinfo->sysinfo_used_percent) {
				char *title;
				
				title = g_strdup_printf (_("SWAP: %ld MB of %ld MB used"),
						used, (unsigned long) swap.total/(1024*1024));
				kiba_object_set_title (KIBA_OBJECT(sysinfo->object), title);
				g_free (title);
			}
			break;
		}
		case KIBA_SYS_NET: {
			glibtop_netlist netlist;
			char **ifnames;
			guint32 i;
			guint64 in = 0;
			// , out = 0;
			int data_per_second;
		
			ifnames = glibtop_get_netlist(&netlist);
		
			for (i = 0; i < netlist.number; ++i) {
				glibtop_netload netload;
				glibtop_get_netload (&netload, ifnames[i]);
				
				if (netload.if_flags & (1 << GLIBTOP_IF_FLAGS_LOOPBACK))
					continue;
				
				in  += netload.bytes_in;
				//       out += netload.bytes_out;
			}
			g_strfreev(ifnames);
			
			data_per_second = ((float) (in - sysinfo->mem.last_in)/1024.0);
			if ((float) data_per_second > (float) sysinfo->max_netload)
				data_per_second = (float) sysinfo->max_netload;
			
			if (data_per_second <= 0) {
				current_percent = 0.0;
			}
			else {
				current_percent = (float) data_per_second/sysinfo->max_netload;
				current_percent = MIN(current_percent, 1.0);
			}
			
			if (sysinfo->device_changed || current_percent != sysinfo->sysinfo_used_percent) {
				char *title;
				if (sysinfo->sysinfo_used_percent == 0.0) {
					title = g_strdup_printf (_("Netload at %i KB/s"), 0);
				}
				else {
					title = g_strdup_printf (_("Netload at %i KB/s"), data_per_second);
				}
				kiba_object_set_title (KIBA_OBJECT(sysinfo->object), title);
				g_free (title);
			}
			sysinfo->mem.last_in = in;
			
			break;
		}
		case KIBA_SYS_ROOT: {
			glibtop_mountlist mountlist;
			glibtop_mountentry *mount_entries;
			glibtop_fsusage fsusage;
			size_t i;
			
			mount_entries = glibtop_get_mountlist(&mountlist, 1);
			
			for(i = 0; i < mountlist.number; ++i) {
				if (strcmp (mount_entries[i].mountdir, "/") != 0)
					continue;
				/* Root Partition found */
				
				glibtop_get_fsusage(&fsusage, mount_entries[i].mountdir);
				
				current_percent = 1.0 - (float)fsusage.bavail/(float)fsusage.blocks;
				
				if (sysinfo->device_changed ||
				    current_percent != sysinfo->sysinfo_used_percent) {
					int total_gb, used_gb;
					char *title;
					
					total_gb = (int) (100*((float)fsusage.blocks)/(512.0*512.0));
					used_gb  = (int) (100*((float)
						(fsusage.blocks-fsusage.bfree))/(512.0*512.0));
						
					title = g_strdup_printf (_("%s: %g GB of %g GB used"),
							mount_entries[i].devname,
							((float)used_gb)/100.0,
							((float)total_gb)/100.0);
					kiba_object_set_title (KIBA_OBJECT(sysinfo->object), title);
					g_free (title);
				}
				break;
			}
			g_free(mount_entries);
			
			break;
		}
		case KIBA_SYS_MEM: {
			unsigned long used;
			glibtop_mem  memory;
			
			glibtop_get_mem (&memory);
			
			used = (unsigned long) (memory.user/(1024*1024));
			current_percent = ((float) used * 1024.0 * 1024.0) / (float) memory.total;
			
			if (sysinfo->device_changed || current_percent != sysinfo->sysinfo_used_percent) {
				char *title;
				
				title = g_strdup_printf (_("RAM: %ld MB of %ld MB used"),
						used, (unsigned long) memory.total/(1024*1024));
				kiba_object_set_title (KIBA_OBJECT(sysinfo->object), title);
				g_free (title);
			}
			break;
		}
		default: {
			int i;
			unsigned int used;
			unsigned long total = 0, idle = 0;
			glibtop_cpu info;
			
			glibtop_get_cpu(&info);
			for (i = 0; i <= glibtop_global_server->ncpu; ++i) {
				total += info.xcpu_total[i] - sysinfo->cpu.last[i].total;
				sysinfo->cpu.last[i].total = info.xcpu_total[i];
				
				idle += info.xcpu_idle[i] - sysinfo->cpu.last[i].idle;
				sysinfo->cpu.last[i].idle = info.xcpu_idle[i];
			}
			used = (unsigned int)(total - idle);
			
			current_percent = (float)used / (float)total;
			
			if (sysinfo->device_changed || current_percent != sysinfo->sysinfo_used_percent) {
				char *title;
				
				title = g_strdup_printf (_("CPU: %i MHz of %i MHz used"),
					(int) (current_percent * sysinfo->cpu.frequency),
					sysinfo->cpu.frequency);
				
				kiba_object_set_title (KIBA_OBJECT(sysinfo->object), title);
				g_free (title);
			}
			break;
		}
	}
	
	if (sysinfo->device_changed)
		sysinfo->device_changed = FALSE;
	
	if (current_percent == sysinfo->sysinfo_used_percent)
		return TRUE;
	
	sysinfo->sysinfo_used_percent = current_percent;
	
	cr = kiba_win_cairo_create (win);
	icon_text = g_strdup_printf ("%.0f%%", sysinfo->sysinfo_used_percent * 100.0);
//TODO font size
	cairo_update_pango_layout (icon_text, "Sans Regular 10",
			cr, &sysinfo->pango_layout);
	cairo_destroy(cr);
	g_free (icon_text);
	
	kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(sysinfo->object));
	
	if (!sysinfo->finger_animate_id)
		sysinfo->finger_animate_id = g_timeout_add (50,
			(GSourceFunc) sysinfo_animate_finger_step, sysinfo);
	
	return TRUE;
}

/**
Rendering
**/

static gint current_level = 0;

static void kiba_sysinfo_draw_surface (cairo_t          *cr,
				       cairo_surface_t  *surface,
				       gint              width,
				       gint              height,
				       void             *data) {
	KibaSysinfo *sysinfo = data;
	char *theme_path;
	gint i;
	
	theme_path = kiba_sysinfo_get_theme_path(sysinfo);
	if (!theme_path) {
		DEBUG_ERROR ("KibaSysinfo: failed to load theme\n");
		return;
	}
	
	for (i = 0; i < G_N_ELEMENTS(sysinfo_theme_data); i++) {
		RsvgHandle *handle = NULL;
		char *icon_file;
		
		if (sysinfo_theme_data[i].level != current_level)
			continue;
		
		icon_file = g_strdup_printf ("%s/%s.svg",
				theme_path,
				sysinfo_theme_data[i].name);
		
		if (!g_file_test(icon_file, G_FILE_TEST_EXISTS)) {
			g_free (icon_file);
			continue;
		}
		
		cairo_save (cr);
		if (sysinfo_theme_data[i].level == KIBA_SYSINFO_HAND_SURFACE) {
			cairo_rotate_centered (cr, 4.0f/3.0f,
				sysinfo->object->allocation.width/2,
				sysinfo->object->allocation.height/2);
			
			cairo_translate (cr,
				sysinfo->object->allocation.width/2,
				sysinfo->object->allocation.height/2);
			cairo_rotate (cr, -M_PI/2.0f);
		}
		
		if (kiba_load_svg (icon_file, &sysinfo->object->allocation.width, &handle)) {
			rsvg_handle_render_cairo (handle, cr);
			rsvg_handle_free (handle);
		}
		cairo_restore (cr);
		
		g_free (icon_file);
	}
	
	g_free (theme_path);
}

static void kiba_sysinfo_update_surface (KibaSysinfo *sysinfo) {
	KibaObject *object = KIBA_OBJECT(sysinfo->object);
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(sysinfo->object);
	KibaWin *win;
	GtkWidget *dock;
	cairo_t *cr;
	gint i;
	
	dock = GTK_WIDGET(object)->parent;
	win = KIBA_WIN(dock->parent);
	
	CairoSurfaceData *surface_data[KIBA_SYSINFO_FOREGROUND+1] = {
		&sysinfo->hand_surface,
		&sysinfo->foreground_surface
	};
	
	cr = kiba_win_cairo_create (win);
	
	cairo_save (cr);
		current_level = KIBA_SYSINFO_BACKGROUND;
		cairo_update_surface (cr, drawable_object->surface_data,
				kiba_sysinfo_draw_surface, sysinfo);
	cairo_restore (cr);
	
	for (i = 0; i < KIBA_SYSINFO_FOREGROUND+1; i++) {
		current_level = i;
		cairo_save (cr);
			cairo_update_surface (cr, surface_data[i],
					kiba_sysinfo_draw_surface, sysinfo);
		cairo_restore (cr);
	}
	
	cairo_destroy (cr);
}

static void kiba_sysinfo_paint (KibaDrawableObject *drawable_object,
			        cairo_t    *cr,
			        gpointer    data) {
	KibaObject *object = KIBA_OBJECT(drawable_object);
	KibaSysinfo *sysinfo = data;
	cairo_surface_t *surface;
	gint unscaled_width;
	gint unscaled_height;
	
	g_return_if_fail (KIBA_IS_OBJECT (object));
	g_return_if_fail (drawable_object->surface_data->surface != NULL);
	
	unscaled_width  = (int) (object->size.width * object->scale_x);
	unscaled_height = (int) (object->size.height * object->scale_y);
	
	if (drawable_object->switch_surface)
		surface = drawable_object->switch_surface;
	else
		surface = drawable_object->surface_data->surface;
	
	cairo_set_source_surface (cr, surface, 0, 0);
	kiba_drawable_object_paint_with_prelight (drawable_object, surface, cr);
	
	if (PANGO_IS_LAYOUT(sysinfo->pango_layout)) {
		PangoRectangle  log;
		float fraction;
		float scale_factor;
		float alpha;
		
		fraction = sysinfo->sysinfo_used_percent;
		alpha = drawable_object->alpha*drawable_object->scroll_alpha;
		
		cairo_save (cr);
		if (sysinfo->flag == KIBA_SYS_NET) {
			if (fraction > 0.9*sysinfo->max_netload)
				cairo_set_source_rgba (cr, drawable_object->scroll_alpha*1.0, 0, 0, 1);
			else
				cairo_set_source_rgba (cr, alpha*sysinfo->text_rgba.r,
						alpha*sysinfo->text_rgba.g,
						alpha*sysinfo->text_rgba.b,
						alpha*sysinfo->text_rgba.a);
		}
		else if (fraction > 0.9)
			cairo_set_source_rgba (cr, drawable_object->scroll_alpha*1.0, 0, 0, 1);
		else
			cairo_set_source_rgba (cr, alpha*sysinfo->text_rgba.r,
						alpha*sysinfo->text_rgba.g,
						alpha*sysinfo->text_rgba.b,
						alpha*sysinfo->text_rgba.a);
		
		scale_factor = unscaled_width/70.0;
		pango_layout_get_pixel_extents (sysinfo->pango_layout, NULL, &log);
		cairo_translate (cr, (unscaled_width - scale_factor*log.width)/2.0,
				      unscaled_height/2 + 1);
		cairo_scale (cr, scale_factor, scale_factor);
		pango_cairo_show_layout (cr, sysinfo->pango_layout);
		cairo_restore (cr);
	}
	
	cairo_save (cr);
		cairo_rotate_centered (cr, sysinfo->sysinfo_last_percent*4.0f/3.0f,
			unscaled_width/2,
			unscaled_height/2);
		cairo_set_source_surface (cr,
			sysinfo->hand_surface.surface, 0, 0);
		cairo_paint_with_alpha (cr, drawable_object->scroll_alpha);
	cairo_restore (cr);
	
	cairo_set_source_surface (cr, sysinfo->foreground_surface.surface, 0, 0);
	cairo_paint_with_alpha (cr, drawable_object->alpha*drawable_object->scroll_alpha);
}

/**
KibaPrefs Callbacks
**/

static void dock_object_size_cb (const char *option_id,
			         gpointer    value,
			         gpointer    data) {
	KibaSysinfo *sysinfo = data;
	
	kiba_sysinfo_set_surface_size (sysinfo);
	kiba_sysinfo_update_surface (sysinfo);
}

static void object_size_factor_cb (const char *option_id,
			           gpointer    value,
			           gpointer    data) {
	KibaSysinfo *sysinfo = data;
	float factor;
	
	factor = g_ascii_strtod((char*) value, NULL);
	
	kiba_object_set_requisition_factor (KIBA_OBJECT(sysinfo->object), factor, factor);
	kiba_sysinfo_set_surface_size (sysinfo);
	kiba_sysinfo_update_surface (sysinfo);
	gtk_widget_queue_resize (sysinfo->object->parent);
}

static void object_alpha_cb (const char *option_id,
			     gpointer    value,
			     gpointer    data) {
	KibaSysinfo *sysinfo = data;
	float alpha;
	
	alpha = g_ascii_strtod((char*) value, NULL);
	
	kiba_drawable_object_set_alpha (KIBA_DRAWABLE_OBJECT(sysinfo->object), alpha);
}

static void device_cb (const char *option_id,
		       gpointer    value,
		       gpointer    data) {
	KibaSysinfo *sysinfo = data;
	float percent;
	
	percent = sysinfo->sysinfo_used_percent;
	
	kiba_sysinfo_set_device (sysinfo, (char*) value);
	
	sysinfo->sysinfo_used_percent = percent;
}

static void max_netload_cb (const char *option_id,
		            gpointer    value,
		            gpointer    data) {
	KibaSysinfo *sysinfo = data;
	
	sysinfo->max_netload = GPOINTER_TO_INT(value);
}

static void text_rgba_cb (const char *option_id,
		          gpointer    value,
		          gpointer    data) {
	KibaSysinfo *sysinfo = data;
	
	sysinfo->text_rgba = kiba_prefs_rgba_from_string((char*) value);
	
	if (sysinfo->object)
		kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(sysinfo->object));
}

static void theme_cb (const char *option_id,
		          gpointer    value,
		          gpointer    data) {
	KibaSysinfo *sysinfo = data;
	
	sysinfo->theme = (char*) value;
	if (sysinfo->object)
		kiba_sysinfo_update_surface (sysinfo);
}


/**
KibaObject Signals
**/

static gboolean kiba_object_button_release_cb (GtkWidget      *widget,
					   GdkEventButton *event,
					   gpointer        data) {
	KibaSysinfo *sysinfo = data;
	
	if (event->button == 3)
		return FALSE;
	
	start_procman (sysinfo);
	
	return FALSE;
}

static void kiba_object_scroll_cb (GtkWidget      *widget,
				   GdkEventScroll *event,
				   gpointer        data) {
	KibaSysinfo *sysinfo = data;
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(sysinfo->object);
	CairoSurfaceData *buf;
	float percent;
	
	percent = sysinfo->sysinfo_used_percent;
	
	if (event->direction == GDK_SCROLL_UP) {
		if (sysinfo->flag != KIBA_SYS_ROOT)
			sysinfo->flag++;
		else
			sysinfo->flag = KIBA_SYS_CPU;
	}
	else {
		if (sysinfo->flag != KIBA_SYS_CPU)
			sysinfo->flag--;
		else
			sysinfo->flag = KIBA_SYS_ROOT;
	}
	
	sysinfo->device_changed = TRUE;
	sysinfo->sysinfo_used_percent = percent;
	
	kiba_drawable_object_switch_surface (drawable_object);
	/* we dont really wanna switch the surface, we just want the animation ;) */
	if (drawable_object->switch_surface)
		drawable_object->surface_data->surface = drawable_object->switch_surface;
}

#endif

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaSysinfo *sysinfo;
	char *device;
	
	sysinfo = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_remove_callback_by_data (file, sysinfo);
	
	/* save device for next start */
	kiba_prefs_write_to_config_file (file, "device", "char",
		sysinfo_device_data[sysinfo->flag]);
	
	kiba_object_destroy (KIBA_OBJECT(sysinfo->object));
	
	g_source_remove (sysinfo->timeout_id);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaSysinfo *sysinfo;
	KibaObject *object;
	gint i;
	
#ifndef USE_SVG
	DEBUG_ERROR ("KibaSysinfo: this Plugin requires librsvg\n");
	return FALSE;
#else
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");	
#endif
	
	glibtop_init();
	
	sysinfo = g_new0 (KibaSysinfo, 1);
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_add_callback (file, "sysinfo_device",
		device_cb, sysinfo, TRUE);
	kiba_prefs_add_callback (file, "sysinfo_text_rgba",
		text_rgba_cb, sysinfo, TRUE);
	kiba_prefs_add_callback (file, "sysinfo_theme",
		theme_cb, sysinfo, TRUE);
	
	sysinfo->pango_layout = NULL;
	
	sysinfo->object = kiba_drawable_object_new (PLUGIN_NAME);
	kiba_add_plugin_object(kiba, plugin, KIBA_OBJECT(sysinfo->object));
	object = KIBA_OBJECT(sysinfo->object);
	kiba_drawable_object_set_paint_function (KIBA_DRAWABLE_OBJECT(object),
		kiba_sysinfo_paint, sysinfo);
	gtk_widget_show (sysinfo->object);
	
	CairoSurfaceData *surface_data[KIBA_SYSINFO_FOREGROUND+1] = {
		&sysinfo->hand_surface,
		&sysinfo->foreground_surface
	};
	
	for (i = 0; i < KIBA_SYSINFO_FOREGROUND+1; i++)
		surface_data[i]->surface = NULL;
	kiba_sysinfo_set_surface_size (sysinfo);
	
	for (i = 0; i <= glibtop_global_server->ncpu; ++i) {
		sysinfo->cpu.frequency += kiba_get_cpu_frequency(i);
	}
	
	sysinfo->sysinfo_used_percent = -1.0;
	sysinfo->mem.last_in = 0;
	sysinfo->device_changed = TRUE;
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) sysinfo);
	
	g_signal_connect (G_OBJECT (sysinfo->object), "button-release-event",
			G_CALLBACK (kiba_object_button_release_cb), sysinfo);
	g_signal_connect (G_OBJECT (sysinfo->object), "scroll-event",
			G_CALLBACK (kiba_object_scroll_cb), sysinfo);
	
	kiba_prefs_add_callback (file, "dock_object_size",
		dock_object_size_cb, sysinfo, FALSE);
	kiba_prefs_add_callback (file, "sysinfo_size_factor",
		object_size_factor_cb, sysinfo, TRUE);
	kiba_prefs_add_callback (file, "sysinfo_alpha",
		object_alpha_cb, sysinfo, TRUE);
	kiba_prefs_add_callback (file, "sysinfo_max_netload",
		max_netload_cb, sysinfo, TRUE);
	
	sysinfo->timeout_id = g_timeout_add (SECOND_INTERVAL,
		(GSourceFunc) kiba_sysinfo_time_handler, (gpointer) sysinfo);
	
	return TRUE;
#endif
}

static PrefMember sysinfo_options[] = {
    { "sysinfo_size_factor", "1.3", KIBA_DOUBLE },
    { "sysinfo_alpha", "0.8", KIBA_DOUBLE },
    { "sysinfo_device", "memory", KIBA_STRING },
    { "sysinfo_text_rgba", "#000000FF", KIBA_STRING },
    { "sysinfo_max_netload", "128", KIBA_INT },
    { "sysinfo_theme", "default", KIBA_STRING }
};

static KibaPluginVtable vtable = {
	"Sysinfo Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
};

KIBA_INIT_PLUGIN (vtable, sysinfo_options, NULL, NULL, NULL);
