/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <libintl.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <X11/Xatom.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-icon.h>
#include <kiba-plugin.h>
#include <kiba-plugin-loader.h>
#include <kiba-prefs.h>
#include <kiba-x11.h>
#include <kiba-screen.h>
#include <kiba-window.h>
#include <kiba-cairo-utils.h>
#include <kiba-macros.h>
#include <kiba-utils.h>
#include <kiba-inline-pixbufs.h>
#include <kiba-menu-items.h>

#define _(String)   dgettext (GETTEXT_PACKAGE, String)
#define PLUGIN_NAME "taskbar"
#define MINIATURE_ICON_SCALE_FACTOR 0.5

typedef struct _KibaTaskbar         KibaTaskbar;
typedef struct _KibaTaskbarIcon     KibaTaskbarIcon;
typedef struct _KibaCustomIcon      KibaCustomIcon;

typedef enum {
	KIBA_TASKBAR_NORMAL,
	KIBA_TASKBAR_DESKTOP,
	KIBA_TASKBAR_SKIPP
} KibaTaskbarWinState;

typedef enum {
	KIBA_TASKBAR_ICON,
	KIBA_TASKBAR_MINIATURE
} KibaTaskbarIconType;

typedef enum {
	KIBA_TASKBAR_SHOW_DESKTOP = 0,
	KIBA_TASKBAR_VIEWPORT_DESKTOP,
	KIBA_TASKBAR_DRAW_MINIATURE,
	KIBA_TASKBAR_GROUPING,
	KIBA_TASKBAR_FLAGS_LAST
}KibaTaskbarFlags;

static const char* kiba_taskbar_flags_data[KIBA_TASKBAR_FLAGS_LAST] = {
	"taskbar_show_desktop",
	"taskbar_viewport_desktop",
	"taskbar_draw_miniature",
	"taskbar_grouping"
};

struct _KibaCustomIcon {
	gchar  *path;
	gchar  *class;
};

struct _KibaTaskbarIcon {
	KibaTaskbar         *taskbar;
	KibaWindow          *window;
	GtkWidget           *object;
	
	GdkPixbuf           *icon_pixbuf;   /* app_window icon */
	GdkPixbuf           *bg_pixbuf;     /* miniature pixbuf */
	PangoLayout         *layout;        /* currently just used for the desktop icon */
	
	KibaTaskbarWinState  state;         /* stores the window state */
	KibaTaskbarIconType  type;
	char                *class;         /* stores the window class */
	char                *name;
	int                  class_member;
	gint                 raise_id;
	
	gboolean             hidden;
};

struct _KibaTaskbar {
	KibaPlugin *plugin;
	
	GList               *taskbar_icons; /* list of all KibaTaskbarIcon's */
	GList               *custom_icons;
	gint                 desktop;
	
	KibaTaskbarFlags     flags;
	KibaRGBA             text_rgba;
	KibaRGBA             border_rgba;
	float               icon_alpha;
	float               border_radius;
	float               border_width;
	GSList              *skip_apps;
	
	gint                 dock_object_size;
	gint                 size;
	gint                 miniature_size;
	float               size_scale;
	float               miniature_size_scale;
};

static void kiba_taskbar_icon_update_surface (KibaTaskbarIcon *icon);

static void kiba_taskbar_icon_new (KibaTaskbar *taskbar,
				   KibaWindow  *window);

static void kiba_taskbar_object_new (KibaPlugin      *plugin,
				     KibaTaskbar     *taskbar,
				     KibaTaskbarIcon *icon);

static void icon_changed_callback (KibaWindow      *window,
				   KibaTaskbarIcon *icon);
static void state_changed_callback (KibaWindow      *window,
				    gint             state,
				    KibaTaskbarIcon *icon);
static void name_changed_callback (KibaWindow      *window,
				   const gchar     *name,
				   KibaTaskbarIcon *icon);

/* Occurs when the KibaObject representing a window in the taskbar changes
   position. */
static void object_position_changed_callback (KibaObject      *object,
			                  GtkAllocation  *allocation,
				              KibaTaskbarIcon *icon);

/**
Helper Functions
**/

static KibaTaskbarIcon* kiba_taskbar_get_icon_for_window (KibaTaskbar *taskbar,
							  void        *window) {
	GList	*l;
		
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		KibaTaskbarIcon *icon = l->data;
		
		if (icon->window == window)
			return icon;
	}
	
	return NULL;
}

static KibaTaskbarIcon* kiba_taskbar_get_active_icon_for_object (KibaTaskbar *taskbar,
								 GtkWidget   *object) {
	GList	*l;
		
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		KibaTaskbarIcon *icon = l->data;
		
		if (icon->class_member == 1 && object == icon->object)
			return icon;
	}
	
	return NULL;
}

static KibaTaskbarIcon* kiba_taskbar_get_desktop (KibaTaskbar *taskbar) {
	GList	*l;
		
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		KibaTaskbarIcon *icon = l->data;
		
		if (icon->state == KIBA_TASKBAR_DESKTOP)
			return icon;
	}
	
	return NULL;
}

static GList* kiba_taskbar_get_sorted_class (KibaTaskbar     *taskbar,
					     KibaTaskbarIcon *class_icon) {
	GList *l, *sorted_list = NULL;
	gboolean member_missing = TRUE;
	
	if (!GET_FLAG(taskbar->flags, KIBA_TASKBAR_GROUPING)) {
		sorted_list = g_list_prepend (sorted_list, class_icon);
		return sorted_list;
	}
	
	while (member_missing) {
		KibaTaskbarIcon *buf_icon = NULL;
		int member_number = 0;
	
		for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
			KibaTaskbarIcon *icon = l->data;
		
			if (icon->class && strcmp (icon->class, class_icon->class) == 0 &&
			    icon->state != KIBA_TASKBAR_DESKTOP &&
			    g_list_find (sorted_list, icon) == NULL &&
			    icon->class_member >= member_number) {
				member_number = icon->class_member;
				buf_icon = icon;
			}
		}
	
		if (buf_icon != NULL)
			sorted_list = g_list_prepend (sorted_list, buf_icon);
		else
			member_missing = FALSE;
	}
	
	return sorted_list;
}

static KibaCustomIcon* kiba_taskbar_use_custom_icon (KibaTaskbar     *taskbar,
						     KibaTaskbarIcon *icon) {
	GList *l;
	
	g_return_val_if_fail (icon->class != NULL, NULL);
	
	for (l = taskbar->custom_icons; l != NULL; l = l->next) {
		KibaCustomIcon *custom_icon = l->data;
		
		if (strcmp (custom_icon->class, icon->class) == 0)
				return custom_icon;
	}
	
	return NULL;
}

static void kiba_taskbar_icon_scroll_icons (KibaTaskbar     *taskbar,
					    gint             direction,
					    KibaTaskbarIcon *old_active,
					    gboolean         focus) {
	KibaTaskbarIcon *new_active = NULL;
	GList *class_list, *l;
	gint i, win_state;
	
	class_list = kiba_taskbar_get_sorted_class (taskbar, old_active);
	
	/* change desktop */
	if (old_active->state == KIBA_TASKBAR_DESKTOP) {
	
		if (!GET_FLAG(taskbar->flags, KIBA_TASKBAR_VIEWPORT_DESKTOP)) {
			int new_workspace;
			int num_workspaces;
			
			num_workspaces = kiba_num_workspaces ();
			
			if (num_workspaces > 1) {
				new_workspace = kiba_get_active_workspace_number (gdk_display) +
							1 + direction;
				
				if (new_workspace > num_workspaces)
					new_workspace = 1;
				else if (new_workspace < 1)
					new_workspace = num_workspaces;
				
				kiba_activate_workspace (new_workspace - 1);
			}
		}
		else {
			int viewport_x, viewport_y, new_viewport_x, num_x, num_y;
			
			kiba_get_num_viewports (&num_x, &num_y);
			viewport_x = num_x - 1;
			viewport_y = num_y - 1;
			
			kiba_get_avtive_viewport (&viewport_x, &viewport_y);
			
			new_viewport_x = viewport_x + direction;
			
			if (new_viewport_x > num_x - 1)
				new_viewport_x = 0;
			else if (new_viewport_x < 0)
				new_viewport_x = num_x - 1;
			
			kiba_change_viewport (new_viewport_x * KIBA_SCREEN_DEFAULT->geometry.width, viewport_y);
		}
		
		return;
	}
	
	if (g_list_length (class_list) <= 1) {
		g_list_free (class_list);
		kiba_focus_window (NULL, old_active->window->gdk_window);
		return;
	}
	
	/* reorder class member */
	for (l = class_list; l != NULL; l = l->next) {
		KibaTaskbarIcon *icon = l->data;
	
		if (direction == -1) {
			if (icon->class_member != 1)
				icon->class_member--;
			else
				icon->class_member = g_list_length (class_list);
		}
		else if (direction == 1) {
			if (icon->class_member != g_list_length (class_list))
				icon->class_member++;
			else
				icon->class_member = 1;
		}
	}
	
	g_list_free (class_list);
	class_list = NULL;
	class_list = kiba_taskbar_get_sorted_class (old_active->taskbar, old_active);
	for (i = 0; i != g_list_length (class_list) && new_active == NULL; i++)
		new_active = g_list_nth_data (class_list, i);
	g_list_free (class_list);
	
	if (new_active == NULL || new_active == old_active || new_active->class_member != 1) {
		kiba_focus_window (NULL, old_active->window->gdk_window);
		return;
	}
	
	new_active->object = old_active->object;
	//TODO: change title when surface is switched
	kiba_object_set_title (KIBA_OBJECT(new_active->object), new_active->name);
	kiba_drawable_object_switch_surface (KIBA_DRAWABLE_OBJECT(new_active->object));
	kiba_taskbar_icon_update_surface (new_active);
	
	gboolean window_on_active_desktop;
	int x,y, wx, wy;
	
	kiba_get_avtive_viewport (&x, &y);
	kiba_get_win_viewport_number (gdk_x11_drawable_get_xid (new_active->window->gdk_window), &wx, &wy);
	
	window_on_active_desktop = x == wx && y == wy &&
		kiba_get_active_workspace_number (gdk_display) ==
		kiba_get_win_workspace_number (gdk_display,
			gdk_x11_drawable_get_xid (new_active->window->gdk_window));
	
	if (focus) {
		win_state = kiba_get_winstate (gdk_display, new_active->window->gdk_window);
	
		if (win_state == IconicState) {
			kiba_deiconify_window (NULL, new_active->window->gdk_window);
			kiba_focus_window (NULL, new_active->window->gdk_window);
		}
		else if (win_state == NormalState)
			kiba_focus_window (NULL, new_active->window->gdk_window);
	}
}

static void kiba_taskbar_get_window_preview (KibaTaskbar     *taskbar,
					     KibaTaskbarIcon *icon) {
	GdkColormap  *colormap;
	gint width, height;
	
	if (icon->bg_pixbuf) {
		g_object_unref (icon->bg_pixbuf);
		icon->bg_pixbuf = NULL;
	}
	
	gdk_window_get_geometry (icon->window->gdk_window, NULL, NULL, &width, &height, NULL);
	
	colormap = gdk_drawable_get_colormap (icon->window->gdk_window);
	icon->bg_pixbuf = gdk_pixbuf_get_from_drawable (NULL,
		icon->window->gdk_window, colormap, 0, 0, 0, 0, width, height);
}

static void kiba_taskbar_get_window_icon (KibaTaskbar     *taskbar,
					  KibaTaskbarIcon *icon) {
	KibaCustomIcon *custom;
	GdkPixbuf *pixbuf = NULL;
	gint size;
	
	/* the desktop may has no icon */
	if (icon->state == KIBA_TASKBAR_DESKTOP)
		return;
	
	size = icon->type == KIBA_TASKBAR_MINIATURE ? taskbar->miniature_size : taskbar->size;
	
	if (custom = kiba_taskbar_use_custom_icon (taskbar, icon)) {
		kiba_load_icon (custom->path, size, &pixbuf);
	}
	if (!pixbuf) {
		icon->icon_pixbuf = kiba_get_icon (
					gdk_x11_drawable_get_xid (icon->window->gdk_window),
					size);
		if (!pixbuf && !icon->icon_pixbuf) {
			icon->icon_pixbuf = gdk_pixbuf_new_from_inline (-1, app_fallback_icon, FALSE, NULL);
			icon->icon_pixbuf = gdk_pixbuf_scale_simple (icon->icon_pixbuf,
						size, size, GDK_INTERP_BILINEAR);
		}
	}
	
	if (!pixbuf)
		return;
	
	if (icon->icon_pixbuf) {
		g_object_unref (icon->icon_pixbuf);
	}
	
	icon->icon_pixbuf = pixbuf;
}

static void kiba_taskbar_icon_get_icon (KibaTaskbar     *taskbar,
					KibaTaskbarIcon *icon) {
	if (!GDK_IS_DRAWABLE (icon->window->gdk_window))
		return;
	
	kiba_taskbar_get_window_icon (taskbar, icon);
	if (icon->type == KIBA_TASKBAR_MINIATURE)
		kiba_taskbar_get_window_preview (taskbar, icon);
}

static void kiba_taskbar_desktop_update_layout (KibaTaskbarIcon *icon) {
	KibaWin *default_win;
	KibaObject *object;
	cairo_t *cr;
	gchar *text, *description;
	
	if (!icon || !GTK_IS_WIDGET(icon->object))
		return;
	
	object = KIBA_OBJECT(icon->object);
	
	default_win = kiba_get_win(kiba_get());
	cr = kiba_win_cairo_create (default_win);
	
	text = g_strdup_printf ("%i", icon->taskbar->desktop);
	description = g_strdup_printf ("Arial Bold %i",
			5*(object->size.width)/12);
	cairo_update_pango_layout (text, description,
			cr, &icon->layout);
	g_free (description);
	g_free (text);
	
	cairo_destroy (cr);
	
	kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(icon->object));
}

static void kiba_taskbar_reload_mindow (GtkWidget *widget,
					gpointer   data) {
	KibaTaskbarIcon *icon = data;
	
	kiba_taskbar_get_window_preview (icon->taskbar, icon);
	kiba_taskbar_icon_update_surface (icon);
}

static void kiba_taskbar_icon_set_object_req_factor (KibaTaskbar     *taskbar,
						     KibaTaskbarIcon *icon) {
	gint width, height;
	
	if (icon->type != KIBA_TASKBAR_MINIATURE || !icon->bg_pixbuf) {
		kiba_object_set_requisition_factor (KIBA_OBJECT(icon->object),
			taskbar->size_scale, taskbar->size_scale);
		return;
	}
	
	width = gdk_pixbuf_get_width (icon->bg_pixbuf);
	height = gdk_pixbuf_get_height (icon->bg_pixbuf);
	
	kiba_object_set_requisition_factor (KIBA_OBJECT(icon->object),
		taskbar->miniature_size_scale*width/height,
		taskbar->miniature_size_scale);
}

static gboolean kiba_taskbar_skip_app (KibaTaskbar *taskbar,
				       GdkWindow   *window) {
	GSList *l;
	Window xwindow;
	XTextProperty text_property;
	Atom cmd_atom;
	char *state;
	
	xwindow = gdk_x11_drawable_get_xid (window);
	
	/* dont skip the desktop */
	state = (char*) kiba_get_wintype (gdk_display, window);
	if (state != NULL && strcmp (state, "_NET_WM_WINDOW_TYPE_DESKTOP") == 0) {
		g_free (state);
		return FALSE;
	}
	g_free (state);
	
	/* skip apps in exclude list */
	cmd_atom = XInternAtom (gdk_display, "WM_CLASS", TRUE);
	if (XGetTextProperty (gdk_display, xwindow, &text_property, cmd_atom)) {
		for (l = taskbar->skip_apps; l != NULL; l = l->next) {
			char *skip_app = l->data;
			
			if (g_strrstr (text_property.value, skip_app)) {
				XFree (text_property.value);
				return TRUE;
			}
		}
	}
	
	if (text_property.value)
		XFree (text_property.value);
	
	return FALSE;
}

static gboolean kiba_taskbar_skip_hints (KibaTaskbar *taskbar,
					 GdkWindow   *window) {
	Atom type;
	gint format;
	gulong nitems;
	gulong bytes_after;
	GdkAtom *atoms = NULL;
	gulong i;
	
	if (GET_FLAG(taskbar->flags, KIBA_TASKBAR_SHOW_DESKTOP)) {
		char *state;
		
		state = (char*) kiba_get_wintype (gdk_display, window);
	
		if (state != NULL && strcmp (state, "_NET_WM_WINDOW_TYPE_DESKTOP") == 0) {
			g_free (state);
			return FALSE;
		}
		g_free (state);
	}
	
	return kiba_window_state_hint (window, "_NET_WM_STATE_SKIP_TASKBAR");
}

static void kiba_taskbar_remove_icon (KibaTaskbar     *taskbar,
				      KibaTaskbarIcon *icon) {
	GList *class_list;
	
	class_list = kiba_taskbar_get_sorted_class (taskbar, icon);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(icon->window),
		G_CALLBACK(icon_changed_callback), icon);
	g_signal_handlers_disconnect_by_func (G_OBJECT(icon->window),
		G_CALLBACK(state_changed_callback), icon);
	g_signal_handlers_disconnect_by_func (G_OBJECT(icon->window),
		G_CALLBACK(name_changed_callback), icon);
	g_signal_handlers_disconnect_by_func (G_OBJECT(icon->object),
		G_CALLBACK(object_position_changed_callback), icon);
	
	/* remove object when there is just one object in class */
	if (icon->object != NULL &&
	   (g_list_length(class_list) == 1 || icon->state == KIBA_TASKBAR_DESKTOP))
		kiba_object_destroy (KIBA_OBJECT(icon->object));
	
	if (GET_FLAG(taskbar->flags, KIBA_TASKBAR_GROUPING) &&
	    icon->class_member == 1 && g_list_length (class_list) > 1)
		kiba_taskbar_icon_scroll_icons (taskbar, -1, icon, FALSE);
	
	taskbar->taskbar_icons = g_list_remove (taskbar->taskbar_icons, icon);
	if (g_list_length(taskbar->taskbar_icons) == 0) {
		g_list_free (taskbar->taskbar_icons);
		taskbar->taskbar_icons = NULL;
	}
	
	if (icon->icon_pixbuf != NULL) {
		g_object_unref (icon->icon_pixbuf);
		icon->icon_pixbuf = NULL;
	}
	
	g_free (icon->class);
	icon->class = NULL;
	
	g_free (icon);
	icon = NULL;
}

static void update_preview_cb (GtkFileChooser *file_chooser,
			       gpointer        data) {
	GtkWidget *preview;
	GdkPixbuf *pixbuf;
	
	preview = GTK_WIDGET(data);
	const char* filename = gtk_file_chooser_get_preview_filename(file_chooser);
	
	if ( filename != NULL ) {
		pixbuf = gdk_pixbuf_new_from_file_at_size(filename, 128, 128, NULL);
	
		gtk_image_set_from_pixbuf(GTK_IMAGE(preview), pixbuf);
		if (pixbuf != NULL)
			gdk_pixbuf_unref(pixbuf);
	
		gtk_file_chooser_set_preview_widget_active(file_chooser, pixbuf != NULL);
	}
}

static void set_custom_icon (GtkWidget *button,
			     gpointer   data) {
	KibaTaskbarIcon *icon = data;
	GtkWidget *dialog, *preview;
	GtkFileFilter *filter;
	GList *l;
	
	KibaPrefsFile *configFile = kiba_get_plugin_config_file(kiba_get(), icon->taskbar->plugin);
	
	filter = gtk_file_filter_new ();
	gtk_file_filter_add_pattern (filter, "*.png");
	gtk_file_filter_add_pattern (filter, "*.xpm");
	gtk_file_filter_add_pattern (filter, "*.svg");
	gtk_file_filter_set_name (filter, "Images");
	dialog = gtk_file_chooser_dialog_new (_("Set Custom Icon"), NULL,
		GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
		GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
	gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), "/usr/share/pixmaps");
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (dialog), filter);
	gtk_file_chooser_set_select_multiple (GTK_FILE_CHOOSER (dialog), FALSE);
	
	preview = gtk_image_new ();
	gtk_file_chooser_set_preview_widget(GTK_FILE_CHOOSER(dialog), preview);
	g_signal_connect(GTK_FILE_CHOOSER (dialog), "update-preview",
			G_CALLBACK(update_preview_cb), preview);
	
	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
		KibaCustomIcon *custom;
		GSList *filenames;
		char *filename;
		char *list_entry;
		
		filenames = gtk_file_chooser_get_filenames (GTK_FILE_CHOOSER (dialog));
		filename = filenames->data;
		
		if (!(custom = kiba_taskbar_use_custom_icon (icon->taskbar, icon))) {
			custom = g_new0 (KibaCustomIcon, 1);
			custom->class = g_strdup (icon->class);
			icon->taskbar->custom_icons =
				g_list_append (icon->taskbar->custom_icons, custom);
		}
		else {
			list_entry = g_strdup_printf ("%s;%s", custom->class, custom->path);
			kiba_prefs_remove_from_list (configFile, "taskbar_custom_icon", list_entry);
			g_free (list_entry);
		}
		custom->path = filename;
		
		list_entry = g_strdup_printf ("%s;%s", custom->class, custom->path);
		kiba_prefs_append_to_list (configFile, "taskbar_custom_icon", list_entry);
		g_free (list_entry);
		g_slist_free (filenames);
	}
	gtk_widget_destroy (dialog);

	for (l = icon->taskbar->taskbar_icons; l != NULL; l = l->next) {
		KibaTaskbarIcon *class_icon = l->data;
		
		if (strcmp (class_icon->class, icon->class) == 0) {
			if (class_icon->icon_pixbuf) {
				g_object_unref (class_icon->icon_pixbuf);
				class_icon->icon_pixbuf = NULL;
			}
			
			
			kiba_taskbar_get_window_icon (class_icon->taskbar, class_icon);
			if (class_icon->object)
				kiba_taskbar_icon_update_surface (class_icon);
		}
	}
}

static void unset_custom_icon (GtkWidget *button,
			       gpointer   data) {
	KibaTaskbarIcon *icon = data;
	KibaCustomIcon  *custom;
	GList *l;
	char *list_entry;
	
	custom = kiba_taskbar_use_custom_icon (icon->taskbar, icon);
	if (!custom)
		return;
	
	KibaPrefsFile *configFile = kiba_get_plugin_config_file(kiba_get(), icon->taskbar->plugin);
	
	list_entry = g_strdup_printf ("%s;%s", custom->class, custom->path);
	kiba_prefs_remove_from_list (configFile, "taskbar_custom_icon", list_entry);
	g_free (list_entry);
	
	icon->taskbar->custom_icons = g_list_remove (icon->taskbar->custom_icons, custom);
	g_free (custom->class);
	g_free (custom);
	
	for (l = icon->taskbar->taskbar_icons; l != NULL; l = l->next) {
		KibaTaskbarIcon *class_icon = l->data;
		
		if (strcmp (class_icon->class, icon->class) == 0) {
			if (class_icon->icon_pixbuf) {
				g_object_unref (class_icon->icon_pixbuf);
				class_icon->icon_pixbuf = NULL;
			}
			
			kiba_taskbar_get_window_icon (class_icon->taskbar, class_icon);
			if (class_icon->object)
				kiba_taskbar_icon_update_surface (class_icon);
		}
	}
}

static void keep_taskbar_icon_in_dock (GtkWidget *button,
				       gpointer   data) {
// TODO
// 	KibaTaskbarIcon *icon = data;
// 	KibaLauncher *launcher = kiba_launcher_get();
// 	KibaObject *object = KIBA_OBJECT(icon->object);
// 	KibaDesktopIcon *dsktop_icon;
// 	FILE *desktop_file;
// 	char *desktop_path;
// 	const char *homedir;
// 	char *icon_path;
// 	char *command;
// 	
// 	if (!launcher)
// 		return;
// 	
// 	homedir = g_get_home_dir ();
// 	
// 	desktop_path = g_strdup_printf ("%s" LAUNCHER_DIR "/%s.desktop", homedir, icon->class);
// 	
// 	icon_path = g_strdup_printf ("%s" LAUNCHER_DIR "/%s.png", homedir, icon->class);
// 	
// 	if (g_file_test (desktop_path, G_FILE_TEST_EXISTS)) {
// 		g_free (desktop_path);
// 		return;
// 	}
// 	
// 	/* writing the desktop file */
// 	/* TODO better name */
// 	desktop_file = fopen (desktop_path, "w");
// 		if (desktop_file == NULL) {
// 			g_free (desktop_path);
// 			return;
// 		}
// 		
// 		command = kiba_get_command_from_pid (kiba_get_win_pid (gdk_display, icon->window->xwindow));
// 		
// 		cairo_surface_write_to_png (object->surface_data->surface, icon_path);
// 		
// 		fprintf (desktop_file,
// 			"[Desktop Entry]\n"
// 			"Encoding=UTF-8\n"
// 			"Version=0.9.2\n"
// 			"Type=Application\n"
// 			"Name=%s\n"
// 			"Comment=None\n"
// 			"Exec=%s\n"
// 			"Icon=%s\n"
// 			"Categories=%s;\n"
// 			"LauncherGroup=%s\n"
// 			"Class=%s\n"
// 			"Description=None\n",
// 			icon->name, command, icon_path,
// 			icon->class, icon->class, icon->class);
// 		
// 		g_free (icon_path);
// 		g_free (command);
// 	fclose (desktop_file);
// 	
// 	/* creating launcher icon for the task */
// 	dsktop_icon = kiba_desktop_icon_new (G_TYPE_MODULE(launcher->plugin));
// 	if (!kiba_desktop_icon_set_file (dsktop_icon, desktop_path)) {
// 		kiba_desktop_icon_destroy (dsktop_icon);
// 		return;
// 	}
// 	kiba_launcher_add (launcher, dsktop_icon);
}

/**
Callbacks
**/

static gint   kiba_taskbar_icon_focus_cb (gpointer data) {
	KibaTaskbarIcon *icon = data;
	
	kiba_focus_window (NULL, icon->window->gdk_window);
	icon->raise_id = 0;
	
	return FALSE;
}

static gint kiba_taskbar_raise_and_focus_window_cb (gpointer data) {
	GdkWindow *window = GDK_WINDOW(data);

	// Move the window to the top
	kiba_raise_window(window);

	// Focus on the window
	kiba_focus_window(NULL, window);

	// Return FALSE to indicate that this function should not be called
	// again
	return FALSE;
}

static gboolean kiba_taskbar_drag_motion_cb (GtkWidget      *widget,
				             GdkDragContext *drag_context,
				             gint            x,
				             gint            y,
				             guint           time,
				             gpointer        data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon  *icon;
	
	icon = kiba_taskbar_get_active_icon_for_object (taskbar, widget);
	if (icon == NULL) return TRUE;
	
	if (icon->raise_id)
		return TRUE;
	
	icon->raise_id = g_timeout_add (1000,   kiba_taskbar_icon_focus_cb, icon);
}

static void kiba_taskbar_drag_leave_cb (GtkWidget      *widget,
				        GdkDragContext *drag_context,
				        guint           time,
				        gpointer        data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon  *icon;
	
	icon = kiba_taskbar_get_active_icon_for_object (taskbar, widget);
	if (icon == NULL) return;
	
	if (!icon->raise_id)
		return;
	
	g_source_remove (icon->raise_id);
	icon->raise_id = 0;
}

static void window_opened_callback (KibaScreen  *screen,
				    KibaWindow  *window,
				    KibaTaskbar *taskbar) {
	kiba_taskbar_icon_new (taskbar, window);
}

static void withdrawn_toggled_callback (KibaScreen  *screen,
				        KibaWindow  *window,
				        KibaTaskbar *taskbar) {
	KibaTaskbarIcon  *icon;
	
	if (GET_FLAG(window->state, KIBA_WINDOW_WITHDRAWN)) {
		icon = kiba_taskbar_get_icon_for_window (taskbar, window);
		if (icon)
			kiba_taskbar_remove_icon (taskbar, icon);
	} else
		kiba_taskbar_icon_new (taskbar, window);

}
static void window_closed_callback (KibaScreen  *screen,
				    KibaWindow  *window,
				    KibaTaskbar *taskbar) {
	KibaTaskbarIcon  *icon;
	
	icon = kiba_taskbar_get_icon_for_window (taskbar, window);
	if (icon == NULL) return;
	
	kiba_taskbar_remove_icon (taskbar, icon);
}

static void active_window_changed_callback (KibaScreen  *screen,
					    KibaWindow  *active_window,
					    KibaTaskbar *taskbar) {
	GList *l;
	KibaTaskbarIcon *win_icon;
	
	win_icon = kiba_taskbar_get_icon_for_window (taskbar, active_window);
	if (win_icon == NULL)
		return;
	
	while (win_icon->class_member > 1) {
		KibaTaskbarIcon *class_leader;
	
		for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
			KibaTaskbarIcon *icon = l->data;
		
			if (icon->window->gdk_window && icon->state != KIBA_TASKBAR_DESKTOP &&
			strcmp (win_icon->class, icon->class) == 0 && icon->class_member == 1) {
				class_leader = icon;
				break;
			}
		}
		if (!class_leader || class_leader == win_icon)
			break;
		
		kiba_taskbar_icon_scroll_icons (taskbar, 1, class_leader, FALSE);
	}
}

static void workspace_changed_callback (KibaScreen  *screen,
					gint         new_active,
					KibaTaskbar *taskbar) {
	
	taskbar->desktop = new_active;
	
	kiba_taskbar_desktop_update_layout(kiba_taskbar_get_desktop(taskbar));
}

static void viewport_changed_callback (KibaScreen   *screen,
				       GdkPoint     *new_active,
				       KibaTaskbar  *taskbar) {
	if (!new_active || taskbar->desktop == new_active->x)
		return;
	
	taskbar->desktop = new_active->x;
	
	kiba_taskbar_desktop_update_layout(kiba_taskbar_get_desktop(taskbar));
}

static void icon_changed_callback (KibaWindow      *window,
				   KibaTaskbarIcon *icon) {
	kiba_taskbar_get_window_icon (icon->taskbar, icon);
	if (icon->class_member == 1)
		kiba_taskbar_icon_update_surface (icon);
}

static void state_changed_callback (KibaWindow      *window,
				    gint             state,
				    KibaTaskbarIcon *icon) {
	gboolean urgent;
	
	if (icon->object) {
		urgent = kiba_object_get_state(KIBA_OBJECT(icon->object), KIBA_OBJECT_URGENT);
		
		if (GET_FLAG(window->state, KIBA_WINDOW_URGENT) && !urgent)
			kiba_object_set_state(KIBA_OBJECT(icon->object), KIBA_OBJECT_URGENT);
		else if (!GET_FLAG(window->state, KIBA_WINDOW_URGENT) && urgent)
			kiba_object_unset_state(KIBA_OBJECT(icon->object), KIBA_OBJECT_URGENT);
		
		if (GET_FLAG(state, KIBA_WINDOW_ICONIFIED) && !icon->hidden) {
			icon->hidden = TRUE;
		}
		else if (!GET_FLAG(state, KIBA_WINDOW_ICONIFIED) && icon->hidden) {
			icon->hidden = FALSE;
		}
	}
}

static void name_changed_callback (KibaWindow      *window,
				   const gchar     *name,
				   KibaTaskbarIcon *icon) {
	g_return_if_fail (name != NULL);
	char  *object_name = g_strdup (name);
	
	while (TRUE) {
		GList *l;
		int    i = 1;
		
		for (l = icon->taskbar->taskbar_icons; l != NULL; l = l->next) {
			KibaTaskbarIcon *check_icon = l->data;
		
			if (strcmp (check_icon->name, object_name) == 0 && check_icon != icon)
				i++;
		}
		
		if (i > 1) {
			int   old_value = 0;
		
			if (g_str_has_suffix (object_name, ")") &&
			    g_strrstr (object_name, " (") != NULL) {
				char *buf, *value;
				
				buf   = g_strrstr (object_name, "(") + 1;
				value = g_strndup (buf, strlen(buf) - 1);
				old_value = atoi (value) - 1;
				g_free (value);
			}
			
			g_free (object_name);
			object_name = NULL;
			object_name = g_strdup_printf ("%s (%i)", name, i + old_value);
		}
		else
			break;
	}
	
	if (icon->name) {
		g_free (icon->name);
		icon->name = NULL;
	}
	icon->name = g_strdup (object_name);
	
	if (icon->object)
		kiba_object_set_title (KIBA_OBJECT(icon->object), object_name);
	
	g_free (object_name);
}

static void object_position_changed_callback (KibaObject      *object,
				       GtkAllocation  *allocation,
				       KibaTaskbarIcon *icon) {
	GList *class_list, *l;
	
	class_list = kiba_taskbar_get_sorted_class (icon->taskbar, icon);
	for (l = class_list; l != NULL; l = l->next) {
		KibaTaskbarIcon *classIcon = l->data;
		// Change the minimize/restore point to be where the kiba-object is.
		kiba_object_set_icon_geometry(object, classIcon->window);
	}
}

static GList* sort_taskbar_icons (KibaDock     *dock,
				  const char   *group,
				  const GList  *icons,
				  gpointer      data) {
	KibaTaskbar *taskbar = data;
	GList *ret = NULL;
	const GList *l;
	
/*TODO:
	* strdown the titles
	* ordering is wrong on startup
	* other sort methods:
		* object size
		* viewport
		* stack order
*/	
	/* sort icons by class */
	for (l = icons; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		KibaTaskbarIcon *icon;
		GList *li, *needle = NULL;
		
		icon = kiba_taskbar_get_active_icon_for_object (taskbar, child->widget);
		if (!icon)
			continue;
		
		for (li = ret; li != NULL; li = li->next) {
			GtkBoxChild *buf_child = li->data;
			KibaTaskbarIcon *buf_icon;
			
			buf_icon = kiba_taskbar_get_active_icon_for_object (taskbar, buf_child->widget);
			if (!buf_icon)
				continue;
			
			if (buf_icon->class[0] > icon->class[0]) {
				needle = li;
				break;
			}
			else if (buf_icon->class[0] == icon->class[0]) {
				gint i;
				gint max;
				
				max = MAX(strlen(icon->class)+1, strlen(buf_icon->class)+1);
				for (i = 1; i < max; i++) {
					if (buf_icon->class[i] > icon->class[i]) {
						needle = li;
						break;
					}
					if (buf_icon->class[i] != icon->class[i])
						break;
				}
			}
		}
		
		ret = g_list_insert_before (ret, needle, l->data);
	}
	
	return ret;
}

/**
KibaPrefs Callbacks
**/

static void dock_object_size_cb (GtkWidget *widget,
			         gint       new_size,
			         gpointer   data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	GList *l;
	
	taskbar->dock_object_size = new_size;
	taskbar->size = taskbar->dock_object_size * taskbar->size_scale;
	taskbar->miniature_size = taskbar->dock_object_size * taskbar->miniature_size_scale;
	
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		icon = l->data;
		
		kiba_taskbar_icon_get_icon (taskbar, icon);
		if (icon->object)
			kiba_taskbar_icon_update_surface (icon);
	}
}

static void object_size_factor_cb (const char *option_id,
			           gpointer    value,
			           gpointer    data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	KibaWin *win;
	KibaDock *dock;
	GList *l;
	
	win = kiba_get_plugin_win(kiba_get(), taskbar->plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	taskbar->size_scale = g_ascii_strtod((char*) value, NULL);
	taskbar->size = taskbar->dock_object_size * taskbar->size_scale;
	
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		icon = l->data;
		
		if (icon->object && icon->type != KIBA_TASKBAR_MINIATURE) {
			kiba_object_set_requisition_factor (KIBA_OBJECT(icon->object),
				taskbar->size_scale, taskbar->size_scale);
			KIBA_DRAWABLE_OBJECT(icon->object)->surface_data->width  = icon->object->allocation.width;
			KIBA_DRAWABLE_OBJECT(icon->object)->surface_data->height = icon->object->allocation.height;
		}
	}
	
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		icon = l->data;
	
		kiba_taskbar_get_window_icon (taskbar, icon);
		if (icon->object && icon->type != KIBA_TASKBAR_MINIATURE)
			kiba_taskbar_icon_update_surface (icon);
	}
	gtk_widget_queue_resize (GTK_WIDGET(dock));
}

static void miniature_size_factor_cb (const char *option_id,
			              gpointer    value,
			              gpointer    data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	KibaWin *win;
	KibaDock *dock;
	GList *l;
	
	win = kiba_get_plugin_win(kiba_get(), taskbar->plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	taskbar->miniature_size_scale = g_ascii_strtod((char*) value, NULL);
	taskbar->miniature_size = taskbar->dock_object_size * taskbar->miniature_size_scale;
	
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		icon = l->data;
		
		if (icon->object && icon->type == KIBA_TASKBAR_MINIATURE) {
			kiba_taskbar_icon_set_object_req_factor (taskbar, icon);
			KIBA_DRAWABLE_OBJECT(icon->object)->surface_data->width  = icon->object->allocation.width;
			KIBA_DRAWABLE_OBJECT(icon->object)->surface_data->height = icon->object->allocation.height;
		}
	}
	
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		icon = l->data;
		
		if (icon->type != KIBA_TASKBAR_MINIATURE)
			continue;
		
		kiba_taskbar_get_window_preview (taskbar, icon);
		if (icon->object)
			kiba_taskbar_icon_update_surface (icon);
	}
	gtk_widget_queue_resize (GTK_WIDGET(dock));
}

static KibaWindow* kiba_screen_get_desktop (KibaScreen *screen) {
	GList *l;
	
	for (l = screen->windows; l != NULL; l = l->next) {
		KibaWindow *win = l->data;
		char *type;
		
		type = kiba_get_wintype (gdk_display, win->gdk_window);	
		if (strcmp (type, "_NET_WM_WINDOW_TYPE_DESKTOP") == 0) {
			g_free (type);
			return win;
		}
		g_free (type);
	}
	
	return NULL;
}

static void taskbar_flags_cb (const char *option_id,
			      gpointer    value,
			      gpointer    data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarFlags modified_flag;
	gint i;
	gboolean enable;
	
	modified_flag = -1;
	for (i = 0; i < G_N_ELEMENTS(kiba_taskbar_flags_data); i++) {
		if (strcmp (option_id, kiba_taskbar_flags_data[i]) == 0)
			modified_flag = i;
	}
	if (modified_flag == -1)
		return;
	
	enable = (gboolean) GPOINTER_TO_INT(value);
	if (enable) {
		if (!GET_FLAG (taskbar->flags, modified_flag))
			SET_FLAG  (taskbar->flags, modified_flag);
	}
	else if (GET_FLAG (taskbar->flags, modified_flag))
		DEL_FLAG  (taskbar->flags, modified_flag);
	
	switch (modified_flag) {
		case KIBA_TASKBAR_SHOW_DESKTOP: {
			KibaTaskbarIcon *desktop;
			
			if (!enable) {
				desktop = kiba_taskbar_get_desktop (taskbar);
				if (desktop) {
					kiba_object_destroy (KIBA_OBJECT(desktop->object));
					taskbar->taskbar_icons =
						g_list_remove (taskbar->taskbar_icons, desktop);
					g_free (desktop);
				}
			}
			else {
				kiba_taskbar_icon_new (taskbar,
					kiba_screen_get_desktop (kiba_screen_get()));
			}
			
			break;
		}
		case KIBA_TASKBAR_DRAW_MINIATURE: {
			GList *l;
			KibaTaskbarIcon *icon;
			
			for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
				icon = l->data;
				
				if (!enable) {
					if (icon->type != KIBA_TASKBAR_MINIATURE ||
					icon->state == KIBA_TASKBAR_DESKTOP)
						continue;
					
					icon->type = KIBA_TASKBAR_ICON;
					if (icon->bg_pixbuf) {
						g_object_unref (icon->bg_pixbuf);
						icon->bg_pixbuf = NULL;
					}
				}
				else {
					if (icon->type == KIBA_TASKBAR_MINIATURE)
						continue;
					
					icon->type = KIBA_TASKBAR_MINIATURE;
				}
				kiba_taskbar_icon_get_icon (taskbar, icon);
				if (icon->object) {
					kiba_taskbar_icon_set_object_req_factor (taskbar, icon);
					KIBA_DRAWABLE_OBJECT(icon->object)->surface_data->width  =
						icon->object->allocation.width;
					KIBA_DRAWABLE_OBJECT(icon->object)->surface_data->height =
						icon->object->allocation.height;
				
					kiba_taskbar_icon_update_surface (icon);
				}
			}
			break;
		}
		case KIBA_TASKBAR_GROUPING: {
			GList *l;
			KibaTaskbarIcon *icon;
			
			if (!enable) {
				for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
					icon = l->data;
					
					if (icon->class_member == 1)
						continue;
					
					icon->class_member = 1;
					kiba_taskbar_object_new (taskbar->plugin, taskbar, icon);
				}
			}
			else {
				GList *classes = NULL;
				
				for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
					GList *li;
					gboolean object_added;
					icon = l->data;
					gint class_member = 1;
					
					if (!icon->object)
						continue;
					
					object_added = FALSE;
					for (li = classes; li != NULL; li = li->next) {
						char *class = li->data;
						
						if (strcmp (class, icon->class) == 0)
							class_member++;
					}
					classes = g_list_append (classes, icon->class);
					icon->class_member = class_member;
					
					if (class_member == 1)
						continue;
					
					kiba_object_destroy (KIBA_OBJECT(icon->object));
					icon->object = NULL;
				}
				g_list_free (classes);
			}
			break;
		}
		case KIBA_TASKBAR_VIEWPORT_DESKTOP: {
			break;
		}
	}
}

static void object_alpha_cb (const char *option_id,
			     gpointer    value,
			     gpointer    data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	GList *l;
	
	taskbar->icon_alpha = g_ascii_strtod((char*) value, NULL);
	
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		icon = l->data;
		
		if (icon->object)
			kiba_drawable_object_set_alpha (KIBA_DRAWABLE_OBJECT(icon->object), taskbar->icon_alpha);
	}
}

static void text_rgba_cb (const char *option_id,
			  gpointer    value,
			  gpointer    data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	GList *l;
	
	taskbar->text_rgba = kiba_prefs_rgba_from_string((char*) value);
	
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		icon = l->data;
		
		if (icon->object)
			kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(icon->object));
	}
}

static void border_rgba_cb (const char *option_id,
			    gpointer    value,
			    gpointer    data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	GList *l;
	
	taskbar->border_rgba = kiba_prefs_rgba_from_string((char*) value);
	
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		icon = l->data;
		
		if (icon->object)
			kiba_taskbar_icon_update_surface (icon);
	}
}

static void border_radius_cb (const char *option_id,
			    gpointer    value,
			    gpointer    data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	GList *l;
	
	taskbar->border_radius = g_ascii_strtod((char*) value, NULL);
	
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		icon = l->data;
		
		if (icon->object)
			kiba_taskbar_icon_update_surface (icon);
	}
}

static void border_width_cb (const char *option_id,
			    gpointer    value,
			    gpointer    data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	GList *l;
	
	taskbar->border_width = g_ascii_strtod((char*) value, NULL);
	
	for (l = taskbar->taskbar_icons; l != NULL; l = l->next) {
		icon = l->data;
		
		if (icon->object)
			kiba_taskbar_icon_update_surface (icon);
	}
}

static void skip_app_cb (const char *option_id,
			 gpointer    value,
			 gpointer    data) {
	KibaTaskbar *taskbar = data;
	
	taskbar->skip_apps = (GSList*) value;
}

static void custom_icon_cb (const char *option_id,
			    gpointer    value,
			    gpointer    data) {
	KibaTaskbar *taskbar = data;
	KibaCustomIcon *icon;
	GSList *sl;
	GList *l;
	
	for (l = taskbar->custom_icons; l != NULL; l = taskbar->custom_icons) {
		icon = l->data;
		
		g_free (icon->class);
		g_free (icon);
		taskbar->custom_icons = g_list_remove (taskbar->custom_icons, l->data);
	}
	g_list_free (taskbar->custom_icons);
	taskbar->custom_icons = NULL;
	
	for (sl = (GSList*) value; sl != NULL; sl = sl->next) {
		gchar *item = sl->data;
		char *path;
		char *wmclass;
		gboolean skip = FALSE;
		
		path = g_strrstr (item, ";");
		if (!path) continue;
		
		wmclass = g_strndup (item, strlen (item) - strlen (path+1) - 1);
		
		/* making sure that the wmclass is associated only once */
		for (l = taskbar->custom_icons; l != NULL; l = l->next) {
			icon = l->data;
			
			if (strcmp (icon->class, wmclass) == 0) {
				skip = TRUE;
			}
		}
		
		if (skip) {
			g_free (wmclass);
			continue;
		}
		
		icon = g_new0 (KibaCustomIcon, 1);
		icon->path = g_strdup(path+1);
		icon->class = wmclass;
		
		taskbar->custom_icons = g_list_append (taskbar->custom_icons, icon);
	}
}


/**
Rendering
**/

static gint current_level = 0;

static void kiba_taskbar_icon_draw_surface (cairo_t          *cr,
					    cairo_surface_t  *surface,
					    gint              width,
					    gint              height,
					    void             *data) {
	KibaTaskbarIcon *icon = data;
	KibaDock *dock;
	const KibaLayoutData *d;
	gint pwidth = 0, pheight = 0;
	
	dock = KIBA_DOCK(icon->object->parent);
	d = &kiba_layout_data[dock->d.position];
	
	if (icon->bg_pixbuf) {
		pwidth = gdk_pixbuf_get_width (icon->bg_pixbuf);
		pheight = gdk_pixbuf_get_height (icon->bg_pixbuf);
		
		cairo_save (cr);
			cairo_paint_shape (cr,
				KIBA_ROUNDED, KIBA_ROUNDED,
				KIBA_ROUNDED, KIBA_ROUNDED,
				icon->taskbar->border_width/2,
				icon->taskbar->border_width/2,
				width-2 - icon->taskbar->border_width/2,
				height-2 - icon->taskbar->border_width/2,
				icon->taskbar->border_radius);
			cairo_set_source_rgba (cr, 1, 1, 1, 1);
			cairo_fill_preserve (cr);
			
			cairo_save (cr);
				cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, 0.0);
				cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);
				cairo_scale (cr, (float) width/pwidth,
						 (float) height/pheight);
				gdk_cairo_set_source_pixbuf (cr, icon->bg_pixbuf, 0, 0);
				cairo_paint (cr);
			cairo_restore (cr);
			
			cairo_set_source_rgba (cr,
				icon->taskbar->border_rgba.r,
				icon->taskbar->border_rgba.g,
				icon->taskbar->border_rgba.b,
				icon->taskbar->border_rgba.a);
			cairo_set_line_width (cr, icon->taskbar->border_width);
			cairo_stroke (cr);
		cairo_restore (cr);
	}
	
	if (icon->icon_pixbuf) {
		if (icon->bg_pixbuf) {
			cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
			if (d->orientation == GTK_ORIENTATION_VERTICAL)
				cairo_translate (cr,
					(d->halignment/2)*(width - height*MINIATURE_ICON_SCALE_FACTOR),
					(d->halignment/2)*(height - height*MINIATURE_ICON_SCALE_FACTOR));
			else
				cairo_translate (cr,
					(d->valignment/2)*(width - height*MINIATURE_ICON_SCALE_FACTOR),
					(d->valignment/2)*height*MINIATURE_ICON_SCALE_FACTOR);
			cairo_scale (cr, MINIATURE_ICON_SCALE_FACTOR, MINIATURE_ICON_SCALE_FACTOR);
		}
		gdk_cairo_set_source_pixbuf (cr, icon->icon_pixbuf, 0, 0);
		cairo_paint (cr);
	}
}

static void kiba_taskbar_icon_update_surface (KibaTaskbarIcon *icon) {
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(icon->object);
	KibaWin *default_win;
	cairo_t *cr;
	
	default_win = kiba_get_win(kiba_get());
	cr = kiba_win_cairo_create (default_win);
	cairo_update_surface (cr, drawable_object->surface_data,
			kiba_taskbar_icon_draw_surface, icon);
	
	cairo_destroy (cr);
	
	kiba_drawable_object_queue_draw (drawable_object);
}

static void kiba_taskbar_icon_paint (KibaDrawableObject *drawable_object,
				     cairo_t    *cr,
				     gpointer    data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	cairo_surface_t *surface;
	
	g_return_if_fail (KIBA_IS_OBJECT (drawable_object));
	g_return_if_fail (drawable_object->surface_data->surface != NULL);
	
	icon = kiba_taskbar_get_active_icon_for_object(taskbar, GTK_WIDGET(drawable_object));
	g_return_if_fail (icon != NULL);
	
	if (drawable_object->switch_surface)
		surface = drawable_object->switch_surface;
	else
		surface = drawable_object->surface_data->surface;
	
	cairo_set_source_surface (cr, surface, 0, 0);
	kiba_drawable_object_paint_with_prelight (drawable_object, surface, cr);
	
//TODO
// 	if (PANGO_IS_LAYOUT(icon->layout)) {
// 		PangoRectangle  log;
// 		gint i;
// 		float alpha;
// 		
// 		pango_layout_get_pixel_extents (icon->layout, NULL, &log);
// 		
// 		cairo_translate (cr,
// 			((*object->size * object->scale_x) - log.width)/2,
// 			((*object->size * object->scale_y) - log.height)/2);
// 		
// 		alpha = object->scroll_alpha*object->alpha;
// 		cairo_set_source_rgba (cr,
// 			alpha*(1.0-taskbar->text_rgba.r),
// 			alpha*(1.0-taskbar->text_rgba.g),
// 			alpha*(1.0-taskbar->text_rgba.b),
// 			alpha*taskbar->text_rgba.a);
// 		for (i = 0; i < 4; i++) {
// 			cairo_save (cr);
// 				cairo_translate (cr, (1 - (i&2)), (1 - 2*(i&1)));
// 				pango_cairo_show_layout (cr, icon->layout);
// 			cairo_restore (cr);
// 		}
// 		
// 		cairo_set_source_rgba (cr,
// 			alpha*taskbar->text_rgba.r,
// 			alpha*taskbar->text_rgba.g,
// 			alpha*taskbar->text_rgba.b,
// 			alpha*taskbar->text_rgba.a);
// 		pango_cairo_show_layout (cr, icon->layout);
// 	}
}

/**
KibaObject Callbacks
**/

static void kiba_taskbar_activate_workspace (GtkWidget *widget,
					     gpointer   data) {
	kiba_activate_workspace (GPOINTER_TO_INT(data) - 1);
}

static void kiba_taskbar_activate_viewport (GtkWidget *widget,
					    gpointer   data) {
	int viewport_x, viewport_y, new_viewport_x, num_x, num_y;
	
	kiba_get_num_viewports (&num_x, &num_y);
	viewport_x = 0;
	viewport_y = num_y - 1;
	
	kiba_get_avtive_viewport (&viewport_x, &viewport_y);
	
	kiba_change_viewport ((GPOINTER_TO_INT(data)-1) * KIBA_SCREEN_DEFAULT->geometry.width,
		viewport_y);
}

static gboolean kiba_object_button_press_cb (GtkWidget      *widget,
					     GdkEventButton *event,
					     gpointer        data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	GtkWidget *menu;
	GtkWidget *item, *image;
	
	if (event->button != 3)
		return TRUE;
	
	icon = kiba_taskbar_get_active_icon_for_object(taskbar, widget);
	
	menu = gtk_menu_new ();
	
	if (icon->bg_pixbuf) {
		item = gtk_image_menu_item_new_with_label (_("Reload Preview"));
		image = gtk_image_new_from_icon_name (GTK_STOCK_REFRESH, GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		g_signal_connect (G_OBJECT (item),
			"activate", G_CALLBACK (kiba_taskbar_reload_mindow), icon);
		gtk_widget_show (item);
	}
	
	if (icon->state != KIBA_TASKBAR_DESKTOP) {
		if (kiba_taskbar_use_custom_icon (icon->taskbar, icon)) {
			item = gtk_menu_item_new_with_label (_("Unset Custom Icon"));
			g_signal_connect (G_OBJECT (item),
				"activate", G_CALLBACK (unset_custom_icon), icon);
		}
		else {
			item = gtk_menu_item_new_with_label (_("Set Custom Icon"));
			g_signal_connect (G_OBJECT (item),
				"activate", G_CALLBACK (set_custom_icon), icon);
		}
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		gtk_widget_show (item);
		
// 		if (kiba_launcher_get()) {
// 			item = gtk_menu_item_new_with_label (_("Keep in Dock"));
// 			g_signal_connect (G_OBJECT (item),
// 				"activate", G_CALLBACK (keep_taskbar_icon_in_dock), icon);
// 			gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
// 			gtk_widget_show (item);
// 		}
	}
	
	kiba_menu_add_separator (GTK_WIDGET (menu));
	
	if (icon->state != KIBA_TASKBAR_DESKTOP) {
		GList *class_list, *l;
		
		class_list = kiba_taskbar_get_sorted_class (icon->taskbar, icon);
		
		for (l = class_list; l != NULL; l = l->next) {
			KibaTaskbarIcon *buf_icon = l->data;
			
			if (strcmp (icon->class, buf_icon->class) == 0 &&
			    buf_icon->window->gdk_window != NULL) {
				GtkWidget *window_entry, *image;
				GdkPixbuf *pixbuf;
				GtkMenu   *winmenu;
				char      *text;
	
				text    = kiba_win_name (gdk_display,
					gdk_x11_drawable_get_xid (buf_icon->window->gdk_window));
				winmenu = (GtkMenu*) gtk_menu_new();
				
				window_entry = gtk_image_menu_item_new_with_label (text);
				g_free (text);
				pixbuf = gdk_pixbuf_scale_simple (buf_icon->icon_pixbuf,
					24, 24, GDK_INTERP_NEAREST);
				image = gtk_image_new_from_pixbuf (pixbuf);
				g_object_unref (pixbuf);
				gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (window_entry), image);
				gtk_menu_shell_append (GTK_MENU_SHELL (menu), window_entry);
				gtk_menu_item_set_submenu (GTK_MENU_ITEM (window_entry),
					GTK_WIDGET(winmenu));
				gtk_widget_show (window_entry);
				
				kiba_load_app_menu (buf_icon->window->gdk_window, winmenu);
			}
		}
		g_list_free (class_list);
	}
	else {
		GtkMenu  *submenu;
		int num_workspaces, active_workspace;
		int num_x = 0, num_y = 0;
		int active_viewport_x = 0, active_viewport_y = 0;
		int i;
		
		num_workspaces = kiba_num_workspaces ();
		active_workspace = kiba_get_active_workspace_number (gdk_display);
		kiba_get_num_viewports (&num_x, &num_y);
		kiba_get_avtive_viewport (&active_viewport_x, &active_viewport_y);
		
		if (num_workspaces > 1) {
			char *workspace;
			
			submenu = (GtkMenu*) gtk_menu_new();
			
			workspace = g_strdup_printf("%s", _("Workspace"));
			item = gtk_image_menu_item_new_with_label (workspace);
			g_free (workspace);
			image = gtk_image_new_from_icon_name ("gnome-fs-desktop", GTK_ICON_SIZE_MENU);
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
			gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
			gtk_menu_item_set_submenu (GTK_MENU_ITEM (item),
				GTK_WIDGET(submenu));
			gtk_widget_show (item);
			
			for (i = 0; i < num_workspaces; i++) {
				
				if (i == active_workspace)
					continue;
				
				workspace = g_strdup_printf("%s %i", _("Workspace"), i+1);
				item = gtk_image_menu_item_new_with_label (workspace);
				g_free (workspace);
				image = gtk_image_new_from_icon_name ("gnome-fs-desktop", GTK_ICON_SIZE_MENU);
				gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
				gtk_menu_shell_append (GTK_MENU_SHELL (submenu), item);
				g_signal_connect (G_OBJECT (item), "activate",
					G_CALLBACK (kiba_taskbar_activate_workspace), GINT_TO_POINTER(i+1));
				gtk_widget_show (item);
			}
		}
		
		if (num_x > 1) {
			char *viewport;
			
			submenu = (GtkMenu*) gtk_menu_new();
			
			viewport = g_strdup_printf("%s", _("Viewport"));
			item = gtk_image_menu_item_new_with_label (viewport);
			g_free (viewport);
			image = gtk_image_new_from_icon_name ("gnome-fs-desktop", GTK_ICON_SIZE_MENU);
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
			gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
			gtk_menu_item_set_submenu (GTK_MENU_ITEM (item),
				GTK_WIDGET(submenu));
			gtk_widget_show (item);
			
			for (i = 0; i < num_x; i++) {
				
				if (i == active_viewport_x)
					continue;
				
				viewport = g_strdup_printf("%s %i", _("Viewport"), i+1);
				item = gtk_image_menu_item_new_with_label (viewport);
				g_free (viewport);
				image = gtk_image_new_from_icon_name ("gnome-fs-desktop", GTK_ICON_SIZE_MENU);
				gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
				gtk_menu_shell_append (GTK_MENU_SHELL (submenu), item);
				g_signal_connect (G_OBJECT (item), "activate",
					G_CALLBACK (kiba_taskbar_activate_viewport), GINT_TO_POINTER(i+1));
				gtk_widget_show (item);
			}
			kiba_menu_add_separator (GTK_WIDGET (menu));
		}
		else if (num_workspaces > 1)
			kiba_menu_add_separator (GTK_WIDGET (menu));
		
		item = gtk_image_menu_item_new_with_label (_("Toggle Show Desktop"));
		image = gtk_image_new_from_icon_name ("gnome-fs-desktop", GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		if (kiba_get_show_desktop ())
			g_signal_connect (G_OBJECT (item), "activate",
				G_CALLBACK (kiba_toggle_showing_desktop), GINT_TO_POINTER(0));
		else
			g_signal_connect (G_OBJECT (item), "activate",
				G_CALLBACK (kiba_toggle_showing_desktop), GINT_TO_POINTER(1));
		gtk_widget_show (item);
	}
	
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
		event->button, event->time);
	
	return TRUE;
}

static gboolean kiba_object_button_release_cb (GtkWidget      *widget,
					       GdkEventButton *event,
					       gpointer        data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	int win_state;
	Window focus_window;
	GList *class_list;
	
	if (event->button != 1)
		return FALSE;
	
	icon = kiba_taskbar_get_active_icon_for_object(taskbar, widget);
	if (!icon) return FALSE;
	
	if (icon->state == KIBA_TASKBAR_DESKTOP) {
		kiba_toggle_showing_desktop (NULL, GINT_TO_POINTER(kiba_get_show_desktop () == FALSE));
		return TRUE;
	}
	
	win_state = kiba_get_winstate (gdk_display, icon->window->gdk_window);
	focus_window = kiba_get_active_window (gdk_display);
	
	class_list = kiba_taskbar_get_sorted_class (icon->taskbar, icon);
	
	if (win_state == IconicState) {
		GList *l;
		
		for (l = class_list; l != NULL; l = l->next) {
			KibaTaskbarIcon *buf = l->data;
			
			if (event->button == 2 ||
			    buf->window->gdk_window == icon->window->gdk_window) {
				kiba_object_set_icon_geometry(KIBA_OBJECT(widget), icon->window);
				
				kiba_deiconify_window (NULL,
					buf->window->gdk_window);

				/* Wait a cycle for a refresh before raising
				 * the window.
				 *
				 * TODO: Find a more elegant way to do this.
				 * In theory, if the user is running too many
				 * window or a slow computer, 20 milliseconds
				 * won't be long enough.
				 */
				g_timeout_add(20,
				              kiba_taskbar_raise_and_focus_window_cb,
				              buf->window->gdk_window);
			}
		}
	}
	else if (win_state == NormalState && gdk_x11_drawable_get_xid (icon->window->gdk_window) == focus_window) {
		GList *l;
		
		kiba_object_set_icon_geometry(KIBA_OBJECT(widget), icon->window);
		
		for (l = class_list; l != NULL; l = l->next) {
			KibaTaskbarIcon *buf = l->data;
			
			if (event->button == 2 || buf->window->gdk_window == icon->window->gdk_window)
				kiba_iconify_window (NULL, buf->window->gdk_window);
		}
	}
	else if (win_state == NormalState)
		kiba_focus_window (NULL, icon->window->gdk_window);
	
	g_list_free (class_list);
	
	return TRUE;
}

static void kiba_object_scroll_cb (GtkWidget      *widget,
				   GdkEventScroll *event,
				   gpointer        data) {
	KibaTaskbar *taskbar = data;
	KibaTaskbarIcon *icon;
	gint direction;
	
	direction = (event->direction == GDK_SCROLL_DOWN ? -1 : 1);
	icon = kiba_taskbar_get_active_icon_for_object(taskbar, widget);
	
	kiba_taskbar_icon_scroll_icons (taskbar, direction, icon, TRUE);
}

/**
Taskbar init Functions
**/

static void kiba_taskbar_object_new (KibaPlugin      *plugin,
				     KibaTaskbar     *taskbar,
				     KibaTaskbarIcon *icon) {
	KibaDock *dock;
	KibaObject *object;
	KibaDrawableObject *drawable_object;
	
	icon->object = kiba_drawable_object_new (PLUGIN_NAME);
	object = KIBA_OBJECT(icon->object);
	drawable_object = KIBA_DRAWABLE_OBJECT(icon->object);
	kiba_drawable_object_set_paint_function (drawable_object,
		kiba_taskbar_icon_paint, taskbar);
	kiba_add_plugin_object(kiba_get(), plugin, object);
	gtk_widget_show (icon->object);
	kiba_object_set_title (KIBA_OBJECT(icon->object), icon->name);
	
	kiba_taskbar_icon_set_object_req_factor (taskbar, icon);
	kiba_object_set_state (object, KIBA_OBJECT_ACTIVE);
	kiba_drawable_object_set_alpha (drawable_object, taskbar->icon_alpha);
	drawable_object->surface_data->width = taskbar->dock_object_size*object->scale_x;
	drawable_object->surface_data->height = taskbar->dock_object_size*object->scale_y;
	kiba_taskbar_icon_update_surface (icon);
	
	/* add object signal handlers */
	g_signal_connect (G_OBJECT (icon->object), "button-press-event",
			G_CALLBACK (kiba_object_button_press_cb), taskbar);
	g_signal_connect (G_OBJECT (icon->object), "button-release-event",
			G_CALLBACK (kiba_object_button_release_cb), taskbar);
	g_signal_connect (G_OBJECT (icon->object), "scroll-event",
			G_CALLBACK (kiba_object_scroll_cb), taskbar);
	g_signal_connect (G_OBJECT (icon->object), "drag-motion",
			G_CALLBACK (kiba_taskbar_drag_motion_cb), taskbar);
	g_signal_connect (G_OBJECT (icon->object), "drag-leave",
			G_CALLBACK (kiba_taskbar_drag_leave_cb), taskbar);
}

static void kiba_taskbar_icon_new (KibaTaskbar *taskbar,
				   KibaWindow  *window) {
	KibaTaskbarIcon *icon;
	gchar *win_state, *name, *type, *class;
	gint window_workspace_number,
             active_workspace_number,
             state;
	
	if (!KIBA_IS_WINDOW(window))
		return;
	
	if (kiba_taskbar_skip_app (taskbar, window->gdk_window) ||
	    kiba_taskbar_skip_hints (taskbar, window->gdk_window))
		return;
	
	type = kiba_get_wintype (gdk_display, window->gdk_window);	
	if (strcmp (type, "_NET_WM_WINDOW_TYPE_NORMAL") == 0)
		state = KIBA_TASKBAR_NORMAL;
	else if (strcmp (type, "_NET_WM_WINDOW_TYPE_DESKTOP") == 0)
		state = KIBA_TASKBAR_DESKTOP;
	else
		state = KIBA_TASKBAR_SKIPP;
	g_free(type);
	
	if (!GET_FLAG(taskbar->flags, KIBA_TASKBAR_SHOW_DESKTOP) && state == KIBA_TASKBAR_DESKTOP)
		return;
	
	if (!window->name)
		return;
	
	class = kiba_get_wmclass (gdk_display, window->xwindow);
	if (!class || strcmp(class, "Kiba-dock") == 0) /* without _NET_WM_CLASS no taskbar icon */
		return;
	
	icon = g_new0 (KibaTaskbarIcon, 1);
	icon->state = state;
	icon->taskbar = taskbar;
	icon->window = window;
	icon->class = class;
	
	if (state == KIBA_TASKBAR_DESKTOP || GET_FLAG(taskbar->flags, KIBA_TASKBAR_DRAW_MINIATURE))
		icon->type = KIBA_TASKBAR_MINIATURE;
	else
		icon->type = KIBA_TASKBAR_ICON;
	
	if (GET_FLAG(taskbar->flags, KIBA_TASKBAR_GROUPING)) {
		GList *class_list;
		
		class_list = kiba_taskbar_get_sorted_class (taskbar, icon);
		icon->class_member = g_list_length (class_list) + 1;
		g_list_free (class_list);
	}
	else
		icon->class_member = 1;
	
	/* intiall icon titel */
	name_changed_callback (icon->window, icon->window->name, icon);
	
	/* check if window is on active workspace */
	window_workspace_number =
		kiba_get_win_workspace_number (gdk_display, icon->window->xwindow);
		active_workspace_number = kiba_get_active_workspace_number (gdk_display);
	
	/* load pixbufs */
	kiba_taskbar_icon_get_icon (taskbar, icon);
	
	taskbar->taskbar_icons = g_list_append (taskbar->taskbar_icons, icon);

	if (icon->class_member == 1) {
		kiba_taskbar_object_new (taskbar->plugin, taskbar, icon);
		kiba_object_set_icon_geometry (KIBA_OBJECT(icon->object),
			window);
		/* add KibaObject signal handler */
		g_signal_connect (G_OBJECT(icon->object), "size-allocate",
		  G_CALLBACK(object_position_changed_callback), icon);
	}
	
	if (icon->state == KIBA_TASKBAR_DESKTOP)
		kiba_taskbar_desktop_update_layout(icon);
	
	/* add window signal handlers */
	g_signal_connect (G_OBJECT(icon->window), "icon-changed",
		  G_CALLBACK(icon_changed_callback), icon);
	g_signal_connect (G_OBJECT(icon->window), "state-changed",
		  G_CALLBACK(state_changed_callback), icon);
	g_signal_connect (G_OBJECT(icon->window), "name-changed",
		  G_CALLBACK(name_changed_callback), icon);
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaTaskbar *taskbar;
	KibaWin *win;
	KibaDock *dock;
	GList *l;
	
	taskbar = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	KibaPrefsFile *configFile = kiba_get_plugin_config_file(kiba, plugin);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_SCREEN_DEFAULT),
		  G_CALLBACK(window_opened_callback), taskbar);
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_SCREEN_DEFAULT),
		  G_CALLBACK(window_closed_callback), taskbar);
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_SCREEN_DEFAULT),
		  G_CALLBACK(withdrawn_toggled_callback), taskbar);
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_SCREEN_DEFAULT),
		  G_CALLBACK(active_window_changed_callback), taskbar);
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_SCREEN_DEFAULT),
		  G_CALLBACK(workspace_changed_callback), taskbar);
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_SCREEN_DEFAULT),
		  G_CALLBACK(viewport_changed_callback), taskbar);
	kiba_prefs_remove_callback_by_data (configFile, taskbar);
	kiba_dock_remove_group_sort_func (dock, PLUGIN_NAME);
	
	for (l = taskbar->taskbar_icons; l != NULL; l = taskbar->taskbar_icons) {
		KibaTaskbarIcon *icon = l->data;
	
		if (KIBA_IS_WINDOW (icon->window)) {
			g_signal_handlers_disconnect_by_func (G_OBJECT(icon->window),
				G_CALLBACK(icon_changed_callback), icon);
			g_signal_handlers_disconnect_by_func (G_OBJECT(icon->window),
				G_CALLBACK(state_changed_callback), icon);
			g_signal_handlers_disconnect_by_func (G_OBJECT(icon->window),
				G_CALLBACK(name_changed_callback), icon);
		}
		if (GDK_IS_PIXBUF (icon->icon_pixbuf)) {
			gdk_pixbuf_unref (icon->icon_pixbuf);
			icon->icon_pixbuf = NULL;
		}

		if (icon->class) {
			g_free (icon->class);
			icon->class = NULL;
		}
		if (icon->name) {
			g_free (icon->name);
			icon->name = NULL;
		}
		
		if (icon->object) {
			g_signal_handlers_disconnect_by_func (G_OBJECT(icon->object),
				G_CALLBACK(object_position_changed_callback), icon);
			kiba_object_destroy (KIBA_OBJECT(icon->object));
			icon->object = NULL;
		}
		
		taskbar->taskbar_icons = g_list_remove (taskbar->taskbar_icons, icon);
		g_free (icon);
		icon = NULL;
	}
	g_list_free (taskbar->taskbar_icons);
	
	for (l = taskbar->custom_icons; l != NULL; l = taskbar->taskbar_icons) {
		KibaCustomIcon *icon = l->data;
		g_free (icon->class);
		g_free (icon);
	}
	g_list_free (taskbar->custom_icons);
	
	g_free (taskbar);
	taskbar = NULL;
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaWin *win;
	KibaDock *dock;
	KibaTaskbar *taskbar;
	KibaPluginLoader *loader;
	GSList *l;
	gint i;
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");	
#endif
	
	taskbar = g_new0 (KibaTaskbar, 1);
	taskbar->custom_icons = NULL;
	taskbar->plugin = plugin;
	
	loader = kiba_plugin_loader_get();
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	g_signal_connect (G_OBJECT (dock), "object-size-changed",
			G_CALLBACK (dock_object_size_cb), taskbar);
	dock_object_size_cb (GTK_WIDGET(dock), dock->d.object_size, taskbar);
	
	KibaPrefsFile *configFile = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_add_callback (configFile, "taskbar_size_factor",
		object_size_factor_cb, taskbar, TRUE);
	kiba_prefs_add_callback (configFile, "taskbar_miniature_size_factor",
		miniature_size_factor_cb, taskbar, TRUE);
	kiba_prefs_add_callback (configFile, "taskbar_alpha",
		object_alpha_cb, taskbar, TRUE);
	kiba_prefs_add_callback (configFile, "taskbar_border_width",
		border_width_cb, taskbar, TRUE);
	kiba_prefs_add_callback (configFile, "taskbar_border_radius",
		border_radius_cb, taskbar, TRUE);
	kiba_prefs_add_callback (configFile, "taskbar_border_rgba",
		border_rgba_cb, taskbar, TRUE);
	kiba_prefs_add_callback (configFile, "taskbar_desktop_text_rgba",
		text_rgba_cb, taskbar, TRUE);
	kiba_prefs_add_callback (configFile, "taskbar_skip_app",
		skip_app_cb, taskbar, TRUE);
	kiba_prefs_add_callback (configFile, "taskbar_custom_icon",
		custom_icon_cb, taskbar, TRUE);
	for (i = 0; i < G_N_ELEMENTS(kiba_taskbar_flags_data); i++) {
		kiba_prefs_add_callback (configFile, kiba_taskbar_flags_data[i],
			taskbar_flags_cb, taskbar, TRUE);
	}
	
	kiba_dock_set_group_sort_func (dock, PLUGIN_NAME,
		(KibaGroupSortFunc) sort_taskbar_icons, (gpointer) taskbar);
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) taskbar);
	
	kiba_screen_for_each_window (KIBA_SCREEN_DEFAULT,
		(GFunc) kiba_taskbar_icon_new, (gpointer) taskbar);
	
	/* add screen signal handlers */
	g_signal_connect (G_OBJECT(KIBA_SCREEN_DEFAULT), "window_opened",
		  G_CALLBACK(window_opened_callback), taskbar);
	g_signal_connect (G_OBJECT(KIBA_SCREEN_DEFAULT), "window_closed",
		  G_CALLBACK(window_closed_callback), taskbar);
	g_signal_connect (G_OBJECT(KIBA_SCREEN_DEFAULT), "withdrawn_toggled",
		  G_CALLBACK(withdrawn_toggled_callback), taskbar);
	g_signal_connect (G_OBJECT(KIBA_SCREEN_DEFAULT), "active_window_changed",
		  G_CALLBACK(active_window_changed_callback), taskbar);
	g_signal_connect (G_OBJECT(KIBA_SCREEN_DEFAULT), "workspace_changed",
		  G_CALLBACK(workspace_changed_callback), taskbar);
	g_signal_connect (G_OBJECT(KIBA_SCREEN_DEFAULT), "viewport_changed",
		  G_CALLBACK(viewport_changed_callback), taskbar);
	
	return TRUE;
}

static PrefMember taskbar_options[] = {
	{ "taskbar_show_desktop", "TRUE", KIBA_BOOLEAN },
	{ "taskbar_viewport_desktop", "TRUE", KIBA_BOOLEAN },
	{ "taskbar_draw_miniature", "FALSE", KIBA_BOOLEAN },
	{ "taskbar_grouping", "TRUE", KIBA_BOOLEAN },
	{ "taskbar_alpha", "0.9", KIBA_DOUBLE },
	{ "taskbar_size_factor", "1.0", KIBA_DOUBLE },
	{ "taskbar_miniature_size_factor", "1.3", KIBA_DOUBLE },
	{ "taskbar_border_width", "2.0", KIBA_DOUBLE },
	{ "taskbar_border_radius", "4.5", KIBA_DOUBLE },
	{ "taskbar_desktop_text_rgba", "#FFFFFFFF", KIBA_STRING },
	{ "taskbar_border_rgba", "#FFFFFFFF", KIBA_STRING },
	{ "taskbar_skip_app", "[cairo-clock]", KIBA_LIST },
	{ "taskbar_custom_icon", "[]", KIBA_LIST }
};

static KibaPluginVtable vtable = {
	"Taskbar Plugin for Kiba-Dock", /* summary */
	NULL, /* description */
 
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load */
	plugin_unload, /* unload */
};

KIBA_INIT_PLUGIN (vtable, taskbar_options, NULL, NULL, NULL);
