/*
 *  Copyright (C) 2008 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <string.h>
#include <pthread.h>
#include <glib.h>

#include <libgnomeui/libgnomeui.h>
#include <libgnomevfs/gnome-vfs-file-info.h>
#include <libgnomevfs/gnome-vfs-utils.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-plugin.h>
#include <kiba-plugin-loader.h>
#include <kiba-utils.h>
#include <kiba-inline-pixbufs.h>

/* Cool-off period between last file modification time and thumbnail creation */
#define THUMBNAIL_CREATION_DELAY_SECS 3

/* Should never be a reasonable actual mtime */
#define INVALID_MTIME 0

typedef struct {
	char *uri;
	GList *callbacks;
	time_t original_file_mtime;
}KibaThumbnailInfo;

typedef void (*KibaThumbnailFactoryCallback) (const char*, gpointer);

typedef struct {
	KibaThumbnailFactoryCallback func;
	gpointer userdata;
}KibaThumbnailFactoryCallbackData;

static GnomeThumbnailFactory* thumb_factory_get (void);

/* Our mutex used when accessing data shared between the main thread and the
   thumbnail thread, i.e. the thumbnail_thread_is_running flag and the
   thumbnails_to_make list. */
static pthread_mutex_t thumbnails_mutex = PTHREAD_MUTEX_INITIALIZER;
/* The currently thumbnailed icon. it also exists in the thumbnails_to_make list
 * to avoid adding it again. Lock thumbnails_mutex when accessing this. */
static KibaThumbnailInfo *currently_thumbnailing = NULL;
/* The list of KibaThumbnailInfo structs containing information about the
   thumbnails we are making. Lock thumbnails_mutex when accessing this. */
static volatile GList *thumbnails_to_make = NULL;
/* List of thumbnails that need to invoke the callbacks.
   Lock thumbnails_mutex when accessing this. */
static volatile GList *thumbnails_to_load = NULL;
/* A flag to indicate whether a thumbnail thread is running, so we don't
   start more than one. Lock thumbnails_mutex when accessing this. */
static volatile gboolean thumbnail_thread_is_running = FALSE;

static gint thumbnail_load_timeout_id = 0;

/*** Private methods ***/

static gboolean __get_file_mtime (const char *file_uri, time_t* mtime) {
	GnomeVFSFileInfo *file_info;
	
	/* gather the info and then compare modification times */
	file_info = gnome_vfs_file_info_new ();
	gnome_vfs_get_file_info (file_uri, file_info, GNOME_VFS_FILE_INFO_FOLLOW_LINKS);

	if (file_info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_MTIME)
		*mtime = file_info->mtime;
	else
		*mtime = INVALID_MTIME;
	
	gnome_vfs_file_info_unref (file_info);
	
	return TRUE;
}

static void __free_thumbnail_info (KibaThumbnailInfo *info) {
	g_list_free(info->callbacks);
	g_free(info->uri);
	g_free(info);
}

/* walk through thumbnails_to_load list and invoke callbacks, this is done in the main thread */
static gint __load_thumbnails (gpointer data) {
	GnomeThumbnailFactory *thumb_factory = thumb_factory_get();
	GList *l;
	
	for (l = (GList*) thumbnails_to_load; l != NULL; l = (GList*) thumbnails_to_load) {
		KibaThumbnailInfo *info = l->data;
		GList *li;
		char *thumbPath;
		
		pthread_mutex_lock (&thumbnails_mutex);
		/*********************************
		 * MUTEX LOCKED
		 *********************************/
		thumbnails_to_load = g_list_remove ((GList*) thumbnails_to_load, info);
		/*********************************
		 * MUTEX UNLOCKED
		 *********************************/
		pthread_mutex_unlock (&thumbnails_mutex);
		
		thumbPath = gnome_thumbnail_factory_lookup (thumb_factory, info->uri, info->original_file_mtime);
		if (!thumbPath) continue;
		for (li = info->callbacks; li != NULL; li = info->callbacks) {
			KibaThumbnailFactoryCallbackData *data = li->data;
			data->func(thumbPath, data->userdata);
			info->callbacks = g_list_remove(info->callbacks, li->data);
			g_free(data);
		}
		__free_thumbnail_info(info);
		g_free(thumbPath);
	}
	
	if (thumbnails_to_make != NULL)
		return TRUE;
	
	thumbnail_load_timeout_id = 0;
	return FALSE;
}

/* thumbnail_thread is invoked as a separate thread to make thumbnails. */
static gpointer __thread_start (gpointer data) {
	GnomeThumbnailFactory *thumb_factory = thumb_factory_get();
	KibaThumbnailInfo *info = NULL;
	KibaWin *win = kiba_get_win(kiba_get());
	GdkPixbuf *pixbuf;
	time_t current_orig_mtime, current_time;
	char *mime_type;
	GList *l;
	
	/* We loop until there are no more thumbails to make, at which point
	   we exit the thread. */
	for (;;) {
		pthread_mutex_lock (&thumbnails_mutex);
		/*********************************
		 * MUTEX LOCKED
		 *********************************/
		
		/* Pop the last thumbnail we just made off the head of the
		   list and free it. I did this here so we only have to lock
		   the mutex once per thumbnail, rather than once before
		   creating it and once after.
		   Don't pop the thumbnail off the queue if the original file
		   mtime of the request changed. Then we need to redo the thumbnail.
		*/
		if (currently_thumbnailing) {
			g_assert (info == currently_thumbnailing);
			thumbnails_to_make = g_list_remove ((GList*) thumbnails_to_make, currently_thumbnailing);
		}
		currently_thumbnailing = NULL;
		
		/* If there are no more thumbnails to make, reset the
		   thumbnail_thread_is_running flag, unlock the mutex, and
		   exit the thread. */
		if (thumbnails_to_make == NULL) {
			thumbnail_thread_is_running = FALSE;
			pthread_mutex_unlock (&thumbnails_mutex);
			pthread_exit (NULL);
		}
		/*********************************
		 * MUTEX UNLOCKED
		 *********************************/
		pthread_mutex_unlock (&thumbnails_mutex);
		
		/* Get the next one to make. We leave it on the list until it
		   is created so the main thread doesn't add it again while we
		   are creating it. */
		info = thumbnails_to_make->data;
		currently_thumbnailing = info;
		current_orig_mtime = info->original_file_mtime;
		
		time (&current_time);
		/* Don't try to create a thumbnail if the file was modified recently.
		   This prevents constant re-thumbnailing of changing files. */ 
		if (current_time < current_orig_mtime + THUMBNAIL_CREATION_DELAY_SECS &&
		    current_time >= current_orig_mtime) {
 			continue;
		}
		
		/* Create the thumbnail. */
		mime_type =  gnome_vfs_get_mime_type(info->uri);
		pixbuf = gnome_thumbnail_factory_generate_thumbnail (thumb_factory,
								     info->uri,
								     mime_type);
		g_free (mime_type);
		if (pixbuf) {
			gnome_thumbnail_factory_save_thumbnail (thumb_factory,
								pixbuf,
								info->uri,
								current_orig_mtime);
			g_object_unref (pixbuf);
		} else {
			gnome_thumbnail_factory_create_failed_thumbnail (thumb_factory, 
									 info->uri,
									 current_orig_mtime);
		}
		
		pthread_mutex_lock (&thumbnails_mutex);
		/*********************************
		 * MUTEX LOCKED
		 *********************************/
		thumbnails_to_load = g_list_append((GList*) thumbnails_to_load, info);
		/*********************************
		 * MUTEX UNLOCKED
		 *********************************/
		pthread_mutex_unlock (&thumbnails_mutex);
	}
}

/*** plugin services ***/

/**
 * kiba_thumbnail_factory_load_icon:
 * @uri: The icon to load.
 * @func: Callback to invoke after the thumbnail was loaded.
 * @userdata: Userdata for @func.
 *
 * Load @uri to thumbnail in a separate thread (if the icon isnt allready added to the factory)
 * and invoke @func in the main thread.
 * Please note that @func may is not called immediately.
 */
static gboolean kiba_thumbnail_factory_load_thumb (const char *uri,
                                                   KibaThumbnailFactoryCallback func,
                                                   gpointer userdata) {
	GnomeThumbnailFactory *thumb_factory = thumb_factory_get();
	KibaThumbnailInfo *info;
	pthread_attr_t thread_attributes;
	pthread_t thumbnail_thread;
	time_t mtime;
	GnomeVFSURI *vfs_uri;
	char *filePath;
	char *escaped_uri;
	char *thumbPath;
	GList *l;
	
	g_return_if_fail(func != NULL);
	
	escaped_uri = gnome_vfs_make_uri_from_input (uri);
	vfs_uri = gnome_vfs_uri_new (escaped_uri);
	g_free (escaped_uri);
	if (vfs_uri != NULL) {
		filePath = gnome_vfs_uri_to_string (vfs_uri, GNOME_VFS_URI_HIDE_NONE);
		gnome_vfs_uri_unref (vfs_uri);
	}
	else {
		filePath = g_strdup(uri);
	}
	
	__get_file_mtime (filePath, &mtime);
	
	/* check if there is already a thumbnail */
	thumbPath = gnome_thumbnail_factory_lookup (thumb_factory, filePath, mtime);
	if (thumbPath) {
		func(thumbPath, userdata);
		g_free(filePath);
		g_free(thumbPath);
		return TRUE;
	}
	
	KibaThumbnailFactoryCallbackData *callbackData = g_new(KibaThumbnailFactoryCallbackData, 1);
	callbackData->func = func;
	callbackData->userdata = userdata;
	
	/* check if the thumbnail is in the to make list */
	for (l = (GList*) thumbnails_to_make; l != NULL; l = l->next) {
		info = l->data;
		if (strcmp(filePath, info->uri) == 0) {
			/* data->callbacks is only used in the main thread, so we can savely access it here ;) */
			info->callbacks = g_list_append (info->callbacks, callbackData);
			g_free(filePath);
			return TRUE;
		}
	}
	
	info = g_new0(KibaThumbnailInfo, 1);
	info->uri = filePath;
	info->callbacks = g_list_append (NULL, callbackData);
	info->original_file_mtime = mtime;
	
	pthread_mutex_lock (&thumbnails_mutex);
	/*********************************
	 * MUTEX LOCKED
	 *********************************/
	thumbnails_to_make = g_list_append ((GList*) thumbnails_to_make, info);
	/*********************************
	 * MUTEX UNLOCKED
	 *********************************/
	pthread_mutex_unlock (&thumbnails_mutex);
	
	if (!thumbnail_load_timeout_id)
		thumbnail_load_timeout_id = g_timeout_add (100, (GSourceFunc) __load_thumbnails, NULL);
	
	if (thumbnail_thread_is_running)
		return TRUE;
	
	/* We create the thread in the detached state, as we don't need/want
	   to join with it at any point. */
	pthread_attr_init (&thread_attributes);
	pthread_attr_setdetachstate (&thread_attributes,  PTHREAD_CREATE_DETACHED);
#ifdef _POSIX_THREAD_ATTR_STACKSIZE
	pthread_attr_setstacksize (&thread_attributes, 128*1024);
#endif
	
	/* We set a flag to indicate the thread is running, so we don't create
	   a new one. We don't need to lock a mutex here, as the thumbnail
	   thread isn't running yet. And we know we won't create the thread
	   twice, as we also check __thread_starter_id before
	   scheduling this idle function. */
	thumbnail_thread_is_running = TRUE;
	pthread_create (&thumbnail_thread, &thread_attributes, __thread_start, NULL);
	
	return TRUE;
}

static GnomeThumbnailFactory* thumb_factory_get (void) {
	static GnomeThumbnailFactory* thumb_factory = NULL;
	
	if (!thumb_factory) {
		thumb_factory = gnome_thumbnail_factory_new (GNOME_THUMBNAIL_SIZE_NORMAL);
	}
	
	return thumb_factory;
}

/*** DBus methods ***/

static gboolean thumbnail_dbus_get_thumb (GObject *plugin,
                                          const gchar *path,
                                          KibaThumbnailFactoryCallback callback,
                                          gpointer user_data,
                                          GError **error) {
	return kiba_thumbnail_factory_load_thumb (path, callback, user_data);
}

static gboolean thumbnail_dbus_get_factory (GObject *plugin,
                                            gpointer *factory_ret,
                                            GError **error) {
	GnomeThumbnailFactory *factory = thumb_factory_get ();
	*factory_ret = (gpointer) factory;
	return TRUE;
}

/*** Plugin methods ***/

static void __plugin_unload (KibaPlugin *plugin) {
	GnomeThumbnailFactory *factory = thumb_factory_get();
	KibaThumbnailInfo *data;
	GList *l;
	
	if (thumbnail_thread_is_running) {
		thumbnail_thread_is_running = FALSE;
		pthread_exit (NULL);
	}
	
	for (l = (GList*) thumbnails_to_make; l != NULL; l = (GList*) thumbnails_to_make) {
		data = l->data;
		__free_thumbnail_info(data);
		thumbnails_to_make = g_list_remove((GList*) thumbnails_to_make, l->data);
	}
	for (l = (GList*) thumbnails_to_load; l != NULL; l = (GList*) thumbnails_to_load) {
		data = l->data;
		__free_thumbnail_info(data);
		thumbnails_to_make = g_list_remove((GList*) thumbnails_to_load, l->data);
	}
	
	if (G_IS_OBJECT(factory))
		g_object_unref(G_OBJECT(factory));
}

static gboolean __plugin_load (KibaPlugin *plugin) {
	return TRUE; /* not much todo here ;) */
}

static PrefMember thumbnail_options[] = {};

static KibaPluginVtable vtable = {
	"Thumbnail plugin for kiba-dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_ONCE_LOADABLE, /* loadable flag */
	
	__plugin_load, /* load  */
	__plugin_unload, /* unload */
};

//TODO we need a remove callback method, maybe the plugin that requests the thumbnail wants to unload!

#include "thumbnail-dbus-glue.h"
KIBA_INIT_PLUGIN (vtable, thumbnail_options, &dbus_glib_thumbnail_dbus_object_info, NULL, NULL);

