/*
 *  Copyright (C) 2007 Nick
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author(s) : Nick
 *              Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <string.h>
#include <stdlib.h>
#include <libintl.h>
#include <gtk/gtk.h>

#include <libgnomeui/libgnomeui.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-icon.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-screen.h>
#include <kiba-debug.h>
#include <kiba-icon-view-win.h>
#include <kiba-dialog.h>
#include <kiba-macros.h>
#include <kiba-image.h>
#include <kiba-utils.h>
#include <kiba-globals.h>

#include "trash-monitor.h"

#define _(String)    dgettext (GETTEXT_PACKAGE, String)
#define PLUGIN_NAME  "trash"

#define ICON_VIEW_ICON_SIZE 48

typedef struct _KibaTrash KibaTrash;

static const char* trash_icon_data[3][3] = {
	{ "gnome-fs-trash-full", "gnome-fs-trash-empty" },
	{ "stock_trash_full",    "stock_trash" },
	{ "trashcan_full", "trashcan_empty" }
};

struct _KibaTrash {
	KibaPlugin *plugin;
	
	GtkWidget *icon;
	
	KibaIconViewWin *view_win;
	GList *files;
	
	TrashMonitor *monitor;
	gboolean is_empty;
	guint monitor_signal_id;
};

typedef struct {
	GtkWidget *box;
	char *path;
	char *name;
}KibaTrashFile;

static void kiba_trash_drag_data_received_cb (GtkWidget        *widget,
					      GdkDragContext   *drag_context,
					      gint              x,
					      gint              y,
					      GtkSelectionData *selection_data,
					      guint             info,
					      guint             time,
					      gpointer          data);
static void __toggle_icon_view(KibaTrash *trash);

/**
Helper Functions
**/

static void __update_item_count (KibaTrash *trash) {
	GtkWidget *widget = GTK_WIDGET(trash->icon);
	GtkWidget *win = widget->parent->parent;
	KibaProgress *progress = KIBA_WIN(win)->progress;
	gchar *title;
	gint trash_files;
	
	trash_files = trash_monitor_get_item_count (trash->monitor);
	trash->is_empty = !trash_files;
	
	if (trash->is_empty)
		kiba_object_set_title (KIBA_OBJECT(trash->icon), _("Empty"));
	else if (!kiba_progress_object_is_running (progress, KIBA_OBJECT(trash->icon))) {
		title = g_strdup_printf ("%s: %i", _("Items"), trash_files);
		kiba_object_set_title (KIBA_OBJECT(trash->icon), title);
		g_free (title);
	}
}

static gint __transfer_callback (GnomeVFSAsyncHandle      *handle,
					  GnomeVFSXferProgressInfo *progress_info,
					  gpointer                  user_data) {
	KibaTrash *trash = user_data;
	GtkWidget *widget = GTK_WIDGET(trash->icon);
	GtkWidget *win = widget->parent->parent;
	KibaProgress *progress = KIBA_WIN(win)->progress;
	
	if (progress_info->files_total != 0) {
		GtkWidget *progressbar;
		gchar *tooltip;
		
		if (progress_info->source_name != NULL) {
			char *short_name;
			
			short_name = g_strrstr(progress_info->source_name, "/");
			if (short_name) short_name++;
			else short_name = progress_info->source_name;
			
			if (strlen (short_name) > 27) {
				char *buf;
			
				buf = g_strndup (short_name, 27);
				tooltip = g_strdup_printf ("%s: %s ...", _("Removing"), buf);
				g_free (buf);
			} else {
				tooltip = g_strdup_printf ("%s: %s", _("Removing"), short_name);
			}
		}
		else {
			char *index, *total;
	
			index = g_strdup_printf ("%i", progress_info->file_index);
			total = g_strdup_printf ("%i", progress_info->files_total);
			tooltip = g_strdup_printf (("%s %s of %s"), _("Removing"), index, total);
			g_free (index);
			g_free (total);
		}
		
		kiba_object_set_tooltip_text(KIBA_OBJECT(trash->icon), tooltip);
		g_free (tooltip);
		kiba_progress_object_set_value (progress, KIBA_OBJECT(trash->icon),
				100 * progress_info->file_index / progress_info->files_total);
	}
	
	if (progress_info->phase == GNOME_VFS_XFER_PHASE_COMPLETED) {
		kiba_object_set_tooltip_text(KIBA_OBJECT(trash->icon), NULL);
		kiba_progress_object_set_value (progress, KIBA_OBJECT(trash->icon), 100);
		__update_item_count (trash);
	}
	
	return 1;
}

static gboolean __confirm_delete_immediately (GdkScreen *screen,
					    gint       num_files,
					    gboolean   all) {
	GtkWidget *dialog, *hbox, *vbox, *image, *label;
	gchar *str, *prompt, *detail;
	int response;
	
	dialog = gtk_dialog_new ();
	gtk_window_set_screen (GTK_WINDOW (dialog), screen);
	atk_object_set_role (gtk_widget_get_accessible (dialog), ATK_ROLE_ALERT);
	gtk_window_set_title (GTK_WINDOW (dialog), _("Delete Immediately?"));
	gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
	
	gtk_widget_realize (dialog);
	gdk_window_set_transient_for (GTK_WIDGET (dialog)->window,
				      gdk_screen_get_root_window (screen));
	gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
	
	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 14);
	
	hbox = gtk_hbox_new (FALSE, 12);
	gtk_container_set_border_width (GTK_CONTAINER (hbox), 5);
	gtk_widget_show (hbox);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox,
			    FALSE, FALSE, 0);
	
	image = gtk_image_new_from_stock (GTK_STOCK_DIALOG_QUESTION,
					  GTK_ICON_SIZE_DIALOG);
	gtk_misc_set_alignment (GTK_MISC (image), 0.5, 0.0);
	gtk_widget_show (image);
	gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
	
	vbox = gtk_vbox_new (FALSE, 12);
	gtk_box_pack_start (GTK_BOX (hbox), vbox, TRUE, TRUE, 0);
	gtk_widget_show (vbox);
	
	if (all) {
		prompt = _("Cannot move items to trash, do you want to delete them immediately?");
		detail = g_strdup_printf (_("None of the %d selected items can be moved to the Trash"), num_files);
	} else {
		prompt = _("Cannot move some items to trash, do you want to delete these immediately?");
		detail = g_strdup_printf (_("%d of the selected items cannot be moved to the Trash"), num_files);
	}
	
	str = g_strconcat ("<span weight=\"bold\" size=\"larger\">",
			   prompt, "</span>", NULL);
	label = gtk_label_new (str);
	gtk_label_set_use_markup (GTK_LABEL (label), TRUE);
	gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
	gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
	gtk_widget_show (label);
	g_free (str);
	
	label = gtk_label_new (detail);
	gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
	gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
	gtk_widget_show (label);
	g_free (detail);
	
	gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_CANCEL,
			       GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_DELETE,
			       GTK_RESPONSE_YES);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog),
					 GTK_RESPONSE_YES);
	
	response = gtk_dialog_run (GTK_DIALOG (dialog));
	
	gtk_object_destroy (GTK_OBJECT (dialog));
	
	return response == GTK_RESPONSE_YES;
}

static void __clear_dialog_response (GtkWidget    *widget,
				   int           response,
				   KibaTrash    *trash) {
	if (response == GTK_RESPONSE_YES) {
		GnomeVFSAsyncHandle *hnd;
		
		trash_monitor_empty_trash (trash->monitor,
			&hnd, __transfer_callback, trash);
	}
	
	gtk_widget_destroy (widget);
}

static void __empty (GtkWidget *widget,
			      gpointer   data) {
	KibaTrash *trash = data;
	GtkWidget *dialog;
	
	dialog = kiba_dialog_confirm(_("Empty all of the items from the trash?"),
	    _("If you choose to empty the trash, all items in it will be permanently lost. Please note that you can also delete them separately."));
	
	g_signal_connect (dialog, "response",
			  G_CALLBACK (__clear_dialog_response), trash);
}

static void __update_icon (KibaTrash *trash) {
	gint i;
	
	for (i = 0; i < G_N_ELEMENTS(trash_icon_data); i++) {
		if (kiba_icon_set_file (KIBA_ICON(trash->icon),
			trash_icon_data[i][trash->is_empty]))
			break;
	}
}

/*** Icon View ***/

static void __place_icon_view (KibaTrash *trash) {
	GtkWidget *dock, *win;
	gint w, h, x, y;
	const KibaLayoutData *d;
	
	dock = GTK_WIDGET(trash->icon)->parent;
	win = dock->parent;
	
	if (!KIBA_IS_WIN(win))
		return;
	if (!KIBA_IS_ICON_VIEW_WIN(trash->view_win))
		return;
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	struct {
		gint  *x, *y;
		gint  *w, *h;
		gint  alignment;
	} magic[2] = {
		{ &x, &y , &w, &h, d->valignment},
		{ &y, &x , &h, &w, d->halignment}
	};
	w = trash->view_win->win->allocation.width;
	h = trash->view_win->win->allocation.height;
	x = trash->icon->allocation.x + GTK_WIDGET(win)->allocation.x;
	y = trash->icon->allocation.y + GTK_WIDGET(win)->allocation.y;
	
	*magic[d->orientation].x -= (*magic[d->orientation].w - trash->icon->allocation.width)/2;
	*magic[d->orientation].y -= (magic[d->orientation].alignment == 0 ?
			- trash->icon->allocation.height : *magic[d->orientation].h);
	
	gdk_window_move (trash->view_win->win->window, x, y);
}

static void __icon_view_size_allocate_cb (GtkWidget     *widget,
						      GtkAllocation *allocation,
						      gpointer       data) {
	KibaTrash *trash = data;
	__place_icon_view (trash);
}

static void __icon_view_child_activated_cb (KibaIconView *icon_view,
					       GtkWidget    *child,
					       gpointer      user_data) {
//FODO
}

static void __icon_view_drag_data_get_cb (GtkWidget        *widget,
					     GdkDragContext   *drag_context,
					     GtkSelectionData *data,
					     guint             info,
					     guint             time,
					     gpointer          user_data) {
	KibaTrashFile *file = user_data;
	
	gtk_selection_data_set(data,
		GDK_SELECTION_TYPE_STRING,
		8,	/* 8 bits per character. */
		(guchar *)file->path,
		strlen(file->path));
}

static void __icon_view_drag_received_cb (GtkWidget        *widget,
					     GdkDragContext   *drag_context,
					     gint              x,
					     gint              y,
					     GtkSelectionData *selection_data,
					     guint             info,
					     guint             time,
					     gpointer          userdata) {
	KibaTrash *trash = userdata;
	kiba_trash_drag_data_received_cb (trash->icon, drag_context, x, y,
		selection_data, info, time, trash);
}

static void __thumbnail_loaded (cairo_surface_t *surface,
                                gint size,
                                gpointer userdata) {
	kiba_image_set_surface (KIBA_IMAGE(userdata), surface, size, size);
}

static void __image_unrealize_cb (GtkWidget *image,
                                  gpointer data) {
	kiba_surface_factory_remove_surface (KIBA_IMAGE(image)->surface);
}

static void image_box_label_set_markup (GtkWidget       *box,
					GtkWidget       *label) {
	char *markup;
	
	markup = g_markup_printf_escaped (
		"<span size=\"small\" style=\"italic\">%s</span>",
		 GTK_LABEL (label)->text);
	gtk_label_set_markup (GTK_LABEL (label), markup);
	g_free (markup);
}

static GtkWidget* __create_icon_view_item (KibaTrashFile *file) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	GtkWidget *image;
	GtkWidget *box;
	GtkWidget *label;
	char *markup;
	
	box = gtk_vbox_new(FALSE, 0);
	
	GtkWidget *image_table;
	image_table = gtk_table_new (1, 3, FALSE);
	
	image = kiba_image_new ();
	kiba_image_set_size (KIBA_IMAGE(image), ICON_VIEW_ICON_SIZE, ICON_VIEW_ICON_SIZE);
	gtk_widget_show (image);
	gtk_table_attach (GTK_TABLE(image_table), image,
				1, 2,
				0, 1,
				GTK_EXPAND | GTK_SHRINK, GTK_EXPAND | GTK_SHRINK,
				0, 0);
	gtk_box_pack_start (GTK_BOX (box), image_table, TRUE, TRUE, 0);
	gtk_widget_show (image_table);
	
	GnomeVFSFileInfo info;
	char *mimeIcon, *mime_type;
	
	if (gnome_vfs_get_file_info (file->path, &info, GNOME_VFS_FILE_INFO_DEFAULT) == GNOME_VFS_OK) {
		mime_type = gnome_vfs_get_mime_type (file->path);
		mimeIcon = gnome_icon_lookup (gtk_icon_theme_get_default(),
						NULL, file->path, NULL,
						&info, mime_type,
						GNOME_ICON_LOOKUP_FLAGS_NONE, NULL);
		g_free(mime_type);
		kiba_surface_factory_load_icon(mimeIcon, ICON_VIEW_ICON_SIZE, __thumbnail_loaded, image);
		g_free(mimeIcon);
	}
	
	label = gtk_label_new(file->name);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
	gtk_label_set_ellipsize (GTK_LABEL(label), PANGO_ELLIPSIZE_END);
	image_box_label_set_markup (box, label);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
	
	gtk_drag_source_set (image,
		GDK_BUTTON1_MASK | GDK_BUTTON2_MASK,
		targets, G_N_ELEMENTS (targets),
		GDK_ACTION_MOVE | GDK_ACTION_COPY);
	g_signal_connect (GTK_OBJECT(image), "unrealize",
		G_CALLBACK(__image_unrealize_cb), NULL);
	g_signal_connect (GTK_OBJECT(image), "drag-data-get",
		G_CALLBACK(__icon_view_drag_data_get_cb), file);
	
	gtk_widget_set_usize (box, ICON_VIEW_ICON_SIZE + 30, -1);
	
	return box;
}

static KibaTrashFile* __get_trash_file (KibaTrash *trash,
                                        const char *path) {
	GList *l;
	for (l = trash->files; l != NULL; l = l->next) {
		KibaTrashFile *file = l->data;
		if (strcmp(file->path, path) == 0)
			return file;
	}
	return NULL;
}

static void __icon_view_remove_trashed_file (KibaTrash *trash,
                                          const char *path) {
	KibaTrashFile *file;
	
	file = __get_trash_file(trash, path);
	if (!file) return;
	
	if (KIBA_IS_ICON_VIEW_WIN(trash->view_win))
		kiba_icon_view_win_remove (trash->view_win, file->box);
	
	trash->files = g_list_remove(trash->files, file);
	g_free(file->path);
	g_free(file);
}

static void __icon_view_add_trashed_file (KibaTrash *trash,
                                          const char *path) {
	KibaTrashFile *file;
	GtkWidget *box;
	
	file = __get_trash_file(trash, path);
	if (file) return;
	
	file = g_new(KibaTrashFile,1);
	file->path = g_strdup(path);
	file->name = g_strrstr(file->path, "/");
	if (file->name) file->name++;
	
	box = __create_icon_view_item (file);
	file->box = box;
	
	kiba_icon_view_win_add (trash->view_win, box);
	
	trash->files = g_list_append(trash->files, file);
}

static void __fill_icon_view(KibaTrash *trash) {
	GList *files = trash_monitor_get_files(trash->monitor);
	GList *l;
	
	for (l = files; l != NULL; l = files) {
		char *path = l->data;
		__icon_view_add_trashed_file(trash,path);
		g_free(path);
		files = g_list_remove(files, l->data);
	}
	g_list_free(files);
}

static void __create_icon_view_win (KibaTrash *trash) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	Kiba *kiba = kiba_get();
	GtkWidget *box;
	GList *l;
	
	trash->view_win = kiba_icon_view_win_new(KIBA_ICON_VIEW_BOX_POSITION_RIGHT, 4, 3, FALSE);
	gtk_window_set_type_hint (GTK_WINDOW(trash->view_win->win), GDK_WINDOW_TYPE_HINT_DOCK);
	gtk_window_stick (GTK_WINDOW(trash->view_win->win));
	gtk_window_set_role (GTK_WINDOW(trash->view_win->win), "kiba-launcher");
	__fill_icon_view(trash);
	
	gtk_drag_dest_set (trash->view_win->icon_view,
		GTK_DEST_DEFAULT_ALL,
		targets, G_N_ELEMENTS (targets),
		GDK_ACTION_COPY | GDK_ACTION_LINK | GDK_ACTION_MOVE);
	gtk_drag_dest_add_uri_targets (trash->view_win->icon_view);
	
	g_signal_connect (GTK_OBJECT(trash->view_win->icon_view), "drag-data-received",
		G_CALLBACK(__icon_view_drag_received_cb), trash);
	g_signal_connect (GTK_OBJECT(trash->view_win->icon_view), "child-activated",
		G_CALLBACK(__icon_view_child_activated_cb), trash);
}

static void __view_win_destroy_cb (GtkWidget *widget,
					       gpointer   user_data) {
	KibaTrash *trash = user_data;
	g_signal_handlers_disconnect_by_func (G_OBJECT(trash->icon),
		  G_CALLBACK(__icon_view_size_allocate_cb), trash);
	g_signal_handlers_disconnect_by_func (G_OBJECT(trash->icon->parent),
		  G_CALLBACK(__icon_view_size_allocate_cb), trash);
	g_signal_handlers_disconnect_by_func (G_OBJECT(trash->icon->parent->parent),
		  G_CALLBACK(__icon_view_size_allocate_cb), trash);
	g_signal_handlers_disconnect_by_func (G_OBJECT(trash->view_win->icon_view),
		  G_CALLBACK(__icon_view_size_allocate_cb), trash);
	
	GList *l;
	for (l = trash->files; l != NULL; l = trash->files) {
		KibaTrashFile *file = l->data;
		__icon_view_remove_trashed_file(trash, file->path);
	}
	
	trash->view_win = NULL;
}

static void __toggle_icon_view(KibaTrash *trash) {
	if (KIBA_IS_ICON_VIEW_WIN(trash->view_win)) {
		gtk_widget_destroy (trash->view_win->win);
	}
	else {
		__create_icon_view_win (trash);
		gtk_widget_show (trash->view_win->win);
		gdk_window_set_role (trash->view_win->win->window, "kiba-trash");
		gdk_window_set_keep_above (trash->view_win->win->window, TRUE);
		__place_icon_view (trash);
		
		g_signal_connect(G_OBJECT (trash->view_win->win), "destroy",
			G_CALLBACK(__view_win_destroy_cb), trash);
		g_signal_connect_after(G_OBJECT (trash->icon), "size-allocate",
			G_CALLBACK(__icon_view_size_allocate_cb), trash);
		g_signal_connect_after(G_OBJECT (trash->icon->parent), "size-allocate",
			G_CALLBACK(__icon_view_size_allocate_cb), trash);
		g_signal_connect_after(G_OBJECT (trash->icon->parent->parent), "size-allocate",
			G_CALLBACK(__icon_view_size_allocate_cb), trash);
		g_signal_connect_after(G_OBJECT (trash->view_win->icon_view), "size-allocate",
			G_CALLBACK(__icon_view_size_allocate_cb), trash);
	}
}

/**
Signal Handler
**/

static void item_count_changed (TrashMonitor *monitor,
				KibaTrash    *trash) {
	__update_item_count (trash);
	__update_icon (trash);
	
	/* update the icon view */
	if (KIBA_IS_ICON_VIEW_WIN(trash->view_win)) {
		GList *files = trash_monitor_get_files(trash->monitor);
		GList *l;
		
		for (l = files; l != NULL; l = l->next) {
			char *path = l->data;
			__icon_view_add_trashed_file(trash, path);
		}
		
		for (l = g_list_copy(trash->files); l != NULL; l = l->next) {
			KibaTrashFile *file = l->data;
			gboolean deletedFile = TRUE;
			GList *li;
			
			for (li = files; li != NULL; li = li->next) {
				char *ipath = li->data;
				if (strcmp(ipath, file->path) == 0) {
					deletedFile = FALSE;
					break;
				}
			}
			
			if (deletedFile)
				__icon_view_remove_trashed_file(trash, file->path);
		}
		
		for (l = files; l != NULL; l = files) {
			g_free(l->data);
			files = g_list_remove(files, l->data);
		}
		g_list_free(files);
	}
}

static void move_to_trash (KibaObject       *object,
			   GdkDragContext   *context,
			   GtkSelectionData *data,
			   KibaTrash        *trash) {
	GList *list, *scan;
	GList *source_uri_list, *target_uri_list, *unmovable_uri_list;
	GnomeVFSResult result;
	
	list = gnome_vfs_uri_list_parse ((gchar *)data->data);
	
	source_uri_list = NULL;
	target_uri_list = NULL;
	unmovable_uri_list = NULL;
	for (scan = g_list_first (list); scan; scan = g_list_next (scan)) {
		GnomeVFSURI *source_uri = scan->data;
		GnomeVFSURI *trash_uri, *target_uri;
		gchar *source_basename;
		
		/* find the trash directory for this file */
		result = gnome_vfs_find_directory (source_uri,
						   GNOME_VFS_DIRECTORY_KIND_TRASH,
						   &trash_uri, TRUE, FALSE, 0);
		if (result != GNOME_VFS_OK) {
			unmovable_uri_list = g_list_prepend (unmovable_uri_list,
							     gnome_vfs_uri_ref (source_uri));
			continue;
		}
		
		source_basename = gnome_vfs_uri_extract_short_name
			(source_uri);
		
		target_uri = gnome_vfs_uri_append_file_name(trash_uri,
							    source_basename);
		g_free (source_basename);
		gnome_vfs_uri_unref (trash_uri);
		
		source_uri_list = g_list_prepend (source_uri_list,
						  gnome_vfs_uri_ref (source_uri));
		target_uri_list = g_list_prepend (target_uri_list,
						  target_uri);
	}
	
	gnome_vfs_uri_list_free(list);
	
	/* we may have added a trash dir, so recheck */
	trash_monitor_recheck_trash_dirs (trash->monitor);
	
	if (source_uri_list) {
		result = gnome_vfs_xfer_uri_list (source_uri_list, target_uri_list,
						  GNOME_VFS_XFER_REMOVESOURCE |
						  GNOME_VFS_XFER_RECURSIVE,
						  GNOME_VFS_XFER_ERROR_MODE_ABORT,
						  GNOME_VFS_XFER_OVERWRITE_MODE_REPLACE,
						  NULL, NULL);
		gnome_vfs_uri_list_free (source_uri_list);
		gnome_vfs_uri_list_free (target_uri_list);
		if (result != GNOME_VFS_OK) {
			DEBUG_ERROR ("Unable to move to trash:\n%s", gnome_vfs_result_to_string (result));
		}
	}
	if (unmovable_uri_list) {
		if (__confirm_delete_immediately (KIBA_GDK_SCREEN,
						g_list_length (unmovable_uri_list),
						source_uri_list == NULL)) {
			result = gnome_vfs_xfer_delete_list (unmovable_uri_list,
							     GNOME_VFS_XFER_ERROR_MODE_ABORT,
							     GNOME_VFS_XFER_RECURSIVE,
							     NULL, NULL);
		} else {
			result = GNOME_VFS_OK;
		}
		gnome_vfs_uri_list_free (unmovable_uri_list);
		if (result != GNOME_VFS_OK) {
			DEBUG_ERROR ("Unable to move to trash:\n%s",
				      gnome_vfs_result_to_string (result));
		}
	}
}

static gboolean kiba_object_button_press_cb (GtkWidget      *widget,
					     GdkEventButton *event,
					     gpointer        data) {
	KibaPlugin *plugin = data;
	KibaTrash *trash;
	GtkWidget *menu;
	GtkWidget *item, *image;
	
	trash = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	if (trash->is_empty)
		return TRUE;
	
	menu = gtk_menu_new ();
	
	item = gtk_image_menu_item_new_with_label (_("Empty Trash"));
	image = gtk_image_new_from_icon_name ("gnome-fs-trash-empty", GTK_ICON_SIZE_MENU);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
	g_signal_connect (G_OBJECT (item),
		"activate", G_CALLBACK (__empty), trash);
	gtk_widget_show (item);
	
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
		event->button, event->time);
	
	return TRUE;
}

static gboolean kiba_object_button_release_cb (GtkWidget      *widget,
					       GdkEventButton *event,
					       gpointer        data) {
	KibaTrash *trash = data;
	char *trash_command;
	
	switch (event->button) {
		case 2: {
			__toggle_icon_view (trash);
			return TRUE;
		}
		case 1:
		default: {
			trash_command = g_strdup_printf ("%s trash:/", file_manager_global);
			kiba_launch_programm (trash_command);
			g_free (trash_command);
			break;
		}
	}
	
	return FALSE;
}

static void icon_theme_changed_cb (GtkIconTheme *icon_theme,
				   gpointer      user_data) {
	KibaTrash *trash = user_data;
	
	__update_icon (trash);
}

/**
KibaPrefs Callbacks
**/

static void dock_object_size_cb (const char *option_id,
			         gpointer    value,
			         gpointer    data) {
	KibaTrash *trash = data;
	
	__update_icon (trash);
}

static void object_size_factor_cb (const char *option_id,
			           gpointer    value,
			           gpointer    data) {
	KibaTrash *trash = data;
	KibaDrawableObject *object;
	float factor;
	
	object = KIBA_DRAWABLE_OBJECT(trash->icon);
	factor = g_ascii_strtod((char*) value, NULL);
	
	kiba_object_set_requisition_factor (KIBA_OBJECT(trash->icon), factor, factor);
	object->surface_data->width  = trash->icon->allocation.width;
	object->surface_data->height = trash->icon->allocation.height;
	__update_icon (trash);
	gtk_widget_queue_resize (trash->icon->parent);
}

static void object_alpha_cb (const char *option_id,
			     gpointer    value,
			     gpointer    data) {
	KibaTrash *trash = data;
	float alpha;
	
	alpha = g_ascii_strtod((char*) value, NULL);
	
	kiba_drawable_object_set_alpha (KIBA_DRAWABLE_OBJECT(trash->icon), alpha);
}

static void kiba_trash_drag_data_received_cb (GtkWidget        *widget,
					      GdkDragContext   *drag_context,
					      gint              x,
					      gint              y,
					      GtkSelectionData *selection_data,
					      guint             info,
					      guint             time,
					      gpointer          data) {
	KibaTrash *trash = data;
	
	if (drag_context->suggested_action != GDK_ACTION_MOVE &&
	    drag_context->suggested_action != GDK_ACTION_COPY)
		return;
	
	move_to_trash (KIBA_OBJECT(widget), drag_context, selection_data, trash);
	gtk_drag_finish (drag_context, TRUE, TRUE, time);
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaTrash  *trash;
	
	trash = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	file = kiba_get_plugin_config_file(kiba, plugin);
	
	kiba_prefs_remove_callback_by_data (file, trash);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(gtk_icon_theme_get_default()),
		G_CALLBACK(icon_theme_changed_cb), trash);
	
	if (KIBA_IS_ICON_VIEW_WIN(trash->view_win))
		__toggle_icon_view(trash);
	
	if (trash->monitor_signal_id)
		g_signal_handler_disconnect (trash->monitor,
			trash->monitor_signal_id);
	trash->monitor_signal_id = 0;
	
	trash_monitor_fini (trash->monitor);
	
	kiba_object_destroy (KIBA_OBJECT(trash->icon));
	
	g_free (trash);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaTrash *trash;
	GnomeVFSResult res;
	GnomeVFSURI *trash_uri;
	GList *l;
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");	
#endif
	
	if (!gnome_vfs_initialized () && !gnome_vfs_init ()) {
		DEBUG_ERROR ("Can not initialise gnome-vfs.\n");
		return FALSE;
	}
	
	trash = g_new0(KibaTrash, 1);
	trash->plugin = plugin;
	
	trash->monitor  = trash_monitor_get (G_TYPE_MODULE(plugin));
	if (!TRASH_IS_MONITOR(trash->monitor)) {
		DEBUG_ERROR ("Failed to get the TrashMonitor.\n");
		g_free (trash);
		return FALSE;
	}
	
	/* add icon to the dock */
	trash->icon = kiba_icon_new_from_file (PLUGIN_NAME, trash_icon_data[0][1]);
	kiba_add_plugin_object(kiba, plugin, KIBA_OBJECT(trash->icon));
	g_signal_connect (G_OBJECT(trash->icon), "drag-data-received",
		G_CALLBACK(kiba_trash_drag_data_received_cb), trash);
	gtk_widget_show (trash->icon);
	
	trash->is_empty = TRUE;
	trash->files = NULL;
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	
	/* Find Trash directories */
	res = gnome_vfs_find_directory (NULL,
		GNOME_VFS_DIRECTORY_KIND_TRASH,
		&trash_uri,
		TRUE, TRUE, 0700);
	if (trash_uri)
		gnome_vfs_uri_unref (trash_uri);
	if (res != GNOME_VFS_OK)
		DEBUG_ERROR ("Unable to find the Trash directory: %s",
			gnome_vfs_result_to_string (res));
	
	__update_item_count (trash);
	__update_icon (trash);
	
	trash->monitor_signal_id = g_signal_connect_after (trash->monitor, "item_count_changed",
			G_CALLBACK (item_count_changed), trash);
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) trash);
	
	/* add widget signal handler */
	g_signal_connect (G_OBJECT (trash->icon), "button-press-event",
			G_CALLBACK (kiba_object_button_press_cb), plugin);
	g_signal_connect (G_OBJECT (trash->icon), "button-release-event",
			G_CALLBACK (kiba_object_button_release_cb), trash);
	
	g_signal_connect (G_OBJECT (gtk_icon_theme_get_default()), "changed",
			G_CALLBACK (icon_theme_changed_cb), trash);
	
	kiba_prefs_add_callback (file, "dock_object_size",
		dock_object_size_cb, trash, FALSE);
	kiba_prefs_add_callback (file, "trash_size_factor",
		object_size_factor_cb, trash, TRUE);
	kiba_prefs_add_callback (file, "trash_alpha",
		object_alpha_cb, trash, TRUE);
	
	return TRUE;
}

static PrefMember trash_options[] = {
    { "trash_size_factor", "1.0", KIBA_DOUBLE },
    { "trash_alpha", "1.0", KIBA_DOUBLE }
};

static KibaPluginVtable vtable = {
	"Trash Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_ONCE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
};

KIBA_INIT_PLUGIN (vtable, trash_options, NULL, NULL, NULL);
