/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <string.h>
#include <gtk/gtk.h>
#include <X11/Xatom.h>

#include <kiba.h>
#include <kiba-plugin.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-container-object.h>
#include <kiba-screen.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>
#include <kiba-x11.h>

#include "na-tray-manager.h"

#define PLUGIN_NAME "tray"
#define TRAY_ICON_SIZE 28

typedef struct {
	KibaPlugin *plugin;
	NaTrayManager *manager;
	GHashTable *icons;
}KibaTray;

typedef struct {
	KibaTray *tray;
	KibaObject *object;
	GtkWidget *icon;
	GtkWidget *box;
	gint idle_redraw_id;
}KibaTrayIcon;

/*** private methods ***/

static gboolean idle_redraw_cb (KibaTrayIcon *tray_icon) {
	tray_icon->idle_redraw_id = 0;
	gtk_widget_hide (tray_icon->box);
	gtk_widget_show (tray_icon->box);
	
	return FALSE;
}

void __force_redraw (KibaTrayIcon *tray_icon) {
	/* Force the icons to redraw their backgrounds.
	* gtk_widget_queue_draw() doesn't work across process boundaries,
	* so we do this instead.
	*/
	if (tray_icon->idle_redraw_id == 0)
		tray_icon->idle_redraw_id = g_idle_add ((GSourceFunc) idle_redraw_cb, tray_icon);
}

static void __remove_object (gpointer key,
                             gpointer value,
                             gpointer data) {
	KibaTray *tray = data;
	KibaTrayIcon *tray_icon = value;
	GtkWidget *widget = GTK_WIDGET(tray_icon->object);
	GtkWidget *child = gtk_bin_get_child(GTK_BIN(widget));
	
	gtk_container_remove(GTK_CONTAINER(tray_icon->box), tray_icon->icon);
	gtk_container_remove(GTK_CONTAINER(tray_icon->object), tray_icon->box);
	kiba_object_destroy(tray_icon->object);
	g_hash_table_remove(tray->icons, key);
	g_free(tray_icon);
}

/** NaTrayManager signal handler */

static void __icon_added (NaTrayManager *manager,
                          GtkWidget *icon,
                          gpointer data) {
	KibaTray *tray = data;
	KibaTrayIcon *tray_icon;
	Kiba *kiba = kiba_get();
	GtkWidget *object;
	GtkWidget *box;
	
	tray_icon = g_hash_table_lookup (tray->icons, icon);
	g_return_if_fail (tray_icon == NULL);
	
	tray_icon = g_new0(KibaTrayIcon, 1);
	tray_icon->tray = tray;
	tray_icon->idle_redraw_id = 0;
	tray_icon->icon = icon;
	
	box = gtk_event_box_new ();
	gtk_container_set_border_width (GTK_CONTAINER(box), 3);
	gtk_widget_show (box);
	gtk_widget_set_size_request (icon, TRAY_ICON_SIZE, TRAY_ICON_SIZE);
	gtk_container_add (GTK_CONTAINER(box), icon);
	gtk_widget_show_all (box);
	
	tray_icon->box = box;
	
	g_hash_table_insert (tray->icons, icon, tray_icon);
	
	object = kiba_container_object_new (PLUGIN_NAME);
	tray_icon->object = KIBA_OBJECT(object);
	kiba_add_plugin_object(kiba, tray->plugin, KIBA_OBJECT(object));
	gtk_widget_show (object);
	
	gtk_container_add (GTK_CONTAINER(object), box);
	
	__force_redraw (tray_icon);
}

static void __icon_removed (NaTrayManager *manager,
                            GtkWidget *icon,
                            gpointer data) {
	KibaTray *tray = data;
	KibaTrayIcon *tray_icon;
	
	tray_icon = g_hash_table_lookup (tray->icons, icon);
	g_return_if_fail (tray_icon != NULL);
	
	__remove_object(icon, tray_icon, tray);
}

static void __icon_message_sent (NaTrayManager *manager,
                                 GtkWidget *icon,
                                 const char *text,
                                 glong id,
                                 glong timeout,
                                 gpointer data) {
	KibaTray *tray = data;
// 	kiba_object_set_tooltip_text(KIBA_OBJECT(object), text);
}

static void __icon_message_cancelled (NaTrayManager *manager,
                                      GtkWidget *icon,
                                      glong id,
                                      gpointer data) {
	KibaTray *tray = data;
// 	kiba_object_set_tooltip_text(KIBA_OBJECT(object), NULL);
}

static void __lost_selection (NaTrayManager *manager,
                              gpointer data) {
	KibaTray *tray = data;
}

/** KibaDock signal handler **/

static void __dock_orientation_changed (KibaDock *dock,
                                        GtkOrientation orientation,
                                        gpointer data) {
	NaTrayManager *manager = data;
	
	na_tray_manager_set_orientation (manager, orientation);
}

/*** Plugin methods ***/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaTray *tray;
	KibaWin *win;
	
	tray = g_object_get_data (G_OBJECT(plugin), "tray");
	
	win = kiba_get_plugin_win(kiba, plugin);
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_WIN_CHILD(win)),
		  G_CALLBACK(__dock_orientation_changed), tray->manager);
	
	g_hash_table_foreach (tray->icons, __remove_object, tray);
	g_hash_table_unref(tray->icons);
	
	g_object_unref (G_OBJECT (tray->manager));
	g_free(tray);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	GtkOrientation orientation;
	KibaTray *tray;
	KibaWin *win;
	KibaDock *dock;
	
	if (na_tray_manager_check_running (KIBA_GDK_SCREEN)) {
		DEBUG_ERROR("Tray allready running!\n");
		return FALSE;
	}
	
	tray = g_new0(KibaTray, 1);
	tray->manager = na_tray_manager_new ();
	tray->plugin = plugin;
	
	if (!na_tray_manager_manage_screen(tray->manager, KIBA_GDK_SCREEN)) {
		DEBUG_ERROR("Failed to manage the screen!\n");
		g_object_unref(G_OBJECT(tray->manager));
		g_free(tray);
		return FALSE;
	}
	
	g_signal_connect (G_OBJECT(tray->manager), "tray-icon-added",
		G_CALLBACK(__icon_added), tray);
	g_signal_connect (G_OBJECT(tray->manager), "tray-icon-removed",
		G_CALLBACK(__icon_removed), tray);
	g_signal_connect (G_OBJECT(tray->manager), "message-sent",
		G_CALLBACK(__icon_message_sent), tray);
	g_signal_connect (G_OBJECT(tray->manager), "message-cancelled",
		G_CALLBACK(__icon_message_cancelled), tray);
	g_signal_connect (G_OBJECT(tray->manager), "lost-selection",
		G_CALLBACK(__lost_selection), tray);
	
	tray->icons = g_hash_table_new (
		NULL, /**<< using default hash function **/
		g_direct_equal); /**<< comparing pointers **/
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	g_signal_connect (G_OBJECT(dock), "orientation-changed",
		G_CALLBACK(__dock_orientation_changed), tray->manager);
	switch (dock->d.position) {
		case KIBA_POSITION_LEFT_EDGE:
		case KIBA_POSITION_RIGHT_EDGE:
			orientation = GTK_ORIENTATION_VERTICAL;
			break;
		default:
			orientation = GTK_ORIENTATION_HORIZONTAL;
			break;
	}
	__dock_orientation_changed (dock, orientation, tray->manager);
	
	g_object_set_data (G_OBJECT(plugin), "tray", tray);
	
	return TRUE;
}

static PrefMember tray_options[] = {
};

static KibaPluginVtable vtable = {
	"Tray Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_ONCE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
};

KIBA_INIT_PLUGIN (vtable, tray_options, NULL, NULL, NULL);
